"""
Quaternion algebras:

    WORK IN PROGRESS!!!!  April 30, 2005  - William Stein -
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.misc import typecheck
import sage.rings.element as element
import sage.rings.ring as ring
import sage.rings.rings as rings
import sage.rings.arith as arith
import sage.algebras.quaternion
import sage.structure.gens as gens
from sage.linalg.vector_space import VectorSpace
from sage.linalg.vector import Vector

ZZ = rings.IntegerRing()
QQ = rings.RationalField()
is_square = arith.is_square

def QuaternionAlgebra(K, a, b, c=0):
    """
    Returns the quaternion algebra over the field K with generators
    i and j and the following relations:
    
                i^2 = a
                j^2 = b
                i*j + j*i = c
                
    A third generator is set to k = i*j.   

    
    INPUT:
        K -- field
        a -- nonzero element of K
        b -- nonzero element of K
        b -- element of K

    EXAMPLES:
    
    We create quaternion algebras over several base fields.
    
        >>> QuaternionAlgebra(RationalField(), -1, -7)
        Quaternion Algebra over Rational Field

        >>> QuaternionAlgebra(NumberField(x**2-5), -1, -7)
        Quaternion Algebra over Number Field with defining polynomial x^2 - 5

        >>> QuaternionAlgebra(GF(3), -1, -7)
        Quaternion Algebra over Finite field of size 3

    Note that both a and b must be nonzero:
    
        >>> QuaternionAlgebra(RationalField(), 1, 0)
        Traceback (most recent call last):
        ...
        ArithmeticError: Both i^2 (=1) and j^2 (=0) must be nonzero.

        >>> QuaternionAlgebra(GF(7), -1, -7)
        Traceback (most recent call last):
        ...
        ArithmeticError: Both i^2 (=6) and j^2 (=0) must be nonzero.

    Note that K must be a field.  This is because we represent
    elements of the quaternion algebra as vectors in a vector space,
    and we support division.  Use the QuaternionOrder_... commands.

         >>> QuaternionAlgebra(IntegerRing(), -1, -1)
         Traceback (most recent call last):
         ...
         TypeError: K (=Integer Ring) must be a field
    
    """
    if K == rings.RationalField():
        return Algebra_QQ(K, a, b, c)
    return Algebra(K, a, b, c)

def QuaternionAlgebraWithDiscs(D1, D2, T):
    """
    The quaternion algebra over Q generated by i and j, where Z[i] and
    Z[j] are quadratic suborders of discriminant D_1 and D_2,
    respectively, and Z[ij - ji] is a quadratic suborder of
    discriminant D_3 = D_1 D_2 - T^2.

    INPUT:
        D1 -- int
        D2 -- int
        T -- int
        
        The integers D_1, D_2 and T must all be even or all odd, and
        D_1, D_2 and D_3 must each be the discriminant of some
        quadratic order, i.e. nonsquare integers = 0, 1 (mod 4).
        
    OUTPUT:
        A quaternion algebra.

    EXAMPLES:
        >>> A = QuaternionAlgebraWithDiscs(-7,-47,1); A
        Quaternion Algebra over Rational Field
        >>> i, j, k = A.gens()
        >>> i**2
        -2 + i
        >>> j**2
        -12 + j
        >>> k**2
        -24 + k
        >>> i.minimal_polynomial()
        x^2 - x + 2
        >>> j.minimal_polynomial()
        x^2 - x + 12

    THIS IS BROKEN:  (todo!!)
        >> k.minimal_polynomial()
        x^2 - x + 24
        >>> i*j
        k
    """
    # adapt types of input arguments
    D1 = ZZ(D1); D2 = ZZ(D2); T = ZZ(T)
    D3 = D1*D2 - T**2
    if is_square(D1) or is_square(D2) or is_square(D3):
        raise ArithmeticError, "Each of D1 (=%s), D2 (=%s), and D3 (=%s) must be nonsquare."%(D1,D2,D3)
    
    # Create the quaternion algebra with
    #     x^2 = D1,   y^2 = D2,  x*y + y*x = 2*T
    A = QuaternionAlgebra(rings.RationalField(), D1, D2, 2*T)
    x,y,z = A.gens()

    # Now set the generators of the algebra to elements i, j, and k
    # with the desired properties.  We choose i and j to have the
    # right charpolys and let k = i*j.   The tricky part of all this
    # is working in the quaternion algebra Q<x,y> with x*y+y*x = 2*T.
    
    # NOTE: I found that using 2*T for the constant works in every
    # example I try but haven't completely written down a formal
    # proof.  That this should be right is very plausible.
    #
    # TODO: ask D. Kohel.
    #
    # Understanding how to construct this funny quaternion algebra
    # inside a usual quaternion algebra took me a long time.  In


    half = QQ('1/2')
    if D1 % 4 == 0:
        i = half*x
    elif D1 % 4 == 1:
        i = half*(x+1)
    else:
        raise ArithmeticError, "D1 (=%s) must be 0 or 1 mod 4"%D1
    if D2 % 4 == 0:
        j = half*y
    elif D2 % 4 == 1:
        j = half*(y+1)
    else:
        raise ArithmeticError, "D2 (=%s) must be 0 or 1 mod 4"%D2
    k = i*j
    
    A.set_gens([i,j,k])
    return A
    
def QuaternionAlgebraRamifiedAt(N):
    raise NotImplementedError

def QuaternionOrder_RamifiedAt(N, M=1):
    raise NotImplementedError

def QuaternionOrderWithDiss(D1, D2, T):
    raise NotImplementedError


class Algebra(ring.Ring, gens.Gens):
    def __init__(self, K, a, b, c=0, gens=None):
        """
        Returns the quaternion algebra over the field K with generators
        i and j and the following relations:
    
                i^2 = a       (!= 0)
                j^2 = b       (!= 0)
                i*j + j*i = c
                
        A third generator is set to k = i*j.   

        EXAMPLES:
        
        """
        typecheck(K,ring.Ring,"K")
        if not K.is_field():
            raise TypeError, "K (=%s) must be a field"%K
        self.__K = K
        ring.Ring.__init__(self)
        self.__constants = (K(a),K(b),K(c))
        if self.__constants[0] == 0 or self.__constants[1] == 0:
            raise ArithmeticError, "Both i^2 (=%s) and j^2 (=%s) must be nonzero."%(
                self.__constants[0], self.__constants[1])
        V = VectorSpace(K,4)
        self.__V = V
        self.__names = ["i", "j", "k"]
        self.__basis = [self(self.__V.gen(i)) for i in range(4)]
        if gens == None:
            self.set_gens([V._1, V._2, V._3])
        else:
            self.set_gens(gens)

    def set_gens(self, g):
        """
        Changes the chosen generators of self.

        This function is mainly used by other easier-to-use constructors
        of quaternion algebras.
        
        INPUT:
            g -- list of three elements of self, such that the
                 K-vector space spanned by 1, g[0], g[1], g[2] is of
                 dimension 4.
                  
        OUTPUT:
            This function changes self by making the elements of g the
            generators.  Elements are printed in terms of the gi.  The
            names used to represent the gi are same as were used
            before calling set_gens.  To change them, use the
            assign_names function.

        EXAMPLES:
            >>> A = QuaternionAlgebra(RationalField(), -1,-7)
            >>> i,j,k = A.gens()
            >>> i
            i
            >>> j**2
            -7
            >>> (k+i)**2
            -8

        The underlying vector that represents j:
            >>> j.vector()
            (0 0 1 0)

        Now if we change generators by swapping i and j,
        then the 0th generator will have square -7, since it
        is really what we called j above.
            >>> A.set_gens([j,i,k+i])
            >>> i
            j

        It's a good idea to either assign the gens to the your
        variables, or change the names of your variables.
            >>> i,j,k = A.gens()
            >>> i**2
            -7
            >>> i
            i
            >>> i*j
            j - k
            >>> k**2
            -8

        The underlying vector does not change:
            >>> i.vector()
            (0 0 1 0)
            
        The difference is that now A.gen(0) is j, A.gen(1) is i,
        and elements are printed in terms of these generators.
        """
        # First adapt to self.
        g = [self(x) for x in g]
        if not len(g) == 3:
            raise ArithmeticError, "There must be exactly three elements in g (=%s)"%g
        B = [self(1).vector()] + [x.vector() for x in g]
        self.__Vbasis = self.__V.subspace_with_basis(B)
        if self.__Vbasis.dimension() != 4:
            raise ArithmeticError, "The g_i together with 1 must span a space of dimension 4, but their span has dimension %s"%self.__Vbasis.dimension()
        self.__gens = g

    def base_field(self):
        return self.__K

    base_ring = base_field

    def constants(self):
        return self.__constants

    def assign_names(self, s):
        typecheck(s,list,"s")
        if len(s) != 3:
            raise IndexError, "s (=%s) must have length 3."
        self.__names = [str(s[i]) for i in range(3)]

    def names(self):
        return self.__names

    def gen(self, i):
        return self.__gens[i]

    def ngens(self):
        return 3
                 
    def __contains__(self, x):
        if not isinstance(x, sage.algebras.quaternion.Quaternion):
            return False
        return x.parent() == self
    
    def __repr__(self):
        return "Quaternion Algebra over %s"%self.__K

    def __call__(self, x):
        if isinstance(x, sage.algebras.quaternion.Quaternion):
            if x.parent() == self:
                return x
            raise TypeError, "coercion of elements between quaternion algebras not defined."
        if isinstance(x, (int, long, rings.Integer, rings.Rational)) or x in self.__K:
            v = self.__V([self.__K(x),0,0,0])
        else:
            v = self.__V(x)
        return sage.algebras.quaternion.Quaternion(self, v)

    def is_atomic_repr(self):
        return False

    def is_field(self):
        return False

    def name(self):
        return "quatalg_%s_%s_%s"%self.constants()

    def characteristic(self):
        return self.__K.characteristic()

    def random(self, bound=0):
        return self(self.__V.random(bound))

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other,  Algebra):
            return -1
        if self.__constants == other.__constants:
            return 0
        return 1

    def vector_space(self):
        return self.__Vbasis

    def basis(self):
        return self.__basis



class Algebra_QQ(Algebra):
    
    def discriminant(self):
        raise NotImplementedError

    def ramified_primes(self):
        return [p for p, _ in arith.factor(self.discriminant())]

    def maximal_order(self):
        """
        Return a maximal order.
        """
        try:
            return self.__maximal_order
        except AttributeError:
            raise NotImplementedError

    def quaternion_order(self, M=1):
        """
        Return a maximal order.
        """
        try:
            return self.__quaternion_order
        except AttributeError:
            raise NotImplementedError

    def is_definite(self):
        raise NotImplementedError

    def is_indefinite(self):
        raise NotImplementedError




class Lattice:
    def __init__(self, R, gens):
        self.__R = R
        self.__gens = gens



class Order(Lattice):
    def __init__(self, R, v):
        self.__R = R
        self.__v = v  # todo

    def __repr__(self):
        return "Quaternion Order of level %s with base ring %s"%(
            self.level(), self.base_ring())

    def level(self):
        raise NotImplementedError

    def base_ring(self):
        return self.__R



class Ideal(Lattice):
    def __init__(self):
        pass

    def __repr__(self):
        return "Quaternion Ideal ..."

    def __mul__(self, other):
        """
        Return the product (=composition) of ideals self and other.
        """
        typecheck(other, QuaternionIdeal, "other")
        raise NotImplementedError

    composite = __mul__  #alias

    def conjugate(self):
        """
        Return the conjugate ideal.
        """
        raise NotImplementedError

    def intersection(self, other):
        """
        Return the intersection of ideals self and other.
        """
        typecheck(other, Ideal, "other")
        raise NotImplementedError


class Lattice_QQ(Lattice):
    def gram_matrix(self):
        """
        Return the Gram matrix with respect to the norm on the basis.
        """

    def reduced_basis(self):
        """
        Return a Minkowski-reduced basis for self.
        """

    def reduced_gram_matrix(self):
        """
        Return the unique Minkowski-reduced Gram matrix.
         """



class Order_QQ(Lattice_QQ, Order):
    def is_definite(self):
        raise NotImplementedError
    
    def left_ideal(self, v):
        """
        Returns the left ideal generated by the sequence of elements
        of the list v.
        """
        raise NotImplementedError
    def right_ideal(self, v):
        """
        Returns the right ideal generated by the sequence of elements
        of the list v.
        """
        raise NotImplementedError

    def prime_ideal(self, p):
        """
        Returns the unique 2-sided prime ideal P this order S over the
        prime pZ. If p is a ramified prime then P properly contains pS
        and need not be principal; otherwise P is equal to p*S.
        """
        raise NotImplementedError

    def commutator_ideal(self):
        """
        Return the two-sides ideal of this order generated by
        elements of the form x*y - y*x.
        """
        raise NotImplementedError

    def left_ideal_classes(self):
        """
        Return a list of representatives for the left
        locally-free ideal classes of self.  The quaternion
        algebra must be definite.
        """
        raise NotImplementedError
    
    def right_ideal_classes(self):
        """
        Return a list of representatives for the right
        locally-free ideal classes of self.  The quaternion
        algebra must be definite.
        """
        raise NotImplementedError
    
    def two_sided_ideal_classes(self):
        """
        Return a list of representatives for the two-sided
        locally-free ideal classes of self.  The quaternion
        algebra must be definite.
        """
        raise NotImplementedError

    def is_isomorphic(self, other):
        raise NotImplementedError

    def units(self):
        """
        Returns a list of representatives for the units in self
        modulo the unit group of the base ring.
        """
        raise NotImplementedError

    def unit_group(self):
        raise NotImplementedError
    
    
        

class Ideal_QQ(Lattice_QQ, Ideal):    

    def left_order(self):
        raise NotImplementedError

    def right_order(self):
        raise NotImplementedError

    def norm(self):
        """
        Return the norm of the ideal, which is the positive
        generator of the image of the norm map in ZZ.
        """
        raise NotImplementedError        


    def is_isomorphic(self, other):
        raise NotImplementedError

    def isomorphism(self, other):
        raise NotImplementedError

    def is_left_isomorphic(self, other):
        raise NotImplementedError        

    def left_isomorphism(self, other):
        raise NotImplementedError        

    def is_right_isomorphic(self, other):
        raise NotImplementedError        

    def right_isomorphism(self, other):
        raise NotImplementedError        

# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.rings import *
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])

