"""
Quaternions
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator
from   sage.misc.misc import typecheck

import sage.rings.element as element
import sage.algebras.quatalg
import sage.rings.rings as rings
import sage.rings.arith
from   sage.rings.coerce import bin_op

class Quaternion(element.Element):
    def __init__(self, Q, x):
        """
        Create the element x of the quaternion algebra Q.
        
        This function should be called only by Q.  To create a
        Quaternion rom x yourself, create Q and then write Q(x).
        """
        #if not isinstance(Q, sage.algebras.quatalg.Algebra):
        #    raise TypeError
        self.__x = x
        self.is_vector = True
        element.Element.__init__(self, Q)

    def __repr__(self):
        s = ""
        v = self.coordinates()
        names = [""] + self.parent().names()
        for i in range(4):
            if v[i] != 0:
                if len(s) > 0:
                    s += " + "
                if len(s) > 0 and v[i] == -1:
                    if len(s) > 0:
                        s = s[:-3]
                    s += " - "
                    if i == 0:
                        s += "1"
                        
                elif v[i] != 1 or i == 0:
                    s += "%s"%v[i]
                    if i > 0:
                        s += "*"
                if i > 0:
                    s += "%s"%names[i]
        if len(s) == 0: s = "0"
        return s

    def __add__(self, other):
        if not isinstance(other, Quaternion):
            return bin_op(self, other, operator.add)
        return Quaternion(self.parent(), self.__x + other.__x)

    def __sub__(self, other):
        if not isinstance(other, Quaternion):
            return bin_op(self, other, operator.sub)
        return Quaternion(self.parent(), self.__x - other.__x)

    def __neg__(self):
        return Quaternion(self.parent(), -self.__x)

    def __mul__(self, other):
        if not isinstance(other, Quaternion):
            return bin_op(self, other, operator.mul)
        A = self.parent()
        a, b, c = A.constants()
        v = self.__x
        w = other.__x
        ## if c == 0:
##             return A([-a*b*v[3]*w[3] + a*v[1]*w[1] + b*v[2]*w[2] + v[0]*w[0],
##                      -b*v[2]*w[3] + b*v[3]*w[2] + v[0]*w[1] + v[1]*w[0],
##                      a*v[1]*w[3] - a*v[3]*w[1] + v[0]*w[2] + v[2]*w[0],
##                      v[0]*w[3] + v[1]*w[2] - v[2]*w[1] + v[3]*w[0]])

        # I found the following formulas by hand (literally).  
        V = v.parent()
        z = v[0]*w + v[1]*V([w[1]*a, w[0], a*w[3], w[2]]) + \
                     v[2]*V([c*w[1]+b*w[2], -b*w[3], w[0] + c*w[3], -w[1]]) + \
                     v[3]*V([-a*b*w[3], c*w[1]+b*w[2], -a*w[1], c*w[3]+w[0]])
        return A(z)
        

    def __invert__(self):
        a, b, c = self.parent().constants()
        v = self.__x
        N = self.norm()
        if N == 0:
            raise ZeroDivisionError, "Quaternion (=%s) is not invertible."%self
        return Quaternion(self.parent(), [v[0]/N,-v[1]/N, -v[2]/N, -v[3]/N])

    def __div__(self, other):
        if not isinstance(other, Quaternion):
            return bin_op(self, other, operator.div)
        return self * ~other

    def __pow__(self, n):
        typecheck(n, (int, long, rings.Integer), "n")
        return sage.rings.arith.generic_power(self, n)

    def __cmp__(self, other):
        if not isinstance(other, Quaternion):
            other = self.parent()(other)
        try:
            other = self.parent()(other)
        except TypeError:
            return -1
        return self.__x.__cmp__(other.__x)

    def characteristic_polynomial(self):
        a, b, c = self.parent().constants()
        assert c!=0
        v = self.__x
        x = rings.polygen(self.parent().base_field())
        return x**2 - 2*v[0]*x + a*b*v[3]**2 - a*v[1]**2 - b*v[2]**2 + v[0]**2

    charpoly = characteristic_polynomial


    def minimal_polynomial(self):
        a, b, c = self.parent().constants()
        assert c!=0
        v = self.__x
        x = rings.polygen(self.parent().base_field())
        if v[1] == 0 and v[2] == 0 and v[3] == 0:
            return x - v[0]
        return x**2 - 2*v[0]*x + a*b*v[3]**2 - a*v[1]**2 - b*v[2]**2 + v[0]**2

    minpoly = minimal_polynomial

    def norm(self):
        try:
            return self.__norm
        except AttributeError:
            a, b, c = self.parent().constants()
            assert c!=0
            v = self.__x
            self.__norm = a*b*v[3]**2 - a*v[1]**2 - b*v[2]**2 + v[0]**2
            return self.__norm
    
    def trace(self):
        raise NotImplementedError # need c info
        return 2*self.__x[0]

    def conjugate(self):
        v = self.__x
        return self.parent()([v[0], -v[1], -v[2], -v[3]])

    def coordinates(self):
        return self.parent().vector_space().coordinates(self.__x)

    def vector(self):
        """
        Returns a reference to the vector that defines this quaternion.
        """
        return self.__x


# doctests
if __name__ ==  '__main__':
    from sage.rings.rings import *
    import doctest, sys
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])

