"""
Elliptic curves
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.groups.abelian as abelian
import sage.rings.arith as arith
import sage.rings.rings as rings
import sage.rings.number_field as number_field
import sage.misc.errors as errors
import sage.misc.misc as misc
import sage.modular.modform as modform
import sage.rings.rings as rings
import sage.tables.elliptic_curves 
import sage.pari.py_pari as py_pari
pari = py_pari.pari
import sage.functions.transcendental as transcendental
import math
factor = arith.factor
sqrt = math.sqrt
exp = math.exp
mul = misc.mul
next_prime = arith.next_prime

try:
    import mwrank
    MWRANK = True
except ImportError:
    MWRANK = False
    pass

oo = rings.infinity       # infinity
Q = rings.RationalField()         # rational numbers
O = rings.O         # big oh
Z = rings.IntegerRing()
C = rings.ComplexField()
R = rings.RealField()
GF = rings.GF

_MAX_HEIGHT=21

class EllipticCurve(object):
    """
    Elliptic curve.

    There are several forms of this function.
    
        -- EllipticCurve([a1,a2,a3,a4,a6]): Elliptic curve with given
           a-invariants.  The invariants are coerced into a the parent
           of the first element.  If all are integers, they are coerced
           into the rational numbers.

        -- EllipticCurve([a4,a6]): Same as above, but a1=a2=a3=0.
        
        -- EllipticCurve(label): Returns the elliptic curve over Q
           from the Cremoa database with the given label.  The label
           is a string, such as"11A" or "37B2".

        -- EllipticCurve(R, [a1,a2,a3,a4,a6]): Create the elliptic curve
           over R with given a-invariants.  Here R can be an arbitrary
           ring.  Note that addition need not be defined.
        

    EXAMPLES:
    We illustrate creating elliptic curves.
    
        >>> EllipticCurve([0,0,1,-1,0])
        Elliptic Curve defined by y^2 + y = x^3 - x  over Rational Field

        >>> EllipticCurve([GF(5)(0),0,1,-1,0])
        Elliptic Curve defined by y^2 + y = x^3 - x  over Finite field of size 5

    Alternatively, one can create the curve over the finite field as follows:
        >>> EllipticCurve(GF(5), [0, 0,1,-1,0])
        Elliptic Curve defined by y^2 + y = x^3 - x  over Finite field of size 5

        >>> E = EllipticCurve([C(0),0,1,-1,0])
        >>> E
        Elliptic Curve defined by y^2 + y = x^3 - x  over Complex Field
        >>> E.j_invariant()
        2988.97297297297297297

        
        >>> E = EllipticCurve(ZZ, [0, 0,1,-1,0])
        >>> E
        Elliptic Curve defined by y^2 + y = x^3 - x  over Integer Ring

    Of course, arithmetic on elliptic curves over Z need not be defined:
        >>> P = E([0,0])
        >>> P + P + P + P
        (2, -3)
        >>> P + P + P + P + P
        Traceback (most recent call last):
        ...
        ArithmeticError: Point (1/4, -5/8) is not on curve.
    """

    def __new__(cls, *args, **kwds):
        if cls != EllipticCurve:
            return object.__new__(cls, *args, **kwds)
        if len(args) == 0 or not isinstance(args[0], (list, str, sage.tables.elliptic_curves.EllipticCurve, rings.Ring)):
            raise TypeError, "Invalid arguments to EllipticCurve constructor."

        if isinstance(args[0],rings.Ring):
            try:
                if args[0] == Q:
                    return EllipticCurve_RationalField(*args, **kwds)
                K = args[0]
                return EllipticCurve_generic(*args, **kwds)
            except ZeroDivisionError:
                raise TypeError, "Invalid arguments to EllipticCurve constructor (missing a-invariants)."
            
        ainvs = args[0]
        if isinstance(ainvs, list):
            if len(ainvs) != 2 and len(ainvs) != 5:
                raise TypeError, "Length of ainvs must be 2 or 5."
        T = False
        try:
            T = isinstance(ainvs, (str, sage.tables.elliptic_curves.EllipticCurve)) or \
               not isinstance(ainvs[0], rings.Element) or \
               ainvs[0].parent() in [Z,Q]
        except:
            pass
        if T:
            return EllipticCurve_RationalField(*args, **kwds)
        return EllipticCurve_generic(*args, **kwds)


    def __init__(self, *args, **kwds):
        raise NotImplementedError


class EllipticCurve_generic(EllipticCurve):
    """
    Elliptic curve over a generic base ring.
    """
    def __init__(self, ainvs, extra=None):
        if extra != None:   # possibility of two arguments
            K, ainvs = ainvs, extra
        else:
            K = ainvs[0].parent()
        assert len(ainvs) == 2 or len(ainvs) == 5
        self.__base_ring = K
        ainvs = [K(x) for x in ainvs]
        if len(ainvs) == 2:
            ainvs = [K(0),K(0),K(0)] + ainvs
        self.__ainvs = ainvs
        if self.discriminant() == 0:
            raise ArithmeticError, "Invariants %s define a singular curve."%ainvs
    
    def __repr__(self):
        """
        String representation of elliptic curve.

        EXAMPLES:
            >>> EllipticCurve([1,2,3,4,5])
            Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 4*x + 5 over Rational Field
        """
        a = self.ainvs()
        #return "y^2 + %s*x*y + %s*y = x^3 + %s*x^2 + %s*x + %s"%\
        #       (a[0], a[2], a[1], a[3], a[4])
        s = "Elliptic Curve defined by "
        s += "y^2 "
        if a[0] == -1:
            s += "- x*y "
        elif a[0] == 1:
            s += "+ x*y "
        elif a[0] != 0:
            s += "+ %s*x*y "%a[0]
        if a[2] == -1:
            s += " - y"
        elif a[2] == 1:
            s += "+ y"
        elif a[2] != 0:
            s += "+ %s*y"%a[2]
        s += " = x^3 "
        if a[1] == -1:
            s += "- x^2 "
        elif a[1] == 1:
            s += "+ x^2 "
        elif a[1] != 0:
            s += "+ %s*x^2 "%a[1]
        if a[3] == -1:
            s += "- x "
        elif a[3] == 1:
            s += "+ x "
        elif a[3] != 0:
            s += "+ %s*x "%a[3]
        if a[4] == -1:
            s += "-1"
        elif a[4] == 1:
            s += "+1"
        elif a[4] != 0:
            s += "+ %s"%a[4]
        s = s.replace("+ -","- ")
        s += " over %s"%self.base_ring()
        return s

    def __contains__(self, P):
        """
        Returns True if and only if P defines is a point on the
        elliptic curve.  P just has to be something that can be
        coerced to a point.

        EXAMPLES:
            >>> E = EllipticCurve([0, 0, 1, -1, 0])
            >>> (0,0) in E
            True
            >>> (1,3) in E
            False
            >>> E = EllipticCurve([GF(7)(0), 1])
            >>> [0,0] in E
            False
            >>> [0,8] in E
            True
            >>> P = E(0,8)
            >>> P
            (0, 1)
            >>> P in E
            True
        """
        if not isinstance(P, Point):
            try:
                P = self(P)
            except ArithmeticError:
                return False
        if P.curve() == self:
            return True
        x, y, a = P[0], P[1], self.ainvs()
        return y**2 + a[0]*x*y + a[2]*y == x**3 + a[1]*x**2 + a[3]*x + a[4]

    def __call__(self, x=oo, y=oo):
        """
        Create a point on the elliptic curve, if possible.  If a the
        point is not on the curve, an ArithmeticError exception is
        raised.

        EXAMPLES:
            >>> E = EllipticCurve([0, 0, 1, -1, 0])
            >>> P = E(0,0)
            >>> P
            (0, 0)
            >>> P.curve()
            Elliptic Curve defined by y^2 + y = x^3 - x  over Rational Field
            >>> E([0,0])
            (0, 0)
            >>> E([1,0,0])
            (1,0,0)
            >>> E(infinity)
            (1,0,0)

        We create points on an elliptic curve over a prime finite field.
            >>> E = EllipticCurve([GF(7)(0), 1])
            >>> E([2,3])
            (2, 3)
            >>> E([0,0])
            Traceback (most recent call last):
            ...
            ArithmeticError: Point (0, 0) is not on curve.

        We create a point on an elliptic curve over a number field.
            >>> x = polygen(RationalField())
            >>> K = NumberField(x**3 + x + 1); a = K.gen()
            >>> E = EllipticCurve([a,a])
            >>> E
            Elliptic Curve defined by y^2  = x^3 + a*x + a over Number Field with defining polynomial x^3 + x + 1
            >>> E = EllipticCurve([K(1),1])
            >>> E
            Elliptic Curve defined by y^2  = x^3 + x +1 over Number Field with defining polynomial x^3 + x + 1
            >>> P = E([a,0])
            >>> P
            (a, 0)
            >>> P+P
            (1,0,0)
        """
        if isinstance(x,Point):
            if x.curve() == self:
                return x
            else:
                x, y = x[0], x[1]
        elif isinstance(x,list):
            if len(x) != 2:
                if len(x) == 3:
                    K = self.base_ring()
                    c = K(x[2])
                    if c != 0:
                        x, y   =   K(x[0])/c, K(x[1])/c
                    else:
                        return Point(self,oo,oo)
                else:
                    raise TypeError, "List must have length 2."
            else:
                x, y = x[0], x[1]
        return Point(self, x, y)

    def __getitem__(self, n):
        """
        """
        raise NotImplementedError, "not implemented."

    def __is_over_RationalField(self):
        return isinstance(self.base_ring(), rings.RationalField)

    def a_invariants(self):
        """
        The a-invariants of this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,5])
            >>> E.a_invariants()
            [1, 2, 3, 4, 5]
            >>> E = EllipticCurve([0,1])
            >>> E
            Elliptic Curve defined by y^2  = x^3 +1 over Rational Field
            >>> E.a_invariants()
            [0, 0, 0, 0, 1]
            >>> E = EllipticCurve([GF(7)(3),5])
            >>> E.a_invariants()
            [0, 0, 0, 3, 5]
        """
        return self.__ainvs

    ainvs = a_invariants

    def b_invariants(self):
        """
        The b-invariants of this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.b_invariants()
            (-4, -20, -79, -21)
            >>> E = EllipticCurve([-4,0])
            >>> E.b_invariants()
            (0, -8, 0, -16)

        ALGORITHM: These are simple functions of the a invariants.

        AUTHOR: William Stein, 2005-04-25
        """
        try:
            return self.__b_invariants
        except AttributeError:
            a1,a2,a3,a4,a6 = tuple(self.ainvs())
            self.__b_invariants = a1*a1 + 4*a2, \
                                  a1*a3 + 2*a4, \
                                  a3**2 + 4*a6, \
                                  a1**2 * a6 + 4*a2*a6 - a1*a3*a4 + a2*a3**2 - a4**2
            return self.__b_invariants

    def c_invariants(self):
        """
        The c-invariants of this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.c_invariants()
            (496, 20008)
            >>> E = EllipticCurve([-4,0])
            >>> E.c_invariants()
            (192, 0)

        ALGORITHM: These are simple functions of the b invariants.

        AUTHOR: William Stein, 2005-04-25
        """
        try:
            return self.__c_invariants
        except AttributeError:
            b2,b4,b6,b8 = self.b_invariants()
            self.__c_invariants = b2**2 - 24*b4,\
                                  -b2**3 + 36*b2*b4 - 216*b6    # note: c6 is wrong in Silverman, but right in Cremona
            return self.__c_invariants

    def base_ring(self):
        """
        Returns the base ring of the elliptic curves.
        
        EXAMPLES:
            >>> E = EllipticCurve([GF(49)(3),5])
            >>> E.base_ring()
            Finite field of size 7^2
            
            >>> E = EllipticCurve([1,1])
            >>> E.base_ring()
            Rational Field

            >>> E = EllipticCurve(ZZ, [3,5])
            >>> E.base_ring()
            Integer Ring
        """
        return self.__base_ring

    base_field = base_ring

    def a1(self):
        """
        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,6])
            >>> E.a1()
            1
        """
        return self.__ainvs[0]
    
    def a2(self):
        """
        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,6])
            >>> E.a2()
            2
        """
        return self.__ainvs[1]
    
    def a3(self):
        """
        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,6])
            >>> E.a3()
            3
        """
        return self.__ainvs[2]
    
    def a4(self):
        """
        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,6])
            >>> E.a4()
            4
        """
        return self.__ainvs[3]
    
    def a6(self):
        """
        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,6])
            >>> E.a6()
            6
        """
        return self.__ainvs[4]




    
    def gens(self):
        raise NotImplementedError
        
    def quadratic_twist(self, D):
        a = self.ainvs()
        return EllipticCurve([0, 0, 0, \
                              -27*D**2*a[0]**4 - 216*D**2*a[0]**2*a[1] + 648*D**2*a[0]*a[2] - 432*D**2*a[1]**2 + 1296*D**2*a[3], \
                              54*D**3*a[0]**6 + 648*D**3*a[0]**4*a[1] - 1944*D**3*a[0]**3*a[2] + 2592*D**3*a[0]**2*a[1]**2 - 3888*D**3*a[0]**2*a[3] - 7776*D**3*a[0]*a[1]*a[2] + 3456*D**3*a[1]**3 - 15552*D**3*a[1]*a[3] + 11664*D**3*a[2]**2 + 46656*D**3*a[4]]).minimal_model()

    def discriminant(self):
        """
        Returns the discriminant of this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve([0,0,1,-1,0])
            >>> E.discriminant()
            37
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.discriminant()
            -161051

            >>> E = EllipticCurve([GF(7)(2),1])
            >>> E.discriminant()
            1
        """
        try:
            return self.__discriminant
        except:
            b2, b4, b6, b8 = self.b_invariants()
            self.__discriminant = -b2**2*b8 - 8*b4**3 - 27*b6**2 + 9*b2*b4*b6
            return self.__discriminant

    def j_invariant(self):
        """
        Returns the j-invariant of this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve([0,0,1,-1,0])
            >>> E.j_invariant()
            110592/37
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.j_invariant()
            -122023936/161051
            >>> E = EllipticCurve([-4,0])
            >>> E.j_invariant()
            1728

            >>> E = EllipticCurve([GF(7)(2),1])
            >>> E.j_invariant()
            1
        """
        try:
            return self.__j_invariant
        except AttributeError:
            c4, _ = self.c_invariants()
            self.__j_invariant = c4**3 / self.discriminant()
            return self.__j_invariant


    def torsion_polynomial(self, n, i=0):
        """
        Returns the n-th torsion polynomial (a.k.a., division polynomial).

        INPUT:
            n -- non-negative integer
            i -- integer, either 0 (default) or 1.
            
        OUTPUT:
            Polynomial -- n-th torsion polynomial, which is a polynomial over
                          the base field of the elliptic curve. 

        EXAMPLES:
            >>> E = EllipticCurve([0,0,1,-1,0])
            >>> E.division_polynomial(1)
            1
            >>> E.division_polynomial(2)
            4*x^3 - 4*x + 1
            >>> E.division_polynomial(3)
            3*x^4 - 6*x^2 + 3*x - 1

            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.torsion_polynomial(0)
            0
            >>> E.torsion_polynomial(1)
            1
            >>> E.torsion_polynomial(2)
            4*x^3 - 4*x^2 - 40*x - 79
            >>> E.torsion_polynomial(3)
            3*x^4 - 4*x^3 - 60*x^2 - 237*x - 21
            >>> E.torsion_polynomial(4)
            8*x^9 - 24*x^8 - 464*x^7 - 2758*x^6 + 6636*x^5 + 34356*x^4 + 53510*x^3 + 99714*x^2 + 351024*x + 459859

            >>> E = EllipticCurve([-4,0])
            >>> E.torsion_polynomial(2)
            4*x^3 - 16*x
            >>> E.torsion_polynomial(5)
            5*x^12 - 248*x^10 - 1680*x^8 + 19200*x^6 - 32000*x^4 + 51200*x^2 + 4096
            >>> E.torsion_polynomial(6)
            12*x^19 - 1200*x^17 - 18688*x^15 + 422912*x^13 - 2283520*x^11 + 9134080*x^9 - 27066368*x^7 + 19136512*x^5 + 19660800*x^3 - 3145728*x

        ALGORITHM: ?

        AUTHOR: Copyright 2005, David Kohel (kohel@maths.usyd.edu.au), 2005-04-25
        """
        n = int(n)
        try:
            return self.__torsion_polynomial[n]
        except AttributeError:
            self.__torsion_polynomial = {}
        except KeyError:
            pass
        E = self; i=int(i)
        if n < 0:
            raise ValueError, "n must be a non-negative integer."
        if i > 1 : 
            raise ValueError, "i must be 0 or 1."
        
        R = rings.PolynomialRing(E.base_ring())
        if i == 1:
            if n == 0:
                f = E.torsion_polynomial(1)
                E.__torsion_polynomial[n] = f 
                return f
            else: 
                x = R.gen()
                psi2 = E.torsion_polynomial(2)
                if n%2 == 0: 
                    f = x * psi2 * (E.torsion_polynomial(n)//psi2)**2 - \
                        E.torsion_polynomial(n+1) * E.torsion_polynomial(n-1)
                    E.__torsion_polynomial[n] = f
                    return f
                else:
                    f = x * E.torsion_polynomial(n)**2 - \
                        (E.torsion_polynomial(n+1)//psi2) * E.torsion_polynomial(n-1)
                    E.__torsion_polynomial[n] = f
                    return f
                
        else:
            
            if n == 0: return R(0)
            if n == 1: return R(1)
            x = R.gen()
            b2, b4, b6, b8 = E.b_invariants()
            psi2 = 4*x**3 + b2*x**2 + 2*b4*x + b6
            if n == 2: 
                f = psi2
                E.__torsion_polynomial[n] = f; return f
            if n == 3:
                f = 3*x**4 + b2*x**3 + 3*b4*x**2 + 3*b6*x + b8
                E.__torsion_polynomial[n] = f; return f
            if n == 4:
                f = psi2 * (2*x**6 + b2*x**5 + 5*b4*x**4 + 10*b6*x**3 \
                    + 10*b8*x**2  + (b2*b8 - b4*b6)*x + b4*b8 - b6**2)
                E.__torsion_polynomial[n] = f; return f
            if n%2 == 0:
                m = n//2 
                if m%2 == 0:
                    f = E.torsion_polynomial(m) * ( \
                        (E.torsion_polynomial(m+2)//psi2) * E.torsion_polynomial(m-1)**2 - \
                        (E.torsion_polynomial(m-2)//psi2) * E.torsion_polynomial(m+1)**2)
                    E.__torsion_polynomial[n] = f; return f
                else:
                    f = psi2 * E.torsion_polynomial(m)*( \
                        E.torsion_polynomial(m+2) * (E.torsion_polynomial(m-1)//psi2)**2 - \
                        E.torsion_polynomial(m-2) * (E.torsion_polynomial(m+1)//psi2)**2)
                    E.__torsion_polynomial[n] = f; return f                    
            else:  
                m = n//2 
                if m%2 == 0:
                    f = psi2 * \
                        E.torsion_polynomial(m+2) * (E.torsion_polynomial(m)//psi2)**3 - \
                        E.torsion_polynomial(m-1) * E.torsion_polynomial(m+1)**3
                    E.__torsion_polynomial[n] = f; return f                    
                else:
                    f = E.torsion_polynomial(m+2) * E.torsion_polynomial(m)**3 - psi2 * \
                        E.torsion_polynomial(m-1)*(E.torsion_polynomial(m+1)//psi2)**3
                    E.__torsion_polynomial[n] = f; return f

    division_polynomial = torsion_polynomial
            
    def weierstrass_model(self):
        a0, a1, a2, a3, a4 = self.ainvs()
        return EllipticCurve([-27*a0**4 - 216*a0**2*a1 + 648*a0*a2 - 432*a1**2 + 1296*a3, \
                              54*a0**6 + 648*a0**4*a1 - 1944*a0**3*a2 + 2592*a0**2*a1**2 -\
                              3888*a0**2*a3 - 7776*a0*a1*a2 + 3456*a1**3 - \
                              15552*a1*a3 + 11664*a2**2 + 46656*a4])
    

class EllipticCurve_RationalField(EllipticCurve_generic):
    """
    Elliptic curve over the Rational Field.
    """
    def __init__(self, ainvs, extra=None):
        if extra != None:   # possibility of two arguments (the first would be the field)
            ainvs = extra   
        if isinstance(ainvs, str):
            X = sage.tables.elliptic_curves.Database().curve(ainvs)
            self.__init_from_database_curve(X)
            return
        elif isinstance(ainvs, sage.tables.elliptic_curves.EllipticCurve):
            self.__init_from_database_curve(ainvs)            
            return
        EllipticCurve_generic.__init__(self, [Q(x) for x in ainvs])
        self.__np = {}
        if self.base_ring() != Q:
            raise TypeError, "Base field (=%s) must be the Rational Field."%self.base_ring()
        
    
    def __init_from_database_curve(self, X):
        EllipticCurve_generic.__init__(self, [Q(a) for a in X.a_invariants])        
        self.__database_curve = X
        self.__base_ring = Q
        self.__analytic_rank = X.rank
        self.__rank = X.rank
        self.__conductor = Z(X.conductor)
        self.__torsion_order = X.torsion
        self.__cremona_label = "%s %s %s"%(X.conductor,X.iso_class,X.number)
        self.__tamagawa_number = X.tamagawa
        self.__minimal_model = self
        if hasattr(X,'modular_degree') and X.modular_degree != None:
            self.__modular_degree = X.modular_degree

    def conductor(self):
        """
        Returns the conductor of the elliptic curve.

        EXAMPLE:
            >>> E = EllipticCurve([1, -1, 1, -29372, -1932937])
            >>> E.conductor()    # known bug in PARI causes this to output 14377604814
            3006
        """
        try:
            return self.__conductor
        except AttributeError:
            self.pari_mincurve() # computes conductor as side effect
        return self.__conductor

    ####################################################################
    #  Access to PARI curves related to this curve.
    ####################################################################

    def __pari_double_prec(self):
        """
        Double the precision of computations with this curve.
        """
        self.pari_curve()
        prec = 2 * self.__pari_curve[1]
        self.__pari_curve = (pari(self.ainvs()).ellinit(prec=prec), prec)
        try:
            del self.__pari_mincurve
        except AttributeError:
            pass
        
    def pari_curve(self):
        """
        Return the PARI curve corresponding to this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve([0, 0,1,-1,0])
            >>> e = E.pari_curve()
            >>> type(e)
            <type '_py_pari.gen'>
            >>> e.type()
            't_VEC'
            >>> e.ellan(10)
            [1, -2, -3, 2, -2, 6, -1, 0, 6, 4]

            >>> E = EllipticCurve(RationalField(), ['1/3', '2/3'])
            >>> e = E.pari_curve()
            >>> e.type()
            't_VEC'
            >>> e[:5]
            [0, 0, 0, 1/3, 2/3]
        """
        try:
            return self.__pari_curve[0]
        except AttributeError:
            self.__pari_curve = (pari(self.a_invariants()).ellinit(prec=10), 10)
        return self.__pari_curve[0]

    def pari_mincurve(self):
        """
        Return the PARI curve corresponding to a minimal model
        for this elliptic curve.

        EXAMPLES:
            >>> E = EllipticCurve(RationalField(), ['1/3', '2/3'])
            >>> e = E.pari_mincurve()
            >>> e[:5]
            [0, 0, 0, 27, 486]
            >>> E.conductor()
            47232
            >>> e.ellglobalred()
            [47232, [1, 0, 0, 0], 2]
        """
        try:
            return self.__pari_mincurve
        except AttributeError:
            if not isinstance(self.base_ring(), rings.RationalField):
                raise NotImplementedError, "This computation is only implemented for elliptic curves over the rational number field."
            e = self.pari_curve()
            R = e.ellglobalred()
            try:
                self.__conductor
            except AttributeError:
                self.__conductor = Z(R[0])
            self.__pari_mincurve = e.ellchangecurve(R[1])
        return self.__pari_mincurve

    def database_curve(self):
        """
        Return the curve in the elliptic curve database isomorphic to
        this curve, if possible.  Otherwise raise a RuntimeError
        exception.  

        EXAMPLES:
        We assume the elliptic curve database is installed.
            >> E = EllipticCurve([1,2,3,4,5])
            >> E.database_curve()
            10351A1: invs=[1, -1, 0, 4, 3], rank=1, tors=1, cp={11: 1, 941: 1}

        We can also create an elliptic curve from a curve in the database:
            >> EllipticCurve(E.database_curve())
            Elliptic Curve defined by y^2 + x*y  = x^3 - x^2 + 4*x + 3 over Rational Field

        The database curve can come directly from the database.
            >> import sage.tables.elliptic_curves as T
            >> db = T.Database()
            >> f = db[37][0][0]
            >> f
            37A1: invs=[0, 0, 1, -1, 0], rank=1, tors=1, cp={37: 1}
            >> EllipticCurve(f)
            Elliptic Curve defined by y^2 + y = x^3 - x  over Rational Field

        NOTES:
        
            1. The model of the curve in the database can be different
               than the Weierstrass model for this curve, since database
               models are always minimal. 
               
            2. The curve C returned from the database is not of type
               sage.ellcurve.EllipticCurve; it has a few data members,
               and very few methods.

        """
        try:
            return self.__database_curve
        except AttributeError:
            try:
                D = sage.tables.elliptic_curves.Database()
            except RuntimeError:
                raise RuntimeError, "The elliptic curve database is not installed."
            F = self.minimal_model()
            v = [int(z) for z in F.a_invariants()]
            for E in D.curves(self.conductor()):
                if v == E.a_invariants:
                    self.__database_curve = E
                    return E
            raise RuntimeError, "Elliptic curve %s not in the database."%self

    def Np(self, p):
        """
        The number of points on E modulo p, where p is a prime
        of good reduction.

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.Np(2)
            5
            >>> E.Np(3)
            5
            >>> E.conductor()
            11
            >>> E.Np(11)
            Traceback (most recent call last):
            ...
            ArithmeticError: p (=11) must be a prime of good reduction
        """
        if not (self.conductor() % p):
            raise ArithmeticError, "p (=%s) must be a prime of good reduction"%p
        return p+1-self.ap(p)

    def __pari_double_prec(self):
        EllipticCurve_generic._EllipticCurve__pari_double_prec(self)
        try:
            del self.__pari_mincurve
        except AttributeError:
            pass
        
    ####################################################################
    #  Access to mwrank
    ####################################################################
    def mwrank_curve(self, verbose=False):
        try:
            return self.__mwrank_curve
        except AttributeError:
            if MWRANK:
                self.__mwrank_curve = mwrank.EllipticCurve(self.ainvs(), verbose=verbose)
            else:
                raise ImportError, "the mwrank package must be installed for requested functionality."
        return self.__mwrank_curve

    def two_descent(self, verbose=True,
                    sel = False,
                    first_limit = 20,
                    second_limit = 5,
                    n_aux = -1,
                    second_descent = 1):
        """
        Compute 2-descent data for this curve.   The actual data is
        part of self.mwrank_curve().

        INPUT:
            verbose     -- (default: True) print what mwrank is doing
            sel         -- (default: False) ??
            first_limit -- (default: 20) ??
            second_limit-- (default: 5)  ??
            n_aux       -- (default: -1) ??
            second_descent -- (default: 1) ??
        """
        self.mwrank_curve().two_descent(verbose=True)


    ####################################################################
    #  Etc.
    ####################################################################

    def anlist(self, n):
        """
        The Fourier coefficients up to and including a_n of the
        modular form attached to this elliptic curve.  The ith element
        of the return list is a[i].
        

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.anlist(3)
            [0, 1, -2, -1]
            
            >>> E = EllipticCurve([0,1])
            >>> E.anlist(20)
            [0, 1, 0, 0, 0, 0, 0, -4, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 8, 0]
        """
        n = int(n)
        try:
            if len(self.__anlist) > n:
                return self.__anlist[:n+1]
        except AttributeError:
            pass
        E = self.pari_mincurve()
        
        # There is some overheard associated with coercing the PARI
        # list back to Python, but it's not bad.  It's better to do it
        # this way instead of trying to eval the whole list, since the
        # int conversion is done very sensibly.  NOTE: This would fail
        # if a_n won't fit in a C int, i.e., is bigger than
        # 2147483648; however, we wouldn't realistically compute
        # anlist for n that large anyways.
        #
        # Some relevant timings:
        #
        # E <--> [0, 1, 1, -2, 0]   389A
        #  E = EllipticCurve([0, 1, 1, -2, 0]);   // SAGE or MAGMA
        #  e = E.pari_mincurve()
        #  f = ellinit([0,1,1,-2,0]);
        #
        #  Computation                                              Time (1.6Ghz Pentium-4m laptop)
        #  time v:=TracesOfFrobenius(E,10000);  // MAGMA            0.120
        #  gettime;v=ellan(f,10000);gettime/1000                    0.046
        #  time v=e.ellan (10000)                                   0.04
        #  time v=E.anlist(10000)                                   0.07
        
        #  time v:=TracesOfFrobenius(E,100000);  // MAGMA           1.620
        #  gettime;v=ellan(f,100000);gettime/1000                   0.676
        #  time v=e.ellan (100000)                                  0.7
        #  time v=E.anlist(100000)                                  0.83

        #  time v:=TracesOfFrobenius(E,1000000);  // MAGMA          20.850
        #  gettime;v=ellan(f,1000000);gettime/1000                  9.238
        #  time v=e.ellan (1000000)                                 9.61
        #  time v=E.anlist(1000000)                                 10.95  (13.171 in cygwin vmware)
        
        #  time v:=TracesOfFrobenius(E,10000000);  //MAGMA          257.850
        #  gettime;v=ellan(f,10000000);gettime/1000      FAILS no matter how many allocatemem()'s!!
        #  time v=e.ellan (10000000)                                139.37  
        #  time v=E.anlist(10000000)                                136.32
        #
        #  The last SAGE comp retries with stack size 40MB,
        #  80MB, 160MB, and succeeds last time.  It's very interesting that this
        #  last computation is *not* possible in GP, but works in py_pari!
        #
        if n >= 2147483648:
            raise RuntimeError, "anlist: n must be < 2147483648."
        
        self.__anlist = [0] + [int(x) for x in E.ellan(n)]
        return self.__anlist

    def analytic_rank(self):
        try:
            return self.__analytic_rank
        except AttributeError:
            raise NotImplementedError
            #x = pari.ellanalyticrank(self)
            #self.__analytic_rank = x[0]
            return self.__analytic_rank

    def rank(self, use_database=False, verbose=False):
        """
        Return the rank of this elliptic curve, assuming no conjectures.

        If we fail to provably compute the rank, raises a RuntimeError
        exception.

        INPUT:
            use_database (bool) -- (default: False), if True, try to
                  look up the regulator in the Cremona database.
            proof (bool) -- (default: True), if True then rank is
                  proven correct; otherwise, computation of rank is not certain.
            verbose -- (default: None), if specified changes the
                       verbosity of mwrank computations.
            
        OUTPUT:
            rank (int) -- the rank of the elliptic curve.

        IMPLEMENTATION: Uses L-functions and Cremona's mwrank C library.
        """
        try:
            return self.__rank
        except AttributeError:
            if use_database:
                try:
                    self.__rank = self.database_curve().rank
                    return self.__rank
                except AttributeError:
                    pass
            N = self.conductor()
            prec = int(4*sqrt(N)) + 10
            if self.root_number() == 1:
                L, err = self.Lseries_at1(prec)            
                if abs(L) > err + 0.0001:  # definitely doesn't vanish
                    misc.verbose("rank 0 because L(E,1)=%s"%L)
                    self.__rank = 0
                    return self.__rank
            else:
                Lprime, err = self.Lseries_deriv_at1(prec)
                if abs(Lprime) > err + 0.0001:  # definitely doesn't vanish
                    misc.verbose("rank 1 because L'(E,1)=%s"%Lprime)
                    self.__rank = 1
                    return self.__rank

            misc.verbose("using mwrank")
            C = self.mwrank_curve()
            C.set_verbose(verbose)
            r = C.rank()
            if not C.certain():                raise RuntimeError, "Unable to compute the rank with certainty (lower bound=%s)."%r
            self.__rank = r
        return self.__rank

    def gens(self, verbose=False):
        """
        Compute and return generators for the Mordell-Weil group E(Q) modulo torsion.
        
        INPUT:
            verbose -- (default: None), if specified changes the
                       verbosity of mwrank computations.
        OUTPUT:
            generators -- List of generators for the Mordell-Weil group.

        IMPLEMENTATION: Uses Cremona's mwrank C library.
        """
        try:
            return self.__gens
        except AttributeError:
            try:
                misc.verbose("Trying to compute rank.")
                r = self.rank()
                misc.verbose("Got r = %s."%r)
                if r == 0:
                    misc.verbose("Rank = 0, so done.")
                    self.__gens = []
                    self.__regulator = float(1)
                    return self.__gens
                if r == 1:
                    misc.verbose("Rank = 1, so using direct search.")
                    h = 5
                    while True:
                        if h > _MAX_HEIGHT:   # overflows 
                            raise RuntimeError, "Small height point not found."
                        G = self.point_search(h, verbose)
                        G = [P for P in G if P.order() == 0]
                        if len(G) > 0:
                            misc.verbose("Direct search succeeded.")
                            G, _, reg = self.saturation(G, verbose=verbose)
                            self.__gens = G
                            self.__regulator = reg
                            return self.__gens
                        h += 2
                    misc.verbose("Direct search FAILED.")
            except RuntimeError:
                pass
            misc.verbose("Calling mwrank.")
            C = self.mwrank_curve(verbose)
            if not (verbose is None):
                C.set_verbose(verbose)
            G = C.gens()
            if not C.certain():
                raise RuntimeError, "Unable to compute the rank, hence generators, with certainty (lower bound=%s).  This could be because Sha[2] is nontrivial."%C.rank()
            self.__gens = [Point(self, x, check=False) for x in G]
        return self.__gens

    def ngens(self):
        return len(self.gens())

    def regulator(self, use_database=False, verbose=None):
        """
        Returns the regulator of this curve, which must be defined
        over Q.

        INPUT:
            use_database -- bool (default: False), if True, try to
                  look up the regulator in the Cremona database.
            verbose -- (default: None), if specified changes the
                  verbosity of mwrank computations.
        """
        try:
            return self.__regulator
        except AttributeError:
            if use_database:
                try:
                    return self.database_curve().regulator
                except:
                    pass
            G = self.gens()
            try:  # in some cases self.gens() efficiently computes regulator.
                return self.__regulator
            except AttributeError:
                pass
            
            C = self.mwrank_curve()
            R = C.regulator()
            if not C.certain():
                raise RuntimeError, "Unable to compute the rank, hence regulator, with certainty (lower bound=%s)."%C.rank()
            self.__regulator = R
        return self.__regulator

    def saturation(self, points, verbose=False, max_prime=0, odd_primes_only=False):
        """
        Given a list of rational points on E, compute the saturation
        in E(Q) of the subgroup they generate.

        INPUT:
            points (list) -- list of points on E
            verbose  (bool) -- (default: False), if True, give verbose output
            max_prime (int) -- (default: 0), saturation is performed
                               for all primes up to max_prime.  If max_prime==0,
                               perform saturation at *all* primes, i.e., compute
                               the true saturation.
            odd_primes_only (bool) -- only do saturation at odd primes

        OUTPUT:
            saturation (list) -- points that form a basis for the saturation
            index (int) -- the index of the group generated by points in their saturation
            regulator (float) -- regulator of saturated points.

        IMPLEMENTATION: Uses Cremona's mwrank package.  With max_prime=0, we call
            mwrank with successively larger prime bounds until the full saturation is
            provably found.  The results of saturation at the previous primes is stored
            in each case, so this should be reasonably fast.
        """
        if not MWRANK:
            raise ImportError, "the mwrank package must be installed for requested functionality."

        if not isinstance(points, list):
            raise TypeError, "points (=%s) must be a list."%points

        v = []
        for P in points:
            if not isinstance(P, Point):
                P = self(P)
            elif P.curve() != self:
                raise ArithmeticError, "point (=%s) must be %s."%(P,self)
            x, y = P.xy()
            d = x.denominator().lcm(y.denominator())
            v.append((x*d, y*d, d))
        c = self.mwrank_curve()
        mw = mwrank.MordellWeil(c, verbose)
        mw.process(v)
        if max_prime == 0:
            repeat_until_saturated = True
            max_prime = 97
        while True:
            ok, index, sat, reg = mw.saturate(max_prime=max_prime, odd_primes_only = odd_primes_only)
            if not ok and repeat_until_saturated:
                max_prime = arith.next_prime(max_prime + 100)
                ok, index, sat, reg = mw.saturate(max_prime=max_prime, odd_primes_only = odd_primes_only)
            else:
                break

        sat = [self(P) for P in sat]
        return sat, index, reg

    def CPS_height_bound(self):
        """
        Return the Cremona-Prickett-Siksek height bound.  This is a
        floating point number B such that if P is a point on the curve,
        then the naive logarithmetic height of P is off from the
        canonical height by at most B.

        WARNING: We assume the model is minimal!
        """
        return self.mwrank_curve().CPS_bound()

    def silverman_height_bound(self):
        """
        Return the Silverman height bound.  This is a floating point
        number B such that if P is a point on the curve, then the
        naive logarithmetic height of P is off from the canonical
        height by at most B.

        Note that the CPS_height_bound is typically much better than
        the Silverman bound.
        """
        return self.mwrank_curve().silverman_bound()
        

    def point_search(self, height_limit, verbose=True):
        """
        Search for points on a curve up to an input bound on the naive logarithmic height.

        INPUT:
            height_limit (float) -- bound on naive height (at most 21, or mwrank overflows)
            verbose  (bool) -- (default: True)

        OUTPUT:
            points (list) -- list of points found

        IMPLEMENTATION: Uses Cremona's mwrank package. 
        """
        if not MWRANK:
            raise ImportError, "the mwrank package must be installed for requested functionality."
        height_limit = float(height_limit)
        if height_limit > _MAX_HEIGHT:
            raise OverflowError, "height_limit (=%s) must be at most %s."%(height_limit,_MAX_HEIGHT)
        c = self.mwrank_curve()
        mw = mwrank.MordellWeil(c, verbose)
        v = mw.search(height_limit, verbose=verbose)
        return [self(P) for P in v]

    def two_torsion_rank(self):
        A = self.torsion_subgroup().invariants()
        if len(A) == 2:
            return 2
        elif len(A) == 1 and A[0] % 2 == 0:
            return 1
        else:
            return 0

    def selmer_rank_bound(self):
        """
        Bound on the rank of the curve, computed using the
        2-selmer group.  This is the rank of the curve
        minus the rank of the 2-torsion, minus a number
        determined by whatever mwrank was able to determine
        related to Sha[2].  Thus in many cases, this is
        the actual rank of the curve.

        EXAMPLE:
        The following is the curve 960D1, which has rank 0,
        but Sha of order 4.
            >>> E = EllipticCurve([0, -1, 0, -900, -10098])
            >>> E.selmer_rank_bound()
            0
            
        It gives 0 instead of 2, because it knows Sha is nontrivial.
        In contrast, for the curve 571A, also with rank 0 and Sha
        of order 4, we get a worse bound:
            >>> E = EllipticCurve([0, -1, 1, -929, -10595])
            >>> E.selmer_rank_bound()
            2
            >>> E.rank()
            0
        """
        try:
            return self.__selmer_rank_bound
        except AttributeError:
            C = self.mwrank_curve()
            self.__selmer_rank_bound = C.selmer_rank_bound()
            return self.__selmer_rank_bound
    
            
    def an(self, n):
        """
        The n-th Fourier coefficient of the modular form corresponding
        to this elliptic curve, where n is a positive integer.
        """
        return int(self.pari_mincurve().ellak(n))

    def ap(self, p):
        """
        The p-th Fourier coefficient of the modular form corresponding
        to this elliptic curve, where p is prime.
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        return int(self.pari_mincurve().ellap(p))

    def quadratic_twist(self, D):
        return EllipticCurve_generic.quadratic_twist(self, D).minimal_model()
    
    def minimal_model(self):
        try:
            return self.__minimal_model
        except AttributeError:
            F = self.pari_mincurve()
            self.__minimal_model = EllipticCurve_RationalField([Q(F[i]) for i in range(5)])
            return self.__minimal_model

    def kodaira_type(self, p):
        """
        Local Kodaira type of the elliptic curve at p.

        1 means good reduction (type I_0), 2, 3 and 4 mean types II, III and IV, respectively,
        4 + nu with nu > 0 means type I_{nu}; finally the opposite values -1, -2, etc. refer
        to the starred types I_0^*, II^*, etc.  
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        try:
            self.__kodaira_type
        except AttributeError:
            self.__kodaira_type = {}
            self.__tamagawa_number = {}
        if not self.__kodaira_type.has_key(p):
            v = self.pari_mincurve().elllocalred(p)
            self.__kodaira_type[p] = str(v[1])
            self.__tamagawa_number[p] = int(v[3])
        return self.__kodaira_type[p]

    def tamagawa_number(self, p):
        """
        The Tamagawa number of the elliptic curve at p.
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        try:
            return self.__tamagawa_number[p]
        except (AttributeError, KeyError):
            self.kodaira_type(p)
            return self.__tamagawa_number[p]

    def tamagawa_product(self):
        """
        Returns the product of the Tamagawa numbers.
        """
        return self.pari_mincurve().ellglobalred()[2].python()

    def real_components(self):
        """
        Returns 1 if there is 1 real component and 2 if there are 2.
        """
        invs = self.weierstrass_model().ainvs()
        x = rings.polygen(self.base_ring())
        f = x**3 + invs[3]*x + invs[4]
        if f.discriminant() > 0:
            return 2
        else:
            return 1

    def period_lattice(self):
        """
        Return a basis for the period lattice of the elliptic curve
        over Q as a 2-tuple.

        The basis has the form [omega_1, omega_2], where
        Im(omega_1/omega_2) > 0 and omega_1 is real.
        
        INPUT:
            -- an elliptic curve
        OUTPUT:
            omega_1 -- complex number
            omega_2 -- complex number
        """
        return tuple(self.pari_curve().omega().python())

    def omega(self):
        """
        Returns the real period.  This is the correct period in the BSD
        conjecture, i.e., it is the least real period * 2 when the period
        lattice is rectangular. 
        """
        return self.period_lattice()[0] * self.real_components()

    def complex_area(self):
        """
        Return the area of a fundamental domain for the period lattice
        of the elliptic curve.
        """
        w1,w2 = self.period_lattice()
        return (w1*w2.imag()).real()

    def Lseries_at1(self, k=0):
        """
        Compute L(E,1) using k terms of the series for L(E,1) form
        page 406 of Henri Cohen's book "A Course in Computational
        Algebraic Number Theory".  If the k argument is not specified,
        then it defaults to sqrt(N), where N is the conductor.

        The real precision of the computation is the precision of Python
        floats.

        INPUT:
            k -- (optional) an integer, defaults to sqrt(N).
            
        OUTPUT:
            float -- L(E,1)
            float -- a bound on the error in the approximation; this
                     is a proveably correct upper bound on the sum
                     of the tail end of the series used to compute L(E,1).

        EXAMPLES:
        

        ALGORITHM:
            1. Compute the root number eps.  If it is -1, return 0.
            
            2. Compute the Fourier coefficients a_n, for n up to and
               including k.
               
            3. Compute the sum
            
                 2 * sum_{n=1}^{k} (a_n / n) * exp(-2*pi*n/Sqrt(N)),
                 
               where N is the conductor of E.
               
            4. Compute a bound on the tail end of the series, which is
            
                 2 * e^(-2 * pi * (k+1) / sqrt(N)) / (1 - e^(-2*pi/sqrt(N))).
                 
               For a proof see [Grigov-Jorza-Patrascu-Patrikis-Stein].
        """
        if self.root_number() == -1:
            return 0
        sqrtN = sqrt(self.conductor())  
        k = int(k)
        if k == 0: k = int(math.ceil(sqrtN))
        an = self.anlist(k)           # list of C ints
        # Compute z = e^(-2pi/sqrt(N))
        pi = 3.14159265358979323846
        z = exp(-2*pi/sqrtN)
        zpow = z
        s = 0.0
        for n in xrange(1,k+1):
            s += (zpow * an[n])/n
            zpow *= z

        error = 2*zpow / (1 - z)
        
        return 2*s, error

    def Lseries_deriv_at1(self, k=0):
        """
        Compute L'(E,1) using k terms of the series for L'(E,1) form
        page 406 of Henri Cohen's book"A Course in Computational
        Algebraic Number Theory".

        The real precision of the computation is the precision of
        Python floats.

        INPUT:

        OUTPUT:

        ALGORITHM:
            1. Compute the root number eps.  If it is 1, return 0.

            2. Compute the Fourier coefficients a_n, for n up to and
               including k.
               
            3. Compute the sum
            
                 2 * sum_{n=1}^{k} (a_n / n) * E_1(2*pi*n/Sqrt(N)),
                 
               where N is the conductor of E, and E_1 is the
               exponential integral function.

            4. Compute a bound on the tail end of the series, which is
            
                 2 * e^(-2 * pi * (k+1) / sqrt(N)) / (1 - e^(-2*pi/sqrt(N))).
                 
               For a proof see [Grigov-Jorza-Patrascu-Patrikis-Stein].  This
               is exactly the same as the bound for the approximation to
               L(E,1) produced by Lseries_at1.
        """
        if self.root_number() == 1: return 0
        k = int(k)
        sqrtN = sqrt(self.conductor())  
        if k == 0: k = int(math.ceil(sqrtN))
        an = self.anlist(k)           # list of C ints
        # Compute z = e^(-2pi/sqrt(N))
        pi = 3.14159265358979323846
        v = transcendental.exponential_integral_1(2*pi/sqrtN, k)
        L = 2*float(sum([ (v[n-1] * an[n])/n for n in xrange(1,k+1)]))
        error = 2*exp(-2*pi*(k+1)/sqrtN)/(1-exp(-2*pi/sqrtN))
        return L, error

    def Lseries(self, s):
        """
        Returns the value of the L-series of the elliptic curve E at s, where s
        must be a real number.  

        NOTES:
        (1) If the conductor of the curve is large, say > 10^12, then this function
        will take a long time, since it uses an O(sqrt(N)) algorithm.
        
        (2) There is an algorithm to compute L(E,s) for any complex s, but it involves
        computing the incomplete Gamma function, which is not implemented in SAGE.

        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,5])
            
            >> E.Lseries(1)
            0.E-19

            The following may have low order bits that are different
            on your machine.
            >> E.Lseries(1.1)    
            0.28549100767814812844289234320
        """
        return self.pari_mincurve().elllseries(s).python()

    def Lambda(self, s, prec):
        """
        Returns the value of the Lambda-series of the elliptic curve E
        at s can be any complex number.  Use Lambda(s) if s is real,
        since it will be much faster.

        WARNING: This function uses sage.functions.transcendental.gamma_inc,
        which currently calls mathematica.

        EXAMPLES:
        """
        s = C(s)
        N = self.conductor()
        pi = transcendental.pi()
        Gamma = transcendental.gamma
        Gamma_inc = transcendental.gamma_inc
        a = self.anlist(prec)
        eps = self.root_number()
        sqrtN = arith.sqrt(N)
        def F(n, t):
            return Gamma_inc(t+1, 2*pi*n/sqrtN) * C(sqrtN/(2*pi*n))**(t+1)
        return sum([a[n]*(F(n,s-1) + eps*F(n,1-s)) for n in xrange(1,prec+1)])

    def Lseries_extended(self, s, prec):
        """
        Returns the value of the L-series of the elliptic curve E
        at s can be any complex number using prec terms of the
        power series expansion.

        WARNING: This function uses sage.functions.transcendental.gamma_inc,
        which currently calls mathematica.

        EXAMPLES:

        """
        try:
            s = C(s)
        except TypeError:
            raise TypeError, "Input argument %s must be coercible to a complex number"%s
        prec = int(prec)
        if abs(s.imag()) < R(0.0000000000001):
            return self.Lseries(s.real())
        N = self.conductor()
        pi = transcendental.pi()
        Gamma = transcendental.gamma
        Gamma_inc = transcendental.gamma_inc
        a = self.anlist(prec)
        eps = self.root_number()
        sqrtN = arith.sqrt(N)
        def F(n, t):
            return Gamma_inc(t+1, 2*pi*n/sqrtN) * C(sqrtN/(2*pi*n))**(t+1)
        return C(N)**(-s/2) * C(2*pi)**s * Gamma(s)**(-1)\
               * sum([a[n]*(F(n,s-1) + eps*F(n,1-s)) for n in xrange(1,prec+1)])
               
    def sigma(self, z, flag=0):
        """
        Returns the value of the Weierstrass sigma function of the lattice
        associated to this elliptic curve E.
        INPUT:
            z -- a complex number
            flag -- 0 - default ???
                    1 - computes an arbitrary determination of log(sigma(z))
                    2, 3 - same using the product expansion instead of theta series.
                           ???
        OUTPUT:
            a complex number

        NOTE: The reason for the ???'s above, is that the PARI documentation for
              ellsigma is very vague.
        """
        return self.pari_curve().ellsigma(z, flag)

    def modular_degree(self):
        """
        Return the modular degree of this elliptic curve.

        NOTE: This function only currently works if this curve is in
        the elliptic curves tables, its a J_0(N)-optimal quotient, and
        you've installed those tables.  In particular, this is
        (currently) not an implementation of Watkins's algorithm.
        """
        try:
            return self.__modular_degree
        except AttributeError:
            N = self.conductor()
            # Try to look up the curve in the database, if it's installed.
            try:
                E = self.database_curve()
            except RuntimeError:
                raise NotImplementedError, "General modular degree algorithm not implemented."
            if hasattr(E,'modular_degree') and E.modular_degree != None:
                self.__modular_degree = E.modular_degree
            else:
                raise NotImplementedError, "Modular degree computation not implemented except for curves in the database."
            return self.__modular_degree

    def modular_parametrization(self):
        """
        Computes and returns ...
        """
        return self.pari_mincurve().elltaniyama()

    def cremona_label(self, space=False):
        try:
            if not space:
                return self.__cremona_label.replace(' ','')
            return self.__cremona_label
        except AttributeError:
            X = self.database_curve()
            self.__cremona_label = "%s %s %s"%(X.conductor,X.iso_class,X.number)
            return self.cremona_label(space)

    def torsion_order(self):
        """
        Return the order of the torsion subgroup.
        """
        try:
            return self.__torsion_order
        except AttributeError:
            self.__torsion_order = self.torsion_subgroup().order()
            return self.__torsion_order

    def torsion_subgroup(self, flag=0):
        """
        Returns the torsion subgroup of this elliptic curve.

        The flag is passed onto PARI and has the same algorithm meaning as there.
        So: If flag = 0, use Doud's algorithm; if flag = 1, use Lutz-Nagell.
        """
        try:
            return self.__torsion
        except AttributeError:
            try:
                G = self.pari_curve().elltors(flag)
            except RuntimeError:
                self.__pari_double_prec()
                return self.torsion_subgroup(flag)
            self.__torsion = abelian.AbelianGroup(G[1].python())
        return self.__torsion

    def newform(self):
        """
        Returns the newform attached to this elliptic curve.
        """
        def comp(n):  # compute series expansion of newform attached E as power series
            v = self.anlist(n)
            return rings.PowerSeries(rings.PowerSeriesRing(rings.Q), v, n)
        M = modform.ModularForms(self.conductor(), 2)
        return modform.Newform(M, comp)

    ## def newform_eval(self, z, prec):
##         """
##         The value of the newform attached to this elliptic curve at
##         the point z in the complex upper half plane, computed using
##         prec terms of the power series expansion.  Note that the power
##         series need not converge well near the real axis.
##         """
##         raise NotImplementedError

    def root_number(self):
        """
        Returns the root number of this elliptic curve.

        This is 1 if the order of vanishing of the L-function L(E,s)
        at 1 is even, and -1 if it is odd.
        """

        try:
            return self.__root_number
        except AttributeError:
            self.__root_number = int(self.pari_mincurve().ellrootno())
        return self.__root_number


    def formal_log(self, prec=20):
        raise NotImplementedError

    def has_cm(self):
        return self.j_invariant() in [0, 54000, -12288000, 1728, \
                                      287496, -3375, 16581375, 8000, \
                                      -32768,  -884736, -884736000,\
                                      -147197952000, -262537412640768000]

    ##########################################################
    # Isogeny class (currently just uses Cremona database.)
    ##########################################################
    def isogeny_class(self):
        """
        Return all curves in the isogeny class of this elliptic curve.

        NOTE: Currently implemented only using the database.
        """
        db = sage.tables.elliptic_curves.Database()
        return [EllipticCurve(X) for X in db.isogeny_class(self.cremona_label(space=False))]

    ##########################################################
    # Galois Representations
    ##########################################################
    def is_surjective(self, p, A=1000):
        """
        Return True if the mod-p representation attached to E
        is surjective, False if it is not, or None if we were
        unable to determine whether it is or not.  

        INPUT:
            p -- int (a prime number)
            A -- int (a bound on the number of a_p to use)

        OUTPUT:
            a 2-tuple:
            -- surjective or (probably) not
            -- information about what it is if not surjective

        EXAMPLES:

        REMARKS:
        
            1.  If p >= 5 then the mod-p representation is surjective
                if and only if the p-adic representation is
                surjective.  When p = 2, 3 there are counterexamples.
                See a very recent paper of Elkies for more details
                when p=3.

            2.  When p <= 3 this function always gives the correct
                result irregardless of A, since it explicitly
                determines the p-division polynomial.
            
        """
        if not arith.is_prime(p):
            raise TypeError, "p (=%s) must be prime."%p

        T = self.torsion_subgroup().order()
        if T % p == 0:
            return False, "%s-torsion"%p

        if p == 2:
            invs = self.weierstrass_model().ainvs()
            R = rings.PolynomialRing(self.base_ring())
            x = R.gen()
            f = x**3 + invs[3]*x + invs[4]
            if not f.is_irreducible():
                return False, '2-torsion'
            if arith.is_square(f.discriminant()):
                return False, "A3"
            return True, None
        
        if p == 3:
            # Algorithm: Let f be the 3-division polynomial, which is
            # a polynomial of degree 4.  Then I claim that this
            # polynomial has Galois group S_4 if and only if the
            # representation rhobar_{E,3} is surjective.  If the group
            # is S_4, then S_4 is a quotient of the image of
            # rhobar_{E,3}.  Since S_4 has order 24 and GL_2(F_3)
            # has order 48, the only possibility we have to consider
            # is that the image of rhobar is isomorphic to S_4.
            # But this is not the case because S_4 is not a subgroup
            # of GL_2(F_3).    If it were, it would be normal, since
            # it would have index 2.  But there is a *unique* normal
            # subgroup of GL_2(F_3) of index 2, namely SL_2(F_3),
            # and SL_2(F_3) is not isomorphic to S_4 (S_4 has a normal
            # subgroup of index 2 and SL_2(F_3) does not.)
            # (What's a simple way to see that SL_2(F_3) is the
            # unique index-2 normal subgroup?  I didn't see an obvious
            # reason, so just used the NormalSubgroups command in MAGMA
            # and it output exactly one of index 2.)

            # Here's Noam Elkies proof for the other direction:
            
            #> Let E be an elliptic curve over Q.  Is the mod-3
            #> representation E[3]  surjective if and only if the
            #> (degree 4) division polynomial has Galois group S_4?  I
            #> can see why the group being S_4 implies the 
            #> representation is surjective, but the converse is not
            #> clear to me.
            # I would have thought that this is the easier part: to
            # say that E[3] is surjective is to say the 3-torsion
            # field Q(E[3]) has Galois group GL_2(Z/3) over Q.  Let
            # E[3]+ be the subfield fixed by the element -1 of
            # GL_2(Z/3).  Then E[3] has Galois group PGL_2(Z/3), which
            # is identified with S_4 by its action on the four
            # 3-element subgroups of E[3].  Each such subgroup is in
            # turn determined by the x-coordinate shared by its two
            # nonzero points.  So, if E[3] is surjective then any
            # permutation of those x-coordinates is realized by some
            # element of Gal(E[3]+/Q).  Thus the Galois group of the
            # division polynomial (whose roots are those
            # x-coordinates) maps surjectively to S_4, which means it
            # equals S_4.
            
            
            f = self.division_polynomial(3)
            if not f.is_irreducible():
                return False, "reducible_3-divpoly"
            n = pari(f).polgalois()[0]
            if n == 24:
                return True, None
            else:
                return False, "3-divpoly_galgroup_order_%s"%n
        
        if self.has_cm():
            return False, "CM"
        an = self.anlist(A)
        ell = 0
        Np = self.conductor() * p
        signs = []
        while True:
            ell = arith.next_prime(ell)
            if ell >= A: break
            if Np % ell != 0:
                a_ell = an[ell]
                if a_ell % p != 0:
                    s = arith.kronecker(a_ell**2 - 4*ell, p)
                    #print ell, s
                    if s == 0: continue
                    if not (s in signs):
                        signs.append(s)
                        if len(signs) == 2:
                            return True, None
            
        # could do something further here...
        return False, signs

    def is_semistable(self):
        if self.base_ring() != Q:
            raise NotImplementedError, "is_semistable only implemented for curves over the rational numbers."
        return self.conductor().is_square_free()

    def non_surjective(self, A=1000):
        """
        Returns a list of primes p such that the mod-p representation
        rho_{E,p} *might* not be surjective (this list usually
        contains 2, because of shortcomings of the algorithm).  If p
        is not in the returned list, then rho_{E,p} is provably
        surjective (see A. Cojocaru's paper).  If the curve has CM
        then infinitely many representations are not surjective, so we
        simply return the sequence [(0,"cm")] and do no further computation.

        INPUT:
            A -- an integer
        OUTPUT:
            list -- if curve has CM, returns [(0,"cm")].  Otherwise, returns a
                    list of primes where mod-p representation very likely
                    not surjective.   At any prime not in this list,
                    the representation is definitely surjective.
        EXAMPLES:
            >>> E = EllipticCurve([0, 0, 1, -38, 90])  # 361A
            >>> E.non_surjective()   # CM curve
            [(0, 'cm')]

            >>> E = EllipticCurve([0, -1, 1, 0, 0]) # X_1(11)
            >>> E.non_surjective()
            [(5, '5-torsion')]

            >>> E = EllipticCurve([0, 0, 1, -1, 0]) # 37A
            >>> E.non_surjective()
            []

            >>> E = EllipticCurve([0,-1,1,-2,-1])   # 141C
            >>> E.non_surjective()
            [(13, [1])]

        ALGORITHM:
            When p<=3 use division polynomials.  For 5 <= p <= B,
            where B is Cojocaru's bound, use the results in Section 2
            of Serre's inventiones paper"Sur Les Representations Modulaires Deg
            Degre 2 de Galqbar Over Q."
        """
        if self.has_cm():
            misc.verbose("cm curve")
            return [(0,"cm")]
        N = self.conductor()
        if self.is_semistable():
            C = 11
            misc.verbose("semistable -- so bound is 11")
        else:
            C = 1 + 4*sqrt(6)*int(N)/3 * sqrt(mul([1+1.0/int(p) for p,_ in factor(N)]))
            misc.verbose("conductor = %s, and bound is %s"%(N,C))
        C = 1 + 4*sqrt(6)*int(N)/3 * sqrt(mul([1+1.0/int(p) for p,_ in factor(N)]))
        misc.verbose("conductor = %s, and bound is %s"%(N,C))
        B = []
        p = 2
        while p <= C:
            t, v = self.is_surjective(p, A=A)
            #misc.verbose("(%s,%s,%s)"%(p,t,v))
            if not t:
                B.append((p,v))
            p = next_prime(p)
        return B

    def eval_modular_form(self, points, prec):
        if not isinstance(points, (list,xrange)):
            try:
                points = list(points)
            except TypeError:
                return self.eval_modular_form([points],prec)
        an = self.pari_mincurve().ellan(prec)
        s = 0
        I = pari("I")
        pi = pari("Pi")
        c = pari(2)*pi*I
        ans = []
        for z in points:
            s = pari(0)
            r0 = (c*z).exp()
            r = r0
            for n in xrange(1,prec):
                s += an[n-1]*r      
                r *= r0
            ans.append(s.python())
        return ans

    ########################################################################
    # Functions related to the BSD conjecture.
    ########################################################################
    def sha_an(self, use_database=False):
        """
        Returns the Birch and Swinnerton-Dyer conjectural order of
        Sha, unless the analytic rank is > 1, in which case this
        function returns 0.

        This result is proved correct if the order of vanishing is 0
        and the Manin constant is <= 2 (this Manin constant condition
        is true if the conductor is squarefree or <= 40000).

        If the optional parameter use_database is True (default:
        False), this function returns the analytic order of Sha as
        listed in Cremona's tables, if this curve appears in Cremona's
        tables (i.e., is a curve over Q of conductor <= 40000).

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])   # 11A  = X_0(11)
            >>> E.sha_an()
            1
            >>> E = EllipticCurve([0, -1, 1, 0, 0])       # X_1(11)
            >>> E.sha_an()
            1

        The smallest conductor curve with nontrivial Sha:
            >>> E = EllipticCurve([1,1,1,-352,-2689])     # 66B3
            >>> E.sha_an()
            4

        The four optimal quotients with nontrivial Sha and conductor <= 1000:
            >>> E = EllipticCurve([0, -1, 1, -929, -10595])       # 571A
            >>> E.sha_an()
            4
            >>> E = EllipticCurve([1, 1, 0, -1154, -15345])       # 681B
            >>> E.sha_an()
            9
            >>> E = EllipticCurve([0, -1, 0, -900, -10098])       # 960D
            >>> E.sha_an()
            4
            >>> E = EllipticCurve([0, 1, 0, -20, -42])            # 960N
            >>> E.sha_an()
            4

        The smallest conductor curve of rank > 1:
            >>> E = EllipticCurve([0, 1, 1, -2, 0])       # 389A (rank 2)
            >>> E.sha_an()
            0

        The following two examples require the mwrank package, for
        computation of a regulator.
            >> e = EllipticCurve([1, 0, 0, -19491080, -33122512122])   # 15834T2
            >> e.sha_an()   # takes about 22 seconds
            25
            >> E = EllipticCurve([0, 0, 1, -1, 0])                     # 37A  (rank 1)
            >> E.sha_an()
            1
        """
        try:
            return self.__sha_an
        except:
            pass
        if use_database:
            try:
                self.__sha_an = int(round(self.database_curve().sha_an))
                return self.__sha_an
            except:
                pass            
        eps = self.root_number()
        if eps == 1:
            L1_over_omega = self.L_ratio()
            if L1_over_omega == 0:
                return 0
            T = self.torsion_subgroup().order()
            Sha = (L1_over_omega * T * T) / Q(self.tamagawa_product())
            try:
                Sha = Z(Sha)
            except ValueError:
                raise RuntimeError, \
                      "There is a bug in sha_an, since the computed conjectural order of Sha is %s, which is not an integer."%Sha
            if not arith.is_square(Sha):
                raise RuntimeError, \
                      "There is a bug in sha_an, since the computed conjectural order of Sha is %s, which is not a square."%Sha
            self.__sha_an = Sha
            return Sha
        
        else:  # rank > 0  (Not provably correct)
            L1, error_bound = self.Lseries_deriv_at1(10*sqrt(self.conductor()) + 10)
            if abs(L1) < error_bound:
                self.__sha_an = 0
                return 0   # vanishes to order > 1, to computed precision
            regulator = self.regulator()   # this could take a *long* time; and could fail...?
            T = self.torsion_subgroup().order()
            omega = self.omega()
            Sha = int(round ( (L1 * T * T) / (self.tamagawa_product() * regulator * omega) ))
            try:
                Sha = Z(Sha)
            except ValueError:
                raise RuntimeError, \
                      "There is a bug in sha_an, since the computed conjectural order of Sha is %s, which is not an integer."%Sha
            if not arith.is_square(Sha):
                raise RuntimeError, \
                      "There is a bug in sha_an, since the computed conjectural order of Sha is %s, which is not a square."%Sha
            self.__sha_an = Sha
            return Sha
            
        
    def L_ratio(self):
        """
        Returns the ratio L(E,1)/Omega as an exact rational
        number. The result is *provably* correct if the Manin constant
        of the associated optimal quotient is <= 2.  This hypothesis
        on the Manin constant is true for all curves of conductor <=
        40000 (by Cremona) and all semistable curves (i.e., squarefree
        conductor).

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])   # 11A  = X_0(11)
            >>> E.L_ratio()
            1/5
            >>> E = EllipticCurve([0, -1, 1, 0, 0])       # X_1(11)
            >>> E.L_ratio()
            1/25
            >>> E = EllipticCurve([0, 0, 1, -1, 0])       # 37A  (rank 1)
            >>> E.L_ratio()
            0
            >>> E = EllipticCurve([0, 1, 1, -2, 0])       # 389A (rank 2)
            >>> E.L_ratio()
            0
            >>> E = EllipticCurve([0, 0, 1, -38, 90])     # 361A (CM curve))
            >>> E.L_ratio()
            0
            >>> E = EllipticCurve([0,-1,1,-2,-1])         # 141C (13-isogeny)
            >>> E.L_ratio()
            1

        WARNING: It's conceivable that machine floats are not large
        enough precision for the computation; if this could be the
        case a RuntimeError is raised.  The curve's real period would
        have to be very small for this to occur.  This does not occur
        for any optimal curve of conductor <= 40000.

        ALGORITHM: Compute the root number.  If it is -1 then L(E,s)
        vanishes to odd order at 1, hence vanishes.  If it is +1, use
        a result about modular symbols and Mazur's "Rational Isogenies"
        paper to determine a provably correct bound (assuming Manin
        constant is <= 2) so that we can determine whether L(E,1) = 0.

        AUTHOR: William Stein, 2005-04-20.
        """
        
        if self.root_number() == -1:
            return 0

        # Even root number.  Decide if L(E,1) = 0.  If E is a modular
        # optimal quotient of J_0(N) elliptic curve, we know that T *
        # L(E,1)/omega is an integer n, where T is the order of the
        # image of the rational torsion point (0)-(oo) in E(Q), and
        # omega is the least real Neron period.  (This is proved in my
        # Ph.D. thesis, but is probably well known.)  We can easily
        # compute omega to very high precision using AGM.  So to prove
        # that L(E,1) = 0 we compute T/omega * L(E,1) to sufficient
        # precision to determine it as an integer.  If eps is the
        # error in computation of L(E,1), then the error in computing
        # the product is (2T/Omega_E) * eps, and we need this to be
        # less than 0.5, i.e.,
        #          (2T/Omega_E) * eps < 0.5,
        # so
        #          eps < 0.5 * Omega_E / (2T) = Omega_E / (4*T).
        #
        # Since in general E need not be optimal, we have to choose
        # eps = Omega_E/(8*t*B), where t is the exponent of E(Q)_tor,
        # and B is a bound on the degree of any isogeny.   A liberal
        # bound on the degrees of cyclic N-isogenies is 200, by Mazur's
        # "Rational Isogenies of Prime Degree" paper, so we take B=200.
        #
        # NOTES: We *do* have to worry about the Manin constant, since
        # we are using the Neron model to compute omega, not the
        # newform.  My theorem replaces the omega above by omega/c,
        # where c is the Manin constant, and the bound must be
        # correspondingly smaller.  If the level is square free, then
        # the Manin constant is 1 or 2, so there's no problem (since
        # we took 8 instead of 4 in the denominator).  If the level
        # is divisible by a square, then the Manin constant could
        # be a divisible by an arbitrary power of that prime, except
        # that Edixhoven claims the primes that appear are <= 7.
        
        t = self.torsion_subgroup().exponent()
        omega = self.period_lattice()[0]
        C = 8*200*t
        eps = omega / C
        if eps < 10**(-15):  # liberal bound on precision of float
            raise RuntimeError, "Insufficient machine precision for computation."
        sqrtN = 2*int(sqrt(self.conductor()))
        k = sqrtN + 10
        while True:
            L1, error_bound = self.Lseries_at1(k)
            if error_bound < eps:
                n = int(round(L1*C/omega))
                quo = Q(n) / Q(C)
                return quo / self.real_components()
            k += sqrtN
            misc.verbose("Increasing precision to %s terms."%k)


    def L1_vanishes(self):
        """
        Returns whether or not L(E,1) = 0. The result is provably
        correct if the Manin constant of the associated optimal
        quotient is <= 2.  This hypothesis on the Manin constant
        is true for all curves of conductor <= 40000 (by Cremona) and
        all semistable curves (i.e., squarefree conductor).

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])   # 11A  = X_0(11)
            >>> E.L1_vanishes()
            False
            >>> E = EllipticCurve([0, -1, 1, 0, 0])       # X_1(11)
            >>> E.L1_vanishes()
            False
            >>> E = EllipticCurve([0, 0, 1, -1, 0])       # 37A  (rank 1)
            >>> E.L1_vanishes()
            True
            >>> E = EllipticCurve([0, 1, 1, -2, 0])       # 389A (rank 2)
            >>> E.L1_vanishes()
            True
            >>> E = EllipticCurve([0, 0, 1, -38, 90])     # 361A (CM curve))
            >>> E.L1_vanishes()
            True
            >>> E = EllipticCurve([0,-1,1,-2,-1])         # 141C (13-isogeny)
            >>> E.L1_vanishes()
            False

        WARNING: It's conceivable that machine floats are not large
        enough precision for the computation; if this could be the
        case a RuntimeError is raised.  The curve's real period would
        have to be very small for this to occur.  

        ALGORITHM: Compute the root number.  If it is -1 then L(E,s)
        vanishes to odd order at 1, hence vanishes.  If it is +1, use
        a result about modular symbols and Mazur's "Rational Isogenies"
        paper to determine a provably correct bound (assuming Manin
        constant is <= 2) so that we can determine whether L(E,1) = 0.

        AUTHOR: William Stein, 2005-04-20.
        """
        return self.L_ratio() == 0
        
    ########################################################################
    # Functions related to bounding the order of Sha (provably correctly!)
    # Heegner points and Kolyvagin's theorem
    ########################################################################
    def two_selmer_shabound(self):
        """
        Returns a bound on the dimension of Sha(E)[2], computed using
        a 2-descent.
        """
        S = self.two_selmer_rank()
        r = self.rank()
        t = self.two_torsion_rank()
        b = S - r - t
        if b % 2 != 0:
            raise ArithmeticError, "There is a bug in two_selmer_shabound since it's %s, but it must be even."%b
        return b
    
    def satisfies_heegner_hypothesis(self, D):
        """
        Returns True precisely when D is a fundamental discriminant
        that satisfies the Heegner hypothesis for this elliptic curve.
        """
        if not number_field.is_fundamental_discriminant(D):
            return False
        if arith.GCD(D, self.conductor()) != 1:
            return False
        K = number_field.QuadraticField(D)
        for p, _ in factor(self.conductor()):
            if len(K.factor_integer(p)) != 2:
                return False
        return True

    def heegner_discriminants(self, bound):
        return [-D for D in xrange(1,bound) if self.satisfies_heegner_hypothesis(-D)]

    def heegner_discriminants_list(self, n):
        """
        List of the first n Heegner discriminants for self.
        """
        v = []
        D = -5
        while len(v) < n:
            while not self.satisfies_heegner_hypothesis(D):
                D -= 1
            v.append(D)
            D -= 1
        return v

    def heegner_point_height(self, D, prec=2):
        """
        Use the Gross-Zagier formula to compute the Neron-Tate
        canonical height over K of the Heegner point corresponding to
        D, as an Interval (since it's computed to some precision using
        L-functions).

        INPUT:
            D (int) -- fundamental discriminant
            prec (int) -- (default: 2), use prec*sqrt(N) + 20 terms
                          of L-series in computations, where N is the
                          conductor.

        OUTPUT:
            Interval that contains the height of the Heegner point.
        """
        # Use interval arithmetic to guaranty correctness.  We assume
        # that alpha is computed to the precision of a float.
        IR = rings.IntervalRing()


        if not self.satisfies_heegner_hypothesis(D):
            raise ArithmeticError, "Discriminant (=%s) must be a fundamental discriminant that satisfies the Heegner hypothesis."%D
        if D == -3 or D == -4:
            raise ArithmeticError, "Discriminant (=%D) must not be -3 or -4."%D
        eps = self.root_number()
        L1_vanishes = self.L1_vanishes()
        if eps == 1 and L1_vanishes:
            return IR(0) # rank even hence >= 2, so Heegner point is torsion.
        alpha = sqrt(abs(D))/(2*self.complex_area())
        F = self.quadratic_twist(D)
        E = self
        k_E = prec*sqrt(E.conductor()) + 20
        k_F = prec*sqrt(F.conductor()) + 20

        MIN_ERR = 1e-10   # we assume that regulator and
                          # discriminant, etc., computed to this accuracy.
                          
        if eps == 1:   # E has even rank
            LF1, err_F = F.Lseries_deriv_at1(k_F)
            LE1, err_E = E.Lseries_at1(k_E)
            err_F = max(err_F, MIN_ERR)
            err_E = max(err_E, MIN_ERR)
            return IR(alpha-MIN_ERR,alpha+MIN_ERR) * IR(LE1-err_E,LE1+err_E) * IR(LF1-err_F,LF1+err_F)

        else:          # E has odd rank
            LE1, err_E = E.Lseries_deriv_at1(k_E)
            LF1, err_F = F.Lseries_at1(k_F)
            err_F = max(err_F, MIN_ERR)
            err_E = max(err_E, MIN_ERR)
            return IR(alpha-MIN_ERR,alpha+MIN_ERR) * IR(LE1-err_E,LE1+err_E) * IR(LF1-err_F,LF1+err_F)
        

    def heegner_index(self, D,  min_p=3, prec=5, verbose=True):
        """
        Assume self has rank 0.
        
        Return (an interval that contains) the square of the odd part
        of the index of the Heegner point in the group of K-rational
        points *modulo torsion* on the twist of the elliptic curve by D.

        If 0 is in the interval of the height of the Heegner point
        computed to the given prec, then this function returns 0.

        INPUT:
            D (int) -- Heegner discriminant
            min_p (int) -- (default: 3) only rule out primes >= min_p
                           dividing the index.  
            verbose (bool) -- (default: True)
            prec (int) -- (default: 5), use prec*sqrt(N) + 20 terms
                          of L-series in computations, where N is the
                          conductor.
        OUTPUT:
            Interval.
        """
        if self.root_number() != 1:
            raise RuntimeError, "The rank must be 0."

        # First compute upper bound on Height of Heegner point.
        # We divide by 2 to get the height over Q on the twist.
        ht = self.heegner_point_height(D, prec=prec)/2
        if 0 in ht:
            return 0
        F = self.quadratic_twist(D)
        h  = ht.upper()
        misc.verbose("Heegner height bound = %s"%h)
        B = F.CPS_height_bound()
        misc.verbose("CPS bound = %s"%B)
        c = h/(min_p**2) + B
        misc.verbose("Search would have to be up to height = %s"%c)
        IR = rings.IntervalRing()
        
        if c > _MAX_HEIGHT:
            misc.verbose("Doing direct computation of MW group.")
            reg = F.regulator(verbose=verbose)
            return ht/IR(reg)

        # Do naive search to eliminate possibility that Heegner point
        # is divisible by p<min_p, without finding Heegner point.
        P = F.point_search(c, verbose=verbose)
        P = [x for x in P if x.order() == 0]
        if len(P) == 0:
            return IR(1)
        S, I, reg = F.saturation(P, verbose=verbose)
        return ht/IR(reg)
        

    def heegner_index_bound(self, D=0,  prec=5, verbose=True, max_height=_MAX_HEIGHT):
        """
        Assume self has rank 0.
        
        Return a list v of primes such that if an odd prime p divides
        the index of the the Heegner point in the group of rational
        points *modulo torsion*, then p is in v.

        If 0 is in the interval of the height of the Heegner point
        computed to the given prec, then this function returns v = 0.
        This does not mean that the Heegner point is torsion, just
        that it is very likely torsion.

        If we obtain no information from a search up to max_height, e.g.,
        if the Siksek et al. bound is bigger than max_height, then
        we return v = -1.

        INPUT:
            D (int) -- (deault: 0) Heegner discriminant; if 0, use the
                       first discriminant < -4 that satisfies the Heegner hypothesis
            verbose (bool) -- (default: True)
            prec (int) -- (default: 5), use prec*sqrt(N) + 20 terms
                          of L-series in computations, where N is the
                          conductor.
            max_height (float) -- should be <= 21; bound on logarithmic naive height
                                  used in point searches.  Make smaller to make this
                                  function faster, at the expense of possibly obtaining
                                  a worse answer.  A good range is between 13 and 21.
                                  
        OUTPUT:
            v -- list or int (bad primes or 0 or -1)
            D -- the discriminant that was used (this is useful if D was
                 automatically selected).
        """
        max_height = min(float(max_height), _MAX_HEIGHT)
        if self.root_number() != 1:
            raise RuntimeError, "The rank must be 0."
        
        if D == 0:
            D = -5
            while not self.satisfies_heegner_hypothesis(D):
                D -= 1
        
        # First compute upper bound on Height of Heegner point.
        ht = self.heegner_point_height(D, prec=prec)
        if 0 in ht:
            return 0, D
        F = self.quadratic_twist(D)
        h  = ht.upper()
        misc.verbose("Heegner height bound = %s"%h)
        B = F.CPS_height_bound()
        misc.verbose("CPS bound = %s"%B)
        H = h
        p = 3
        while True:
            c = h/(2*p**2) + B
            if c < max_height:
                break
            if p > 100:
                break
            p = next_prime(p)
        misc.verbose("Using p = %s"%p)

        if c > max_height:
            misc.verbose("No information by searching only up to max_height (=%s)."%c)
            return -1, D
            
        misc.verbose("Searching up to height = %s"%c)
        IR = rings.IntervalRing()
        eps = 10e-5

        def _bound(P):
            """
            We will use this function below in two places.  It
            bounds the index using a nontrivial point.
            """
            assert len(P) == 1
            
            S, I, reg = F.saturation(P, verbose=verbose)
            h = IR(reg-eps,reg+eps)            
            ind2 = ht/(h/2)
            misc.verbose("index squared = %s"%ind2)
            ind = ind2.sqrt()
            misc.verbose("index = %s"%ind)
            # Compute upper bound on square root of index.
            if ind.length() < 1:
                t, i = ind.is_int()
                if t:   # unique integer in interval, so we've found exact index squared.
                    return arith.prime_divisors(i), D
            raise RuntimeError, "Unable to compute bound for e=%s, D=%s (try increasing precision)"%(self,D)
        
        # First try a quick search, in case we get lucky and find
        # a generator.
        P = F.point_search(13, verbose=verbose)
        P = [x for x in P if x.order() == 0]
        if len(P) > 0:
            return _bound(P)

        # Do search to eliminate possibility that Heegner point is
        # divisible by primes up to p, without finding Heegner point.
        P = F.point_search(c, verbose=verbose)
        P = [x for x in P if x.order() == 0]
        if len(P) == 0:
            # We've eliminated the possibility of a divisor up to p.
            return arith.prime_range(3,p), D
        else:
            return _bound(P)
        

    def shabound_kolyvagin(self, D=0, regulator=None,
                           ignore_nonsurj_hypothesis=False):
        """
        Given a fundamental discriminant D (=-3,-4) that satisfies the
        Heegner hypothesis, return a list of primes so that
        Kolyvagin's theorem (as in Gross's paper) implies that any
        prime divisor of #Sha is in this list.

        INPUT:
            D -- (optional) a fundamental discriminant < -4 that satisfies the
                 Heegner hypothesis for E; if not given, use the first such D
                 
            regulator -- (optional) regulator of E(K); if not given, will
                         be computed (which could take a long time)

                         
            ignore_nonsurj_hypothesis (optional: default False) --
                      If True, then gives the bound coming from Heegner point
                      index, but without any hypothesis on surjectivity
                      of the mod-p representation.
                      

        OUTPUT:
            bound and index

        More precisely:
        
                0 -- if E/K has complex multiplication or analytic rank >= 2
            or 
                B -- list of primes such that if p divides Sha(E/K), then p
                     is in B.

            and

                I -- the odd part of the index of the Heegner point in the full
                     group of K-rational points on E.  (If E has CM, returns 0.)

        REMARKS:
            (1) We do not have to assume that the Manin constant is 1
                (or a power of 2).  If the Manin constant were
                divisible by a prime, that prime would get included in
                the list of bad primes.

            (2) We assume the Gross-Zagier theorem is True under the
                hypothesis that gcd(N,D) = 1, instead of the stronger
                hypothesis gcd(2*N,D)=1 that is in the original
                Gross-Zagier paper.  That Gross-Zagier is true when
                gcd(N,D)=1 is"well-known" to the experts, but doesn't
                seem to written up well in the literature.

            (3) Correctness of the computation is guaranteed using
                interval arithmetic, under the assumption that the
                regulator, square root, and period lattice are
                computed to precision at least 10^(-10), i.e., they are
                correct up to addition or a real number with absolute
                value less than 10^(-10).
        """
        if self.has_cm():
            return 0, 0
        
        if D == 0:
            D = -5
            while not self.satisfies_heegner_hypothesis(D):
                D -= 1
                
        if not self.satisfies_heegner_hypothesis(D):
            raise ArithmeticError, "Discriminant (=%s) must be a fundamental discriminant that satisfies the Heegner hypothesis."%D
        if D == -3 or D == -4:
            raise ArithmeticError, "Discriminant (=%s) must not be -3 or -4."%D
        eps = self.root_number()
        L1_vanishes = self.L1_vanishes()
        if eps == 1 and L1_vanishes:
            return 0, 0        # rank even hence >= 2, so Kolyvagin gives nothing.
        alpha = sqrt(abs(D))/(2*self.complex_area())
        F = self.quadratic_twist(D)
        E = self
        k_E = 2*sqrt(E.conductor()) + 10
        k_F = 2*sqrt(F.conductor()) + 10
        #k_E = 2
        #k_F = 2

        # Use interval arithmetic to guaranty correctness.  We assume
        # that alpha is computed to the precision of a float.
        IR = rings.IntervalRing()

        MIN_ERR = 1e-10   # we assume that regulator and
                          # discriminant, etc., computed to this accuracy.
        tries = 0                         
        while True:
            tries += 1
            if tries >= 6:
                raise RuntimeError, "Too many precision increases in shabound_kolyvagin"
            if eps == 1:   # E has even rank
                misc.verbose("Conductor of twist = %s"%F.conductor())
                LF1, err_F = F.Lseries_deriv_at1(k_F)
                LE1, err_E = E.Lseries_at1(k_E)
                err_F = max(err_F, MIN_ERR)
                err_E = max(err_E, MIN_ERR)
                if regulator != None:
                    hZ = regulator/2
                else:
                    hZ = F.regulator(use_database=True)/2
                #print  alpha * LE1 * LF1 / hZ
                I = IR(alpha) * IR(LE1-err_E,LE1+err_E) * IR(LF1-err_F,LF1+err_F) / hZ
                #print I

            else:          # E has odd rank

                if regulator != None:
                    hZ = regulator/2
                else:
                    hZ = self.regulator(use_database=True)/2
                LE1, err_E = E.Lseries_deriv_at1(k_E)
                LF1, err_F = F.Lseries_at1(k_F)
                err_F = max(err_F, MIN_ERR)
                err_E = max(err_E, MIN_ERR)
                #I = alpha * LE1 * LF1 / hZ

                I = IR(alpha) * IR(LE1-err_E,LE1+err_E) * IR(LF1-err_F,LF1+err_F) / hZ

            misc.verbose('interval = %s'%I)
            t, n = I.is_int()
            if t:
                break
            elif I.length() < 1:
                raise RuntimeError, "Problem in shabound_kolyvagin; square of index is not an integer -- D=%s, I=%s."%(D,I)
            misc.verbose("Doubling bounds")
            k_E *= 2
            k_F *= 2
        # end while

        # We include 2 since Kolyvagin (in Gross) says nothing there
        if n == 0:  return 0, 0  # no bound
        F = factor(n)
        B = [2]
        for p, e in factor(n):
            if p > 2:
                if e%2 != 0:
                    raise RuntimeError, "Problem in shabound_kolyvagin; square of index is not a perfect square!  D=%s, I=%s, n=%s, e=%s."%(D,I,n,e)
                B.append(p)
            else:
                n /= 2**e  # replace n by its odd part
        if not ignore_nonsurj_hypothesis:
            for p, _ in self.non_surjective():   
                B.append(p)
        B = list(set([int(x) for x in B]))
        B.sort()
        return B, n
    

    def shabound_kato(self):
        """
        Returns a list p of primes such tha theorems of Kato's and
        others (e.g., as explained in a paper/thesis of Grigor Grigorov)
        imply that if p divides #Sha(E) then p is in the list.

        If L(E,1) = 0, then Kato's theorem gives no information, so
        this function returns False.

        THEOREM (Kato): Suppose p >= 5 is a prime so the p-adic
        representation rho_{E,p} is surjective.  Then ord_p(#Sha(E))
        divides ord_p(L(E,1)/Omega_E).

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])   # 11A  = X_0(11)
            >>> E.shabound_kato()
            [2, 3, 5]
            >>> E = EllipticCurve([0, -1, 1, 0, 0])       # X_1(11)
            >>> E.shabound_kato()
            [2, 3, 5]
            >>> E = EllipticCurve([1,1,1,-352,-2689])     # 66B3
            >>> E.shabound_kato()
            [2, 3]

        For the following curve one really has 25 | #Sha (by Grigorov-Stein paper):
            >> E = EllipticCurve([1, -1, 0, -332311, -73733731])   # 1058D1
            >> E.shabound_kato()
            [2, 3, 5]
            >> E.non_surjective()
            []

        For this one, Sha is divisible by 7.
            >> E = EllipticCurve([0, 0, 0, -4062871, -3152083138])   # 3364C1
            >> E.shabound_kato()
            [2, 3, 7]

        No information about curves of rank > 0:
            >>> E = EllipticCurve([0, 0, 1, -1, 0])       # 37A  (rank 1)
            >>> E.shabound_kato()
            False
        """
        if self.has_cm():
            return False
        if self.L1_vanishes():
            return False
        B = [2,3]
        for p, _ in self.non_surjective():   # for p >= 5, mod-p surj => p-adic surj
            if p > 3:
                B.append(p)

        # The only other p that might divide B are those that divide
        # the integer 2*#E(Q)_tor^2 * L(E,1)/omega.  So we compute
        # that to sufficient precision to determine it.  Note that
        # we have to assume the Manin constant is <=2 in order to provably
        # compute L(E,1)/omega.
        for p, n in factor(self.sha_an()):
            if n >= 2:    # use parity of Sha
                B.append(int(p))
        B = list(set(B))
        B.sort()
        return B

    def shabound(self):
        """
        Compute a provably correct bound on the order of the Shafarevich-Tate
        group of this curve. The bound is a either False (no bound) or a list
        B of primes such that any divisor of Sha is in this list.
        """
        if self.L1_vanishes():
            B = self.shabound_kolyvagin()
        else:
            B = self.shabound_kato()
        return B
    

    ##############################################################################
    # Formal Groups
    ##############################################################################

    # 	def	weierstrass_p(self):
    #         # TODO: add allowing negative valuations for power series
    #         return 1/t**2 + a1/t + rings.frac(1,12)*(a1-8*a2) -a3*t \
    #                - (a4+a1*a3)*t**2  + O(t**3)


    def padic_sigma(self, p, prec=20):
        """
        Returns the p-adic sigma function of the elliptic curve
        as a power series in t to precision prec.
        """
        R = rings.PowerSeriesRing(variable="t")
        t = R.g0
        a1, a2, a3, a4, a6 = self.ainvs()
        misc.todo("Implement genuine computation of sigma.")
        return t + a1/2 * t**2 + (a1**2+a2)/3 * t**3 + (a1**3+2*a1*a2+2*a3)/4 * t**4 + O(t**5)

    def formal_w(self, prec=20):
        """
        Return the formal power series
        
                     w(t) = t^3 + ...

        to precision O(t^prec) of Proposition IV.1.1 of [Silverman
        AEC1].  This is the formal expansion of w = -1/y about the
        formal parameter t = -x/y at infinity.

        We compute w using the recursive procedure (4.1) on page 18
        of Bluher's `A leisurely introduction to formal groups and
        elliptic curves', which I downloaded from
        
            http://www.math.uiuc.edu/Algebraic-Number-Theory/0076/
            
        """
        try:
            pr, w = self.__formal_w
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = w.parent().gen()
            return w + O(t**prec)
        R = rings.PowerSeriesRing(self.base_ring(), "t")
        t = R.gen()
        w = t**3 + O(t**prec)
        a1, a2, a3, a4, a6 = self.ainvs()
        for n in range(4,prec):
            if n<pr:
                w[n] = self.__formal_w[1][n]   # already know it
            else:
                w[n] = a1*w[n-1] + \
                       a2*w[n-2] + \
                       a3*sum([w[i]*w[n-i] for i in range(3,n-2)]) + \
                       a4*sum([w[i]*w[n-1-i] for i in range(3,n-3)]) + \
                       a6*sum([w[i]*w[j]*w[n-i-j] for i in range(3,n-2) \
                               for j in range(3,n-2) if n-i-j >= 3])
        self.__formal_w = (prec, w)
        return w

    def formal_x(self, prec=20):
        """
        Return the formal power series x(t) in terms of the local
        parameter t = -x/y at infinity.
        """
        prec = max(prec,0)
        try:
            pr, x = self.__formal_x
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = x.parent().gen()
            return x + O(t**prec)
        w = self.formal_w(prec+5)
        t = w.parent().gen()
        x = t/w   
        self.__formal_x = (prec, x)
        return x
    
    def formal_y(self, prec=20):
        """
        Return the formal power series y(t) in terms of the local
        parameter t = -x/y at infinity.
        """
        prec = max(prec,0)
        try:
            pr, y = self.__formal_y
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = y.parent().gen()
            return y + O(t**prec)
        w = self.formal_w(prec+6)
        t = w.parent().gen()
        y = -1/w   
        self.__formal_y = (prec, y)
        return y

    def formal_inverse(self, prec=20):
        prec = max(prec,0)
        try:
            pr, inv = self.__formal_inverse
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = inv.parent().gen()
            return inv + O(t**prec)
        x = self.formal_x(prec)
        y = self.formal_y(prec)
        a1, _, a3, _, _ = self.ainvs()
        inv = x / ( y + a1*x + a3)          # page 114 of Silverman, AEC I
        self.__formal_inverse = (prec, inv)
        return inv

    def formal_n_isogeny(self, n, prec=20):
        prec = max(prec,0)
        try:
            pr, phi = self.__formal_n_isogeny[n]
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = phi.parent().gen()
            return phi + O(t**prec)
        try:
            _ = self.__formal_n_isogeny
        except:
            self.__formal_n_isogeny = {}
        misc.todo()

    def formal_group(self, prec=10):
        prec = max(prec,0)
        try:
            pr, F = self.__formal_group
        except AttributeError:
            pr = -1
        if prec <= pr:
            return F
        R1 = rings.PowerSeriesRing(self.base_ring(),"t1")
        R2 = rings.PowerSeriesRing(R1,"t2")
        t1 = R1.gen().add_bigoh(prec)
        t2 = R2.gen().add_bigoh(prec)
        w = self.formal_w(prec)
        def tsum(n):
            return sum([t2**m * t1**(n-m-1) for m in range(n)])
        lam = sum([tsum(n)*w[n] for n in range(3,prec)])
        w1 = R1(w.coeffs()).add_bigoh(prec)
        nu = w1 - lam*t1
        a1, a2, a3, a4, a6 = self.ainvs()
        lam2 = lam*lam
        lam3 = lam2*lam
        t3 = -t1 - t2 - \
             (a1*lam + a3*lam2 + a2*nu + 2*a4*lam*nu + 3*a6*lam2*nu)/  \
             (1 + a2*lam + a4*lam2 + a6*lam3)
        inv = self.formal_inverse(prec)
        F = inv(t3)
        self.__formal_group = (prec, F)
        return F

    def formal_mult(self, n, prec=10):
        R = rings.PowerSeriesRing(self.base_ring(),"t")
        t = R._0
        if n == 1:
            return t +O(t**prec)
        if n == -1:
            return self.formal_inverse(prec)
        if n < 0:
            return self.formal_inverse(prec)(self.formal_mult(-n,prec))
        F = self.formal_group(prec)
        g = F.parent().base_ring().gen()
        for m in range(2,n+1):
            g = F(g)        
        return R(g.coeffs())

    ##############################################################################
    # End Formal Groups
    ##############################################################################

    ####################################################################
    # End elliptic curve over Q
    ####################################################################

    
class Point:
    """
    Point on an elliptic curve.
    """

    def __init__(self, curve, x=oo, y=oo, check=True):
        if not isinstance(curve, EllipticCurve):
            raise TypeError, "Argument 1 must be an elliptic curve."
        self.__curve = curve
        self.is_vector = True
        K = curve.base_ring()
        if isinstance(x, (list,tuple)):
            if len(x) == 1:
                x, y = x[0], 1
            elif len(x) == 2:
                x, y = x[0], x[1]
            else:
                if x[2] == 0:
                    x = rings.infinity
                else:
                    try:
                        z = K(x[2])
                        x, y = K(x[0])/z, K(x[1])/z
                    except TypeError:
                        raise ArithmeticError, "Point %s is not on curve."%self
        if x is oo:
            self.__x, self.__y = oo, oo
            return
        try:
            x = K(x)
            y = K(y)
        except TypeError:
            raise ArithmeticError, "Point %s is not on curve."%((x,y),)
        self.__x, self.__y = x, y
        if y*y + curve.a1()*x*y + curve.a3()*y != \
               x**3 + curve.a2()*x**2 + curve.a4()*x + curve.a6():
            raise ArithmeticError, "Point %s is not on curve."%self

    def __repr__(self):
        if self.is_infinity():
            return "(1,0,0)"
##         if isinstance(self.base_ring(), rings.RationalField):
##             d = self.__x.denominator().lcm(self.__y.denominator())
##         else:
##             d = self.base_ring()(1)
        return "(%s, %s)"%(self.__x, self.__y)

    def base_ring(self):
        return self.__curve.base_ring()
    base_field = base_ring

    def order(self):
        """
        Return the order of this point on the elliptic curve.
        If the point has infinite order, returns 0.

        EXAMPLE:
            >>> E = EllipticCurve([0,0,1,-1,0])
            >>> P = E([0,0]); P
            (0, 0)
            >>> P.order()
            0

            >>> E = EllipticCurve([0,1])
            >>> P=E([-1,0])
            >>> P.order()
            2
        """
        if self.is_infinity():
            return 1
        return self.__curve.pari_curve().ellorder([self.__x, self.__y])

    def xy(self):
        return self.__x, self.__y

    def x(self):
        return self.__x

    def y(self):
        return self.__y

    def curve(self):
        return self.__curve

    def __getitem__(self, n):
        if n==0:
            return self.__x
        elif n==1:
            return self.__y
        else:
            raise IndexError, "Index must be 0 or 1"

    def is_infinity(self):
        return self.__x is rings.infinity
    
    def __add__(self, right):
        """Add self to right."""
        if self.curve() != right.curve():
            raise errors.MathError("Points must be on the same curve.")

        # Use Prop 7.1.7 of Cohen "A Course in Computational Algebraic Number Theory"
        if self.is_infinity():
            return right
        if right.is_infinity():
            return self
        a1, a2, a3, a4, a6 = self.__curve.ainvs()
        x1, y1 = self.__x, self.__y
        x2, y2 = right.__x, right.__y
        if x1 == x2 and y1 == -y2 - a1*x2 - a3:
            return Point(self.__curve) # point at infinity
        if x1 == x2 and y1 == y2:
            m = (3*x1*x1 + 2*a2*x1 + a4 - a1*y1) / (2*y1 + a1*x1 + a3)
        else:
            m = (y1-y2)/(x1-x2)
        x3 = -x1 - x2 - a2 + m*(m+a1)
        y3 = -y1 - a3 - a1*x3 + m*(x1-x3)
        return Point(self.__curve, x3, y3)

    def __sub__(self, right):
        return self + (-right)

    def __neg__(self):
        if self.is_infinity():
            return self
        E, x, y = self.__curve, self.__x, self.__y
        return Point(self.__curve, x, -y - E.a1()*x - E.a3())

    def __mul__(self, m):
        return self.__rmul__(m)

    def __rmul__(self, m):
        m = int(m)
        if m<0:
            return (-m)*(-self)
        if m==1:
            return self
        if m==0:
            return Point(self.__curve)
        power = Point(self.__curve)
        i = 0
        apow2 = self
        while ((m>>i) > 0):
            if((m>>i) & 1):
                power += apow2
            apow2 += apow2
            i += 1
        return power

    def height(self):
        """
        The height of the point.
        """
        return self.curve().pari_curve().ellheight([self[0], self[1]])

    def sigma(self, p):
        k = rings.Qp(p)
        if self.is_infinity():
            return k(0)
        sigma = self.__curve.sigma(p)
        return sigma(k(-self.__x/self.__y))


    def padic_height(self, p):
        """Returns the p-adic cyclotomic height of the point.

        padic_height(p)

        Input:
           p: a prime number
        Output:
           p-adic cyclotomic height, which is a single p-adic number.

        Algorithm:
        We compute this height using the following formula, which
        is valid for points that are in the intersection of the
        identity component of the Neron model with the kernel of
        reduction modulo p:

     h(P) = 1/2 * sum_{ell!=p} sup(0,-ord_ell(x(P)))  + log_p(sigma_p(-x(P)/y(P)) / e),

        where P=(a/e^2, b/e^3) with gcd(a,e)=1, and
        where the first sum is over primes ell that don't equal p.
        If P isn't in the subgroup mentioned above, let n be a positive
        integer so that n*P is in that subgroup.  Then we return
        h(n*P)/(n**2), which does not depend on the choice of n, and
        is defined using the above formula.

        """

        if not arith.is_prime(p):
            raise errors.UsageError("p must be a prime number.")
        E = self.curve()
        if E.conductor() % p == 0:
            raise errors.UsageError("p must be a prime of good reduction.")

        ap = E.ap(p)
        if ap == 0:
            raise errors.UsageError("p must a prime of good ordinary reduction.")
        
        Np = p+1 - E.ap(p)
        c = 1
        for q, _ in arith.factor(E.conductor()):
            c = arith.lcm(c, E.tamagawa_number(q))

        n = Np*c
        misc.verbose("Multiplying point by %s."%n)
        R = n*self
            
        x, y = R.xy()
        sigma = E.padic_sigma(p)
        Qp = rings.pAdicField(p)
        print "x = ", x
        print "-x/y = ", Qp(-x/y)
        d = x.denominator()
        e = d.sqrt()
        if e*e != d:
            raise errors.MathError("The denominator is not a perfect square!?")
        t = Qp(-x/y)
        return (sigma(t)/Qp(e)).log()



        w = Qp(0)
        for ell, ord in d.factor():
            if ell != p:
                w += ord * Qp(ell).log()
        # note that there is no 1/2, since we factored the square root.
        t = Qp(-x/y)
        print "sigm = ", sigma(t)
        print "e = ", e
        hP = w + (sigma(t)/e).log()
        print "w = ", w
        print "sigma(t) = ", sigma(t)
        print "sigma(t)/e = ", sigma(t)/e
        print "(sigma(t)/e).log() = ", (sigma(t)/e).log()
        print "sum = ", w + (sigma(t)/e).log()

        return hP/Qp(n**2)



def curves(conductors):
    """
    Return iterator over all known curves (in database) with conductor
    in the list of conductors.
    """
    if isinstance(conductors, (int,long,rings.Element)):
        conductors = [conductors]
    db = sage.tables.elliptic_curves.Database()
    for N in conductors:
        for E in db.curves(N):
            yield EllipticCurve(E)

def optimal_curves(conductors):
    """
    Return iterator over all known optimal curves (in database) with
    conductor in the list of conductors.
    """
    if isinstance(conductors, (int,long,rings.Element)):
        conductors = [conductors]
    db = sage.tables.elliptic_curves.Database()
    for N in conductors:
        for E in db.optimal_curves(N):
            yield EllipticCurve(E)

def allcurves(min_cond=11):
    """
    Iterate through all curves in the elliptic curves database with
    conductor >= min_cond.
    """
    db = sage.tables.elliptic_curves.Database()
    for N in xrange(min_cond, 40001):
        for E in db.curves(N):
            yield EllipticCurve(E)
                
if __name__ ==  '__main__':
    from sage.rings.rings import *
    ZZ = IntegerRing()
    QQ = IntegerRing()
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
