"""
Vectors
"""
#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

#import Numeric

import sage.linalg.matrix
import sage.linalg.vector_space 
import sage.misc.misc as misc
import sage.rings.element as element

class Vector:
    def __init__(self, parent):
        if False and not isinstance(parent, \
                sage.linalg.vector_space.VectorSpace):
            raise TypeError, "parent must be a VectorSpace"
        self._parent = parent
        self.__degree = parent.degree()
        
    def __repr__(self):
        d = self.__degree
        if d == 0: return "()"
        # compute column widths
        S = [str(x) for x in self.list()]
        #width = max([len(x) for x in S])
        s = "("
        for i in xrange(d):
            if i == d-1:
                sep = ""
            else:
                sep=" "
            entry = S[i]
            #if i > 0:
            #    entry = " "*(width-len(entry)) + entry
            s += entry + sep
        s = s + ")"
        return s
        
    def __str__(self):
        return self.__repr__()
    
    ###################################################
    ## Access:
    ##   These *must* be overridden by each subclass
    ##
    ## In general, get and set are meant to be more efficient
    ## than getitem and setitem, because they do not do
    ## any error checking or coercion.  Use with care.
    ###################################################
    def __getitem__(self, i):
        raise NotImplementedError
    
    def get(self, i):
        """
        get is meant to be more efficient
        than getitem, because it does not do
        any error checking.
        """
        return self[i]
    
    def __setitem__(self, i, x):
        raise NotImplementedError
        
    def set(self, i, x):
        """
        set is meant to be more efficient
        than setitem, because it does not do
        any error checking or coercion.  Use with care.
        """
        self[i] = x
        
    ###################################################
    ## Properties
    ###################################################
    def base_field(self):
        return self._parent.base_field()
    base_ring = base_field
        
    def degree(self):
        return self.__degree
    
    def parent(self):
        return self._parent

    ###################################################
    ## Functions
    ###################################################
    def copy(self):
        return self.parent()(entries=self.entries(), \
                copy=True, coerce_entries=False)
        
    def denominator(self):
        R = self.base_ring()
        if self.degree() == 0: return 1
        x = self.list()
        d = x[0].denominator()
        for y in x:
            d = d.lcm(y.denominator())
        return d
        
    def dict(self):
        e = {}
        for i in xrange(self.degree()):
            c = self[i]
            if c != 0:
                e[i] = c
        return e
        
    def entries(self):
        # This *must* be defined in the derived class.
        # It is the underlying representation of elements.
        raise NotImplementedError
        
    def height(self):
        raise NotImplementedError
        
    def is_dense(self):
        return self.parent().is_dense()
   
    def is_sparse(self):
        return self.parent().is_sparse()
        
    def is_vector(self):
        return True
        
    def list(self):
        return [self[i] for i in xrange(self.degree())]
        
    def nonzero_positions(self):
        """
        Returns the sorted list of integers i such that self[i] != 0.
        """
        z = self.base_ring()(0)
        v = self.list()
        return [i for i in xrange(self.degree()) if v[i] != z]
    
    #def norm(self):
    #        return Numeric.math.sqrt(Numeric.sum(self.numeric_array()))
        
    #def numeric_array(self, typecode=Numeric.Float64):
    #    return Numeric.array(self.list(), typecode=typecode)
        
    def rational_reconstruction(self):
        raise NotImplementedError
        
    def str_latex(self):
        """
        Print a latex representation of self.
        For example, if self is [1,2,3,4], the following 
        latex is generated: (1,2,3,4)
        """
        s = str(self)
        return str(self).replace("[","(").replace("]",")")
        
        
    ###################################################
    ## Arithmetic
    ###################################################
    def __abs__(self):
        return self.norm()
    
    def __add__(self, right):
        if not isinstance(right, Vector):
            raise TypeError, "right must be a Vector"
        if self._parent == right._parent:
            V = self._parent
        elif self._parent.ambient_space() == right._parent.ambient_space():
            V = self._parent.ambient_space()
        else:
            raise TypeError, "self and right must have compatible parents"
        X = [self[i] + right[i] for i in xrange(self.degree())]
        return V(X, coerce_entries=False, copy=False, check_element=False)
            
    def __cmp__(self, right):
        if not isinstance(right, Vector):
            right = self._parent(right)
        if not (self._parent is right._parent) and \
            not (self._parent.ambient_space() is right._parent.ambient_space()):
            return -1   
        for i in xrange(self.degree()):
            c = self[i].__cmp__(right[i])
            if c: return c
        return 0
        
    def __invert__(self):
        raise NotImplementedError
        
    def _matrix_multiply(self, A):
        """
        Return the product self*A.
        """
        return A.vector_matrix_multiply(self)
        
    def _scalar_multiply(self, s):
        """
        return the product s*self.
        """
        s = self.base_field()(s)
        return self._parent([x*s for x in self.list()], \
                    copy=False, coerce_entries=False, check_element=False)
        
    def __mod__(self, p):
        raise NotImplementedError
    
    def __mul__(self, right):
        if isinstance(right, sage.linalg.matrix.Matrix):
            # vector x matrix multiply
            return self._matrix_multiply(right)
        try:
            right = self.base_ring()(right)
        except TypeError:
            raise TypeError, "unable to multiply %s by %s"%(self, right)
        return self._scalar_multiply(right)
        
    def __rmul__(self, left):
        return self._scalar_multiply(left)
        
    def __neg__(self):
        return self._scalar_multiply(-1)
        
    def __pos__(self):
        return self
        
    def __pow__(self, n):
        raise NotImplementedError
        
    def __sub__(self, right):
        if not isinstance(right, Vector):
            raise TypeError, "right must be a Vector"
        if self._parent == right._parent:
            V = self._parent
        elif self._parent.ambient_space() == right._parent.ambient_space():
            V = self._parent.ambient_space()
        else:
            raise TypeError, "self and other must have compatbile parents"
        X = [self[i] - right[i] for i in xrange(self.degree())]
        return V(X, coerce_entries=False, copy=False, check_element=False)
            

#############################################
## Generic dense vector
############################################# 
class Vector_generic_dense(Vector):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Vector.__init__(self, parent)
        R = self.base_ring()
        if entries == 0:
            zero = R(0)
            entries = [zero for _ in xrange(self.degree())]
        else:
            if not isinstance(entries, list):
                raise TypeError, "entries must be a list"
            if len(entries) != self.degree():
                raise ArithmeticError, "entries must be a list of length %s"%\
                            self.degree()
            if coerce_entries:
                try:
                    entries = [R(x) for x in entries]
                except TypeError:
                    raise TypeError, "Unable to coerce entries to %s"%R
            elif copy:
                # Make a copy
                entries = list(entries) 
        self.__entries = entries
        
    def __getitem__(self, i):
        """
        """
        i = int(i)
        #if not isinstance(i, int):
        #    raise TypeError, "index must an integer"
        if i < 0 or i >= self.degree():
            raise IndexError, "index (i=%s) must be between 0 and %s"%(i,
                            self.degree()-1)
        return self.__entries[i]
    
    def __setitem__(self, i, value):
        """
        Set entry i of self to value.
        """
        i = int(i)
        #if not isinstance(i, int):
        #    raise TypeError, "index must an integer"
        if i < 0 or i >= self.degree():
            raise IndexError, "index (i=%s) must be between 0 and %s"%(i,
                            self.degree()-1)
        self.__entries[i] = self.base_ring()(value)

    def entries(self):
        return self.__entries
        
    def list(self):
        return self.__entries
        
        
#############################################
## Generic sparse vector
#############################################  
def _sparse_dot_product0(v, w):
    """
    v and w are sorted lists of pairs (i,x), where i is the position
    and x is a ring element.
    """
    s = 0
    i = 0
    j = 0
    while i < len(v) and j < len(w):
        a = v[i]; b = w[j]
        if a[0] < b[0]:
            i += 1
            continue
        elif a[0] > b[0]:
            j += 1
            continue
        s += a[1] * b[1]
        i += 1
        j += 1
    return s
    
def _sparse_dot_product(v, w):
    """
    v and w are dictionaries with integer keys.
    """
    x = set(v.keys()).intersection(set(w.keys()))
    return sum([v[k]*w[k] for k in x])
    
class Vector_generic_sparse(Vector):
    """
    A generic sparse vector is a dictionary with keys ints i
    and entries in the base ring.
    WARNING: In creation, we do not check that the i pairs satisfy
             0 <= i < degree
    """
    def __init__(self, parent, 
                 entries=0, 
                 coerce_entries=True, 
                 copy=True):
        Vector.__init__(self, parent)
        R = self.base_ring()
        if entries == 0:
            entries = {}
        else:
            if isinstance(entries, list):
                if len(entries) != self.degree():
                    raise TypeError, "entries has the wrong length"
                x = entries
                entries = {}
                for i in xrange(self.degree()):
                    if x[i] != 0:
                        entries[i] = x[i]
                copy = False
            if not isinstance(entries, dict):
                raise TypeError, "entries must be a dict"
            if coerce_entries:
                try:
                    for k, x in entries.iteritems():
                        entries[k] = R(x)
                except TypeError:
                    raise TypeError, "Unable to coerce entries to %s"%R
            elif copy:
                # Make a copy
                entries = dict(entries)
        self.__entries = entries
        self.__zero = R(0)
    
    def __getitem__(self, i):
        #if not isinstance(i, int):
        i = int(i)
            #raise TypeError, "index must an integer"
        if i < 0 or i >= self.degree():
            raise IndexError, "index (i=%s) must be between 0 and %s"%(i,
                            self.degree()-1)
        if self.__entries.has_key(i):
            return self.__entries[i]
        return self.__zero
        
    def get(self, i):
        """
        Like __getitem__ but with no type or bounds checking.
        Returns 0 if access is out of bounds.
        """
        i = int(i)
        if self.__entries.has_key(i):
            return self.__entries[i]
        return self.__zero
        
    def set(self, i, x):
        """
        Like __setitem__ but with no type or bounds checking.
        """
        i = int(i)
        if x == 0:
            if self.__entries.has_key(i):
                del self.__entries[i]
            return
        self.__entries[i] = x
    
    def __setitem__(self, i, value):
        """
        """
        i = int(i)
        #if not isinstance(i, int):
        #    raise TypeError, "index must an integer"
        if i < 0 or i >= self.degree():
            raise IndexError, "index (i=%s) must be between 0 and %s"%(i,
                            self.degree()-1)
        self.set(i, value)
           
    def denominator(self):
        R = self.base_ring()
        x = self.entries()
        if len(x) == 0:
            return 1
        Z = x.iteritems()
        d = Z.next()[1].denominator()
        for _, y in Z:
            d = d.lcm(y.denominator())
        return d
        
    def entries(self):
        return self.__entries
        
    def dict(self):
        return self.__entries
        
    def nonzero_positions(self):
        """
        Returns the set of pairs (i,j) such that self[i,j] != 0.
        """
        K = self.__entries.keys()
        K.sort()
        return K
    

if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
