"""
Vector space
"""
#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

# System imports
import random
import weakref

# Manin imports
import sage.rings.rings as rings
import sage.rings.ring as ring
import sage.structure.gens as gens
import vector
import matrix_space

class VectorSpace(object):
    """
    Create a Vector Space.
    
    Two create an ambient space over a field with given dimension
    using the calling syntax
        VectorSpace(field, dimension, sparse=True)
    and to create a subspace use
        VectorSpace(list_of_vectors),
    where list_of_vectors is a list of vectors in a VectorSpace.
    
    EXAMPLES:
    First we create some ambient spaces.
        >>> from sage.rings.rings import RationalField, CyclotomicField
        >>> V = VectorSpace(RationalField(), 389); V
        Full Vector space of degree 389 over Rational Field
        >>> V.dimension()
        389
        >>> K = CyclotomicField(7)
        >>> VK = VectorSpace(K, 3); VK
        Full Vector space of degree 3 over Cyclotomic Field of order 7 and degree 6
        >>> K.gen(0)*VK.gen(1)
        (0 zeta_7 0)
            
    Next we create some subspaces.
        >>> W = VectorSpace([VK.gen(1) + K.gen(0)*VK.gen(2)]); W
        Vector space of degree 3, dimension 1 over Cyclotomic Field of order 7 and degree 6
        Basis matrix:
        [     0      1 zeta_7]
    
        >>> V = VectorSpace(RationalField(), 3)
        >>> W = VectorSpace([V.gen(0)]); W
        Vector space of degree 3, dimension 1 over Rational Field
        Basis matrix:
        [1 0 0]
        >>> v0 = V([1,2,3])
        >>> v1 = V([4,5,6])
        >>> v2 = V([7,8,9])
        >>> W = VectorSpace([v0,v1,v2]); W
        Vector space of degree 3, dimension 2 over Rational Field
        Basis matrix:
        [ 1  0 -1]
        [ 0  1  2]

    We illustrate most of the  functions on an ambient space:
        >>> V = VectorSpace(RationalField(), 3)
        >>> V.ambient_space()
        Full Vector space of degree 3 over Rational Field
        >>> V == V.ambient_space()
        True
        >>> V == VectorSpace(CyclotomicField(3), 3)
        False
        >>> V.base_field()
        Rational Field
        >>> V.basis()
        [(1 0 0), (0 1 0), (0 0 1)]
        >>> V.coordinates(V._0 + 2*V._1 - 3*V._2)
        [1, 2, -3]
        >>> V.degree()
        3
        >>> V.dimension()
        3
        >>> V.gen(0)
        (1 0 0)
        >>> V.is_dense()
        True
        >>> V.is_sparse()
        False
        >>> V.matrix()
        [1 0 0]
        [0 1 0]
        [0 0 1]
        >>> V.ngens()
        3
        >>> V.random_element() in V
        True
        >>> V.vector([1,2,3])
        (1 2 3)
        >>> V.vector_space(100)
        Full Vector space of degree 100 over Rational Field
        >>> V.zero_vector()
        (0 0 0)
        
    We create a sparse ambient vector space, and demonstrate each
    of the above functions on it.
        >>> V = VectorSpace(RationalField(), 3, sparse=True)
        >>> V.ambient_space()
        Full Vector space of degree 3 over Rational Field
        >>> V == V.ambient_space()
        True
        >>> V == VectorSpace(CyclotomicField(3), 3)
        False
        >>> V.base_field()
        Rational Field
        >>> V.basis()
        [(1 0 0), (0 1 0), (0 0 1)]
        >>> V.coordinates(V._0 + 2*V._1 - 3*V._2)
        [1, 2, -3]
        >>> V.degree()
        3
        >>> V.dimension()
        3
        >>> V.gen(0)
        (1 0 0)
        >>> V.is_dense()
        False
        >>> V.is_sparse()
        True
        >>> V.matrix()
        [1 0 0]
        [0 1 0]
        [0 0 1]
        >>> V.matrix().is_sparse()
        True
        >>> V.ngens()
        3
        >>> V.random_element() in V
        True
        >>> v = V.vector([1,2,3]); v
        (1 2 3)
        >>> v.is_sparse()
        True
        >>> W = V.vector_space(100); W
        Full Vector space of degree 100 over Rational Field
        >>> W.is_sparse()
        True
        
    We create a subspace of a dense vector space, and illustrate each
    function on it:
        >>> V0 = VectorSpace(RationalField(), 3)
        >>> V = VectorSpace([V0([1,2,3]), V0([4,5,6]), V0([7,8,9])]); V
        Vector space of degree 3, dimension 2 over Rational Field
        Basis matrix:
        [ 1  0 -1]
        [ 0  1  2]
        >>> V.ambient_space()
        Full Vector space of degree 3 over Rational Field
        >>> V == V.ambient_space()
        False
        >>> V.base_field()
        Rational Field
        >>> V.generators()
        [(1 2 3), (4 5 6), (7 8 9)]
        >>> V.coordinates(V([1, 0, -1]))
        [1, 0]
        >>> V.coordinates(V([1, 0, 0]))
        Traceback (most recent call last):
        ...
        ArithmeticError: v is not in self
        >>> V.degree()
        3
        >>> V.dimension()
        2
    
    Note that the ith generator of a subspace is not the ith element
    of the list of vectors that generators, but is the free generator,
    i.e., the ith basis vector.
        >>> w = V.gen(0); w
        (1 0 -1)
        >>> w.is_sparse()
        False
        >>> V.is_dense()
        True
        >>> V.matrix()
        [ 1  0 -1]
        [ 0  1  2]
        >>> V.matrix().is_sparse()
        False
        >>> V.ngens()
        2
        >>> V.random_element() in V
        True
        >>> V.vector([1,2,3])
        (1 2 3)
        >>> V([1,2,3])
        (1 2 3)
        >>> V.vector([1,0,0])
        Traceback (most recent call last):
        ...
        ArithmeticError: v is not in self
        >>> V.zero_vector()
        (0 0 0)
        
    We create a subspace of a sparse vector space, and illustrate each
    function on it:
        >>> V0 = VectorSpace(RationalField(), 3, sparse =True)
        >>> V0.is_sparse()
        True
        >>> V = VectorSpace([V0([1,2,3]), V0([4,5,6]), V0([7,8,9])]); V
        Vector space of degree 3, dimension 2 over Rational Field
        Basis matrix:
        [ 1  0 -1]
        [ 0  1  2]
        >>> V.is_sparse()
        True
        >>> V.ambient_space()
        Full Vector space of degree 3 over Rational Field
        >>> V == V.ambient_space()
        False
        >>> V.base_field()
        Rational Field
        >>> V.generators()
        [(1 2 3), (4 5 6), (7 8 9)]
        >>> V.coordinates(V([1, 0, -1]))
        [1, 0]
        >>> V.coordinates(V([1, 0, 0]))
        Traceback (most recent call last):
        ...
        ArithmeticError: v is not in self
        >>> V.degree()
        3
        >>> w = V.gen(0); w
        (1 0 -1)
        >>> w.is_sparse()
        True
        >>> V.is_dense()
        False
        >>> V.matrix()
        [ 1  0 -1]
        [ 0  1  2]
        >>> V.matrix().is_sparse()
        True
        >>> V.ngens()
        2
        >>> V.random_element() in V
        True
        >>> V.vector([1,2,3])
        (1 2 3)
        >>> V.vector([1,0,0])
        Traceback (most recent call last):
        ...
        ArithmeticError: v is not in self
        >>> V.zero_vector()
        (0 0 0)
    """
    def __new__(cls, *args, **kwds):
        if cls != VectorSpace:
            return object.__new__(cls, *args, **kwds)
        if len(args) == 0:
            raise TypeError, "VectorSpace requires at least 2 arguments."
        if isinstance(args[0], list):
            return VectorSpace_subspace(*args, **kwds)
        if isinstance(args[0], ring.Ring) :
            return VectorSpace_ambient(*args, **kwds)
        raise TypeError, "VectorSpace -- give either a list of vectors or a base field and a dimension"

    def __init__(self, *args, **kwds):
        raise NotImplementedError
        
                   
class VectorSpace_generic(VectorSpace, gens.Gens):
    def __init__(self,  base_field,             
                        degree,
                        sparse=False):
        """
        Create the space of all vectors of given degree over base_field.
        INPUT:
            base_field -- a field
            degree -- int >= 0, the degree of the vector space 
                      (number of components of vectors).
            sparse -- whether or not matrices are given a sparse representation 
                      (default to False)
        """
        if not isinstance(base_field, rings.Ring):
            raise TypeError, "base_field must be a rings.Ring"
            
        if not base_field.is_field():
            raise TypeError, "base_field must be a field"

        #if not isinstance(degree, int):
        degree = int(degree)
    
        if degree < 0:
            raise ArithmeticError, "degree must be nonnegative"

        self.__base_field = base_field
        self.__degree = degree
        self.__is_sparse = sparse
        self.__set_vector_class()
        
    def __add__(self, other):
        if not isinstance(other, VectorSpace):
            raise TypeError, "other must be a Vector Space"
        if self.ambient_space() != other.ambient_space():
            raise ArithmeticError, "self and other must have the same ambient space"
        return VectorSpace(self.basis() + other.basis(), self.ambient_space())
                
    def __call__(self, entries=0, coerce_entries=True,
                       copy=True, check_element=True):
        return self.vector(entries, coerce_entries, copy, check_element)
        
    def __cmp__(self, right):
        if self is right:
            return 0
        if not isinstance(right, VectorSpace):
            return -1
        n = self.dimension().__cmp__(right.dimension())
        if n: return n
        n = self.base_ring().__cmp__(right.base_ring())
        if n: return n
        n = self.is_sparse().__cmp__(right.is_sparse())
        if n: return n
        if self.is_ambient():
            if not right.is_ambient():
                return 1
            else:
                return 0
        # now self not ambient
        if right.is_ambient():
            return -1
        # have to compare underlying basis (this could be slow if they *are* equivalent)
        if self.basis() == right.basis():
            return 0
        return 1
        #return self.matrix().__cmp__(right.matrix())
        
    def __contains__(self, x):
        raise NotImplementedError
 
    def __repr__(self):
        if self.is_sparse():
            s = "sparse"
        else:
            s = "dense"
        return "Full VectorSpace of %s by %s %s matrices over %s"%(
                    self.__nrows, self.__ncols, s, self.__base_field)
          
    def __set_vector_class(self):
        if self.is_sparse():
            self.__vector_class = vector.Vector_generic_sparse
            return
        self.__vector_class = vector.Vector_generic_dense
            
    def ambient_space(self):
        return NotImplementedError
        
    def base_field(self):
        return self.__base_field
        
    def base_ring(self):
        return self.__base_field
        
    def basis(self):
        raise NotImplementedError
    echelonized_basis = basis
    
    def _echelon_coordinates(self, v):
        """
        Write v in terms of the echelon basis for self.

        Returns a list c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        if not isinstance(v, vector.Vector):
            raise TypeError, "v must be a Vector"
        if v.parent().ambient_space() != self.ambient_space():
            raise ArithmeticError, "v is not in self"
        nz = v.nonzero_positions()
        if len(nz) == 0:
            return []
        c = []
        for b in self.echelonized_basis():
            j = b.nonzero_positions()[0]
            c.append(v[j])
            # NOTE: It is *very* important that the product be written b*v[j] instead
            # of v[j]*b below.  Otherwise the arithmetic goes through some coercision
            # tests, which test inclusion, which in some cases call the coordinates
            # function and when get a huge (but not infinite) wasteful recursion.
            v -= b*v[j]    
        if v != 0:
            raise ArithmeticError, "v is not in self"%v
        return c

    def coordinates(self, v):
        """
        Write v in terms of the basis for self.

        Returns a list c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        return self._echelon_coordinates(v)

    def degree(self):
        return self.__degree
        
    def dimension(self):
        raise NotImplementedError
        
    def gen(self, n):
        return self.basis()[n]
        
    def is_ambient(self):
        raise NotImplementedError
    
    def is_dense(self):
        return not self.__is_sparse
        
    def is_full(self):
        return self.degree() == self.dimension()
        
    def is_sparse(self):
        return self.__is_sparse
        
    def matrix(self):
        try:
            return self.__matrix
        except AttributeError:
            MAT = matrix_space.MatrixSpace(self.base_field(),
                            len(self.basis()), self.degree(),
                            sparse = self.is_sparse())
            self.__matrix = MAT(self.basis())
            return self.__matrix
        
    def ngens(self):
        return self.dimension()
 
    def random_element(self, X=[-2,-1,1,2], prob=1.0, coerce=True):
        """
        Returns a random element of self.
        """
        if not isinstance(X, list):
            raise TypeError, "X must be a list"
        if not isinstance(prob, float):
            raise TypeError, "prob must be a float"
        if not isinstance(coerce, bool):
            raise TypeError, "coerce must be a bool"
        R = self.base_field()
        if coerce:
            X = [R(a) for a in X]
        v = self(0)
        for i in xrange(self.dimension()):
            if random.random() <= prob:
                v += self.gen(i)*random.choice(X)
        return v

    random = random_element
 
    def subspace(self, gens):
        """
        Create a subspace of self.

        INPUT:
            gens -- a list of vector in self
        OUTPUT:
            VectorSpace -- the subspace spanned by the vectors in the list gens.
            The basis for the subspace is always put in reduced row echelon form.
        EXAMPLES:
            >>> import sage.rings.rings as rings
            >>> V = VectorSpace(rings.RationalField(), 3)
            >>> B = V.basis()
            >>> W = V.subspace([B[0]+B[1], 2*B[1]-B[2]])
            >>> W
            Vector space of degree 3, dimension 2 over Rational Field
            Basis matrix:
            [   1    0  1/2]
            [   0    1 -1/2]
        """
        
        if not isinstance(gens, list):
            raise TypeError, "gens must be a list"
        for v in gens:
            if not (v in self):
                raise TypeError, "each element of gens must be in self"
        return VectorSpace(gens, self.ambient_space())

    def subspace_with_basis(self, basis):
        """
        Create a subspace of self with given basis.

        INPUT:
            basis -- a list of linearly independent vectors

        OUTPUT:
            VectorSpace_subspace_with_basis -- the subspace with given basis.
            The basis for the subspace is always put in reduced row echelon form.
        EXAMPLES:
            >>> import sage.rings.rings as rings
            >>> V = VectorSpace(rings.RationalField(), 3)
            >>> B = V.basis()
            >>> W = V.subspace_with_basis([B[0]+B[1], 2*B[1]-B[2]])
            >>> W
            Vector space of degree 3, dimension 2 over Rational Field
            User basis matrix:
            [ 1  1  0]
            [ 0  2 -1]
        """
        
        if not isinstance(basis, list):
            raise TypeError, "basis must be a list"
        for v in basis:
            if not (v in self):
                raise TypeError, "each element of basis must be in self"
        W = VectorSpace_subspace_with_basis(basis, self.ambient_space())
        if W.dimension() != len(basis):
            raise ArithmeticError, \
                   "vectors in basis are not linearly independent."
        return W
            
    def vector(self, x=0, coerce_entries=True, copy=True, check_element=True):
        if isinstance(x, (int, rings.Integer)) and x==0:
            return self.zero_vector()
        if isinstance(x, vector.Vector):
            if x.parent() == self:
                return x.copy()
            x = x.list()
        w = self.__vector_class(self, x, coerce_entries, copy)
        if check_element and not self.is_ambient():
            self.coordinates(w)
        return w
     
    def vector_space(self, degree):
        return VectorSpace_ambient(self.__base_field, 
                degree, sparse=self.is_sparse())
 
    def zero_vector(self):
        return self.__vector_class(self, 0)
        
       
_objs = {}
class _uniq(object):
    def __new__(cls, base_field, degree, sparse=False):
        key = (base_field, degree, sparse)
        if _objs.has_key(key):
            m = _objs[key]()
            if m != None: return m
        O = object.__new__(cls)
        _objs[key] = weakref.ref(O)
        return O
    
class VectorSpace_ambient(VectorSpace_generic, _uniq):
    """
    Ambient vector space:
            base_field -- a field
            degree -- int >= 0, the degree of the vector space 
                      (number of components of vectors).
            sparse -- whether or not vectors are given a sparse representation 
                      (default to False)
    """
    def __init__(self, base_field, degree, sparse=False):
        VectorSpace_generic.__init__(self, base_field, degree, sparse)
        
    def __contains__(self, x):
        return isinstance(x, vector.Vector) and \
            x.parent().ambient_space() == self
    
    def __repr__(self):
        return "Full Vector space of degree %s over %s"%(self.degree(), self.base_field())
        
    def ambient_space(self):
        return self
        
    def basis(self):
        """
        EXAMPLES:
            >>> from sage.rings.rings import RationalField
            >>> V = VectorSpace(RationalField(), 2); V
            Full Vector space of degree 2 over Rational Field
            >>> V.basis()
            [(1 0), (0 1)]
        """
        try:
            return self.__basis
        except AttributeError:
            ZERO = self(0)
            self.__basis = []
            for i in xrange(self.degree()):
                v = ZERO.copy()
                v[i] = 1
                self.__basis.append(v)
        return self.__basis

    echelonized_basis = basis
        
    def dimension(self):
        return self.degree()
        
    def is_ambient(self):
        return True
    
class VectorSpace_subspace(VectorSpace_generic):
    """
    """
    def __init__(self, vectors, ambient_space=None):
        """
        """
        if not isinstance(vectors, list):
            raise TypeError, "vectors must be a list of vectors"
        if ambient_space == None:
            if len(vectors) == 0:
                raise TypeError, "ambient_space must be specified if list of vectors is empty"
            ambient_space = vectors[0].parent().ambient_space()           
        if not isinstance(ambient_space, VectorSpace_ambient):
            raise TypeError, "ambient_space must be an ambient_space Vector Space"
            
        for v in vectors:
            if not isinstance(v, vector.Vector) or \
                    not v in ambient_space:
                raise TypeError, "each entry must be a vector in the ambient_space space"
        self.__generators = list(vectors)
        for v in self.__generators:
            v._parent = self
        
        VectorSpace_generic.__init__(self, ambient_space.base_field(), 
                ambient_space.degree(), ambient_space.is_sparse())
        
        self.__ambient_space = ambient_space
        
    def __contains__(self, v):
        if v.parent() == self:
            return True
        try:
            c = self.coordinates(v)
        except ArithmeticError:
            return False
        return True
    
    def __repr__(self):
        return "Vector space of degree %s, dimension %s over %s\n"%(
                self.degree(), self.dimension(), self.base_field()) + \
                "Basis matrix:\n%s"%self.matrix()
                #"Echelonized basis:\n%s"%"\n".join([str(x) for x in self.basis()])
                
    def ambient_space(self):
        return self.__ambient_space
                
    def basis(self):   # must compute basis matrix as side effect
        try:
            return self.__basis
        except AttributeError:
            # Make a matrix whose rows are the generators, then
            # put it in reduced row echelon form.  The basis is
            # then the set of nonzero rows of this echelon matrix.
            MAT = matrix_space.MatrixSpace(self.base_field(),
                        len(self.generators()), self.degree(),
                        sparse = self.is_sparse())
            A = MAT(self.generators())
            E = A.reduced_row_echelon_form()
            r = E.rank()
            self.__basis = [E.row(i) for i in xrange(r)]
            for v in self.__basis:
                v._parent = self
        return self.__basis

    echelonized_basis = basis

    def dimension(self):
        try:
            return self.__dimension
        except AttributeError:
            self.__dimension = len(self.basis())
        return self.__dimension
        
    def generators(self):
        return self.__generators
        
    def is_ambient(self):
        return False
    
class VectorSpace_subspace_with_basis(VectorSpace_subspace):
    def __init__(self, basis, ambient=None):
        VectorSpace_subspace.__init__(self, basis, ambient)
    
    def __repr__(self):
        return "Vector space of degree %s, dimension %s over %s\n"%(
                self.degree(), self.dimension(), self.base_field()) + \
                "User basis matrix:\n%s"%self.matrix()

    def basis(self):
        try:
            return self.__basis
        except AttributeError:
            B = self.echelonized_basis()
            if len(B) != len(self.generators()):
                raise ArithmeticError, "generators do not form a basis"
            self.__basis = self.generators()
            return self.__basis
    user_basis = basis
            
    def echelonized_basis(self):
        return VectorSpace_subspace.basis(self)
            
    def _user_to_echelon_matrix(self):
        try:
            return self.__user_to_echelon_matrix
        except AttributeError:
            rows = sum([self._echelon_coordinates(b) for b in self.generators()], [])
            M = matrix_space.MatrixSpace(self.base_field(),
                                         self.dimension(), sparse = self.is_sparse())
            self.__user_to_echelon_matrix = M(rows)
        return self.__user_to_echelon_matrix
        

    def _echelon_to_user_matrix(self):
        try:
            return self.__echelon_to_user_matrix
        except AttributeError:
            self.__echelon_to_user_matrix = ~self._user_to_echelon_matrix()
            return self.__echelon_to_user_matrix

    def coordinates(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a list c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        # First found the coordinates of v wrt echelon basis.
        w = VectorSpace_generic.coordinates(self, v)
        # Next use transformation matrix from echelon basis to
        # user basis.
        T = self._echelon_to_user_matrix()
        return T.linear_combination_of_rows(w)
        
            
                
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
