"""
Preparses input from the interpreter

Modified input.

  -- All ^'s (not in strings) are replaced by **'s.

  -- If M is a variable and i an integer,
     then M.i is replaced by M._i, so generators can be
     accessed as in MAGMA.

  -- quit alone on a line quits.

  -- load to load in scripts

  -- Most int literals n are replaced by ZZ(n)
     and all float literals x by RR(x).  Thus 2/3 is a
     rational number.  If they are in []'s, they
     aren't replaced (see comment below -- need to
     improve this!)

  -- 
All other input is processed normally.

I rewrote the SAGE/IPython preparser and uploaded it in the newest version of SAGE.  (2005-04-15). 
  
It automatically converts *most* integer literals to SAGE Integer's and  
decimal literals to SAGE Reals.  It does not convert indexes into 
1-d arrays, since those have to be ints.   
 
I also extended the load command so it *really* works exactly 
like the SAGE interpreter, so e.g., ^ for exponentiation is 
allowed.  Also files being loaded can themselves load other files. 
Finally, I added an "attach" command, e.g., 
    attach 'file' 
that works kind of like attach in MAGMA.  Whenever you enter a blank 
line in the SAGE interpreter, *all* attached files that have changed 
are automatically reloaded.  Moreover, the attached files work according 
to the SAGE interpreter rules, i.e., ^ --> **, etc.  
 
I also fixed it so ^ is not replaced by ** inside strings. 
 
Finally, I added back the M.n notation for the n-th generator 
of object M, again like in MAGMA.

EXAMPLE:

sage: 2/3
      2/3
sage: type(2/3)
      <type 'rational.Rational'>
sage: a = 49928420832092
sage: type(a)
      <type 'integer.Integer'>
sage: a.factor()
      [(2, 2), (11, 1), (1134736837093, 1)]
sage: v = [1,2,3]
sage: type(v[0])
      <type 'integer.Integer'>
sage: w = [5]    #<--- could be an index, so not converted to Integer
sage: w = [5/7] # this is the main gotcha..
sage: w
      [0]
sage: type(w[0])
      <type 'int'>

If we don't make potential list indices int's, then lots of stuff
breaks, or users have to type v[int(7)], which is insane.  
A fix would be to only not make what's in the brackets an
Integer if what's before the bracket is a valid identifier,
so the w = [5] above would work right. 

sage: s = "x^3 + x + 1"
sage: s
      'x^3 + x + 1'
sage: pari(s)
      x^3 + x + 1
sage: f = pari(s)
sage: f^2
      x^6 + 2*x^4 + 2*x^3 + x^2 + 2*x + 1
sage: V = VectorSp
VectorSpace                      VectorSpace_subspace
VectorSpace_ambient              VectorSpace_subspace_with_basis
VectorSpace_generic
sage: V = VectorSpace(Q,3)
sage: V.0
      (1 0 0)
sage: V.1
      (0 1 0)
sage: s = "This. Is. It."
sage: print s
This. Is. It.

"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

__author__ = 'William Stein <was@math.harvard.edu>'
__version__= '0.2.0'
__license__ = 'GPL'
__date__   = 'March 2005'

import os

# IPython has a prefilter() function that analyzes each input line. We redefine
# it here to first pre-process certain forms of input

# The prototype of any alternate prefilter must be like this one (the name
# doesn't matter):
# - line is a string containing the user input line.
# - continuation is a parameter which tells us if we are processing a first line of
#   user input or the second or higher of a multi-line statement.

def isalphadigit_(s):
    return s.isalpha() or s.isdigit() or s=="_"

def last_bracket_is_after_identifier(line, i):
    """
    Return True if and only if the previous bracket before line[i]
    is after a valid identifier followed possibly by some space.
    """
    line = line[:i]
    j = line.rfind("[")
    if j == -1:
        return False
    line = line[:j].rstrip()
    if len(line) == 0:
        return False
    c = line[len(line)-1]
    return c == "]" or c == ")" or isalphadigit_(c) 


def custom_subst(line):
    line = line.split("\n")[0]   # xreadlines leaves the '\n' at end of line
    
    # Wrap all integers with Z(), so they are SAGE integers.
    
    def wrap_num(i, line, is_real, num_start):
        if is_real:
            O = "RR('"; C = "')"
        else:
            O = "ZZ("; C = ")"
        line = line[:num_start] + O + line[num_start:i]  \
               + C + line[i:]
        return line, len(O+C)
    
    i = 0
    num_start = -1
    in_single_quote = False
    in_double_quote = False
    bracket_depth = 0
    seen_comma = False
    in_number = False
    is_real = False
    while i < len(line):
        if bracket_depth > 0 and line[i] == ",":
            seen_comma = True
            
        
        if in_number:
            if line[i] == ".": # and (i +1 == len(line) or (i +1 < len(line) and line[i+1].isdigit())):
                is_real = True
            elif not line[i].isdigit():
                # end of a number
                # Do we wrap?
                #print line[num_start:i], bracket_depth, seen_comma
                #if bracket_depth == 0 or (bracket_depth > 0 and seen_comma):
                if i < len(line) and line[i]=="e" or line[i]=="E":
                    i += 1   # skip wrapping
                elif bracket_depth == 0 or (bracket_depth > 0 and \
                                          not last_bracket_is_after_identifier(line, i)):
                    line, n = wrap_num(i, line, is_real, num_start)
                    i += n
                else:
                    i += 1
                in_number = False
                is_real = False
                continue
            
        elif line[i] == "." and i > 0 and i < len(line)-1 and \
                 not in_single_quote and not in_double_quote and \
                 isalphadigit_(line[i-1]) and line[i+1].isdigit():
            # Generators like in MAGMA: replace all ".<number>" by "._<number>"
            # If . is preceeded by \, then replace "\." by ".".            
            line = line[:i] + "._" + line[i+1:]
            i += 1

        # exponents can be either ^ or **
        if line[i] == "^" and not in_single_quote and not in_double_quote:
            line = line[:i] + "**" + line[i+1:]
            i += 2
            continue

        if line[i] == "[" and not (in_single_quote or in_double_quote):
            bracket_depth += 1
        elif line[i] == "]" and not (in_single_quote or in_double_quote):
            bracket_depth -= 1
            if bracket_depth == 0:
                seen_comma = False
                
        if line[i] == "'":
            if not in_double_quote and not in_single_quote:
                in_single_quote = True
                i += 1
                continue
            elif in_single_quote:
                in_single_quote = False
                i += 1
                continue
            
        if line[i] == '"':
            if not in_double_quote and not in_single_quote:
                in_double_quote = True
                i += 1
                continue
            elif in_double_quote:
                in_double_quote = False
                i += 1
                continue
            
            
        if not in_number and not in_single_quote and not in_double_quote \
               and line[i].isdigit() and \
               (i == 0 or (i > 0 and not isalphadigit_(line[i-1]))):
            in_number = True
            num_start = i
        i += 1

    if in_number:
        line, _ = wrap_num(i, line, is_real, num_start)

    # Time command like in MAGMA:
    if line.lstrip()[:5] == "time ":
        if line[-1:] == ";":
            line = line[:-1]
        line = '__time__=misc.cputime(); %s; print "Time:", misc.cputime(__time__), "seconds"'%\
               line.lstrip()[4:]

    return line



def compile_file(name,):
    name2 = "%s.syc"%name
    assert isinstance(name, str)
    if os.path.exists(name2) and \
           os.path.getmtime(name) <= os.path.getmtime(name2):
        # WE CAN'T safely DO this caching if there are any
        # recursive includes
        if open(name).read().find('load') == -1:
            return name2
    f = open(name)
    F = []
    A = f.readlines()
    i = 0
    while i < len(A):
        L = A[i]
        if L[:5] == "load ":
            try:
                name_load = eval(L[5:])
            except:
                pass
            else:
                if isinstance(name_load, str):
                    try:
                        name2_load = compile_file(name_load)
                        G = open(name2_load)
                    except IOError:
                        print "File '%s' not found, so skipping load"%name
                        i += 1
                        continue
                    else:
                        A = A[:i] + G.readlines() + A[i+1:]
                        continue
        F.append(custom_subst(L))
        i += 1
    # end while

    open(name2,"w").write("\n".join(F))
    return name2


attached = {}

def do_prefilter_paste(line, continuation):
    """
    Alternate prefilter for input.
    """
    global attached

    # This is so it's OK to have lots of blank space at the
    # beginning of any non-continuation line. 
    if not continuation:
        line = line.lstrip()
    line = line.rstrip()

    if not continuation and line == "":
        for F, tm in attached.iteritems():
            if os.path.exists(F) and os.path.getmtime(F) > tm:
                # Reload F.
                line = '%run -i "' + compile_file(F) + '"'
                attached[F] = os.path.getmtime(F)
                print "reloading '%s'"%F
                

    # Get rid of leading >>> so that pasting of examples from the documentation
    # works.  This is like MAGMA's SetLinePrompt(false).
    if not continuation:
        while True:
            strip = False
            if line[:3] == ">>>":
                line = line[3:].lstrip()
                strip = True
            elif line[:5] == "sage:":
                line = line[5:].lstrip()
                strip = True
            if not strip:
                break
            else:
                line = line.lstrip()
        
    # 'quit' alone on a line to quit.
    if line == 'quit':
        line = '%Quit'

    #################################################################
    # A "load" command, like \r file in PARI or load "file" in MAGMA
    #################################################################
    if line[:5] == 'load ':
        # The -i so the file is run with the same environment,
        # e.g., including the "from sage import *"
        try:
            name = eval(line[5:])
        except:
            pass
        else:
            if isinstance(name, str):
                line = '%run -i "' + compile_file(name) + '"'

    # IDEA:
    # I could make an attach command like in MAGMA.
    # The file to attach would be any file that could
    # be loaded.  At attach time it would be loaded as above.
    # It would also be put in a list that is a variable
    # with scope this interpreter.py file.  Each time
    # prefilter_paste is called and continuation is False,
    # check all attached file names and if any have changed,
    # compile the file, then use %run to load them again
    # as above.  This will be like the MAGMA attach, but
    # in some ways better.   It will be very nice for interactive
    # code development.
    if line[:7] == 'attach ':
        # The -i so the file is run with the same environment,
        # e.g., including the "from sage import *"
        try:
            name = eval(line[7:])
        except:
            pass
        else:
            if isinstance(name, str):
                if not os.path.exists(name):
                    print "No file '%s'"%name
                    line = ""
                else:
                    line = '%run -i "' + compile_file(name) + '"'
                    attached[name] = os.path.getmtime(name)

    line = custom_subst(line)
    return line
    

def prefilter_paste(self, line, continuation):
    """
    Alternate prefilter for input.
    """
    try:
        line2 = do_prefilter_paste(line, continuation)
    except None:
        print "WARNING: An error occured in the SAGE parser while"
        print "parsing the following line:"
        print line
        print "Please report this as a bug (include the output of typing '%hist')."
        line2 = line
        
    from IPython.iplib import InteractiveShell
    return InteractiveShell._prefilter(self, line2, continuation)
        

            
# Rebind this to be the new IPython prefilter:
from IPython.iplib import InteractiveShell
InteractiveShell.prefilter = prefilter_paste

# Clean up the namespace.
del InteractiveShell,prefilter_paste



