"""
Miscellaneous functions for timing, verbosity, etc.
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator, os, sys, time, weakref, random

try:
    SAGE_HOME = os.environ["SAGE_HOME"]
except KeyError:
    if os.name == 'nt':
        # windows 
        os.environ['SAGE_HOME'] = 'c:\\program files\\SAGE\\'
    else:
        SAGE_HOME = "%s/.sage/"%os.environ["HOME"]
    if not os.path.exists(SAGE_HOME):
        os.makedirs(SAGE_HOME)
    
LOGFILE = "%s/log/sage_log"%SAGE_HOME



#################################################################
# Functions to help with interfacing with CXX code that
# uses the GMP library
#################################################################
def to_gmp_hex(n):
    return hex(n).replace("L","").replace("0x","")

def cputime(t=0):
    """
    Return the time in CPU second since SAGE started, or with optional
    argument t, return the time since time t.
    INPUT:
        t -- (optional) float, time in CPU seconds
    OUTPUT:
        float -- time in CPU seconds
    """
    return time.clock() - t

def clock(cmd):
    t=cputime()
    eval(compile(cmd,"clock",'single'))
    return cputime(t)

LEVEL=0  # default
def verbose(mesg="", t=0, level=1, caller_name=None):
    if level>LEVEL:
        return cputime()
    if t != 0 and mesg=="":
        mesg = "Finished."
    # see recipe 14.7 in Python Cookbook
    if caller_name == None:
        caller_name = sys._getframe(1).f_code.co_name 
        if caller_name == "?: ":
            caller_name = ""
    s = "VERBOSE%s (%s): %s"%(level, caller_name, mesg)
    if t!=0:
        s = s + " (time = %s)"%cputime(t)
    print s
    #open(LOGFILE,"a").write(s+"\n")
    return cputime()

def todo(mesg=""):
    caller_name = sys._getframe(1).f_code.co_name
    raise NotImplementedError, "%s: todo -- %s"%(caller_name, mesg)

def set_verbose(level):
    """
    Set the verbosity level.

    INPUT:
       int level: an integer between 0 and 2, inclusive.
    OUTPUT:
       changes the state of the verbosity flag.
    """
    global LEVEL
    LEVEL = level

def get_verbose():
    global LEVEL
    return LEVEL


def version():
    import sage.version
    return 'SAGE Version %s, Export Date: %s'%(sage.version.version, sage.version.date)

def generic_cmp(x,y):
    if x<y:
        return -1
    elif x==y:
        return 0
    return 1
    
def mul(x, z=None):
    """
    Return the product of the elements in the list x.
    If optimal argument z is not given, start the product
    with the first element of the list, otherwise use z.

    >>> mul([1,2,34])
    68
    >>> mul([2,3], 5)
    30
    """
    if z == None:
        if len(x) == 0:
            return 1
        elif len(x) == 1:
            return x[0]
        return reduce(operator.mul, x[1:], x[0])
    else:
        if len(x) == 0:
            return z
        return reduce(operator.mul, x, z)

# alternative name for mul
prod = mul

def add(x, z=0):
    return reduce(operator.add, x, z)

def add_nonempty(x):
    return reduce(operator.add, x[1:], x[0])

def union(x, y=None):
    if y==None:
        if len(x) == 0:
            return []
        z = x[0]
        for i in range(1,len(x)):
            z = union(z,x[i])
        return z
    else:
        z = x + y
        z.sort()
        w=[]
        for i in range(len(z)):
            if i==len(z)-1 or z[i]!=z[i+1]:
                w.append(z[i])
        return w

def uniq(x):
    """
    Return a sublist of the elements in the list x that is sorted and
    is such that the entries in the sublist are unique.
    """
    assert isinstance(x,list), "Argument must be a list."
    d={}
    for a in x:
        d[a] = None
    return d.keys()
        
# Compute a nice string representation of a linear
# combination of some formal symbols
def repr_lincomb(symbols, lincomb):
    s = ""
    first = True
    i=0
    for c in lincomb:
        b = symbols[i]
        if c != 0:
            if c == 1:
                if not first:
                    coeff = " + "
                else:
                    coeff = ""
            elif c == -1:
                if not first:
                    coeff = " - "
                else:
                    coeff = "-"
            else:
                if not first:
                    coeff = " + %s*"%c
                else:
                    coeff = "%s*"%c
            s += "%s%s"%(coeff, b)
            first = False
        i += 1
    if first:
        s = "0"
#     else:
#         if isinstance(lincomb[0], (int, long, rings.Integer, rings.Rational)):
#             s = s.replace("+ -","- ")
    return s

    
#################################################################
# Generally useful
#################################################################

def parent(x):
    """Returns the parent of x, if x has a parent, otherwise
    return None."""
    try:
        return x.parent()
    except AttributeError:
        return None


def assert_attribute(x, attr, init=None):
    """
    If the object x has the attribute attr, do nothing.
    If not, set x.attr to init.
    """
    if x.__dict__.has_key(attr): return
    if attr[:2] == "__":
        z = str(x.__class__).split("'")
        if len(z) > 1:
            z = z[1]
        else:
            z = z[0]
        attr = "_" + z[len(x.__module__)+1:] + attr
    x.__dict__[attr] = init

#################################################################
# Useful but hard to classify
#################################################################

def random_sublist(X, s):
    """
    Return a random sublist of the list X where the chance
    of including a particular element is s.

    INPUT:
        X -- list
        s -- floating point number between 0 and 1
    OUTPUT:
        list
    """
    return [a for a in X if random.random() <= s]



def powerset(X):
    """
    Iterator over the *list* of all subsets of the iterable X,
    in no particular order.  Each list appears exactly once,
    up to order.

    INPUT:
        X -- an iterable
    OUTPUT:
        iterator of lists
    EXAMPLES:
        >>> list(powerset([1,2,3]))
        [[], [1], [2], [1, 2], [3], [1, 3], [2, 3], [1, 2, 3]]
        >>> for z in powerset([0,[1,2]]): print z,
        [] [0] [[1, 2]] [0, [1, 2]]

    The reason we return lists instead of sets is that the elements
    of sets must be hashable and many structures on which one wants
    the powerset consist of non-hashable objects.
    """
    pairs = [(2**i, x) for i, x in enumerate(X)]
    for n in xrange(2**len(pairs)):
        yield [x for m, x in pairs if m&n]

#################################################################
# Type checking
#################################################################
def typecheck(x, C, var="x"):
    if not isinstance(x, C):
        raise TypeError, "%s (=%s) must be of type %s."%(var,x,C)

#################################################################
# System information
#################################################################
def is_64bit():
    import sys
    # TODO: There's probably a better way to do this..
    return sys.maxint == 9223372036854775807

#################################################################
# This will likely eventually be useful.
#################################################################

class lazy_prop(object):
    def __init__(self, calculate_function):
        self._calculate = calculate_function
        self.__doc__ = calculate_function.__doc__

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        value = self._calculate(obj)
        setattr(obj, self._calculate.func_name, value)
        return value

def prop(f):
    return property(f, None, None, f.__doc__)


#################################################################
# For documentation tests.
#################################################################
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])

