from sage.rings.rings import Infinity, infinity
from sage.rings.integer import Integer
from sage.rings.integer_ring import Z
from sage.rings.rational_field import Q

class Cusp:
    
    """
    The set of cusps.

    A X{cusp} is either a rational number of infinity, i.e., an element
    of P1(Q).   A Cusp is stored as a pair (a,b), where gcd(a,b)=1
    and a,b are of type sage.rings.integers.Integer.
    """

    def __init__(self, a, b=1):
        """
        Create the cusp a/b in P^1(Q), where if b=0 it is the point at
        infinity.
        
        >>> Cusp(2,3)
        2/3
        >>> Cusp(3,6)
        1/2
        >>> Cusp(1,0)
        Infinity
        >>> Cusp(infinity)
        Infinity
        >>> Cusp(5)
        5
        >>> Cusp(Q("1/2"))             # rational number
        1/2
        >>> Cusp(1.5)
        Traceback (most recent call last):
        ...
        TypeError: Unable to coerce 1.5,1 to a Cusp
        """
        if isinstance(a, Infinity):
            self.__a = Z(1)
            self.__b = Z(0)
            return
        if a in Q:
            self.__a = a.numer()
            self.__b = a.denom()
            return
        
        if isinstance(a, (int,long)) and isinstance(b, (int,long)):
            a = Z(a)
            b = Z(b)
            
        if isinstance(a, Integer) and isinstance(b,Integer):
            g = a.gcd(b)
            self.__a = a/g
            self.__b = b/g
            return
        else:
            raise TypeError, "Unable to coerce %s,%s to a Cusp"%(a,b)

    def numerator(self):
        """Return the numerator of the cusp a/b.

        >>> x=Cusp(6,9); x
        2/3
        >>> x.numerator()
        2
        """
        return self.__a

    def denominator(self):
        """Return the denominator of the cusp a/b.

        >>> x=Cusp(6,9); x
        2/3
        >>> x.denominator()
        3
        """        
        return self.__b

    def __repr__(self):
        if self.__b.is_zero():
            return "Infinity"
        if self.__b != Z(1):
            return "%s/%s"%(self.__a,self.__b)
        else:
            return str(self.__a)

    def __neg__(self):
        return Cusp(-self.__a, self.__b)

    def is_gamma0_equiv(self, other, N):
        """
        Return whether self and other are equivalent modulo Gamma_0(N).
        
        Keyword arguments:
            other -- a cusp
            N -- an integer (specifies the group Gamma_0(N))

        Return True and an integer a if this cusp is equivalent to
        other modulo the action of Gamma_0(N).  Here a is the upper
        left entry of a matrix in Gamma_0(N) that sends self to other
        (or is it other way around??).  If the two cusps self and
        other are not congruent, then return (False, None).
        
        """
        N = Integer(N)
        u1 = self.__a
        v1 = self.__b
        u2 = other.__a
        v2 = other.__b
        if (u1,v1) != (Integer(0),Integer(1)):
            if v1 in [Integer(0),Integer(1)]:
                s1 = Integer(1)
            else:
                s1 = u1.inverse_mod(v1)
        else:
            s1 = 0
        if (u2,v2) != (Integer(0),Integer(1)):
            if v2 in [Integer(0),Integer(1)]:
                s2 = 1
            else:
                s2 = u2.inverse_mod(v2)
        else:
            s2 = 0
        g = (v1*v2).gcd(N)
        a = s1*v2 - s2*v1
        if a%g != Integer(0):
            return (False, None)

        # Now we know the cusps are equivalent.  Use the proof of Prop 2.2.3
        # of Cremona to find a matrix in Gamma_0(N) relating them.
        dum, s2, r2 = u2.xgcd(-v2)
        assert dum.is_one()
        dum, s1, r1 = u1.xgcd(-v1)
        assert dum.is_one()
        a = s1*v2 - s2*v1
        assert (a%g).is_zero()
        # solve x*v1*v2 + a = 0 (mod N).
        d,x0,y0 = (v1*v2).xgcd(N)          # x0*v1*v2 + y0*N = d = g.
        # so x0*v1*v2 - g = 0 (mod N)
        x = -x0 * (a/g)
        # now  x*v1*v2 + a = 0 (mod N)
        s1p = s1+x*v1
        return (True, (u2*s1p-r2*v1)%N)

    def is_gamma1_equiv(self, other, N):
        """
        """
        raise NotImplementedError

    def is_chi_equiv(self, other, chi):
        """
        """
        raise NotImplementedError        

    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
