"""
Dirichlet characters
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import weakref

import sage.rings.arith as arith
import sage.misc.errors as errors
import sage.misc.misc as misc
import sage.rings.rings as rings
import sage.structure.gens as gens

def TrivialCharacter(N, base_ring=rings.RationalField()):
    return DirichletGroup(N, base_ring)(1)

class DirichletCharacter:
    """
    A DirichletCharacter is the extension of a homomorphism
    (Z/NZ)^* --> R^*, for some ring R, to a map Z/NZ --> R,
    got by sending those x with (N,x)>1 to 0.

    Create with
        DirichletCharacter(parent, values_on_gens)
    INPUT:
        parent -- DirichletGroup, a group of Dirichlet characters
        values_on_gens -- list of ring elements, the values of the
                          Dirichlet character on the chosen generators
                          of (Z/NZ)^*.
    OUTPUT:
        DirichletCharacter -- a Dirichlet character
    """
    def __init__(self, parent, values_on_gens):
        assert len(values_on_gens) == len(parent.unit_gens()), \
               "wrong number of values(=%s) on unit gens (want %s)"%( \
               values_on_gens,len(parent.unit_gens()))
        R = parent.base_ring()
        self.__values_on_gens = [R(x) for x in values_on_gens]
        self.__parent = parent
        self.__modulus = parent.modulus()
        self.__values = None
        self.__is_trivial = None
        self.__order = None
        self.__conductor = None
        self.__decomp = None
        self.__is_odd = None

    def __eval_at_minus_one(self):
        """
        Efficiently evalute the character at -1 using knowledge of its
        order.   This is potentially much more efficient than computing
        the value of -1 directly using dlog and a large power of the
        image root of unity.

        We use the following.
        Proposition: Suppose eps is a character mod p^n, where p is a prime.
        Then eps(-1) = -1 if and only if 
               p = 2 and the factor of eps at 4 is nontrivial
           or
               p > 2 and 2 does not divide eulerphi(p^n)/order(eps).
        """
        try:
            return self.__value_at_minus_one
        except AttributeError:
            D = self.decomposition()
            val = 1
            for e in D:
                if e.modulus() % 2 == 0:
                    val *= e.__values_on_gens[0]
                elif (arith.eulerphi(e.parent().modulus()) / e.order()) % 2 != 0:
                    val *= -1
            self.__value_at_minus_one = val
        return self.__value_at_minus_one

    def __call__(self, m):
        m = int(m%self.__modulus)
        if self.__values == None:
            if m == self.__modulus - 1:
                return self.__eval_at_minus_one()
            self.values()
        return self.__values[m]

    def change_base(self, R):
        """
        Returns the base extension of self to the ring R.
        """
        G = self.parent().change_base(R)
        return G(self)

    def __cmp__(self, other):
        if not isinstance(other, DirichletCharacter):
            if isinstance(other,(int,rings.Integer)) and other == 1 and self.is_trivial():
                return True
            return -1
        return misc.generic_cmp(self.__values_on_gens, other.__values_on_gens)

    def __hash__(self):
        return hash(str(self.__values_on_gens))

    def __invert__(self):
        values_on_gens = [1/x for x in self.__values_on_gens]
        return DirichletCharacter(self.__parent, values_on_gens)

    def __mul__(self, other):
        if other==None:
            return None
        if self.__parent == other.parent():
            values_on_gens = [self.__values_on_gens[i]*other.__values_on_gens[i]
                              for i in range(len(self.__values_on_gens))]
            return DirichletCharacter(self.__parent, values_on_gens)
        raise NotImplementedError, \
              "multiplication of Dirichlet characters with different parents not implemented"

    def __pow__(self, exp):
        vals = [z**exp for z in self.__values_on_gens]
        return DirichletCharacter(self.__parent, vals)

    def __repr__(self):
        #return "DirichletCharacter(%s, %s)"%(self.parent(), self.__values_on_gens)
        #if self.is_trivial(): return "1"
        return ("%s"%self.__values_on_gens).replace(" ","")

    def __rmul__(self, left):
        if left==None:
            return None
        raise errors.UsageError("multiplication not defined.")

    def base_ring(self):
        """
        Returns the base ring of the parent of self.
        """
        return self.__parent.base_ring()

    def bernoulli(self, k):
        """
        Returns the generalized Bernoulli number B_{k,eps}.
        
        Let eps be this character (not necessarily primitive), and
        let k>=0 be an integer weight.  This function computes
        the (generalized) Bernoulli number B_{k,eps}, e.g., as defined 
        on page 44 of Diamond-Im:
        
          sum_{a=1}^{N} eps(a) t*e^(at)/(e^(N*t)-1) 
                 = sum_{k=0}^{\infty} B_{k,eps}/{k!}*t^k.
                 
        where N is the modulus of eps. 
        """
        try:
            self.__bernoulli
        except AttributeError:
            self.__bernoulli = {}
        if self.__bernoulli.has_key(k):
            return self.__bernoulli[k]
        N = self.modulus()
        K = self.base_ring()
        
        if N != 1 and self(-1) != K((-1)**k):
            return K(0)

        # The following is very slow, since poly rings over a very slow field
        # are very slow...
        if False:
            R = rings.PowerSeriesRing(K, variable="t")
            t = R.gen()
            prec = k+2   # todo: fix this
            F = sum([(self(a) * t * (a*t).exp(prec)) / ((N*t).exp(prec) - 1) \
                     for a in range(1,N)])
            self.__bernoulli[k] = F[k]*arith.factorial(k)
            return self.__bernoulli[k]


        # This is better since it computes the same thing, but requires
        # no arith in a poly ring over a number field.
        prec = k+2
        R = rings.PowerSeriesRing(rings.RationalField())
        t = R.gen()
        # g(t) = t/(e^{Nt}-1)
        g = t/((N*t).exp(prec) - 1)

        # h(n) = g(t)*e^{nt}
        h = [0] + [g * ((n*t).exp(prec)) for n in range(1,N+1)]   

        ber = sum([self(a)*h[a][k] for a in range(1,N+1)]) * arith.factorial(k)
        
        self.__bernoulli[k] = ber        
        return self.__bernoulli[k]

    def change_base_ring(self, R):
        """
        Tries to compute the Dirichlet character over R that has the same
        values as self, and if successful returns that character.  
        """
        if self.base_ring() == R:
            return self
        return DirichletGroup(self.modulus(),R)(self)

    def conductor(self):
        """
        Computes and returns the conductor of self.
        """
        if self.__conductor != None:
            return self.__conductor
        if self.modulus() == 1 or self.is_trivial():
            self.__conductor = 1
            return self.__conductor
        F = arith.factor(self.modulus())
        if len(F) > 1:
            self.__conductor = misc.mul([d.conductor() for d in self.decomposition()])
            return self.__conductor
        p = F[0][0]
        # When p is odd, and x =/= 1, the conductor is the smallest p**r such that
        #   Order(x) divides EulerPhi(p**r) = p**(r-1)*(p-1).
        # For a given r, whether or not the above divisiblity holds
        # depends only on the factor of p**(r-1) on the right hand side.
        # Since p-1 is coprime to p, this smallest r such that the
        # divisibility holds equals Valuation(Order(x),p)+1.
        self.__conductor = p**(arith.valuation(self.order(),p) + 1)
        if p == 2 and F[0][1] > 2 and self.__values_on_gens[1].order() != 1:
            self.__conductor *= 2;
        return self.__conductor

    def decomposition(self):
        """
        Return the decomposition of self as a product of Dirichlet characters 
        of prime power modulus, where the prime powers exactly divide the
        modulus of self.
        """
        if self.__decomp != None:
            return self.__decomp
        D = self.parent().decomposition()
        vals = [[z] for z in self.__values_on_gens]
        R = self.base_ring()
        if self.modulus()%8 == 0:   # 2 factors at 2.
            vals[0].append(vals[1][0])
            del vals[1]
        self.__decomp = [D[i](vals[i]) for i in range(len(D))]
        return self.__decomp

    def extend(self, M):
        """
        Returns the extension of self to a Dirichlet character
        modulo the multiple M of the modulus. which must also be a
        """
        if M % self.modulus() != 0:
            raise errors.UsageError("M(=%s) must be a multiple of the modulus(=%s)"%(M,self.modulus()))
        H = DirichletGroup(M, self.base_ring())
        return H(self)

    def is_even(self):
        """
        Return True if and only if self(-1) == 1.
        """
        return not self.is_odd()

    def is_odd(self):
        """
        Return True if and only if self(-1) != 1. 
        """
        if self.__is_odd != None:
            return self.__is_odd
        self.__is_odd = (self(-1) != self.base_ring()(1))
        return self.__is_odd

    def is_trivial(self):
        """
        Returns true if self is the trivial character, i.e., has
        order 1.
        """
        if self.__is_trivial != None:
            return self.__is_trivial
        self.__is_trivial = True
        R = self.base_ring()
        for x in self.__values_on_gens:
            if x != R(1):
                self.__is_trivial = False
                return False
        return self.__is_trivial

    def maximize_base_ring(self):
        """
        Let
        
               eps : (Z/N)^* ----> Q(zeta_n)
               
        be a Dirichlet character.  This function returns an equal
        Dirichlet character
        
               chi : (Z/N)^* ----> Q(zeta_m)
               
        where m is LCM(n, exponent of (Z/N)^*).

        """
        g = rings.IntegerModRing(self.modulus()).unit_group_exponent()
        if g == 1:
            g = 2
        z = self.base_ring().zeta()
        n = z.order()
        m = arith.LCM(g,n)
        if n == m:
            return self
        K = rings.CyclotomicField(m)
        return self.change_base_ring(K)

    def minimize_base_ring(self):
        """
        Return a Dirichlet character that equals this one, but over as
        small a subfield (or subring) of the base ring as possible.

        NOTE: This function is currently only implemented when the
        base ring is a number field.
        """

        if self.is_trivial(): return self
        if isinstance(self.base_ring(),rings.RationalField): return self
        
        if isinstance(self.base_ring(),rings.NumberField):
            if arith.eulerphi(self.order()) == self.base_ring().degree():
                return self
            K = rings.CyclotomicField(self.order())
            return self.change_base(K)

        raise NotImplementedError, "minimize_base_ring is currently " + \
              "only implemented when the base ring is a number field."

    def modulus(self):
        """
        Returns the modulus of self.
        """
        return self.__modulus

    def order(self):
        """
        Returns the order of self.
        """
        if self.__order != None:
            return self.__order
        self.__order = int(arith.LCM([x.order() for x in self.__values_on_gens]))
        return self.__order

    def parent(self):
        """
        Returns the parent of self.
        """
        return self.__parent

    def restrict(self, M):
        """
        Returns the restriction of self to a Dirichlet character
        modulo the divisor M of the modulus, which must also be a
        multiple of the conductor of self.
        """
        assert self.modulus()%M == 0, \
               "M(=%s) must divide the modulus(=%s)"%(M,self.modulus())
        assert M%self.conductor() == 0, \
               "conductor(=%s) must divide M(=%s)"%(self.conductor(),M)
        H = DirichletGroup(M, self.base_ring())
        return H(self)

    def values(self):
        """
        Returns a list of the values of self on each integer between 0
        and the modulus of self.
        """
        if self.__values != None:
            return self.__values
        # Build cache of all values of the Dirichlet character.
        # I'm going to do it this way, since in my app the modulus
        # is *always* small and we want to evaluate the character
        # *a* *lot*.
        R = self.parent().base_ring()
        zero = R(0)
        mod = self.__modulus
        x = [zero for _ in range(mod)]
        if self.is_trivial():  # easy special case
            for n in range(mod):
                # todo: optimization idea -- factor phi(n), then
                # "cross out" all multiples of each prime factor of phi(n).
                if arith.GCD(n,mod) == 1:
                    x[n] = 1
            self.__values = x
            return
        #end
        # we lift the gens to the int type here, since it is a lot faster
        gens = [z.lift() for z in self.parent().unit_gens()]
        exponents = [0 for _ in range(len(gens))]
        Z = self.parent().integers_mod()
        n = 1
        last = [g.order()-1 for g in self.parent().unit_gens()]
        stop = last[:]; stop[len(stop)-1] += 1
        value = R(1)
        val_on_gen = self.__values_on_gens
        only_int_vals = True
        tmp = []
        for z in val_on_gen:
            if z == R(1):
                tmp.append(1)
            elif z == R(-1):
                tmp.append(-1)
            else:
                only_int_vals = False
                break
        #end
        if only_int_vals:
            val_on_gen = tmp
            value = 1
        while exponents != stop:
            # record character value on n
            x[int(n)] = value
            # iterate:
            #   increase the exponent vector by 1,
            #   increase n accordingly, and increase value
            exponents[0] += 1   # inc exponent
            value *= val_on_gen[0]  # inc value
            n *= gens[0]
            n %= mod
            i = 0
            while i < len(exponents)-1 and exponents[i] > last[i]:
                exponents[i] = 0
                # now increment position i+1:
                exponents[i+1] += 1
                value *= val_on_gen[i+1]
                n *= gens[i+1]
                n %= mod
                i += 1
            #end
        #end
        if only_int_vals:
            x = [R(z) for z in x]
        self.__values = x
        return self.__values

    def values_on_gens(self):
        """
        Returns a list of the values of self on each of the minimal
        generators of (Z/NZ)^*, where N is the modulus of self.
        """
        return self.__values_on_gens

# This is so there is only one DirichletGroup with given parameters
# at a time.
_objsDirichletGroup = {}
class _uniqDirichletGroup(object):
    def __new__(cls, modulus, base_ring=None, zeta=None, zeta_order=None):
        key = (base_ring, modulus, zeta, zeta_order)
        if _objsDirichletGroup.has_key(key):
            x = _objsDirichletGroup[key]()
            if x != None: return x
        R = object.__new__(cls)
        _objsDirichletGroup[key] = weakref.ref(R)
        return R

class DirichletGroup(_uniqDirichletGroup, gens.Gens):
    """
    The group of Dirichlet character mod modulus with values in the
    subgroup <zeta> of the multiplicative group of the base_ring.  If
    the base_ring is omitted then we use Q(zeta_n), where n is the
    exponent of (Z/NZ)^*.  If zeta is omitted then we compute and use
    a maximal-order zeta in base_ring, if possible.

    Creation function:
        DirichletGroup(modulus, base_ring=None, zeta=None, zeta_order=None)
    INPUT:
        modulus -- int
        base_ring -- Ring (optional), where characters take their values
                     (should be an integral domain).
        zeta -- Element (optional), element of base_ring; zeta is a root of unity
        zeta_order -- int (optional), the order of zeta

    OUTPUT:
        DirichletGroup -- a group of Dirichlet characters.

    NOTES:
        Uniqueness -- If a group is created with the same parameters
        as another DirichletGroup still in memory, then the same group
        is returned instead of a new group defined by the same
        parameters.

    EXAMPLES:
    The default base ring is a cyclotomic field of order the exponent
    of (Z/NZ)^*.
        >>> DirichletGroup(20)
        Group of Dirichlet characters of modulus 20 over Cyclotomic Field of order 4 and degree 2
        
    We create the group of Dirichlet character mod 20 with values
    in the rational numbers:
        >>> G = DirichletGroup(20, rings.RationalField()); G
        Group of Dirichlet characters of modulus 20 over Rational Field
        >>> G.order()
        4
        >>> G.base_ring()
        Rational Field

    The elements of G print as lists giving the values of the
    character on the generators of (Z/NZ)^*:
        >>> list(G)
        [[1,1], [-1,1], [1,-1], [-1,-1]]

    Next we construct the group of Dirichlet character mod 20, but
    with values in Q(zeta_n):
        >>> G = DirichletGroup(20)
        >>> G.list()
        [[1,1], [-1,1], [1,zeta_4], [-1,zeta_4], [1,-1], [-1,-1], [1,-zeta_4], [-1,-zeta_4]]

    We next compute several invariants of G:
        >>> G.gens()
        [[-1,1], [1,zeta_4]]
        >>> G.unit_gens()
        [11, 17]
        >>> G.zeta()
        zeta_4
        >>> G.zeta_order()
        4

    In this example we create a Dirichlet character with values in a
    number field.  We have to give zeta, but not its order.
        >>> R = rings.PolynomialRing(rings.RationalField()); x = R.gen()
        >>> K = rings.NumberField(x**4 + 1); a = K.gen(0)
        >>> G = DirichletGroup(5, K, a); G
        Group of Dirichlet characters of modulus 5 over Number Field with defining polynomial x^4 + 1
        >>> G.list()
        [[1], [a^2], [-1], [-a^2]]
    """
    def __init__(self, modulus, base_ring=None, zeta=None, zeta_order=None):
        if base_ring == None:
            assert zeta==None and zeta_order==None, \
                   "zeta and zeta_order must be None if base_ring not specified."
            e = rings.IntegerModRing(modulus).unit_group_exponent()
            base_ring = rings.CyclotomicField(e)
        if zeta == None:
            if isinstance(base_ring(1),int) or isinstance(base_ring(1),long):
                zeta = base_ring(-1)
                zeta_order = 2
            else:
                zeta = base_ring.zeta()
                zeta_order = zeta.order()
        elif zeta_order == None:
            zeta_order = zeta.order()
        #end
        assert zeta != None, "Bug in creation of Dirichlet group."
        assert zeta_order != None, "Bug in creation of Dirichlet group."
        assert base_ring != None, "Bug in creation of Dirichlet group."
        self.__base_ring = base_ring
        self.__zeta = zeta
        self.__zeta_order = zeta_order
        self.__modulus = modulus
        self.__integers = rings.IntegerModRing(modulus)
        self.__gens = None
        self.__order = None
        self.__decomp = None

    def change_base(self, R):
        """
        Returns the Dirichlet group over R with the same modulus as self.
        """
        return DirichletGroup(self.modulus(), R)

    def __call__(self, x):
        if isinstance(x, (int,rings.Integer)) and x==1:
            R = self.base_ring()
            x = [R(1) for _ in self.unit_gens()]
        if isinstance(x, list):  # list of values on each unit generator
            return DirichletCharacter(self, x)
        elif isinstance(x, DirichletCharacter):  # coercion
            if x.parent() == self:
                return x
            assert self.modulus() % x.conductor() == 0, "conductor must divide modulus"
            a = []
            R = self.base_ring()
            for u in self.unit_gens():
                v = u.lift()
                # have to do this, since e.g., unit gens mod 11 are not units mod 22.
                while arith.GCD(x.modulus(),int(v)) != 1:   
                    v += self.modulus()
                a.append(R(x(v)))
            return self(a)
        raise NotImplementedError, "No coercion of %s into %s defined."%(x, self)

    def __cmp__(self, other):
        if not isinstance(other, DirichletGroup):
            return -1
        if other.modulus() < self.modulus():
            return -1
        if other.modulus() > self.modulus():
            return 1
        if other.base_ring() < self.base_ring():
            return -1
        if other.base_ring() > self.base_ring():
            return 1
        if other.__zeta != self.__zeta:
            return -1
        return 0

    def __len__(self):
        return self.order()

    def __repr__(self):
        return "Group of Dirichlet characters of modulus %s over %s"%\
               (self.modulus(),self.base_ring())

    def primitive_character(self):
        """
        Returns the primitive character associated to self.
        """
        return self.restrict(self.conductor())

    def base_ring(self):
        """
        Returns the base ring of self.
        """
        return self.__base_ring

    def decomposition(self):
        """
        Returns the Dirichlet groups of prime power modulus
        corresponding to primes dividing modulus.
        """
        if self.__decomp != None:
            return self.__decomp
        R = self.base_ring()
        self.__decomp = [DirichletGroup(p**r,R) for p, r \
                           in arith.factor(self.modulus())]
        return self.__decomp

    def galois_orbits(self):
        """
        Return a list of the Galois orbits of Dirichlet characters
        in self.

        The Galois group is the absolute Galois group of the prime
        subfield of Frac(R).
        """
        try:
            return self.__galois_orbits
        except AttributeError:
            pass
        L = list(self)
        G = []
        n = self.zeta_order()
        R = self.base_ring()
        p = R.characteristic()
        if p == 0:
            Auts = [e for e in xrange(2,n) if arith.GCD(e,n) == 1]
        else:
            # The nontrivial automorphisms in characteristic p are
            # k-th powering for 
            #         k = p, p^2, ..., p^(r-1),
            # where p^r = 1 (mod n), so r is the mult order of p modulo n.
            r = rings.IntegerModRing(n)(p).order() 
            Auts = [p**m for m in xrange(1,r)]
        while len(L) > 0:
            eps = L[0]
            orbit = [eps]
            del L[0]
            for s in Auts:
                chi = eps**s
                try:
                    L.remove(chi)
                    orbit.append(chi)
                except ValueError:
                    pass
            G.append(orbit)
        self.__galois_orbits = G
        return G

    def gen(self, n=0):
        n = int(n)
        g = self.gens()
        if n<0 or n>=len(g):
            raise IndexError, "n(=%s) must be between 0 and %s"%(n,len(g)-1)
        return g[n]

    def gens(self):
        """
        Returns generators of self.
        """
        if self.__gens != None:
            return self.__gens
        self.__gens = []
        ug = self.unit_gens()
        R = self.base_ring()
        one = [R(1) for i in range(len(ug))]
        zeta = self.zeta()
        ord = self.zeta_order()
        for i in range(len(ug)):
            vals = one[:]
            vals[i] = zeta**(ord/arith.GCD(ord,ug[i].order()))
            self.__gens.append(self(vals))
        return self.__gens

    def integers_mod(self):
        """
        Returns the group of integers Z/NZ where N is the
        modulus of self.
        """
        return self.__integers

    def modulus(self):
        """
        Returns the modulus of self.
        """
        return self.__modulus

    def ngens(self):
        """
        Returns the number of generators of self.
        """
        return len(self.gens())

    def order(self):
        if self.__order != None:
            return self.__order
        self.__order = 1
        for g in self.gens():
            self.__order *= int(g.order())
        return self.__order

    def unit_gens(self):
        """
        Returns the minimal generators for the units of (Z/NZ)^*,
        where N is the modulus of self.
        """
        return self.__integers.unit_gens()

    def zeta(self):
        """
        Returns the chosen root zeta of unity in the base ring R.
        """
        return self.__zeta

    def zeta_order(self):
        """
        Returns the order of the chosen root zeta of unity in the base
        ring R.
        """
        return self.__zeta_order


if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])

