"""
Hecke algebras and modules 
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import math
import sage.rings.arith as arith
import sage.modular.dims as dims
import sage.misc.misc as misc

class UniversalHeckeAlgebra:

    def __init__(self, weight):
        self.__weight = weight

    def __repr__(self):
        return "Universal Hecke algebra of weight %s"%self.__weight

class UniversalHeckeOperator:

    def __init__(self, parent, x):
        self.__parent = parent
        self.__x = x

    def __repr__(self):
        return "Universal Hecke operator %s in %s"%\
               (self.__parent, self.__x)


class HeckeAlgebra(UniversalHeckeAlgebra):

    def __init__(self, hecke_module, anemic=False):
        assert isinstance(hecke_module, HeckeModule)
        self.__hecke_module = hecke_module
        self.__bound = None

    def __repr__(self):
        return "Hecke algebra acting on %s"%self.__hecke_module

    def hecke_module(self):
        return self.__hecke_module

    def __call__(self, n):
        return HeckeOperator(self, n)

    def decomposition(self):
        return self.__hecke_module.decomposition()

    def generators(self):
        raise NotImplementedError

    def discriminant(self):
        raise NotImplementedError

    def bound(self):
        """
        An integer n so that the Hecke operators T_1,...,T_n generate
        the Hecke algebra as a Z-module.
        """
        if self.__bound != None:
            return self.__bound
        M = self.hecke_module()
        k = M.weight()
        N = M.level()
        self.__bound = int(math.ceil(k*dims.idxG0(N)/12.0))
        return self.__bound

    
class HeckeOperator(UniversalHeckeOperator):

    def __init__(self, parent, x):
        assert isinstance(parent, HeckeAlgebra)
        self.__parent = parent
        self.__x = x
        self.__matrix = None
        
    def __repr__(self):
        return "Hecke operator defined by %s on %s"%\
               (self.__x, self.__parent)
    
    def matrix(self):
        if self.__matrix != None:
            return self.__matrix
        self.__matrix = self.__parent.hecke_module().\
                        hecke_matrix(self.__x)
        return self.__matrix
    
    def charpoly(self):
        return self.matrix().charpoly()
    
    def fcp(self):
        return self.matrix().fcp()
    
    def trace(self):
        return self.matrix().trace()
    
    def det(self):
        return self.matrix().det()

        
class HeckeModule:
    
    def __init__(self):
        # true is anemic decomposition, and false is non-anemic one
        self.__decomposition = {True:None, False:None}  
        
    def dimension(self):
        raise NotImplementedError

    def set_factor_number(self, i):
        self.__factor_number = i

    def factor_number(self):
        try:
            return self.__factor_number
        except:
            return -1
    
    def level(self):
        raise NotImplementedError
    
    def weight(self):
        raise NotImplementedError
    
    def base_ring(self):
        raise NotImplementedError
    
    def hecke_matrix(self, n):
        raise NotImplementedError
    
    def vector_space(self):
        raise NotImplementedError

    def decomposition(self, anemic=True):
        if not isinstance(anemic, bool):
            raise TypeError, "anemic must be of type bool."
        if self.__decomposition[anemic] != None:
            return self.__decomposition[anemic]
        time = misc.verbose("Decomposing %s"%self)
        T = self.hecke_algebra()
        D = []
        U = [self.vector_space()]
        p = 2
        while len(U) > 0 and p <= T.bound():
            misc.verbose(mesg="p=%s"%p,t=time)
            if anemic:
                while arith.GCD(p, self.level()) != 1:
                    p = arith.next_prime(p)
            misc.verbose("Decomposition using p=%s"%p)
            t = T(p).matrix()
            Uprime = []
            for V in U:
                for W, is_irred in t.decompose_subspace(V):
                    if is_irred:
                        D.append(self(W))
                    else:
                        Uprime.append(W)
                    #end
                #end
            #end
            p = arith.next_prime(p)
            U = Uprime
        #end
        for W in U:
            D.append(self(W))
        for A in D:
            if anemic:
                A.__is_splittable_anemic = False
                A.__is_splittable = False
            else:
                A.__is_splittable = False
        self.__is_splittable = len(D) > 1
        if anemic:
            self.__is_splittable_anemic = len(D) > 1
            
        self.__decomposition[anemic] = D
        self.__decomposition[anemic].sort()
        for i in range(len(D)):
            self.__decomposition[anemic][i].set_factor_number(i)
        return self.__decomposition[anemic]
        
    def is_splittable(self):
        """
        Returns true if and only if only it is possible to split
        off a nontrivial generalized eigenspace of self as the
        kernel of some Hecke operator.
        """
        if not hasattr(self, "__is_splittable"):
            self.decomposition(anemic=False)
        return self.__is_splittable
        
    def is_splittable_anemic(self):
        """
        Returns true if and only if only it is possible to split
        off a nontrivial generalized eigenspace of self as the
        kernel of some Hecke operator of index coprime to the level.
        """
        if not hasattr(self,"__is_splittable_anemic"):
            self.decomposition(anemic=True)
        return self.__is_splittable_anemic

