"""
Boundary modular symbols.
"""

#*****************************************************************************
#       MANIN: Modular Forms Software, (c) William Stein, 2004
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

class BoundarySymbol:

    """
    Specific boundary modular symbol.
    """

    def __init__(self, parent, c, i=-1):
        """
        Create a boundary symbol.

        INPUT:
        CBoundarySymbols parent -- a space of boundary modular symbols
        c is either:
           -- a coefficient in the base ring of parent
           -- a vector over the base ring of parent
        int i -- if c is a coefficient, i gives the index into the
                 nonzero cusps list, except if c=0 in which case i
                 must be -1.
        """
        self.__parent = parent
        if isinstance(c, linalg.Vector):
            self.__vector = c
            return 
        self.__c = c   # coefficient
        self.__i = i   # index into list of nonzero symbols (-1 if c=0)
        n = max(self.__number_known(),i+1)
        self.__vector = linalg.VectorSpace(parent.base_field(),n)(0)
        if c != 0:
            self.__vector[i] = c

    def __number_known(self):
        """
        Returns the number of inequalivalent boundary classes that
        have been seen so far.
        """
        return len(self.__parent._BoundarySymbols__nonzero_classes)

    def __repr__(self):
        """
        Returns a string representation for self for printing purposes.
        """
        basis = ["[%s]"%x for x in self.__parent._BoundarySymbols__nonzero_classes]
        lincomb = self.__vector
        return misc.repr_lincomb(basis, lincomb)

    def __add__(self, other):
        if not isinstance(other, BoundarySymbol):
            return coerce.bin_op(self, other, operator.add)
        return BoundarySymbol(self.__parent, self.vector() + other.vector())

    def __radd__(self, left):
        if not isinstance(left, BoundarySymbol):
            return coerce.bin_op(left, self, operator.add)
        return BoundarySymbol(self.__parent, self.vector() + left.vector())

    def __sub__(self, other):
        if not isinstance(other, BoundarySymbol):
            return coerce.bin_op(self, other, "-")
        return BoundarySymbol(self.__parent, self.vector() - other.vector())

    def __rsub__(self, left):
        if not isinstance(left, BoundarySymbol):
            return coerce.bin_op(left, self, "-")
        return BoundarySymbol(self.__parent, left.vector() - self.vector())

    def __mul__(self, other):
        return BoundarySymbol(self.__parent, self.vector()*other)

    def __rmul__(self, left):
        return BoundarySymbol(self.__parent, self.vector()*left)

    def __neg__(self):
        return BoundarySymbol(self.__parent, -self.vector())

    def parent(self):
        return self.__parent

    def vector(self):
        """
        Returns the vector on the basis for boundary symbols that
        defines self.

        INPUT:
           BoundarySymbol self -- arbitrary boundary symbol

        OUTPUT:
           linalg.Vector -- underlying vector
        """
        n = self.__number_known()
        if len(self.__vector) < n:
            v = linalg.VectorSpace(self.__parent.base_field(),n)(0)
            for i in range(len(self.__vector)):
                if self.__vector[i] != 0:
                    v[i] = self.__vector[i]
            self.__vector = v
        return self.__vector

