"""nodoctest
Generic spaces of modular symbols.

All the spaces of modular symbols derive from this class.  This space
should never be directly constructed.
"""

#*****************************************************************************
#       MANIN: Modular Forms Software, (c) William Stein, 2004
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import math
import random
import weakref

import sage.linalg.vector_space as vector_space
import sage.misc.misc as misc
import sage.modular.dims as dims
import sage.modular.hecke as hecke
import sage.rings.rings as rings
import sage.structure.gens as gens
import sage.structure.function as function

import element


class ModularSymbolsSpace(hecke.HeckeModule, gens.Gens):
    def __init__(self, group, weight, character, sign, base_field):
        self.__group = group
        self.__weight = weight
        self.__character = character
        self.__sign = sign
        self.__base_field = base_field
        hecke.HeckeModule.__init__(self)
        
    def __cmp__(self, other):
        """
        Compare self and other.
        """
        if not isinstance(other, ModularSymbolsSpace):
            return -1
        if self.__group < other.__group:
            return -1
        elif self.__group > other.__group:
            return 1
        elif self.__weight < other.__weight:
            return -1
        elif self.__weight > other.__weight:
            return 1
        elif self.__character < other.__character:
            return -1
        elif self.__character > other.__character:
            return 1
        elif self.__sign < other.__sign:
            return -1
        elif self.__sign > other.__sign:
            return 1
        elif self.__base_field < other.__base_field:
            return -1
        elif self.__base_field > other.__base_field:
            return 1
        if self.is_ambient() and other.is_ambient():
            return 0
        if self.ambient_space() < other.ambient_space():
            return -1
        elif self.ambient_space() > other.ambient_space():
            return 1
        
        # ambients same and at most one of the two spaces is ambient
        if self.dimension() < other.dimension():
            return -1
        elif self.dimension() > other.dimension():
            return 1
            
        # both have same dimension and are in same ambient space, 
        # but neither is ambient
        if self.vector_space() == other.vector_space():
            return 0
               
        n = 2  # not equal -- distinguish using charpolys 
        while True:
            f = self.T(n).charpoly()
            g = other.T(n).charpoly()
            c = f.__cmp__(g)
            if c != 0:
                return c
            n += 1
            if n > self.sturm_bound():
                return 0
 
    def base_field(self):
        return self.__base_field

    def base_ring(self):
        return self.__base_field
    
    def basis(self):
        """
        Returns a basis for self.
        """
        # We do not cache, since that would create circular references,
        # which are best avoided.
        V = self.vector_space()
        return [element.ModularSymbolElement(self, V.gen(i)) for \
                        i in range(V.dimension())]

    def character(self):
        return self.__character
                    
    def dimension(self):
        raise NotImplementedError
        
    def eigenvalues(self):
        """
        Return a_n, psi, i, where a_n is the n-th eigenvalue, 
           psi : Z --> V is a map 
        and i:V-->K is an isomorphism of V with a
        number field (or Q), such that the composition iso(psi(n)) is
        the eigenvalue of the n-th Hecke operator acting on a fixed
        element of self (which must be new and non-splittable).

        The reason we separate psi and i, is that in general the values
        of psi take very little space to write down and are easier to
        compute, whereas the values i(psi(n)) can be huge (the idea to
        separate these maps was suggested to William Stein by John 
        Cremona).  The function an is simply the composite of i 
        and psi.
        
        INPUT:
            self -- a simple space of modular symbols
            
        OUTPUT:
            an  -- a map from positives integers to a field K.
            psi -- a map from the positive integers to
                   a vector space V
            i   -- a map from V to a field K.
            
        EXAMPLES:
            >>> import modsym
            >>> M = modsym.ModularSymbols(11)      # notice sign=0
            >>> S = M.cuspidal_subspace()
            >>> an, phi, i = S.eigenvalues()
            >>> phi(2)
            [-2, 0]
            >>> i(phi(2))
            -2
            >>> an(2)
            -2
            >>> E = EisensteinSubspace(M)
            >>> an, phi, i = E.eigenvalues()
            >>> an(2)
            3
            >>> M = ModularSymbols(43, sign=1)
            >>> D = M.decomposition()
            >>> E = D[0]
            >>> E.is_eisenstein()
            True
            >>> an, phi, i = E.eigenvalues()
            >>> an(7)
            8
            >>> C = D[2]; C.is_cuspidal()
            True
            >>> an, phi, i = C.eigenvalues()
            >>> phi(2)
            ...
            >>> i(phi(2))
            a
            >>> a2 = an(2); a2
            a
            >>> a2.parent()
            ...

        ALGORITHM:
        In the algorithm below, we let T denote the image of the Hecke
        algebra in End(M), which we compute efficiently as discussed 
        above.  If the sign of M is 0, replace M by the kernel of *-1:

        1. Verify that input is valid: i.e., that it is new and simple.

        2. Reduce to the prime case using the relations.  Remark:
           Multiplying matrices is O(n^(2+?)), whereas multiplying
           elements of a number field is O(n^2) (I think), so it would
           seem best to reduce to the prime case first, to
           avoid computing Hecke operators T_n, with n composite.  
           This is unclear though.

        3. Choose any (preferably "sparse", if possible) nonzero
           element v in M.

        4. Let psi be the map T --> M, given by psi(t) = v*t.

        5. Choose random elements t in T until we find one such that
           the iterates v, v*t, v*t^2, ..., etc. span M.  Let t be
           the one we find.  Such a t will exist (at least over an
           infinite field) because T is semisimple and M is assumed
           simple.

        6. Compute the characteristic polynomial f(x) of t, and let
           K=Q[x]/(f(x)) be the number field generated by a root 
           of f(x).

        7. Let e:M --> K be the map such that 
        
                            e(v*t^i)=x^i (mod f(x))
                            
           for i = 0,1,2,...,deg(f)-1. Since the t^i*v span M, this
           determines e.  To compute e, the main thing is to compute
           the change of basis matrix from the standard basis for M to
           the basis {v*t^i}.  This change of basis matrix is the inverse
           of the matrix whose rows are the v*t^i for i=0,...,deg(f)-1.

        8. Finally we have
        
                      a_n = e(psi(T_n)) = e(v*T_n)
        
           for Hecke operators T_n, where the a_n are eigenvalues.

        Remarks:
        (1) Space complexity:  The map psi is easy to compute and store
            space-wise, as are the values of psi on many T.
            The map e is likely to involve big numbers, since it's
            the iterates of a vector under a matrix, which can be
            quite large.   Also to compute e on an arbitrary element,
            we have to write it in terms of the iterates of v, which
            means inverting a matrix with large entries.  Thus, and
            this seems unavoidable, it takes a lot of space to store
            e and compute its values.  For many applications, e.g.,
            databases, it is better to store a matrix that defines
            e and the images under psi of lots of T_n.
            
        (2) How can we find a minimal collection of information from 
            which we can compute the map n |--> psi(T_n)?  Do we need 
            the whole modular symbols presentation?  No, we need only the
            image of each generating Manin symbol in M under
            projection.  The Hecke operators are then given by the
            standard Manin symbols formulas, where we reduce all
            resulting Manin symbols to their image in M.  
        """

        try:
            K, W, v0 = self.__eigenvalue_function
            def i(v):    return K(W.coordinates(v))
            def comp(x): return i(psi(x))
            def psi(n):  return self.hecke_operator(n)(v0).vector()
            return i, comp, psi
        except AttributeError:
            pass

        if self.dimension() == 0:
            raise ArithmeticError, "self must have positive dimension"
        if not self.is_new():
            raise ArithmeticError, "self must be new"
        #if self.is_splittable():
        #    raise ArithmeticError, "self must not be splittable"

        dim = self.dimension()
        if self.is_cuspidal() and self.sign() == 0: 
            dim /= 2

        pi = self.projection()
        M = self.ambient_space()
        e = -1
        for i in xrange(M.dimension()):
            v0 = M.gen(i)
            v = pi(v0)
            if v != 0:
                e = i
                break
        assert e != -1, "there is a bug in eigenvalue"

        m = 2
        t = M.hecke_operator(m)
        while True:
            X = [v0]
            for i in xrange(1,dim):
                X.append(X[len[X]-1]*t)
            assert len(X) == dim
            Y = [pi(w) for w in X]
            A = self.subspace(Y)
            if A.dimension() == dim:
                break
            m += 1
            t += random.choice([-1,1]) * self.hecke_operator(m).matrix()

        # NOT DONE -- and I'm tired....
        W = vector_space.VectorSpace(V)
        
        # 6. Charpoly of t
        if self.is_cuspidal() and self.sign() == 0:
            f = t.restrict(W).charpoly()
        else:
            f = t.charpoly()
        K = rings.NumberField(f)
        x = K.gen()

        # 7. Compute e
        # Change of basis matrix:
        # C = ~(linalg.Matrix(V).transpose())    # ~ is inverse of matrix
        
        # The following records the appropriate change of basis
        # matrix, and by using coordinates this works in general, even
        # if V is a proper subspace (e.g., when the sign is 0).  Also
        # any optimization to the coordinates function will improve
        # this automatically.
        
        # 8. Define eigenvalue function
        self.__eigenvalues = {}
        #def an(m):
        #    if self.__eigenvalues.has_key(m): return self.__eigenvalues[m]
        #    self.__eigenvalues[m] = e(psi(m))
        #    return self.__eigenvalues[m]
        self.__eigenvalue_function = (K, W, v0)
        def i(v):    return K(W.coordinates(v.vector()))
        def comp(x): return i(psi(x))
        def psi(n):  return v0*self.hecke_matrix(n)
        return comp, psi, i

    def eigenvalue(self, n):
        """
        Returns an eigenvalue of T_n acting on self, where
        self must be new and non-splittable.

        The eigenvalue is of the n-th Hecke operator acting on a fixed
        eigenvector of self, which is consistent between different
        calls of this function.
        """
        n = int(n)
        #if not isinstance(n, int):
        #    raise TypeError, "n must be an int"
        try:
            return self.__eigenvalues[n]
        except:
            self.__eigenvalues = {}
        try:
            an = self.__eigenvalue_function[0](n)
        except:
            self.eigenvalues()
            an = self.__eigenvalue_function[0](n)
        self.__eigenvalues[n] = an
        return an

    def gen(self, n):
        """
        The n-th generator of self.  
        INPUT:
            n -- int, an integer between 0 to the dimension minus 1.
        RETURN:
            -- The n-th basis element (ModularSymbol)
        NOTE:
            To avoid circular references, this function creates 
            the n-th basis ModularSymbol each time it is called from
            the n-th generator of the underlying vector space.
        """
        if n <0 or n>=self.dimension():
            raise IndexError
        return element.ModularSymbolElement(self, self.vector_space().gen(n))

    def group(self):
        """
        Returns the group of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           CongruenceSubgroup -- the congruence subgroup that this is a space
                              of modular symbols for.

        ALGORITHM:
           The group is recorded when this space is created.

        EXAMPLES:
        >>> m = ModularSymbols(20)
        >>> m.group()
        Gamma_0(20)
        """
        return self.__group

    def hecke_algebra(self, anemic=False):
        try:
            return self.__hecke_algebra
        except:
            self.__hecke_algebra = hecke.HeckeAlgebra(self, anemic)
            return self.__hecke_algebra
        
    def hecke_matrix(self, n):
        """
        The matrix of the n-th Hecke operator acting on the 
        basis for self.
        """
        assert n >= 1, "n must be positive."
        try:
            self._hecke_matrices
        except:
            self._hecke_matrices = {}
        if not self._hecke_matrices.has_key(n):
            self._compute_hecke_matrix(n)
        return self._hecke_matrices[n] 
    
    def hecke_operator(self, n):
        """
        Returns the n-th Hecke operator as a Matrix Function.
        INPUT:
           ModularSymbols self -- Hecke equivariant space of
                                   modular symbols
           int n -- an integer at least 1.
        OUTPUT:
           function.MatrixFunction -- the matrix function that
                    represents the n-th Hecke operator on the chosen
                    basis for self.
        """
        n = int(n)
        #if not isinstance(n, int):
        #    raise TypeError, "n must be an int"
            
        return function.MatrixFunction(self.hecke_matrix(n), self, self, \
                                       "Hecke operator T_%s on %s"%(n,self)) 
    
    def is_ambient(self):
        raise NotImplementedError
        
    def is_new(self):
        misc.verbose("** fake is_new")
        return True
        
    def is_cuspidal(self):
        misc.verbose("** fake is_cuspidal")
        return True
        
    def level(self):
        """
        Returns the level of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           int -- the level

        ALGORITHM:
           The level is recorded when self is created.

        EXAMPLES:
        >>> m = ModularSymbols(20)
        >>> m.level()
        20
        """
        return self.__group.level()
        
    def ngens(self):
        """
        The number of generators of self.
        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
        OUTPUT:
           int -- the number of generators, which is the same as the
                  dimension of self.
        ALGORITHM:
           Call the dimension function.
        EXAMPLES:
        >>> m = modsym.ModularSymbols(33)
        >>> m.ngens()
        9
        >>> m.dimension()
        9
        >>> modsym.ModularSymbols(100, weight=2, sign=1).ngens()
        18
        """
        return self.dimension()

    def projection(self):
        """
        Return the projection map from the ambient space to self.
        INPUT:
           ModularSymbols self -- A decomposition factor of an
                                   ambient space.
        OUTPUT:
           function.MatrixFunction -- The output is a matrix function,
                         which defines projection from the ambient
                         space to self.
        ALGORITHM:
           Let B be the matrix whose columns are got by concatenating
           together a basis for the factors of the ambient space.
           Then the projection matrix onto self is the submatrix of
           B^(-1) got from the rows corresponding to self, i.e., if
           the basis vectors for self appear as columns n through m of
           B, then the projection matrix is got from rows n through m
           of B^(-1).  This is because projection with respect to the
           B basis is just given by an m-n+1 row slice P of a diagonal
           matrix D with 1's in the n through m positions, so
           projection with respect to the standard basis is given by
           P*B^(-1), which is just rows n through m of B^(-1).
        """
        # Compute the Hecke-stable projection map pi from the ambient
        # space of M to M, which is easy because M is new and we
        # probably don't know M unless we've already decomposed the
        # new subspace.  (Computing the projection map is the same as
        # write the ambient space as a direct sum of M and its
        # Hecke-stable complement, which is the old subspace plus the
        # other new factors.)  With the projection map in hand, we
        # can compute Hecke operators directly on M fairly quickly
        # without having to compute them on the whole ambient space.
        # (I sort of learned this trick from Joe Wetherell, or at
        # least he was aware of it when I mentioned it to him in an
        # airport once.  It's also sort of like a trick Cremona uses
        # in his book for elliptic curves.)
        try:
            return self.__projection
        except:
            i = self.factor_number()
            if i == -1:
                raise NotImplementedError,\
                      "Computation of projection only implemented "+\
                      "for decomposition factors."
            A = self.ambient()
            B = A.__decomposition_matrix_inverse()
            n = sum([A[j].dimension() for j in range(i)])
            C = B.matrix_from_columns(range(n,n+self.dimension()))
            pi = function.MatrixFunction(C, A, self, \
                                         "Projection"%self)
            self.__projection = pi
            return self.__projection

    def qeigenform(self, prec):
        """
        Returns the q-expansion to precision prec of a newform
        associated to self, where self must be new, cuspidal,
        and non-splittable.
        """
        try:
            f = self.__qeigenform
        except:
            if not self.is_cuspidal():
                raise ArithmeticError, "self must be cuspidal."
            a2 = self.eigenvalue(2)
            R = rings.PowerSeriesRing(a2.parent(), "q")
            q = R.gen(0)
            f = q + a2*q**2 + rings.O(q**3)
        if f.prec() < prec:
            R = f.parent()
            ext = [self.eigenvalue(n) for n in range(f.prec(),prec)]
            f = R(list(f) + ext)
            self.__qeigenform = f.O(prec)
            return self.__qeigenform
        else:
            return f.O(prec)

    def sign(self):
        """
        Returns the sign of self.

        For efficiency reasons, it is often useful to compute in the
        (largest) quotient of modular symbols where the * involution
        acts as +1, or where it acts as -1.
        

        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
           
        OUTPUT:
           int -- the sign of self, either -1, 0, or 1.
                  -1 -- largest quotient where * acts as -1,
                  +1 -- largest quotient where * acts as +1, 
                   0 -- full space of modular symbols (no quotient).

        ALGORITHM:
           Call the dimension function.

        EXAMPLES:
        >>> m = ModularSymbols(33)
        >>> m.dimension()
        6
        >>> m.sign()
        1
        >>> m = ModularSymbols(33, sign=0)
        >>> m.sign()
        0
        >>> m.dimension()
        9
        >>> m = ModularSymbols(33, sign=-1)
        >>> m.sign()
        -1
        >>> m.dimension()
        3
        """
        return self.__sign
        
    def sturm_bound(self):
        """
        Returns the Sturm bound for this space of modular symbols.
        
        Returns a positive integer n = (k/12)*[SL_2(Z):Gamma_0(N)] 
        such that the Hecke operators  T1,...,Tn acting on cusp forms
        generate the Hecke algebra as a Z-module when
        the character is trivial or quadratic.  Otherwise, T1,...,Tn 
        generate the Hecke algebra at least as a Z[eps]-module, 
        where Z[eps] is the ring generated by the values of eps.
        Alternatively, this is a bound such that if two cusp forms
        associated to this space of modular symbols are congruent
        modulo (lambda, q^n), then they are congruent modulo lambda.
        
        REFERENCE: See the Agashe-Stein appendix to 
        Lario and Schoof's "Some computations with Hecke rings and 
        deformation rings", Experimental Math., 11 (2002), no. 2, 303-311.
        This result originated in the paper Sturm, "On the congruence
        of modular forms", Springer LNM 1240, 275--280, 1987.
        
        REMARK:
        Kevin Buzzard pointed out to me (William Stein) in Fall 2002 that the 
        above bound is fine for Gamma1 with character, as one sees by taking a 
        power of f.  More precisely, if f = 0 (mod p) for first s coefficients,
        then f^r = 0 (mod p) for first s*r coefficents.   Since weight of f^r 
        is r*weight(f), it follows that if s >= sturm bound for Gamma_0 at
        weight(f), then f^r has valuation large enough to be forced to be 0 
        at r*weight(f) by Sturm bound (which is valid if we choose r right).  
        Thus f = 0 (mod p).   Conclusion:  For Gamma_1 with fixed character, 
        the Sturm bound is *exactly* the same as for Gamma_0.    

        A key point is that we are finding Z[eps] generators for the Hecke 
        algebra here, not Z-generators.  So if one wants generators for the 
        Hecke algebra over Z, this bound is wrong.
        """
        
        try:
            return self.__sturm_bound
        except:
            self.__sturm_bound = \
                  int(math.ceil((self.weight() + dims.idxG0(self.level()))/12.0))
        return self.__sturm_bound
        
    def T(self, n):
        """
        Returns the n-th Hecke operator T_n.  This function is a
        synonym for for hecke_operator.
        """
        return self.hecke_operator(n)
        
    def vector_space(self):
        """
        Returns the underlying vector space of self.
        NOTES:
        This is always an ambient vector space over the base ring of
        dimension equal to the dimension of self.  For the embedded
        subspace of the ambient space's vector space, use the
        embedded_subspace function.
        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
        OUTPUT:
           VectorSpace
        ALGORITHM:
           Create and return the ambient vector space over the base
           field of dimension equal to the dimension of self.
        EXAMPLES:
        >>> import modsym
        >>> m = modsym.ModularSymbols(23)
        >>> m.vector_space()
        Full Vector space of degree 5 over Rational Field
        
        >.>> a = m[0]
        >.>> a.vector_space()
        Full Vector space of degree 1 over Rational Field
        """

        try:
            return self.__vector_space
        except:
            self.__vector_space = vector_space.VectorSpace(
                        self.base_field(),self.dimension())
        return self.__vector_space


    def weight(self):
        """
        Returns the weight of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           int -- the weight

        ALGORITHM:
           The weight is recorded when self is created.

        EXAMPLES:
        >>> m = ModularSymbols(20, weight=2)
        >>> m.weight()
        2
        """
        return self.__weight
    


if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
