"""nodoctest
Ambient spaces of modular symbols with a nontrivial Dirichlet character.

TODO: Maybe this should be deleted??!?
"""

#*****************************************************************************
#       MANIN: Modular Forms Software, (c) William Stein, 2004
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import weakref

import ambient
import sage.modular.congroup as congroup


class ModularSymbolsWithCharacter(ambient.ModularSymbolsAmbient):
    def __init__(self, weight, character, sign=0, base_field=None):
        """
        Initialize an ambient space of modular symbols with character.
        
        INPUT::
            weight -- int, the weight
            character -- 
            sign -- int, either -1, 0, or 1
            base_field -- rings.Ring (defaults to the rational numbers)

        """
        if base_field==None: base_field=character.base_field()
        if character.base_field() != base_field:
            character = character.change_base(base_field)
        ModularSymbolsSpace.__init__(self,  weight, \
                                     congroup.Gamma1(character.modulus()), \
                                     character, sign, base_field)
            





class BoundarySpaceWithCharacter:

    """
    Space of boundary modular symbols.
    """

    def __init__(self, group, weight=2, character=1, \
                 sign=0, base_field=rings.RationalField()):
        """
        Create the space of boundary symbols with given parameters.

        INPUT:
           BoundarySymbols self -- the space we are creating
           int level -- an integer that is >= 1
           int weight -- an integer that is >= 2
           dirichlet.DirichletCharacter character -- a Dirichlet
                   character whose modulus is a divisor of level.
           int sign -- either -1, 0, or 1
           rings.Ring base_field -- a field
        """
        assert weight >= 2, "weight must be at least 2."
        assert sign in [-1,0,1], "sign must be -1, 0, or 1."

        self.__group = group
        self.__level = group.level()
        self.__weight = weight
        self.__character = character
        self.__sign = sign
        self.__base_field = base_field
        self.__nonzero_classes = []
        self.__zero_classes = []  # classes that are killed by eps or sign relations
        
    def __repr__(self):
        return "Space of boundary symbols of weight %s, level %s, character %s and sign %s"%\
               (self.__weight, self.__level, self.__character, self.__sign)
    
    def group(self):
        return self.__group

    def level(self):
        return self.__group.level()
    
    def weight(self):
        return self.__weight
    
    def character(self):
        return self.__character
    
    def sign(self):
        return self.__sign
    
    def base_field(self):
        return self.__base_field
    
    def __call__(self, x):
        """
        Return the BoundarySymbol in self associated to x.

        INPUT:
           x -- an object of one of the following types:
                ModularSymbol 
                ManinSymbol 
                linalg.Vector
                Cusp, or something coercible to Cusp
                
           Minimal error checking is performed; e.g., this function
           does not check that the level of x is compatible with the
           level of self.
           
        OUTPUT:
           BoundarySymbol, or raises an errors.CoercionError upon
           failure.
           
        """
        if isinstance(x, ModularSymbol):
            # Return sum of the elements in v
            # Can't just use "sum" command, since it would
            # start with coercion of 0 into BoundarySymbol, which
            # would corresponding to the cusp 0 instead of the 0 element.
            return sum([c*self(m) for c, m in x.manin_symbol()],\
                       BoundarySymbol(self, 0))
        
        elif isinstance(x, ManinSymbol):
            g = x.lift_to_sl2z(self.__level)
            b = self(Cusp(g[0],g[2]))
            if x.i() == 0:
                return b - self(Cusp(g[1],g[3]))
            return b
        
        elif isinstance(x, linalg.Vector):
            return BoundarySymbol(self, x)
        
        if not isinstance(x,Cusp):
            try:
                x = Cusp(x)
            except:
                raise errors.CoercionError("No coercion of %s into %s defined."%(x,self))
        # Finally coerce a cusp into the class of the cusp.
        
        # Check if the cusp is in a nonzero cusp class we have seen
        # so far, and if so return the approproate boundary symbol.
        t, v = self.__is_known_nonzero_cusp(x)
        if t:
            return v
        # Check if the cusp is equivalent to a zero cusp class.
        if self.__is_known_zero_cusp(x):
            return BoundarySymbol(self, 0)
        # Check if the cusp class is killed by the character or sign relations,
        # and if so add it as a representative to the list of zero cusp classes.
        if self.__cusp_class_is_killed_by_character_relations(x) or \
              self.__cusp_class_is_killed_by_sign_relations(x):
            self.__zero_classes.append(x)
            return BoundarySymbol(self, 0)
        # The cusp is nonzero and not equivalent to any previous cusps.
        # Append the cusp to the end of the list of nonzero cusp classes.
        self.__nonzero_classes.append(x)
        return BoundarySymbol(self, 1, len(self.__nonzero_classes)-1)

    def __is_known_nonzero_cusp(self, x):
        for i in range(len(self.__nonzero_classes)):
            b = self.__nonzero_classes[i]
            # a is upper left corner of matrix sending [b] to [a].
            t, a = b.is_gamma0_equiv(x, self.__level)
            if t:   # x is equivalent to b
                if self.__character == 1:
                    c = 1
                else:
                    c = self.__character(a)
                return (True, BoundarySymbol(self, c, i))
            if self.__sign != 0:
                t, a = b.is_gamma0_equiv(-x, self.__level)
                if t:
                    if self.__character == 1:
                        c = self.__sign
                    else:
                        c = self.__character(a)
                    return (True, BoundarySymbol(self, c, i))
        return (False, None)

    def __is_known_zero_cusp(self, x):
        for b in self.__zero_classes:
            t, _ = b.is_gamma0_equiv(x, self.__level)
            if t:
                return True
        return False

    def __cusp_class_is_killed_by_character_relations(self, x):
        """Determine if the character relation kills this cusp class."""
        #if self.__character.is_trivial():
        #    return False
        eps = self.__character
        u = x.numerator()
        v = x.denominator()
        N = self.__level
        g = arith.GCD(N,v)
        z = N/g
        for j in range(g):
            alp = 1-j*z
            if arith.GCD(alp, N) == 1:
                if (v*(1-alp)) % N == 0 and (u*(1-alp)) % g == 0:
                    if eps(alp) != 1:
                        return True
        return False

    def __cusp_class_is_killed_by_sign_relations(self, x):
        """Determine if the sign relation kills this cusp class."""
        if self.__sign != 0:
            t, a = x.is_gamma0_equiv(-x,self.__level)
            if t and self.__character(a) != self.__sign:
                return True
        return False
        
