"""
py_pari -- Python interface to PARI C-library.

This module allows you to use Python *instead* of gp, with minimal
performance penalty, and all the advantages associated with Python
being a real programming language:

      - Easier memory management:
            Garbage collect all PARI objects that have no more Python references to them

                >>> garbage_collect()     
                (0, 0)
            
            Also, if the "PARI stack overflows", then the amount of
            memory used by PARI is automatically doubled, all the PARI
            objects are correctly updated, and your computation
            continues automatically.

            TODO: ** actually this currently kills all the current objects!
      
      - Exception handling
      
      - Class definitions
      
      - Object serialization of any PARI object whose string
        representation evaluates to the object, e.g.,
            >>> a = pari(5)
            >>> import pickle
            >>> s = pickle.dumps(a); 
            >>> z = pickle.loads(s);
            >>> print z
            5
            >>> type(z)
            <type 'py_pari.gen'>
            >>> a == z
            True
            >>> a is z
            False

WARNINGS:

  * Never import this module twice.  Your program will freeze.
  
  * PARI commands run at the C-level and cannot be interupted
    from the Python interpreter with ctrl-C until they are finished.
    
"""

import gc

cdef int PREC
PREC=32

cdef extern from "stdlib.h":
    ctypedef int size_t
    void free(void *ptr)

cdef extern from "string.h":
    void *memmove(void *dest, void *src, size_t n)
    void *memcpy(void *dest, void *src, size_t n)

cdef extern from "Python.h":
    void PyMem_Free(void *p)
    void* PyMem_Malloc(int)
    void* PyMem_Realloc(void *p, size_t n)

cdef extern from 'setjmp.h':
    struct __jmp_buf_tag:
        pass
    ctypedef __jmp_buf_tag jmp_buf
    int setjmp (jmp_buf __env)
    int longjmp (jmp_buf __env, int val)

ctypedef unsigned long ulong

cdef extern from 'pari/pari.h':
    ctypedef long* GEN
    ctypedef void entree   # fake -- not used
    ctypedef void FILE   # fake -- not used
    ctypedef void pariFILE   # fake -- not used
    ctypedef void va_list   # fake -- not used
    ctypedef void pari_timer   # fake -- not used    
    ctypedef void stackzone   # fake -- not used    

    # misc ...
    long    typ(GEN x)
    long    itos(GEN x)
    double  gtodouble(GEN x)
    GEN     stoi(long s)
    #GEN     dbltor(double s)
    # These types are actually an enum type, but I can't get Pyrex to "properly"
    # wrap enums.  It doesn't matter as long as they are treated as ints by pyrexc.
    extern int t_INT, t_REAL, t_INTMOD, t_FRAC, t_COMPLEX, t_PADIC, t_QUAD, \
           t_POLMOD, t_POL, t_SER, t_RFRAC, t_QFR, t_QFI, t_VEC, t_COL,  \
           t_MAT, t_LIST, t_STR, t_VECSMALL

    extern int BYTES_IN_LONG
    
    ctypedef unsigned long pari_sp
    pari_sp avma, bot, top, zonetop


    # Flx.c 
    GEN     Fl_to_Flx(ulong x, long sv)
    GEN     Flm_id(long n)
    GEN     Flm_to_FlxV(GEN x, long sv)
    GEN     Flm_to_FlxX(GEN x, long v,long w)
    GEN     Flm_to_ZM(GEN z)
    GEN     Flv_to_Flx(GEN x, long vs)
    GEN     Flv_to_ZC(GEN z)
    GEN     Flv_to_ZV(GEN z)
    GEN     Flv_polint(GEN xa, GEN ya, ulong p, long vs)
    GEN     Flv_roots_to_pol(GEN a, ulong p, long vs)
    GEN     Flx_Fl_mul(GEN y, ulong x, ulong p)
    GEN     Flx_to_Flv(GEN x, long N)
    GEN     Flx_to_ZX(GEN z)
    GEN     Flx_to_ZX_inplace(GEN z)
    GEN     Flx_add(GEN x, GEN y, ulong p)
    GEN     Flx_deriv(GEN z, ulong p)
    GEN     Flx_div_by_X_x(GEN a, ulong x, ulong p, ulong *rem)
    GEN     Flx_divrem(GEN x, GEN y, ulong p, GEN *pr)
    ulong   Flx_eval(GEN x, ulong y, ulong p)
    GEN     Flx_extgcd(GEN a, GEN b, ulong p, GEN *ptu, GEN *ptv)
    ulong   Flx_extresultant(GEN a, GEN b, ulong p, GEN *ptU, GEN *ptV)
    GEN     Flx_gcd(GEN a, GEN b, ulong p)
    GEN     Flx_gcd_i(GEN a, GEN b, ulong p)
    GEN     Flx_invmontgomery(GEN T, ulong p)
    int     Flx_is_squarefree(GEN z, ulong p)
    GEN     Flx_mul(GEN x, GEN y, ulong p)
    GEN     Flx_neg(GEN x, ulong p)
    GEN     Flx_neg_inplace(GEN x, ulong p)
    GEN     Flx_normalize(GEN z, ulong p)
    GEN     Flx_pow(GEN x, long n, ulong p)
    GEN     Flx_recip(GEN x)
    GEN     Flx_red(GEN z, ulong p)
    GEN     Flx_rem_montgomery(GEN x, GEN mg, GEN T, ulong p)
    GEN     Flx_rem(GEN x, GEN y, ulong p)
    GEN     Flx_renormalize(GEN x, long l)
    ulong   Flx_resultant(GEN a, GEN b, ulong p)
    GEN     Flx_shift(GEN a, long n)
    GEN     Flx_sqr(GEN x, ulong p)
    GEN     Flx_sub(GEN x, GEN y, ulong p)
    long    Flx_valuation(GEN x)
    GEN     FlxM_to_ZXM(GEN z)
    GEN     FlxV_Flv_innerprod(GEN V, GEN W, ulong p)
    GEN     FlxV_to_Flm(GEN v, long n)
    GEN     FlxV_to_ZXC(GEN x)
    GEN     FlxX_add(GEN P, GEN Q, ulong p)
    GEN     FlxX_shift(GEN a, long n)
    GEN     FlxX_to_Flm(GEN v, long n)
    GEN     FlxX_to_ZXX(GEN B)
    GEN     FlxYqQ_pow(GEN x, GEN n, GEN S, GEN T, ulong p)
    GEN     Flxq_inv(GEN x,GEN T,ulong p)
    GEN     Flxq_invsafe(GEN x, GEN T, ulong p)
    GEN     Flxq_mul(GEN y,GEN x,GEN T,ulong p)
    GEN     Flxq_pow(GEN x, GEN n, GEN T, ulong p)
    GEN     Flxq_powers(GEN x, long l, GEN T, ulong p)
    GEN     Flxq_sqr(GEN y,GEN T,ulong p)
    GEN     FlxqX_normalize(GEN z, GEN T, ulong p)
    GEN     FlxqX_Flxq_mul(GEN P, GEN U, GEN T, ulong p)
    GEN     FlxqX_red(GEN z, GEN T, ulong p)
    GEN     FlxqX_mul(GEN x, GEN y, GEN T, ulong p)
    GEN     FlxqX_safegcd(GEN P, GEN Q, GEN T, ulong p)
    GEN     FlxqX_sqr(GEN x, GEN T, ulong p)
    GEN     FlxqX_divrem(GEN x, GEN y, GEN T, ulong p, GEN *pr)
    GEN     FlxqXQ_pow(GEN x, GEN n, GEN S, GEN T, ulong p)
    GEN     Z_to_Flx(GEN x, ulong p, long v)
    GEN     ZM_to_Flm(GEN x, ulong p)
    GEN     ZV_to_Flv(GEN x, ulong p)
    GEN     ZX_to_Flx(GEN x, ulong p)
    GEN     ZXV_to_FlxV(GEN v, ulong p)
    GEN     ZXX_to_FlxX(GEN B, ulong p, long v)
    GEN     polx_Flx(long sv)
    GEN     zero_Flx(long sv)

     # alglin1.c 

    GEN     Flm_Flv_mul(GEN x, GEN y, ulong p)
    GEN     Flm_deplin(GEN x, ulong p)
    GEN     Flm_indexrank(GEN x, ulong p)
    GEN     Flm_inv(GEN x, ulong p)
    GEN     Flm_ker(GEN x, ulong p)
    GEN     Flm_ker_sp(GEN x, ulong p, long deplin)
    GEN     Flm_mul(GEN x, GEN y, ulong p)
    GEN     FlxqM_ker(GEN x, GEN T, ulong p)
    GEN     FpC_FpV_mul(GEN x, GEN y, GEN p)
    GEN     FpM_FpV_mul(GEN x, GEN y, GEN p)
    GEN     FpM_deplin(GEN x, GEN p)
    GEN     FpM_image(GEN x, GEN p)
    GEN     FpM_intersect(GEN x, GEN y, GEN p)
    GEN     FpM_inv(GEN x, GEN p)
    GEN     FpM_invimage(GEN m, GEN v, GEN p)
    GEN     FpM_ker(GEN x, GEN p)
    GEN     FpM_mul(GEN x, GEN y, GEN p)
    long    FpM_rank(GEN x, GEN p)
    GEN     FpM_indexrank(GEN x, GEN p)
    GEN     FpM_suppl(GEN x, GEN p)
    GEN     FqM_gauss(GEN a, GEN b, GEN T, GEN p)
    GEN     FqM_ker(GEN x, GEN T, GEN p)
    GEN     FqM_suppl(GEN x, GEN T, GEN p)
    GEN     QM_inv(GEN M, GEN dM)
    GEN     ZM_inv(GEN M, GEN dM)
    void    appendL(GEN x, GEN t)
    GEN     cget1(long l, long t)
    GEN     concat(GEN x, GEN y)
    GEN     concatsp(GEN x, GEN y)
    GEN     concatsp3(GEN x, GEN y, GEN z)
    GEN     deplin(GEN x)
    GEN     det(GEN a)
    GEN     det0(GEN a,long flag)
    GEN     det2(GEN a)
    GEN     dethnf(GEN x)
    GEN     dethnf_i(GEN mat)
    GEN     detint(GEN x)
    GEN     diagonal(GEN x)
    GEN     eigen(GEN x, long prec)
    GEN     extract(GEN x, GEN l)
    GEN     extract0(GEN x, GEN l1, GEN l2)
    GEN     gaddmat(GEN x, GEN y)
    GEN     gaddmat_i(GEN x, GEN y)
    GEN     gauss(GEN a, GEN b)
    GEN     gaussmodulo(GEN M, GEN D, GEN Y)
    GEN     gaussmodulo2(GEN M, GEN D, GEN Y)
    GEN     gscalcol(GEN x, long n)
    GEN     gscalcol_i(GEN x, long n)
    GEN     gscalcol_proto(GEN z, GEN myzero, long n)
    GEN     gscalmat(GEN x, long n)
    GEN     gscalsmat(long x, long n)
    GEN     gtomat(GEN x)
    GEN     gtrans(GEN x)
    GEN     gtrans_i(GEN x)
    int     hnfdivide(GEN A, GEN B)
    GEN     idmat(long n)
    GEN     idmat_intern(long n,GEN myun,GEN myzero)
    GEN     image(GEN x)
    GEN     image2(GEN x)
    GEN     imagecompl(GEN x)
    GEN     indexrank(GEN x)
    GEN     inverseimage(GEN mat, GEN y)
    long    isdiagonal(GEN x)
    long    isscalarmat(GEN x, GEN s)
    GEN     ker(GEN x)
    GEN     keri(GEN x)
    GEN     matextract(GEN x, GEN l1, GEN l2)
    GEN     matimage0(GEN x,long flag)
    GEN     matker0(GEN x, long flag)
    GEN     matmuldiagonal(GEN x, GEN d)
    GEN     matmultodiagonal(GEN x, GEN y)
    GEN     matsolvemod0(GEN M, GEN D, GEN Y,long flag)
    GEN     mattodiagonal(GEN m)
    GEN     mattodiagonal_i(GEN m)
    long    rank(GEN x)
    GEN     row(GEN A, long x1)
    GEN     row_i(GEN A, long x0, long x1, long x2)
    GEN     rowextract_i(GEN A, long x1, long x2)
    GEN     rowextract_ip(GEN A, GEN p, long x1, long x2)
    GEN     rowextract_p(GEN A, GEN p)
    GEN     sindexrank(GEN x)
    GEN     sum(GEN v, long a, long b)
    GEN     suppl(GEN x)
    GEN     vconcat(GEN A, GEN B)
    GEN     vec_ei(long n, long i)
    GEN     vec_Cei(long n, long i, GEN c)
    GEN     vecextract_i(GEN A, long y1, long y2)
    GEN     vecextract_ip(GEN A, GEN p, long y1, long y2)
    GEN     vecextract_p(GEN A, GEN p)

    # alglin2.c 

    GEN     QuickNormL1(GEN x,long prec)
    GEN     QuickNormL2(GEN x,long prec)
    GEN     ZM_to_zm(GEN z)
    GEN     adj(GEN x)
    GEN     assmat(GEN x)
    GEN     caract(GEN x, int v)
    GEN     caract2(GEN p, GEN x, int v)
    GEN     caradj(GEN x, long v, GEN *py)
    GEN     caradj0(GEN x, long v)
    GEN     carhess(GEN x, long v)
    GEN     charpoly0(GEN x, int v,long flag)
    GEN     conjvec(GEN x,long prec)
    GEN     gconj(GEN x)
    GEN     gnorm(GEN x)
    GEN     gnorml1(GEN x,long prec)
    GEN     gnorml2(GEN x)
    GEN     gtrace(GEN x)
    GEN     hess(GEN x)
    GEN     hnf(GEN x)
    GEN     hnfall(GEN x)
    GEN     hnflll(GEN x)
    GEN     hnflll_i(GEN A, GEN *ptB, int remove)
    GEN     hnfmod(GEN x, GEN detmat)
    GEN     hnfmodid(GEN x,GEN p)
    GEN     hnfmodidpart(GEN x, GEN p)
    GEN     hnfperm(GEN x)
    GEN     intersect(GEN x, GEN y)
    GEN     jacobi(GEN a, long prec)
    GEN     matfrobenius(GEN M, long flag)
    GEN     matrixqz(GEN x, GEN pp)
    GEN     matrixqz0(GEN x, GEN pp)
    GEN     matrixqz2(GEN x)
    GEN     matrixqz3(GEN x)
    GEN     signat(GEN a)
    GEN     smith(GEN x)
    GEN     smith2(GEN x)
    GEN     smithall(GEN x, GEN *ptU, GEN *ptV)
    GEN     smithclean(GEN z)
    GEN     sqred(GEN a)
    GEN     sqred1(GEN a)
    GEN     sqred1intern(GEN a)
    GEN     sqred3(GEN a)
    GEN     zm_to_ZM(GEN z)
    GEN     zx_to_ZX(GEN z)

    # anal.c 

    void    addhelp(entree *ep, char *s)
    void    delete_named_var(entree *ep)
    long    delete_var()
    entree  *fetch_named_var(char *s, int doerr)
    long    fetch_user_var(char *s)
    long    fetch_var()
    GEN     flisexpr(char *t)
    GEN     flisseq(char *t)
    void    freeep(entree *ep)
    entree  *gp_variable(char *s)
    long    hashvalue(char *s)
    entree* install(void *f, char *name, char *code)
    entree  *is_entry(char *s)
    void    kill0(entree *ep)
    GEN     lisexpr(char *t)
    GEN     lisseq(char *t)
    long    manage_var(long n, entree *ep)
    void    name_var(long n, char *s)
    GEN     readseq(char *c, int strict)
    GEN     strtoGENstr(char *s)
    GEN     type0(GEN x)

    # arith1.c 

    GEN     bestappr0(GEN x, GEN a, GEN b)
    GEN     bestappr(GEN x, GEN k)
    long    carrecomplet(GEN x, GEN *pt)
    long    cgcd(long a,long b)
    void    check_quaddisc(GEN x, long *s, long *r, char *f)
    GEN     chinese(GEN x, GEN y)
    GEN     chinois(GEN x, GEN y)
    GEN     classno2(GEN x)
    GEN     classno(GEN x)
    long    clcm(long a,long b)
    GEN     compimag(GEN x, GEN y)
    GEN     compimagraw(GEN x, GEN y)
    GEN     compraw(GEN x, GEN y)
    GEN     compreal(GEN x, GEN y)
    GEN     comprealraw(GEN x, GEN y)
    GEN     contfrac0(GEN x, GEN b, long flag)
    GEN     fibo(long n)
    GEN     Fp_gener_fact(GEN p, GEN fa)
    GEN     Fp_gener(GEN p)
    GEN     fundunit(GEN x)
    GEN     gboundcf(GEN x, long k)
    GEN     gcarrecomplet(GEN x, GEN *pt)
    GEN     gcarreparfait(GEN x)
    GEN     gcf2(GEN b, GEN x)
    GEN     gcf(GEN x)
    GEN     gener(GEN m)
    GEN     gfundunit(GEN x)
    GEN     ggener(GEN m)
    long    gisanypower(GEN x, GEN *pty)
    GEN     gisfundamental(GEN x)
    GEN     gisprime(GEN x, long flag)
    GEN     gispseudoprime(GEN x, long flag)
    GEN     gispsp(GEN x)
    GEN     gkrogs(GEN x, long y)
    GEN     gkronecker(GEN x, GEN y)
    GEN     gmillerrabin(GEN n, long k)
    GEN     gnextprime(GEN n)
    GEN     gprecprime(GEN n)
    GEN     gracine(GEN a)
    GEN     gregula(GEN x, long prec)
    GEN     hclassno(GEN x)
    long    hil0(GEN x, GEN y, GEN p)
    long    hil(GEN x, GEN y, GEN p)
    long    isanypower(GEN x, GEN *y)
    long    isfundamental(GEN x)
    long    ispower(GEN x, GEN k, GEN *pty)
    long    isprimeAPRCL(GEN N)
    long    isprime(GEN x)
    long    isprimeSelfridge(GEN x)
    long    ispseudoprime(GEN x, long flag)
    long    ispsp(GEN x)
    long    krois(GEN x, long y)
    long    kronecker(GEN x, GEN y)
    long    krosi(long s, GEN x)
    long    kross(long x, long y)
    long    krouu(ulong x, ulong y)
    GEN     lcmii(GEN a, GEN b)
    GEN     mpfact(long n)
    GEN     mpfactr(long n, long prec)
    GEN     Fp_inv(GEN a, GEN m)
    GEN     Fp_invsafe(GEN a, GEN m)
    GEN     Fp_pow(GEN a, GEN n, GEN m)
    GEN     Fp_sqrt(GEN a, GEN p)
    GEN     Fp_sqrtn(GEN a, GEN n, GEN p, GEN *zetan)
    GEN     nucomp(GEN x, GEN y, GEN l)
    GEN     nudupl(GEN x, GEN l)
    GEN     nupow(GEN x, GEN n)
    GEN     order(GEN x)
    GEN     pnqn(GEN x)
    GEN     powraw(GEN x, long n)
    GEN     powrealraw(GEN x, long n)
    GEN     primeform(GEN x, GEN p, long prec)
    GEN     Qfb0(GEN x, GEN y, GEN z, GEN d, long prec)
    GEN     qfbclassno0(GEN x,long flag)
    GEN     qfbimagsolvep(GEN Q, GEN n)
    GEN     qfbred0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
    GEN     qfbsolve(GEN Q, GEN n)
    GEN     qfi(GEN x, GEN y, GEN z)
    GEN     qfr(GEN x, GEN y, GEN z, GEN d)
    GEN     quaddisc(GEN x)
    GEN     racine(GEN a)
    GEN     redimag(GEN x)
    GEN     redreal(GEN x)
    GEN     redrealnod(GEN x, GEN isqrtD)
    GEN     regula(GEN x, long prec)
    GEN     rhoreal(GEN x)
    GEN     rhorealnod(GEN x, GEN isqrtD)
    GEN     seq_umul(ulong a, ulong b)
    GEN     sqcompimag(GEN x)
    GEN     sqcompreal(GEN x)
    ulong   Fl_sqrt(ulong a, ulong p)
    ulong   Fl_gener_fact(ulong p, GEN fa)
    ulong   Fl_gener(ulong p)
    GEN     znstar(GEN x)

    # arith2.c 

    GEN     addprimes(GEN primes)
    GEN     auxdecomp(GEN n, long all)
    long    bigomega(GEN n)
    GEN     binaire(GEN x)
    long    bittest(GEN x, long n)
    GEN     boundfact(GEN n, long lim)
    GEN     core(GEN n)
    GEN     corepartial(GEN n, long l)
    GEN     core0(GEN n,long flag)
    GEN     core2(GEN n)
    GEN     core2partial(GEN n, long l)
    GEN     coredisc(GEN n)
    GEN     coredisc0(GEN n,long flag)
    GEN     coredisc2(GEN n)
    GEN     decomp(GEN n)
    GEN     decomp_small(long n)
    GEN     divisors(GEN n)
    GEN     factorint(GEN n, long flag)
    GEN     gbigomega(GEN n)
    GEN     gbitand(GEN x, GEN y)
    GEN     gbitneg(GEN x, long n)
    GEN     gbitnegimply(GEN x, GEN y)
    GEN     gbitor(GEN x, GEN y)
    GEN     gbittest(GEN x, GEN n)
    GEN	gbittest3(GEN x, GEN n, long c)
    GEN     gbitxor(GEN x, GEN y)
    GEN     gboundfact(GEN n, long lim)
    GEN     gissquarefree(GEN x)
    GEN     gmu(GEN n)
    GEN     gnumbdiv(GEN n)
    GEN     gomega(GEN n)
    GEN     gphi(GEN n)
    GEN     gsumdiv(GEN n)
    GEN     gsumdivk(GEN n,long k)
    char*   initprimes(ulong maxnum)
    long    issquarefree(GEN x)
    ulong   maxprime()
    void    maxprime_check(ulong c)
    long    mu(GEN n)
    GEN     numbdiv(GEN n)
    long    omega(GEN n)
    GEN     phi(GEN n)
    GEN     prime(long n)
    GEN     primepi(GEN x)
    GEN     primes(long n)
    GEN     removeprimes(GEN primes)
    GEN     smallfact(GEN n)
    GEN     sumdiv(GEN n)
    GEN     sumdivk(GEN n,long k)

    # base1.c 

    GEN     bnfnewprec(GEN nf, long prec)
    GEN     bnrnewprec(GEN bnr, long prec)
    void    check_pol_int(GEN x, char *s)
    GEN     check_units(GEN x, char *f)
    void    checkbid(GEN bid)
    GEN     checkbnf(GEN bnf)
    GEN     checkbnf_discard(GEN bnf)
    void    checkbnr(GEN bnr)
    void    checkbnrgen(GEN bnr)
    void    checkid(GEN x, long N)
    GEN     checknf(GEN nf)
    GEN     checknfelt_mod(GEN nf, GEN x, char *s)
    void    checkprimeid(GEN bid)
    void    checkrnf(GEN rnf)
    GEN     galois(GEN x, long prec)
    GEN     galoisapply(GEN nf, GEN aut, GEN x)
    GEN     get_bnf(GEN x,int *t)
    GEN     get_bnfpol(GEN x, GEN *bnf, GEN *nf)
    GEN     get_nf(GEN x,int *t)
    GEN     get_nfpol(GEN x, GEN *nf)
    GEN     get_primeid(GEN x)
    GEN     glambdak(GEN nfz, GEN s, long prec)
    int     gpolcomp(GEN p1, GEN p2)
    GEN     gsmith(GEN x)
    GEN     gsmith2(GEN x)
    GEN     gzetak(GEN nfz, GEN s, long prec)
    GEN     gzetakall(GEN nfz, GEN s, long flag, long prec)
    GEN     initalg(GEN x, long prec)
    GEN     initalgred(GEN x, long prec)
    GEN     initalgred2(GEN x, long prec)
    GEN     initzeta(GEN pol, long prec)
    GEN     mathnf0(GEN x,long flag)
    GEN     matsnf0(GEN x,long flag)
    long    nf_get_r1(GEN nf)
    long    nf_get_r2(GEN nf)
    void    nf_get_sign(GEN nf, long *r1, long *r2)
    long    nfgetprec(GEN x)
    GEN     nfinit0(GEN x,long flag, long prec)
    GEN     nfnewprec(GEN nf, long prec)
    GEN     nfnewprec_i(GEN nf, long prec)
    GEN     rootsof1(GEN x)
    GEN     tschirnhaus(GEN x)

    # base2.c 

    GEN     allbase(GEN f, int flag, GEN *dx, GEN *dK, GEN *index, GEN *ptw)
    GEN     base(GEN x, GEN *y)
    GEN     base2(GEN x, GEN *y)
    void    checkmodpr(GEN modpr)
    GEN     compositum(GEN pol1, GEN pol2)
    GEN     compositum2(GEN pol1, GEN pol2)
    GEN     discf(GEN x)
    GEN     discf2(GEN x)
    GEN     factoredbase(GEN x, GEN p, GEN *y)
    GEN     factoreddiscf(GEN x, GEN p)
    GEN     ff_to_nf(GEN x, GEN modpr)
    GEN     fix_relative_pol(GEN nf, GEN x, int chk_lead)
    GEN     gcdpm(GEN f1,GEN f2,GEN pm)
    long    idealval(GEN nf,GEN ix,GEN vp)
    GEN     idealprodprime(GEN nf, GEN L)
    GEN     indexpartial(GEN P, GEN DP)
    GEN     modprX(GEN x, GEN nf,GEN modpr)
    GEN     modprX_lift(GEN x, GEN modpr)
    GEN     modprM(GEN z, GEN nf,GEN modpr)
    GEN     modprM_lift(GEN z, GEN modpr)
    GEN     nf_to_ff_init(GEN nf, GEN *pr, GEN *T, GEN *p)
    GEN     nf_to_ff(GEN nf, GEN x, GEN modpr)
    GEN     nfbasis(GEN x, GEN *y,long flag,GEN p)
    GEN     nfbasis0(GEN x,long flag,GEN p)
    GEN     nfdiscf0(GEN x,long flag, GEN p)
    GEN     nfreducemodideal(GEN nf,GEN x,GEN ideal)
    GEN     nfreducemodpr(GEN nf, GEN x, GEN modpr)
    GEN     polcompositum0(GEN pol1, GEN pol2,long flag)
    GEN     primedec(GEN nf,GEN p)
    GEN     rnfbasis(GEN bnf, GEN order)
    GEN     rnfdet(GEN nf, GEN order)
    GEN     rnfdet2(GEN nf, GEN A, GEN I)
    GEN     rnfdiscf(GEN nf, GEN pol)
    GEN     rnfequation(GEN nf, GEN pol2)
    GEN     rnfequation0(GEN nf, GEN pol2, long flall)
    GEN     rnfequation2(GEN nf, GEN pol)
    GEN     rnfhermitebasis(GEN bnf, GEN order)
    long    rnfisfree(GEN bnf, GEN order)
    GEN     rnflllgram(GEN nf, GEN pol, GEN order,long prec)
    GEN     rnfpolred(GEN nf, GEN pol, long prec)
    GEN     rnfpolredabs(GEN nf, GEN pol, long flag)
    GEN     rnfpseudobasis(GEN nf, GEN pol)
    GEN     rnfsimplifybasis(GEN bnf, GEN order)
    GEN     rnfsteinitz(GEN nf, GEN order)
    GEN     smallbase(GEN x, GEN *y)
    GEN     smalldiscf(GEN x)
    long    val_fact(ulong n, ulong p)
    GEN     zk_to_ff_init(GEN nf, GEN *pr, GEN *T, GEN *p)
    GEN     zk_to_ff(GEN x, GEN modpr)
    GEN     zkmodprinit(GEN nf, GEN pr)

    # base3.c 

    GEN     _algtobasis(GEN nf, GEN x)
    GEN     _algtobasis_cp(GEN nf, GEN x)
    GEN     _basistoalg(GEN nf, GEN x)
    GEN     algtobasis(GEN nf, GEN x)
    GEN     algtobasis_i(GEN nf,GEN x)
    GEN     arch_to_perm(GEN arch)
    GEN     basistoalg(GEN nf, GEN x)
    GEN     element_div(GEN nf, GEN x, GEN y)
    GEN     element_inv(GEN nf, GEN x)
    GEN     element_invmodideal(GEN nf, GEN x, GEN ideal)
    GEN     element_mul(GEN nf,GEN x,GEN y)
    GEN     element_muli(GEN nf,GEN x,GEN y)
    GEN     element_mulid(GEN nf, GEN x, long i)
    GEN     element_mulvec(GEN nf, GEN x, GEN v)
    GEN     element_pow(GEN nf,GEN x,GEN k)
    GEN     element_pow_mod_p(GEN nf, GEN x, GEN n, GEN p)
    GEN     element_powmodideal(GEN nf,GEN x,GEN k,GEN ideal)
    GEN     element_powmodidele(GEN nf,GEN x,GEN k,GEN idele,GEN structarch)
    GEN     element_sqr(GEN nf,GEN x)
    GEN     element_sqri(GEN nf, GEN x)
    long    element_val(GEN nf, GEN x, GEN vp)
    GEN     eltmul_get_table(GEN nf, GEN x)
    GEN     ideallist(GEN nf,long bound)
    GEN     ideallist0(GEN nf,long bound, long flag)
    GEN     ideallistarch(GEN nf, GEN list, GEN arch)
    GEN     ideallistarch0(GEN nf, GEN list, GEN arch,long flag)
    GEN     ideallistarchgen(GEN nf, GEN list, GEN arch)
    GEN     ideallistunit(GEN nf,long bound)
    GEN     ideallistunitarch(GEN bnf,GEN list,GEN arch)
    GEN     ideallistunitarchgen(GEN bnf,GEN list,GEN arch)
    GEN     ideallistunitgen(GEN nf,long bound)
    GEN     ideallistzstar(GEN nf,long bound)
    GEN     ideallistzstargen(GEN nf,long bound)
    GEN     idealstar0(GEN nf, GEN x,long flag)
    int     isnfscalar(GEN x)
    GEN     lllreducemodmatrix(GEN x,GEN y)
    GEN     nfdiveuc(GEN nf, GEN a, GEN b)
    GEN     nfdivrem(GEN nf, GEN a, GEN b)
    GEN     nfmod(GEN nf, GEN a, GEN b)
    GEN     nfreducemodidele(GEN nf,GEN g,GEN idele,GEN structarch)
    GEN     reducemodinvertible(GEN x, GEN y)
    GEN     reducemodmatrix(GEN x, GEN y)
    GEN     reducemodHNF(GEN x, GEN y, GEN *Q)
    GEN     set_sign_mod_idele(GEN nf, GEN x, GEN y, GEN idele, GEN sarch)
    GEN     smithrel(GEN H, GEN *newU, GEN *newUi)
    GEN     vecmodii(GEN a, GEN b)
    GEN     zarchstar(GEN nf,GEN x,GEN arch)
    GEN     zideallog(GEN nf,GEN x,GEN bigideal)
    GEN     zidealstar(GEN nf, GEN x)
    GEN     zidealstarinit(GEN nf, GEN x)
    GEN     zidealstarinitall(GEN nf, GEN x,long flun)
    GEN     zidealstarinitgen(GEN nf, GEN x)
    GEN     znlog(GEN x, GEN g)
    GEN     zsigne(GEN nf,GEN alpha,GEN arch)
    GEN     zsigns(GEN nf,GEN alpha)

    # base4.c 

    int     Z_ishnfall(GEN x)
    GEN     element_divmodpr(GEN nf, GEN x, GEN y, GEN modpr)
    GEN     element_invmodpr(GEN nf, GEN y, GEN modpr)
    GEN     element_mulmodpr(GEN nf, GEN x, GEN y, GEN modpr)
    GEN     element_powmodpr(GEN nf, GEN x, GEN k, GEN modpr)
    GEN     element_reduce(GEN nf, GEN x, GEN ideal)
    GEN     ideal_two_elt(GEN nf, GEN ix)
    GEN     ideal_two_elt0(GEN nf, GEN ix, GEN a)
    GEN     ideal_two_elt2(GEN nf, GEN x, GEN a)
    GEN     idealadd(GEN nf, GEN x, GEN y)
    GEN     idealaddmultoone(GEN nf, GEN list)
    GEN     idealaddtoone(GEN nf, GEN x, GEN y)
    GEN     idealaddtoone0(GEN nf, GEN x, GEN y)
    GEN     idealappr(GEN nf, GEN x)
    GEN     idealappr0(GEN nf, GEN x, long fl)
    GEN     idealapprfact(GEN nf, GEN x)
    GEN     idealchinese(GEN nf, GEN x, GEN y)
    GEN     idealcoprime(GEN nf, GEN x, GEN y)
    GEN     idealdiv(GEN nf, GEN x, GEN y)
    GEN     idealdiv0(GEN nf, GEN x, GEN y,long flag)
    GEN     idealdivexact(GEN nf, GEN x, GEN y)
    GEN     idealdivpowprime(GEN nf, GEN x, GEN vp, GEN n)
    GEN     idealmulpowprime(GEN nf, GEN x, GEN vp, GEN n)
    GEN     idealfactor(GEN nf, GEN x)
    GEN     idealhermite(GEN nf, GEN x)
    GEN     idealhermite2(GEN nf, GEN a, GEN b)
    GEN     idealhnf0(GEN nf, GEN a, GEN b)
    GEN     idealintersect(GEN nf, GEN x, GEN y)
    GEN     idealinv(GEN nf, GEN ix)
    GEN     ideallllred(GEN nf,GEN ix,GEN vdir,long prec)
    GEN     idealred_elt(GEN nf, GEN I)
    GEN     ideallllred_elt(GEN nf, GEN I, GEN vdir)
    GEN     idealmul(GEN nf, GEN ix, GEN iy)
    GEN     idealmul0(GEN nf, GEN ix, GEN iy, long flag, long prec)
    GEN     idealmulh(GEN nf, GEN ix, GEN iy)
    GEN     idealmulprime(GEN nf,GEN ix,GEN vp)
    GEN     idealmulred(GEN nf, GEN ix, GEN iy, long prec)
    GEN     idealnorm(GEN nf, GEN x)
    GEN     idealpow(GEN nf, GEN ix, GEN n)
    GEN     idealpow0(GEN nf, GEN ix, GEN n, long flag, long prec)
    GEN     idealpowred(GEN nf, GEN ix, GEN n,long prec)
    GEN     idealpows(GEN nf, GEN ideal, long iexp)
    long    idealtyp(GEN *ideal, GEN *arch)
    GEN     ideleaddone(GEN nf, GEN x, GEN idele)
    int     ishnfall(GEN x)
    int     isidentity(GEN x)
    GEN     hnfall_i(GEN A, GEN *ptB, long remove)
    long    isideal(GEN nf,GEN x)
    long    isinvector(GEN v, GEN x, long n)
    GEN     minideal(GEN nf,GEN ix,GEN vdir,long prec)
    GEN     mul_content(GEN cx, GEN cy)
    GEN     nfdetint(GEN nf,GEN pseudo)
    GEN     nfhermite(GEN nf, GEN x)
    GEN     nfhermitemod(GEN nf, GEN x, GEN detmat)
    GEN     nfkermodpr(GEN nf, GEN x, GEN modpr)
    GEN     nfmodprinit(GEN nf, GEN pr)
    GEN     nfsmith(GEN nf, GEN x)
    GEN     nfsolvemodpr(GEN nf, GEN a, GEN b, GEN modpr)
    GEN     prime_to_ideal(GEN nf, GEN vp)
    GEN     principalideal(GEN nf, GEN a)
    GEN     principalidele(GEN nf, GEN a, long prec)
    GEN     vecdiv(GEN x, GEN y)
    GEN     vecinv(GEN x)
    GEN     vecmul(GEN x, GEN y)
    GEN     vecpow(GEN x, GEN n)

    # base5.c 

    GEN     lift_to_pol(GEN x)
    GEN     matalgtobasis(GEN nf, GEN x)
    GEN     matbasistoalg(GEN nf, GEN x)
    GEN     rnfalgtobasis(GEN rnf, GEN x)
    GEN     rnfbasistoalg(GEN rnf, GEN x)
    GEN     rnfelementabstorel(GEN rnf, GEN x)
    GEN     rnfelementdown(GEN rnf, GEN x)
    GEN     rnfelementreltoabs(GEN rnf, GEN x)
    GEN     rnfelementup(GEN rnf, GEN x)
    GEN     rnfidealabstorel(GEN rnf, GEN x)
    GEN     rnfidealdown(GEN rnf, GEN x)
    GEN     rnfidealhermite(GEN rnf, GEN x)
    GEN     rnfidealmul(GEN rnf,GEN x,GEN y)
    GEN     rnfidealnormabs(GEN rnf, GEN x)
    GEN     rnfidealnormrel(GEN rnf, GEN x)
    GEN     rnfidealreltoabs(GEN rnf, GEN x)
    GEN     rnfidealtwoelement(GEN rnf,GEN x)
    GEN     rnfidealup(GEN rnf, GEN x)
    GEN     rnfinitalg(GEN nf,GEN pol,long prec)

    # bibli1.c 

    GEN     ZM_zc_mul(GEN x, GEN y)
    GEN     ZM_zm_mul(GEN x, GEN y)
    GEN     T2_from_embed(GEN x, long r1)
    GEN     algdep(GEN x, long n, long prec)
    GEN     algdep0(GEN x, long n, long bit,long prec)
    GEN     algdep2(GEN x, long n, long bit)
    GEN     factoredpolred(GEN x, GEN p)
    GEN     factoredpolred2(GEN x, GEN p)
    GEN     kerint(GEN x)
    GEN     kerint1(GEN x)
    GEN     lindep(GEN x, long prec)
    GEN     lindep0(GEN x, long flag,long prec)
    GEN     lindep2(GEN x, long bit)
    GEN     lll(GEN x, long prec)
    GEN     lllgen(GEN x)
    GEN     lllgram(GEN x, long prec)
    GEN     lllgramall(GEN x, long alpha, long flag)
    GEN     lllgramgen(GEN x)
    GEN     lllgramint(GEN x)
    GEN     lllgramintern(GEN x, long alpha, long flag, long prec)
    GEN     lllgramkerim(GEN x)
    GEN     lllgramkerimgen(GEN x)
    GEN     lllint(GEN x)
    GEN     lllint_i(GEN x, long alpha, int gram, GEN *h, GEN *ptfl, GEN *ptB)
    GEN     lllint_ip(GEN x, long alpha)
    GEN     lllintern(GEN x, long D, long flag, long prec)
    GEN     lllintpartial(GEN mat)
    GEN     lllintpartial_ip(GEN mat)
    GEN     lllkerim(GEN x)
    GEN     lllkerimgen(GEN x)
    GEN     matkerint0(GEN x,long flag)
    GEN     minim(GEN a, GEN borne, GEN stockmax)
    GEN     nf_get_LLL(GEN nf)
    GEN     qfrep0(GEN a, GEN borne, long flag)
    GEN     qfminim0(GEN a, GEN borne, GEN stockmax,long flag, long prec)
    GEN     minim2(GEN a, GEN borne, GEN stockmax)
    GEN     ordred(GEN x)
    GEN     perf(GEN a)
    GEN     polred(GEN x)
    GEN     polred0(GEN x, long flag, GEN p)
    GEN     polred2(GEN x)
    GEN     polredabs(GEN x)
    GEN     polredabs0(GEN x, long flag)
    GEN     polredabs2(GEN x)
    GEN     polredabsall(GEN x, long flun)
    GEN     qflll0(GEN x, long flag, long prec)
    GEN     qflllgram0(GEN x, long flag, long prec)
    GEN     smallpolred(GEN x)
    GEN     smallpolred2(GEN x)
    char    *stackmalloc(size_t N)
    GEN     zncoppersmith(GEN P0, GEN N, GEN X, GEN B)

    # bibli2.c 

    GEN     binome(GEN x, long k)
    int     cmp_pol(GEN x, GEN y)
    int     cmp_prime_ideal(GEN x, GEN y)
    int     cmp_prime_over_p(GEN x, GEN y)
    int     cmp_vecint(GEN x, GEN y)
    GEN     convol(GEN x, GEN y)
    GEN     cyclo(long n, long v)
    GEN     dirdiv(GEN x, GEN y)
    GEN     dirmul(GEN x, GEN y)
    GEN     dirzetak(GEN nf, GEN b)
    long    gen_search(GEN x, GEN y, int flag, int (*cmp)(GEN,GEN))
    GEN     gen_setminus(GEN set1, GEN set2, int (*cmp)(GEN,GEN))
    GEN     gen_sort(GEN x, int flag, int (*cmp)(GEN,GEN))
    GEN     genrand(GEN N)
    GEN     getheap()
    long    getrand()
    long    getstack()
    long    gettime()
    GEN     gprec(GEN x, long l)
    GEN     gprec_trunc(GEN x, long pr)
    GEN     gprec_w(GEN x, long pr)
    GEN     ggrando(GEN x, long n)
    GEN     gtor(GEN x, long l)
    GEN     gtoset(GEN x)
    GEN     indexlexsort(GEN x)
    GEN     indexsort(GEN x)
    GEN     laplace(GEN x)
    GEN     legendre(long n, long v)
    GEN     lexsort(GEN x)
    GEN     mathilbert(long n)
    GEN     matqpascal(long n, GEN q)
    GEN     modreverse_i(GEN a, GEN T)
    GEN     numtoperm(long n, GEN x)
    int     pari_compare_int(int *a,int *b)
    int     pari_compare_long(long *a,long *b)
    GEN     permtonum(GEN x)
    GEN     polint(GEN xa, GEN ya, GEN x, GEN *dy)
    GEN     polrecip(GEN x)
    GEN     polymodrecip(GEN x)
    GEN     roots_to_pol(GEN a, long v)
    GEN     setintersect(GEN x, GEN y)
    long    setisset(GEN x)
    GEN     setminus(GEN x, GEN y)
    long    setrand(long seed)
    long    setsearch(GEN x, GEN y, long flag)
    GEN     setunion(GEN x, GEN y)
    GEN     sindexlexsort(GEN x)
    GEN     sindexsort(GEN x)
    GEN     sort(GEN x)
    long    tablesearch(GEN T, GEN x, int (*cmp)(GEN,GEN))
    GEN     tayl(GEN x, long v, long precdl)
    GEN     tchebi(long n, long v)
    GEN     vecbinome(long n)
    GEN     vecsort(GEN x, GEN k)
    GEN     vecsort0(GEN x, GEN k, long flag)

    # buch1.c 

    GEN     buchimag(GEN D, GEN gcbach, GEN gcbach2, GEN gCO)
    GEN     buchreal(GEN D, GEN gsens, GEN gcbach, GEN gcbach2, GEN gRELSUP, long prec)
    GEN     cgetalloc(long t, size_t l)
    GEN     quadclassunit0(GEN x, long flag,GEN data, long prec)
    GEN     quadhilbert(GEN D, GEN flag, long prec)
    GEN     quadray(GEN bnf, GEN f, GEN flag, long prec)


    # buch2.c 

    GEN     bnfclassunit0(GEN P,long flag,GEN data,long prec)
    GEN     bnfinit0(GEN P,long flag,GEN data,long prec)
    GEN     bnfmake(GEN sbnf,long prec)
    GEN     buchall(GEN P, double bach, double bach2, long nbrelpid, long flun, long prec)
    GEN     buchfu(GEN bignf)
    GEN     check_and_build_obj(GEN S, int tag, GEN (*build)(GEN))
    GEN     classgrouponly(GEN P,GEN data,long prec)
    GEN     isprincipal(GEN bignf, GEN x)
    GEN     isprincipalall(GEN bignf, GEN x,long flall)
    GEN     isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag)
    GEN     isprincipalforce(GEN bignf,GEN x)
    GEN     isprincipalgen(GEN bignf, GEN x)
    GEN     isprincipalgenforce(GEN bignf,GEN x)
    GEN     isunit(GEN bignf, GEN x)
    GEN     quick_isprincipalgen(GEN bnf, GEN x)
    GEN     regulator(GEN P,GEN data,long prec)
    GEN     signunits(GEN bignf)
    GEN     smallbuchinit(GEN pol,double bach,double bach2,long nbrelpid,long prec)
    GEN     zsignunits(GEN bnf, GEN archp, int add_zu)

    # buch3.c 

    GEN     bnrclass0(GEN bignf, GEN ideal, long flag)
    GEN     bnrconductor(GEN arg0,GEN arg1,GEN arg2,GEN flag)
    GEN     bnrconductorofchar(GEN bnr,GEN chi)
    GEN     bnrdisc0(GEN arg0, GEN arg1, GEN arg2, long flag)
    GEN     bnrdisclist0(GEN bnf,GEN borne, GEN arch, long all)
    GEN     bnrinit0(GEN bignf,GEN ideal,long flag)
    long    bnrisconductor(GEN arg0,GEN arg1,GEN arg2)
    GEN     buchnarrow(GEN bignf)
    GEN     buchray(GEN bignf,GEN ideal)
    GEN     buchrayinit(GEN bignf,GEN ideal)
    GEN     buchrayinitgen(GEN bignf,GEN ideal)
    long    certifybuchall(GEN bnf)
    GEN     conductor(GEN bnr,GEN subgroup,long all)
    GEN     decodemodule(GEN nf, GEN fa)
    GEN     discrayabs(GEN bnr,GEN subgroup)
    GEN     discrayabscond(GEN bnr,GEN subgroup)
    GEN     discrayabslist(GEN bnf,GEN listes)
    GEN     discrayabslistarch(GEN bnf, GEN arch, long bound)
    GEN     discrayabslistlong(GEN bnf, long bound)
    GEN     discrayrel(GEN bnr,GEN subgroup)
    GEN     discrayrelcond(GEN bnr,GEN subgroup)
    GEN     idealmodidele(GEN bnr, GEN x)
    GEN     isprincipalray(GEN bignf, GEN x)
    GEN     isprincipalrayall(GEN bignf, GEN x,long flall)
    GEN     isprincipalraygen(GEN bignf, GEN x)
    GEN     rayclassno(GEN bignf,GEN ideal)
    GEN     rayclassnolist(GEN bnf,GEN listes)
    GEN     rnfconductor(GEN bnf, GEN polrel, long flag)
    GEN     rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
    GEN     rnfnormgroup(GEN bnr, GEN polrel)
    GEN     subgrouplist0(GEN bnr, GEN indexbound, long all)

    # buch4.c 

    GEN     bnfisnorm(GEN bnf,GEN x,long flag,long PREC)
    GEN     rnfisnorm(GEN S, GEN x, long flag)
    GEN     rnfisnorminit(GEN bnf, GEN relpol, int galois)
    GEN     bnfissunit(GEN bnf,GEN suni,GEN x)
    GEN     bnfsunit(GEN bnf,GEN s,long PREC)
    long    nfhilbert(GEN bnf,GEN a,GEN b)
    long    nfhilbert0(GEN bnf,GEN a,GEN b,GEN p)
    long    nfhilbertp(GEN bnf,GEN a,GEN b,GEN p)
    long    qpsoluble(GEN pol,GEN p)
    long    qpsolublenf(GEN bnf,GEN pol,GEN p)
    long    zpsoluble(GEN pol,GEN p)
    long    zpsolublenf(GEN bnf,GEN pol,GEN p)

    # elliptic.c 

    GEN     addell(GEN e, GEN z1, GEN z2)
    GEN     akell(GEN e, GEN n)
    GEN     anell(GEN e, long n)
    GEN     apell(GEN e, GEN p)
    GEN     apell2(GEN e, GEN p)
    GEN     bilhell(GEN e, GEN z1, GEN z2, long prec)
    GEN     coordch(GEN e, GEN ch)
    GEN     ellap0(GEN e, GEN p, long flag)
    GEN     elleisnum(GEN om, long k, long flag, long prec)
    GEN     elleta(GEN om, long prec)
    GEN     ellheight0(GEN e, GEN a, long flag,long prec)
    GEN     ellinit0(GEN x,long flag,long prec)
    GEN     ellminimalmodel(GEN E, GEN *ptv)
    long    ellrootno(GEN e, GEN p)
    GEN     ellsigma(GEN om, GEN z, long flag, long prec)
    GEN     elltors0(GEN e, long flag)
    GEN     ellwp0(GEN e, GEN z, long flag, long prec, long PREC)

    GEN     ellzeta(GEN om, GEN z, long prec)
    GEN     ghell(GEN e, GEN a, long prec)
    GEN     ghell2(GEN e, GEN a, long prec)
    GEN     globalreduction(GEN e1)
    GEN     initell(GEN x, long prec)
    GEN     localreduction(GEN e, GEN p1)
    GEN     lseriesell(GEN e, GEN s, GEN A, long prec)
    GEN     mathell(GEN e, GEN x, long prec)
    int     oncurve(GEN e, GEN z)
    GEN     ordell(GEN e, GEN x, long prec)
    GEN     orderell(GEN e, GEN p)
    GEN     pointch(GEN x, GEN ch)
    GEN     pointell(GEN e, GEN z, long prec)
    GEN     powell(GEN e, GEN z, GEN n)
    GEN     smallinitell(GEN x)
    GEN     subell(GEN e, GEN z1, GEN z2)
    GEN     taniyama(GEN e)
    GEN     torsell(GEN e)
    GEN     weipell(GEN e, long precdl)
    GEN     zell(GEN e, GEN z, long prec)

    # es.c 
    GEN     GENtocanonicalstr(GEN x)
    GEN     GENtoGENstr(GEN x)
    char*   GENtoTeXstr(GEN x)
    char*   GENtostr(GEN x)
    GEN     Str(GEN g)
    GEN     Strchr(GEN g)
    GEN     Strexpand(GEN g)
    GEN     Strtex(GEN g)
    void    brute(GEN g, char format, long dec)
    void    bruteall(GEN g, char f, long d, long sp)
    void    bruterr(GEN x,char format,long dec)
    void    error0(GEN g)
    void    etatpile(unsigned int n)
    char*   expand_tilde(char *s)
    int     file_is_binary(FILE *f)
    void    flusherr()
    void    fprintferr(char* pat, ...)
    void    killallfiles(int check)
    int     killfile(pariFILE *f)
    GEN     lisGEN(FILE *f)
    void    matbrute(GEN g, char format, long dec)
    pariFILE* newfile(FILE *f, char *name, int type)
    void    os_close(long fd)
    char*   os_getenv(char *s)
    long    os_open(char *s, int mode)
    void    os_read(long fd, char ch[], long s)
    void    (*os_signal(int sig, void (*f)(int)))(int)
    void    outbeaut(GEN x)
    void    outbeauterr(GEN x)
    void    outbrute(GEN x)
    void    outerr(GEN x)
    void    outmat(GEN x)
    void    output(GEN x)
    void    outsor(GEN x)
    void    outtex(GEN x)
    char*   pGENtostr(GEN g, long flag)
    void    pari_fclose(pariFILE *f)
    pariFILE*   pari_fopen(char *s, char *mode)
    pariFILE*   pari_safefopen(char *s, char *mode)
    char*   pari_strdup(char *s)
    char*   pari_strndup(char *s, long n)
    char*   pari_unique_filename(char *s)
    void    pari_unlink(char *s)
    void    pariflush()
    void    pariputc(char c)
    void    pariputs(char *s)
    void    pariputsf(char *format, ...)
    int     popinfile()
    #void    print(GEN g)   # syntax error
    void    print1(GEN g)
    void    printp(GEN g)
    void    printp1(GEN g)
    void    printtex(GEN g)
    GEN     readbin(char *name, FILE *f)
    void    sor(GEN g, char fo, long dd, long chmp)
    void    switchin(char *name)
    void    switchout(char *name)
    void    texe(GEN g, char format, long dec)
    pariFILE* try_pipe(char *cmd, int flag)
    char*   type_name(long t)
    void    voir(GEN x, long nb)
    #void    vpariputs(char* format, va_list args)
    void    write0(char *s, GEN g)
    void    write1(char *s, GEN g)
    void    writebin(char *name, GEN x)
    void    writetex(char *s, GEN g)

    # galconj.c 

    GEN     checkgal(GEN gal)
    GEN     galoisconj(GEN nf)
    GEN     galoisconj0(GEN nf, long flag, GEN d, long prec)
    GEN     galoisconj2(GEN x, long nbmax, long prec)
    GEN     galoisconj4(GEN T, GEN den, long flag, long karma)
    GEN     galoisexport(GEN gal, long format)
    GEN     galoisfixedfield(GEN gal, GEN v, long flag, long y)
    GEN     galoisidentify(GEN gal)
    GEN     galoisinit(GEN nf, GEN den, long karma)
    GEN     galoisisabelian(GEN gal, long flag)
    GEN     galoispermtopol(GEN gal, GEN perm)
    GEN     galoissubgroups(GEN G)
    GEN     galoissubfields(GEN G, long flag, long v)
    long    numberofconjugates(GEN T, long pdepart)
    GEN     vandermondeinverse(GEN L, GEN T, GEN den, GEN prep)
    # gen1.c 

    GEN     gadd(GEN x, GEN y)
    GEN     gaddsg(long x, GEN y)
    GEN     gdiv(GEN x, GEN y)
    GEN     gmul(GEN x, GEN y)
    GEN     gsqr(GEN x)
    GEN     gsub(GEN x, GEN y)

    # gen2.c 
    GEN     gopgs2(GEN (*f)(GEN, GEN), GEN y, long s)
    GEN     gopsg2(GEN (*f)(GEN, GEN), long s, GEN y)
    void    gopsgz(GEN (*f)(GEN, GEN), long s, GEN y, GEN z)
    int     opgs2(int (*f)(GEN, GEN), GEN y, long s)

    long    ZX_valuation(GEN x, GEN *Z)
    GEN     cgetimag()
    GEN     cgetp(GEN x)
    GEN     cvtop(GEN x, GEN p, long l)
    GEN     cvtop2(GEN x, GEN y)
    GEN     from_Kronecker(GEN z, GEN pol)
    GEN     gabs(GEN x, long prec)
    void    gaffect(GEN x, GEN y)
    void    gaffsg(long s, GEN x)
    GEN     gclone(GEN x)
    int     gcmp(GEN x, GEN y)
    int     gcmpsg(long x, GEN y)
    int     gcmp0(GEN x)
    int     gcmp1(GEN x)
    int     gcmp_1(GEN x)
    GEN     gcvtop(GEN x, GEN p, long r)
    int     gegal(GEN x, GEN y)
    int     gegalsg(long s, GEN x)
    long    gexpo(GEN x)
    long    ggval(GEN x, GEN p)
    long    glength(GEN x)
    GEN     gmax(GEN x, GEN y)
    GEN     gmin(GEN x, GEN y)
    GEN     gneg(GEN x)
    GEN     gneg_i(GEN x)
    GEN     greffe(GEN x, long l, long use_stack)
    int     gsigne(GEN x)
    GEN     gtofp(GEN z, long prec)
    GEN     gtolist(GEN x)
    long    gtolong(GEN x)
    int     lexcmp(GEN x, GEN y)
    GEN     listconcat(GEN list1, GEN list2)
    GEN     listcreate(long n)
    GEN     listinsert(GEN list, GEN object, long index)
    void    listkill(GEN list)
    GEN     listput(GEN list, GEN object, long index)
    GEN     listsort(GEN list, long flag)
    GEN     matsize(GEN x)
    GEN     normalize(GEN x)
    GEN     normalizepol(GEN x)
    GEN     normalizepol_i(GEN x, long lx)
    long    polvaluation(GEN x, GEN *z)
    long    polvaluation_inexact(GEN x, GEN *Z)
    GEN     pureimag(GEN x)
    GEN     quadtoc(GEN x, long l)
    long    sizedigit(GEN x)
    long    u_lval(ulong x, ulong p)
    long    u_lvalrem(ulong x, ulong p, ulong *py)
    long    u_pvalrem(ulong x, GEN p, ulong *py)
    GEN     to_Kronecker(GEN P, GEN Q)
    GEN     vecmax(GEN x)
    GEN     vecmin(GEN x)
    long    Z_lval(GEN n, ulong p)
    long    Z_lvalrem(GEN n, ulong p, GEN *py)
    long    z_pval(long n, GEN p)
    long    Z_pval(GEN n, GEN p)
    long    Z_pvalrem(GEN x, GEN p, GEN *py)

    # gen3.c 

    GEN     _toser(GEN x)
    GEN     Mod0(GEN x, GEN y,long flag)
    GEN     ceil_safe(GEN x)
    GEN     ceilr(GEN x)
    GEN     centerlift(GEN x)
    GEN     centerlift0(GEN x,long v)
    GEN     coefs_to_col(long n, ...)
    GEN     coefs_to_int(long n, ...)
    GEN     coefs_to_pol(long n, ...)
    GEN     coefs_to_vec(long n, ...)
    GEN     compo(GEN x, long n)
    GEN     deg1pol(GEN x1, GEN x0,long v)
    GEN     deg1pol_i(GEN x1, GEN x0,long v)
    long    degree(GEN x)
    GEN     denom(GEN x)
    GEN     deriv(GEN x, long v)
    GEN     derivpol(GEN x)
    GEN     derivser(GEN x)
    GEN     diviiround(GEN x, GEN y)
    GEN     divrem(GEN x, GEN y, long v)
    GEN     gand(GEN x, GEN y)
    GEN     gceil(GEN x)
    GEN     gcvtoi(GEN x, long *e)
    GEN     gdivent(GEN x, GEN y)
    GEN     gdiventres(GEN x, GEN y)
    GEN     gdivgs(GEN x, long s)
    GEN     gdivmod(GEN x, GEN y, GEN *pr)
    GEN     gdivround(GEN x, GEN y)
    GEN     geq(GEN x, GEN y)
    GEN     geval(GEN x)
    GEN     gfloor(GEN x)
    GEN     gfloor2n(GEN x, long s)
    GEN     gfrac(GEN x)
    GEN     gge(GEN x, GEN y)
    GEN     ggprecision(GEN x)
    GEN     ggt(GEN x, GEN y)
    GEN     gimag(GEN x)
    GEN     ginv(GEN x)
    GEN     gle(GEN x, GEN y)
    GEN     glt(GEN x, GEN y)
    GEN     gmod(GEN x, GEN y)
    GEN     gmodulcp(GEN x,GEN y)
    GEN     gmodulo(GEN x,GEN y)
    GEN     gmodulsg(long x, GEN y)
    GEN     gmodulss(long x, long y)
    GEN     gmul2n(GEN x, long n)
    GEN     gmulsg(long s, GEN y)
    GEN     gne(GEN x, GEN y)
    GEN     gnot(GEN x)
    GEN     gor(GEN x, GEN y)
    GEN     gpolvar(GEN y)
    long    gprecision(GEN x)
    GEN     gram_matrix(GEN M)
    GEN     greal(GEN x)
    GEN     grndtoi(GEN x, long *e)
    GEN     ground(GEN x)
    GEN     gshift(GEN x, long n)
    GEN     gsubst(GEN x, long v, GEN y)
    GEN     gsubstpol(GEN x, GEN v, GEN y)
    GEN     gtocol(GEN x)
    GEN     gtopoly(GEN x, long v)
    GEN     gtopolyrev(GEN x, long v)
    GEN     gtoser(GEN x, long v)
    GEN     gtovec(GEN x)
    GEN     gtovecsmall(GEN x)
    GEN     gtrunc(GEN x)
    int     gvar(GEN x)
    int     gvar2(GEN x)
    int     gvar9(GEN x)
    GEN     hqfeval(GEN q, GEN x)
    GEN     imag_i(GEN x)
    GEN     int2n(long n)
    GEN     integ(GEN x, long v)
    int     iscomplex(GEN x)
    int     isexactzero(GEN g)
    int     isexactzeroscalar(GEN g)
    int     isinexactreal(GEN x)
    long    isint(GEN n, long *ptk)
    int     ismonome(GEN x)
    GEN     lift(GEN x)
    GEN     lift0(GEN x,long v)
    GEN     lift_intern0(GEN x,long v)
    GEN     truncr(GEN x)
    GEN     mulmat_real(GEN x, GEN y)
    GEN     numer(GEN x)
    long    padicprec(GEN x, GEN p)
    GEN     polcoeff0(GEN x,long n,long v)
    GEN     polcoeff_i(GEN x, long n, long v)
    long    poldegree(GEN x,long v)
    GEN     poleval(GEN x, GEN y)
    GEN     pollead(GEN x,long v)
    long    precision(GEN x)
    GEN     precision0(GEN x,long n)
    GEN     qf_base_change(GEN q, GEN M, int flag)
    GEN     qfeval(GEN q, GEN x)
    GEN     real_i(GEN x)
    GEN     real2n(long n, long prec)
    GEN     recip(GEN x)
    GEN     round0(GEN x, GEN *pte)
    GEN     roundr(GEN x)
    GEN     scalarpol(GEN x, long v)
    GEN     scalarser(GEN x, long v, long prec)
    GEN     simplify(GEN x)
    GEN     simplify_i(GEN x)
    GEN     truecoeff(GEN x, long n)
    GEN     trunc0(GEN x, GEN *pte)
    GEN     u2toi(ulong a, ulong b)
    GEN     zerocol(long n)
    GEN     zeromat(long m, long n)
    GEN     zeropadic(GEN p, long e)
    GEN     zeropol(long v)
    GEN     zeroser(long v, long prec)
    GEN     zerovec(long n)

    # groupid.c

    long    group_ident(GEN G, GEN S)

    # ifactor1.c 

    long    BSW_psp(GEN N)
    long    is_357_power(GEN x, GEN *pt, ulong *mask)
    long    is_odd_power(GEN x, GEN *pt, ulong *curexp, ulong cutoffbits)
    long    millerrabin(GEN n, long k)
    GEN     nextprime(GEN n)
    GEN     plisprime(GEN N, long flag)
    GEN     precprime(GEN n)

    # init.c 

    long    TIMER(pari_timer *T)
    void    TIMERstart(pari_timer *T)
    long    allocatemoremem(size_t newsize)
    GEN     changevar(GEN x, GEN y)
    void    disable_dbg(long val)
    GEN     dummycopy(GEN x)
    void    err(long numerr, ...)
    #void   *err_catch(long errnum, jmp_buf *penv)
    void    err_leave(void **v)
    GEN     forcecopy(GEN x)
    void    freeall()
    GEN     gcopy(GEN x)
    GEN     gcopy_i(GEN x, long lx)
    GEN     gerepile(pari_sp ltop, pari_sp lbot, GEN q)
    void    gerepileall(pari_sp av, int n, ...)
    void    gerepileallsp(pari_sp av, pari_sp tetpil, int n, ...)
    void    gerepilecoeffs(pari_sp av, GEN x, int n)
    void    gerepilecoeffssp(pari_sp av, pari_sp tetpil, long *g, int n)
    GEN     gerepilecopy(pari_sp av, GEN x)
    void    gerepilemany(pari_sp av, GEN* g[], int n)
    void    gerepilemanysp(pari_sp av, pari_sp tetpil, GEN* g[], int n)
    GEN     gerepileupto(pari_sp av, GEN q)
    GEN     gerepileuptoint(pari_sp av, GEN q)
    GEN     gerepileuptoleaf(pari_sp av, GEN q)
    char*   gpmalloc(size_t bytes)
    char*   gprealloc(void *pointer,size_t size)
    void    gunclone(GEN x)
    void    killbloc(GEN x)
    void    msgTIMER(pari_timer *T, char *format, ...)
    void    msgtimer(char *format, ...)
    GEN     newbloc(long n)
    void    pari_init(size_t parisize, ulong maxprime)
    GEN     reorder(GEN x)
    void    stackdummy(GEN x, long l)
    GEN     stackify(GEN x)
    stackzone* switch_stack(stackzone *z, long n)
    long    taille(GEN x)
    long    taille2(GEN x)
    long    timer()
    long    timer2()

    # members.c 

    GEN     member_a1(GEN x)
    GEN     member_a2(GEN x)
    GEN     member_a3(GEN x)
    GEN     member_a4(GEN x)
    GEN     member_a6(GEN x)
    GEN     member_area(GEN x)
    GEN     member_b2(GEN x)
    GEN     member_b4(GEN x)
    GEN     member_b6(GEN x)
    GEN     member_b8(GEN x)
    GEN     member_bnf(GEN x)
    GEN     member_c4(GEN x)
    GEN     member_c6(GEN x)
    GEN     member_clgp(GEN x)
    GEN     member_codiff(GEN x)
    GEN     member_cyc(GEN clg)
    GEN     member_diff(GEN x)
    GEN     member_disc(GEN x)
    GEN     member_e(GEN x)
    GEN     member_eta(GEN x)
    GEN     member_f(GEN x)
    GEN     member_fu(GEN x)
    GEN     member_futu(GEN x)
    GEN     member_gen(GEN x)
    GEN     member_group(GEN x)
    GEN     member_index(GEN x)
    GEN     member_j(GEN x)
    GEN     member_mod(GEN x)
    GEN     member_nf(GEN x)
    GEN     member_no(GEN clg)
    GEN     member_omega(GEN x)
    GEN     member_orders(GEN x)
    GEN     member_p(GEN x)
    GEN     member_pol(GEN x)
    GEN     member_reg(GEN x)
    GEN     member_roots(GEN x)
    GEN     member_sign(GEN x)
    GEN     member_t2(GEN x)
    GEN     member_tate(GEN x)
    GEN     member_tufu(GEN x)
    GEN     member_tu(GEN x)
    GEN     member_w(GEN x)
    GEN     member_zk(GEN x)
    GEN     member_zkst(GEN bid)

    # mp.c 

    int     absi_cmp(GEN x, GEN y)
    int     absi_equal(GEN x, GEN y)
    int     absr_cmp(GEN x, GEN y)
    GEN     addii_sign(GEN x, long sx, GEN y, long sy)
    GEN     addir_sign(GEN x, long sx, GEN y, long sy)
    GEN     addrr_sign(GEN x, long sx, GEN y, long sy)
    GEN     addsi_sign(long x, GEN y, long sy)
    GEN     addsr(long x, GEN y)
    GEN     addss(long x, long y)
    void    affir(GEN x, GEN y)
    void    affrr(GEN x, GEN y)
    GEN     bezout(GEN a, GEN b, GEN *u, GEN *v)
    long    cbezout(long a,long b,long *uu,long *vv)
    void    cgiv(GEN x)
    int     cmpii(GEN x, GEN y)
    int     cmprr(GEN x, GEN y)
    int     cmpsi(long x, GEN y)
    int     cmpui(ulong x, GEN y)
    GEN     dbltor(double x)
    GEN     diviiexact(GEN x, GEN y)
    GEN     diviuexact(GEN x, ulong y)
    GEN     divir(GEN x, GEN y)
    GEN     divis(GEN y, long x)
    GEN     divis_rem(GEN x, long y, long *rem)
    GEN     diviu_rem(GEN y, ulong x, ulong *rem)
    GEN     divri(GEN x, GEN y)
    GEN     divrr(GEN x, GEN y)
    GEN     divrs(GEN x, long y)
    GEN     divsi(long x, GEN y)
    GEN     divsr(long x, GEN y)
    GEN     dvmdii(GEN x, GEN y, GEN *z)
    int     egalii(GEN x, GEN y)
    GEN     floorr(GEN x)
    GEN     gcdii(GEN x, GEN y)
    GEN     int_normalize(GEN x, long known_zero_words)
    int     invmod(GEN a, GEN b, GEN *res)
    ulong   invrev(ulong b)
    ulong   Fl_inv(ulong x, ulong p)
    GEN     ishiftr(GEN x, long n)
    GEN     modii(GEN x, GEN y)
    void    modiiz(GEN x, GEN y, GEN z)
    void    mpdivz(GEN x, GEN y, GEN z)
    GEN     mulii(GEN x, GEN y)
    GEN     mulir(GEN x, GEN y)
    GEN     mulrr(GEN x, GEN y)
    GEN     mulsi(long x, GEN y)
    GEN     mulsr(long x, GEN y)
    GEN     mulss(long x, long y)
    GEN     mului(ulong x, GEN y)
    GEN     mulur(ulong x, GEN y)
    GEN     muluu(ulong x, ulong y)
    long    pari_rand31()
    GEN     randomi(GEN x)
    int     ratlift(GEN x, GEN m, GEN *a, GEN *b, GEN amax, GEN bmax)
    GEN     resmod2n(GEN x, long n)
    double  rtodbl(GEN x)
    GEN     shifti(GEN x, long n)
    GEN     shifti3(GEN x, long n, long flag)
    GEN     sqri(GEN x)
    #define sqrti(x) sqrtremi((x),NULL)
    GEN     sqrtremi(GEN S, GEN *R)
    GEN     subsr(long x, GEN y)
    GEN     truedvmdii(GEN x, GEN y, GEN *z)
    ulong   umodiu(GEN y, ulong x)
    long    vals(ulong x)

    # nffactor.c 

    GEN     nffactor(GEN nf,GEN x)
    GEN     nffactormod(GEN nf,GEN pol,GEN pr)
    int     nfisgalois(GEN nf, GEN x)
    GEN     nfroots(GEN nf,GEN pol)
    GEN     rnfcharpoly(GEN nf,GEN T,GEN alpha,int n)
    GEN     rnfdedekind(GEN nf,GEN T,GEN pr)
    GEN     unifpol(GEN nf,GEN pol,long flag)

    # part.c 

    GEN     numbpart(GEN x)

    # perm.c 

    GEN     abelian_group(GEN G)
    GEN     bitvec_alloc(long n)
    void    bitvec_clear(GEN bitvec, long b)
    void    bitvec_set(GEN bitvec, long b)
    GEN     bitvec_shorten(GEN bitvec, long n)
    long    bitvec_test(GEN bitvec, long b)
    GEN     cyclicgroup(GEN g, long s)
    GEN     cyclicperm(long l, long d)
    GEN     cyc_pow(GEN cyc, long exp)
    GEN     cyc_pow_perm(GEN cyc, long exp)
    GEN     dicyclicgroup(GEN g1, GEN g2, long s1, long s2)
    GEN     group_abelianHNF(GEN G, GEN L)
    GEN     group_abelianSNF(GEN G, GEN L)
    long    group_domain(GEN G)
    GEN     group_elts(GEN G, long n)
    GEN     group_export(GEN G, long format)
    long    group_isA4S4(GEN G)
    long    group_isabelian(GEN G)
    GEN     group_leftcoset(GEN G, GEN g)
    long    group_order(GEN G)
    long    group_perm_normalize(GEN N, GEN g)
    GEN     group_quotient(GEN G, GEN H)
    GEN     group_rightcoset(GEN G, GEN g)
    GEN     group_subgroups(GEN G)
    GEN     groupelts_center(GEN S)
    GEN     groupelts_abelian_group(GEN S)
    int     perm_commute(GEN p, GEN q)
    GEN     perm_cycles(GEN v)
    GEN     perm_identity(long l)
    GEN     perm_inv(GEN x)
    GEN     perm_mul(GEN s, GEN t)
    long    perm_order(GEN perm)
    GEN     perm_pow(GEN perm, long exp)
    GEN     quotient_group(GEN C, GEN G)
    GEN     quotient_perm(GEN C, GEN p)
    GEN     vec_to_vecsmall(GEN z)
    GEN     vecperm_orbits(GEN v, long n)
    GEN     vecsmall_append(GEN V, long s)
    long    vecsmall_coincidence(GEN u, GEN v)
    GEN     vecsmall_concat(GEN u, GEN v)
    GEN     vecsmall_const(long n, long c)
    GEN     vecsmall_copy(GEN x)
    int     vecsmall_lexcmp(GEN x, GEN y)
    long    vecsmall_pack(GEN V, long base, long mod)
    int     vecsmall_prefixcmp(GEN x, GEN y)
    GEN     vecsmall_prepend(GEN V, long s)
    GEN     vecsmall_shorten(GEN v, long n)
    void    vecsmall_sort(GEN V)
    GEN     vecsmall_to_col(GEN z)
    GEN     vecsmall_to_vec(GEN z)
    GEN     vecsmall_uniq(GEN V)
    GEN     vecvecsmall_indexsort(GEN x)
    GEN     vecvecsmall_sort(GEN x)
    long    vecvecsmall_search(GEN x, GEN y, long flag)

    # polarit1.c 

    long    Flx_nbfact(GEN z, ulong p)
    long    Flx_nbroots(GEN f, ulong p)
    GEN     FpX_degfact(GEN f, GEN p)
    long    FpX_is_irred(GEN f, GEN p)
    long    FpX_is_squarefree(GEN f, GEN p)
    long    FpX_is_totally_split(GEN f, GEN p)
    GEN     FpX_factor(GEN f, GEN p)
    long    FpX_nbfact(GEN f, GEN p)
    long    FpX_nbroots(GEN f, GEN p)
    GEN     FpXQX_gcd(GEN P, GEN Q, GEN T, GEN p)
    GEN     FqX_factor(GEN x, GEN T, GEN p)
    GEN     FqX_gcd(GEN P, GEN Q, GEN T, GEN p)
    long    FqX_is_squarefree(GEN P, GEN T, GEN p)
    long    FqX_nbfact(GEN u, GEN T, GEN p)
    long    FqX_nbroots(GEN f, GEN T, GEN p)
    GEN     FpX_rand(long d, long v, GEN p)
    GEN     FpX_roots(GEN f, GEN p)
    GEN     apprgen(GEN f, GEN a)
    GEN     apprgen9(GEN f, GEN a)
    GEN     factcantor(GEN x, GEN p)
    GEN     factmod(GEN f, GEN p)
    GEN     factmod9(GEN f, GEN p, GEN a)
    GEN     factormod0(GEN f, GEN p,long flag)
    GEN     factorpadic0(GEN f,GEN p,long r,long flag)
    GEN     factorpadic2(GEN x, GEN p, long r)
    GEN     factorpadic4(GEN x, GEN p, long r)
    GEN     ffinit(GEN p,long n, long v)
    int     gdvd(GEN x, GEN y)
    long    hensel_lift_accel(long n, long *pmask)
    GEN     init_Fq(GEN p, long n, long v)
    GEN     padicsqrtnlift(GEN a, GEN n, GEN S, GEN p, long e)
    int     poldvd(GEN x, GEN y, GEN *z)
    GEN     poldivrem(GEN x, GEN y, GEN *pr)
    GEN     poldivrem_i(GEN x, GEN y, GEN *pr, long vx)
    GEN     rootmod(GEN f, GEN p)
    GEN     rootmod0(GEN f, GEN p,long flag)
    GEN     rootmod2(GEN f, GEN p)
    GEN     rootpadic(GEN f, GEN p, long r)
    GEN     rootpadicfast(GEN f, GEN p, long e)
    GEN     rootpadiclift(GEN T, GEN S, GEN q, long e)
    GEN     rootpadicliftroots(GEN f, GEN S, GEN q, long e)
    GEN     roots2(GEN pol,long PREC)
    GEN     rootsold(GEN x, long l)
    GEN     simplefactmod(GEN f, GEN p)

    # polarit2.c 

    GEN     Newton_exponents(long e)
    GEN     Q_content(GEN x)
    GEN     Q_denom(GEN x)
    GEN     Q_div_to_int(GEN x, GEN c)
    GEN     Q_muli_to_int(GEN x, GEN d)
    GEN     Q_primitive_part(GEN x, GEN *ptc)
    GEN     Q_primpart(GEN x)
    GEN     Q_remove_denom(GEN x, GEN *ptd)
    GEN     RgX_extgcd(GEN x, GEN y, GEN *U, GEN *V)
    GEN     ZX_squff(GEN f, GEN *ex)
    GEN     centermod(GEN x, GEN p)
    GEN     centermod_i(GEN x, GEN p, GEN ps2)
    GEN     centermodii(GEN x, GEN p, GEN po2)
    GEN     concat_factor(GEN f, GEN g)
    GEN     content(GEN x)
    GEN     deg1_from_roots(GEN L, long v)
    GEN     discsr(GEN x)
    GEN     divide_conquer_prod(GEN x, GEN (*mul)(GEN,GEN))
    GEN     factor(GEN x)
    GEN     factor0(GEN x,long flag)
    GEN     factorback(GEN fa,GEN nf)
    GEN     factorback0(GEN fa,GEN e, GEN nf)
    GEN     factorbackelt(GEN fa, GEN e, GEN nf)
    GEN     factpol(GEN x, long hint)
    GEN     gbezout(GEN x, GEN y, GEN *u, GEN *v)
    GEN     gcd0(GEN x, GEN y,long flag)
    GEN     gdeflate(GEN x, long v, long d)
    GEN     gdivexact(GEN x, GEN y)
    GEN     ggcd(GEN x, GEN y)
    GEN     ginvmod(GEN x, GEN y)
    GEN     gisirreducible(GEN x)
    GEN     glcm(GEN x, GEN y)
    GEN     glcm0(GEN x, GEN y)
    GEN     hensel_lift_fact(GEN pol, GEN Q, GEN T, GEN p, GEN pe, long e)
    GEN     leftright_pow(GEN,GEN,void*,GEN (*sqr)(void*,GEN),GEN (*mul)(void*,GEN,GEN))
    GEN     leftright_pow_u(GEN x, ulong n, void *data, GEN (*sqr)(void*,GEN), GEN (*mul)(void*,GEN,GEN))
    long    logint(GEN B, GEN y, GEN *ptq)
    GEN     newtonpoly(GEN x, GEN p)
    GEN     nfgcd(GEN P, GEN Q, GEN nf, GEN den)
    GEN     nfisincl(GEN a, GEN b)
    GEN     nfisisom(GEN a, GEN b)
    GEN     nfrootsQ(GEN x)
    GEN     poldeflate(GEN x0, long *m)
    GEN     poldeflate_i(GEN x0, long d)
    GEN     poldisc0(GEN x, long v)
    GEN     polfnf(GEN a, GEN t)
    GEN     polhensellift(GEN pol, GEN fct, GEN p, long exp)
    GEN     polinflate(GEN x0, long d)
    GEN     polresultant0(GEN x, GEN y,long v,long flag)
    GEN     polsym(GEN x, long n)
    GEN     primitive_part(GEN x, GEN *c)
    GEN     primpart(GEN x)
    GEN     pseudorem(GEN x, GEN y)
    GEN     quadgen(GEN x)
    GEN     quadpoly(GEN x)
    GEN     quadpoly0(GEN x, long v)
    GEN     reduceddiscsmith(GEN pol)
    GEN     resultant2(GEN x, GEN y)
    GEN     resultantducos(GEN x, GEN y)
    GEN     roots_from_deg1(GEN x)
    GEN     sort_factor(GEN y, int (*cmp)(GEN,GEN))
    GEN     sort_factor_gen(GEN y, int (*cmp)(GEN,GEN))
    GEN     sort_vecpol(GEN a, int (*cmp)(GEN,GEN))
    GEN     srgcd(GEN x, GEN y)
    long    sturmpart(GEN x, GEN a, GEN b)
    GEN     subresall(GEN u, GEN v, GEN *sol)
    GEN     subresext(GEN x, GEN y, GEN *U, GEN *V)
    GEN     sylvestermatrix(GEN x,GEN y)
    GEN     vecbezout(GEN x, GEN y)
    GEN     vecbezoutres(GEN x, GEN y)

    # polarit3.c 

    ulong   Fl_pow(ulong x, ulong n, ulong p)
    GEN     Fp_pows(GEN A, long k, GEN N)
    GEN     Fp_powu(GEN x, ulong k, GEN p)
    GEN     FpM_red(GEN z, GEN p)
    GEN     FpM_to_mod(GEN z, GEN p)
    GEN     FpV_polint(GEN xa, GEN ya, GEN p)
    GEN     FpV_red(GEN z, GEN p)
    GEN     FpV_roots_to_pol(GEN V, GEN p, long v)
    GEN     FpV_to_mod(GEN z, GEN p)
    GEN     FpX_FpXQ_compo(GEN f,GEN x,GEN T,GEN p)
    GEN     FpX_FpXQV_compo(GEN f,GEN x,GEN T,GEN p)
    GEN     FpX_Fp_add(GEN y,GEN x,GEN p)
    GEN     FpX_Fp_mul(GEN y,GEN x,GEN p)
    GEN     FpX_add(GEN x,GEN y,GEN p)
    GEN     FpX_center(GEN T,GEN mod)
    GEN     FpX_chinese_coprime(GEN x,GEN y,GEN Tx,GEN Ty,GEN Tz,GEN p)
    GEN     FpX_divrem(GEN x, GEN y, GEN p, GEN *pr)
    GEN     FpX_eval(GEN x,GEN y,GEN p)
    GEN     FpX_extgcd(GEN x, GEN y, GEN p, GEN *ptu, GEN *ptv)
    GEN     FpX_factorff_irred(GEN P, GEN Q, GEN l)
    void    FpX_ffintersect(GEN P,GEN Q,long n,GEN l,GEN *SP,GEN *SQ,GEN MA,GEN MB)
    GEN     FpX_ffisom(GEN P,GEN Q,GEN l)
    GEN     FpX_gcd(GEN x, GEN y, GEN p)
    GEN     FpX_mul(GEN x,GEN y,GEN p)
    GEN     FpX_neg(GEN x,GEN p)
    GEN     FpX_normalize(GEN z, GEN p)
    GEN     FpX_red(GEN z, GEN p)
    GEN     FpX_sqr(GEN x,GEN p)
    GEN     FpX_sub(GEN x,GEN y,GEN p)
    GEN     FpX_to_mod(GEN z, GEN p)
    GEN     FpXQ_charpoly(GEN x, GEN T, GEN p)
    GEN     FpXQ_div(GEN x,GEN y,GEN T,GEN p)
    GEN     FpXQ_ffisom_inv(GEN S,GEN Tp, GEN p)
    GEN     FpXQ_inv(GEN x,GEN T,GEN p)
    GEN     FpXQ_invsafe(GEN x, GEN T, GEN p)
    GEN     FpXQ_matrix_pow(long n, long m, GEN y, GEN P, GEN l)
    GEN     FpXQ_minpoly(GEN x, GEN T, GEN p)
    GEN     FpXQ_mul(GEN y,GEN x,GEN T,GEN p)
    GEN     FpXQ_pow(GEN x, GEN n, GEN T, GEN p)
    GEN     FpXQ_powers(GEN x, long l, GEN T, GEN p)
    GEN     FpXQ_sqr(GEN y, GEN T, GEN p)
    GEN     FpXQ_sqrtn(GEN a, GEN n, GEN T, GEN p, GEN *zetan)
    GEN     FpXQX_mul(GEN x, GEN y, GEN T, GEN p)
    GEN     FpXQX_red(GEN z, GEN T, GEN p)
    GEN     FpXQX_sqr(GEN x, GEN T, GEN p)
    GEN     FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
    GEN     FpXQX_divrem(GEN x, GEN y, GEN T, GEN p, GEN *pr)
    GEN     FpXQX_safegcd(GEN P, GEN Q, GEN T, GEN p)
    GEN     FpXQXV_prod(GEN V, GEN Tp, GEN p)
    GEN     FpXQYQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
    GEN     FpXV_FpV_innerprod(GEN V, GEN W, GEN p)
    GEN     FpXV_prod(GEN V, GEN p)
    GEN     FpXV_red(GEN z, GEN p)
    GEN     FpXX_red(GEN z, GEN p)
    GEN     FpX_rescale(GEN P, GEN h, GEN p)
    GEN     FpY_FpXY_resultant(GEN a, GEN b0, GEN p)
    GEN     Fq_inv(GEN x, GEN T, GEN p)
    GEN     Fq_invsafe(GEN x, GEN T, GEN p)
    GEN     Fq_add(GEN x, GEN y, GEN T, GEN p)
    GEN     Fq_mul(GEN x, GEN y, GEN T, GEN p)
    GEN     Fq_neg(GEN x, GEN T, GEN p)
    GEN     Fq_neg_inv(GEN x, GEN T, GEN p)
    GEN     Fq_pow(GEN x, GEN n, GEN T, GEN p)
    GEN     Fq_red(GEN x, GEN T, GEN p)
    GEN     Fq_sub(GEN x, GEN y, GEN T, GEN p)
    GEN     FqM_to_FlxM(GEN x, GEN T, GEN pp)
    GEN     FqV_roots_to_pol(GEN V, GEN T, GEN p, long v)
    GEN     FqV_red(GEN z, GEN T, GEN p)
    GEN     FqV_to_FlxC(GEN v, GEN T, GEN pp)
    GEN     FqX_Fq_mul(GEN P, GEN U, GEN T, GEN p)
    GEN     FqX_div(GEN x, GEN y, GEN T, GEN p)
    GEN     FqX_divrem(GEN x, GEN y, GEN T, GEN p, GEN *z)
    GEN     FqX_normalize(GEN z, GEN T, GEN p)
    GEN     FqX_red(GEN z, GEN T, GEN p)
    GEN     FqX_rem(GEN x, GEN y, GEN T, GEN p)
    GEN     FqX_mul(GEN x, GEN y, GEN T, GEN p)
    GEN     FqX_sqr(GEN x, GEN T, GEN p)
    GEN     QXQ_inv(GEN A, GEN B)
    GEN     ZX_caract(GEN A, GEN B, long v)
    GEN     ZX_disc(GEN x)
    int     ZX_is_squarefree(GEN x)
    GEN     ZX_resultant(GEN A, GEN B)
    GEN     ZX_QX_resultant(GEN A, GEN B)
    GEN     ZX_s_add(GEN y,long x)
    long    brent_kung_optpow(long d, long n)
    GEN     modulargcd(GEN a,GEN b)
    GEN     rescale_pol(GEN P, GEN h)
    GEN     unscale_pol(GEN P, GEN h)
    GEN     stopoly(ulong m, ulong p, long v)
    GEN     stopoly_gen(GEN m, GEN p, long v)
    int     u_pow(int p, int k)
    long    u_val(ulong n, ulong p)

    # RgX.c 

    int     is_rational(GEN x)
    int     RgX_is_rational(GEN x)
    GEN     RgM_to_RgXV(GEN x, long v)
    GEN     RgM_to_RgXX(GEN x, long v,long w)
    GEN     RgM_zc_mul(GEN x, GEN y)
    GEN     RgM_zm_mul(GEN x, GEN y)
    GEN     RgV_to_RgX(GEN x, long v)
    GEN     RgV_zc_mul(GEN x, GEN y)
    GEN     RgV_zm_mul(GEN x, GEN y)
    GEN     RgX_divrem(GEN x,GEN y,GEN *r)
    GEN     RgX_mul(GEN x,GEN y)
    GEN     RgX_mulspec(GEN a, GEN b, long na, long nb)
    GEN     RgX_powers(GEN a, GEN T, long l)
    GEN     RgXQX_divrem(GEN x,GEN y,GEN T,GEN *r)
    GEN     RgXQX_mul(GEN x,GEN y,GEN T)
    GEN     RgXQX_red(GEN P, GEN T)
    GEN     RgXQX_RgXQ_mul(GEN x, GEN y, GEN T)
    GEN     RgX_Rg_mul(GEN y, GEN x)
    GEN     RgX_RgX_compo(GEN f, GEN x, GEN T)
    GEN     RgX_shift(GEN x, long n)
    GEN     RgX_sqr(GEN x)
    GEN     RgX_sqrspec(GEN a, long na)
    GEN     RgXV_to_RgM(GEN v, long n)
    GEN     RgX_to_RgV(GEN x, long N)
    GEN     RgXX_to_RgM(GEN v, long n)
    GEN     RgXY_swap(GEN x, long n, long w)

    # rootpol.c 

    GEN     cleanroots(GEN x,long l)
    int     isrealappr(GEN x, long l)
    GEN     roots(GEN x,long l)
    GEN     roots0(GEN x,long flag,long l)

    #subcyclo.c 

    GEN     galoissubcyclo(GEN N, GEN sg, long flag, long v)
    GEN     polsubcyclo(long n, long d, long v)
    GEN     subcyclo(long n, long d, long v)
    GEN     znstar_small(GEN zn)

    # subfields.c 

    GEN     subfields(GEN nf,GEN d)
    GEN     subfields0(GEN nf,GEN d)

    # subgroup.c 

    void    forsubgroup(entree *oep, GEN cyc, GEN bound, char *och)
    GEN     subgrouplist(GEN cyc, GEN bound)

    # stark.c 

    GEN     bnrL1(GEN bnr, GEN sbgrp, long flag, long prec)
    GEN     bnrrootnumber(GEN bnr, GEN chi, long flag, long prec)
    GEN     bnrstark(GEN bnr, GEN subgroup, long prec)

    # sumiter.c 

    GEN     direuler(void *E, GEN (*eval)(GEN,void*), GEN ga, GEN gb, GEN c)
    GEN     direuler0(entree *ep, GEN a, GEN b, char *ch, GEN c)
    GEN     divsum(GEN num,entree *ep, char *ch)
    void    fordiv(GEN a, entree *ep, char *ch)
    void    forpari(entree *ep, GEN a, GEN b, char *ch)
    void    forprime(entree *ep, GEN a, GEN b, char *ch)
    void    forstep(entree *ep, GEN a, GEN b, GEN s, char *ch)
    void    forvec(entree *ep, GEN x, char *ch, long flag)
    GEN     forvec_start(GEN x, long flag, GEN *d, GEN (**next)(GEN,GEN))
    GEN     intnum(void *E, GEN (*e)(GEN, void*), GEN a,GEN b, GEN tab, long prec)
    long    intnumstep(long prec)
    GEN     intnumromb0(entree *ep, GEN a, GEN b, char *ch, long flag, long prec)
    GEN     intnum0(entree *ep, GEN a, GEN b, char *ch, GEN tab, long prec)
    GEN     intcirc0(entree *ep, GEN a, GEN R, char *ch, GEN tab, long prec)
    GEN     intmellininv0(entree *ep, GEN sig, GEN x, char *ch, GEN tab, long prec)
    GEN     intmellininvshort(GEN sig, GEN x, GEN tab, long prec)
    GEN     intlaplaceinv0(entree *ep, GEN sig, GEN x, char *ch, GEN tab, long prec)
    GEN     intnuminit(GEN a, GEN b, long m, long prec)
    GEN     intnuminitgen0(entree *ep, GEN a, GEN b, char *ch, long m, long flag, long prec)
    GEN     intfuncinit0(entree *ep, GEN a, GEN b, char *ch, long flag, long m, long prec)
    # GEN     intnumdoub0(entree *epx, GEN a, GEN b, entree *epy, char *chc, char *chd, char *chf, GEN tabext, GEN tabint, long prec) 
    GEN     intfoursin0(entree *ep, GEN a, GEN b, GEN x, char *ch, GEN tab, long prec)
    GEN     intfourcos0(entree *ep, GEN a, GEN b, GEN x, char *ch, GEN tab, long prec)
    GEN     intfourexp0(entree *ep, GEN a, GEN b, GEN x, char *ch, GEN tab, long prec)
    GEN     sumnum(void *E, GEN (*f)(GEN,void*), GEN a,GEN sig,GEN tab,long flag,long prec)
    GEN     sumnum0(entree *ep, GEN a, GEN sig, char *ch, GEN tab, long flag, long prec)
    GEN     sumnumalt(void *E, GEN (*f)(GEN,void*),GEN a,GEN s,GEN tab,long flag,long prec)
    GEN     sumnumalt0(entree *ep, GEN a, GEN sig, char *ch, GEN tab, long flag, long prec)
    GEN     sumnuminit(GEN sig, long m, long sgn, long prec)
    GEN     matrice(GEN nlig, GEN ncol,entree *ep1, entree *ep2, char *ch)
    GEN     polzag(long n, long m)
    GEN     polzagreel(long n, long m, long prec)
    GEN     prodeuler(void *E, GEN (*eval)(GEN,void*), GEN ga, GEN gb, long prec)
    GEN     prodeuler0(entree *ep, GEN a, GEN b, char *ch, long prec)
    GEN     prodinf(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     prodinf0(entree *ep, GEN a, char *ch, long flag, long prec)
    GEN     prodinf1(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     produit(entree *ep, GEN a, GEN b, char *ch, GEN x)
    GEN     somme(entree *ep, GEN a, GEN b, char *ch, GEN x)
    GEN     sumalt(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     sumalt0(entree *ep, GEN a, char *ch,long flag, long prec)
    GEN     sumalt2(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     sumpos(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     sumpos2(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     sumpos0(entree *ep, GEN a, char *ch, long flag,long prec)
    GEN     suminf(void *E, GEN (*eval)(GEN,void*), GEN a, long prec)
    GEN     suminf0(entree *ep, GEN a, char *ch, long prec)
    GEN     vecteur(GEN nmax, entree *ep, char *ch)
    GEN     vecteursmall(GEN nmax, entree *ep, char *ch)
    GEN     vvecteur(GEN nmax, entree *ep, char *ch)
    GEN     zbrent0(entree *ep, GEN a, GEN b, char *ch, long prec)
    GEN     zbrent(void *E, GEN (*eval)(GEN,void*), GEN a, GEN b, long prec)

    # thue.c 

    GEN     bnfisintnorm(GEN x, GEN y)
    GEN     thue(GEN thueres, GEN rhs, GEN ne)
    GEN     thueinit(GEN pol, long flag, long prec)

    # trans1.c 

    GEN     Pi2n(long n, long prec)
    GEN     PiI2(long prec)
    GEN     PiI2n(long n, long prec)
    void    consteuler(long prec)
    void    constpi(long prec)
    GEN     exp_Ir(GEN x)
    GEN     gcos(GEN x, long prec)
    GEN     gcotan(GEN x, long prec)
    GEN     gexp(GEN x, long prec)
    GEN     glog(GEN x, long prec)
    GEN     gpow(GEN x, GEN n, long prec)
    GEN     gpowgs(GEN x, long n)
    GEN     gsin(GEN x, long prec)
    void    gsincos(GEN x, GEN *s, GEN *c, long prec)
    GEN     gsqrt(GEN x, long prec)
    GEN     gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
    GEN     gtan(GEN x, long prec)
    GEN     log0(GEN x,long flag, long prec)
    GEN     mpcos(GEN x)
    GEN     mpeuler(long prec)
    GEN     mpexp(GEN x)
    GEN     mpexp1(GEN x)
    GEN     mplog(GEN x)
    GEN     mplog2(long prec)
    GEN     mppi(long prec)
    GEN     mpsin(GEN x)
    void    mpsincos(GEN x, GEN *s, GEN *c)
    GEN     sqrtr(GEN x)
    GEN     sqrtnr(GEN x, long n)
    GEN     palog(GEN x)
    GEN     powgi(GEN x, GEN n)
    GEN     teich(GEN x)

    # trans2.c 

    GEN     bernfrac(long n)
    GEN     bernreal(long n, long prec)
    GEN     bernvec(long nomb)
    GEN     gach(GEN x, long prec)
    GEN     gacos(GEN x, long prec)
    GEN     garg(GEN x, long prec)
    GEN     gash(GEN x, long prec)
    GEN     gasin(GEN x, long prec)
    GEN     gatan(GEN x, long prec)
    GEN     gath(GEN x, long prec)
    GEN     gch(GEN x, long prec)
    GEN     ggamd(GEN x, long prec)
    GEN     ggamma(GEN x, long prec)
    GEN     glngamma(GEN x, long prec)
    GEN     gpsi(GEN x, long prec)
    GEN     gsh(GEN x, long prec)
    GEN     gth(GEN x, long prec)
    void    mpbern(long nomb, long prec)

    # trans3.c 

    GEN     agm(GEN x, GEN y, long prec)
    GEN     dilog(GEN x, long prec)
    GEN     eint1(GEN x, long prec)
    GEN     eta(GEN x, long prec)
    GEN     eta0(GEN x, long flag,long prec)
    GEN     gerfc(GEN x, long prec)
    GEN     gpolylog(long m, GEN x, long prec)
    void    gpolylogz(long m, GEN x, GEN y)
    GEN     gzeta(GEN x, long prec)
    GEN     hyperu(GEN a, GEN b, GEN gx, long prec)
    GEN     incgam(GEN a, GEN x, long prec)
    GEN     incgam0(GEN a, GEN x, GEN z,long prec)
    GEN     incgam1(GEN a, GEN x, long prec)
    GEN     incgam2(GEN a, GEN x, long prec)
    GEN     incgam3(GEN a, GEN x, long prec)
    GEN     incgamc(GEN a, GEN x, long prec)
    GEN     hbessel1(GEN n, GEN z, long prec)
    GEN     hbessel2(GEN n, GEN z, long prec)
    GEN     ibessel(GEN n, GEN z, long prec)
    GEN     jbessel(GEN n, GEN z, long prec)
    GEN     jbesselh(GEN n, GEN z, long prec)
    GEN     nbessel(GEN n, GEN z, long prec)
    GEN     jell(GEN x, long prec)
    GEN     kbessel(GEN nu, GEN gx, long prec)
    GEN     kbessel0(GEN nu, GEN gx, long flag,long prec)
    GEN     kbessel2(GEN nu, GEN x, long prec)
    GEN     polylog(long m, GEN x, long prec)
    GEN     polylog0(long m, GEN x, long flag, long prec)
    GEN     polylogd(long m, GEN x, long prec)
    GEN     polylogdold(long m, GEN x, long prec)
    GEN     polylogp(long m, GEN x, long prec)
    GEN     szeta(long x, long prec)
    GEN     theta(GEN q, GEN z, long prec)
    GEN     thetanullk(GEN q, long k, long prec)
    GEN     trueeta(GEN x, long prec)
    GEN     veceint1(GEN nmax, GEN C, long prec)
    GEN     vecthetanullk(GEN q, long k, long prec)
    GEN     weber0(GEN x, long flag,long prec)
    GEN     wf(GEN x, long prec)
    GEN     wf2(GEN x, long prec)

    GEN     padicfieldslist(GEN p, GEN m, GEN d, long flag)


cdef extern from 'pari/anal.h':
    struct __z:
        jmp_buf env
    ctypedef __z gp_data
    extern gp_data* GP_DATA

cdef extern from 'pari/paripriv.h':
    void  pari_sig_init(void (*f)(int))


# Initialize the relevant part of GP for exception trapping.
cdef gp_data __GP_DATA
GP_DATA = &__GP_DATA


#cdef void py_pari_sighandler(int sig):
#    os_signal(sig, gp_sighandler)
#    print sig
#    return
#cdef extern from "py_pari_c.h":
#    void py_pari_sighandler(int)

#cdef extern from "signal.h":
#    void (*signal(int sig, void (*f)(int)))(int)

## cdef sigjmp_buf _jb
## def si_err():
##     cdef int n
##     n = sigsetjmp(_jb,1)
##     if n:
##         raise KeyboardInterrupt

## cdef void sigint_handler(int n):
##     #si_err()
##     os_signal(2, sigint_handler) 
##     if n == 0: return
##     print "DEBUG: Signal %s raised"%n
##     longjmp(GP_DATA.env, -1)
##     #siglongjmp(_jb, -1)

cdef size_t fix_size(size_t a):
    cdef size_t b
    b = a - (a & (BYTES_IN_LONG-1))     # sizeof(long) | b <= a  
    if b < 1024:
        b = 1024
    return b

cdef size_t xxx_init_stack(pari_sp size) except -1:
    cdef size_t s, old_size
    cdef pari_sp new_top, new_bot
    
    global top, bot, avma

    # Decide on size
    if size == 0:
        size = 2*(top - bot)
    s = fix_size(size)

    # Save old size
    if bot:
        old_size = top - bot
    else:
        old_size = 0

    # Alocate memory for new stack using Python's memory allocator.
    # As explained in the python/C api reference, using this instead
    # of malloc is much better (and more platform indepenedent, etc.)
    new_bot = <pari_sp> PyMem_Malloc(s)
    
    if not new_bot:   # If there isn't enough memory, try to allocate less.
        s = old_size
        while not new_bot:
            s = s/2   
            if not s:
                raise MemoryError, "Unable to allocate any more memory for PARI."
            new_bot = <pari_sp> PyMem_Malloc(s)
            
    new_top = new_bot + s

    # Copy all the data over from the old stack to the new one.


    # Free the memory occupied by the old stack.

    #PyMem_Free(<void*>bot)


cdef int init_stack(size_t size) except -1:
    #print "init_stack -- size = ", size
    cdef size_t s

    global top, bot, avma

    # Decide on size
    s = fix_size(size)

    # Alocate memory for new stack using Python's memory allocator.
    # As explained in the python/C api reference, using this instead
    # of malloc is much better (and more platform indepenedent, etc.)
    bot = <pari_sp> PyMem_Malloc(s)
    if not bot:
        raise MemoryError, "Unable to allocate %s bytes memory for PARI."%(<long>size)
    top = bot + s
    avma = top

def xxx_stack_resize(mb=0):
    """
    Resize stack to be mb in size, where mb is a number of megabytes.
    If mb = 0, double the current stack size.   Unlike in gp/PARI, you
    can both increase and *decrease* the stack size.

    ALGORITHM:  This function does the following.
        1. Allocates a new stack of the requested size
           using the Python memory allocator.
        2. Copies the current stack contents to the new
           stack using memcpy.
        4. Updates the references of all Python/PARI objects
           to point to their location on the new stack. 
        4. Frees the memory occupied by the old stack.
    """
    cdef pari_sp new_bot, new_top, new_avma
    cdef size_t amount, old_size, s

    global top, bot, avma

    am = 1048576*mb

    old_size = top-bot

    if am == 0:
        amount = old_size
    else:
        amount = am - old_size

    if <long> amount < 0:
        if <long>(old_size+amount) < <long>(top-avma):
            raise MemoryError, "Cannot shrink PARI stack that small, since some PARI objects would be destroyed."

    # Change the amount of memory used by PARI
    s = fix_size(old_size + amount)
    new_bot = <pari_sp> PyMem_Malloc(s)
    if not new_bot:
        raise MemoryError, "Unable to allocate %s bytes memory for PARI."%(<long>amount)
    # Record the top of the new stack
    new_top = new_bot + s
    new_avma = new_top - (top-avma)

    # Copy the contents of the old stack onto the new one.
    memcpy(<void*>new_avma, <void*>avma, (top-avma))

    # Update all the gens to reflect their new position in memory.
    # Again, this is only needed because PARI is upside down.
    shift_gens(new_top)

    # Free old stack
    #PyMem_Free(<void*> bot)

    # Update the avma pointer and current top.
    avma = new_avma
    top = new_top
    bot = new_bot

def stack_resize(mb=0):
    """
    Resize stack to be mb in size, where mb is a number of megabytes.
    If mb = 0, double the current stack size.   Unlike in gp/PARI, you
    can both increase and *decrease* the stack size.

    ALGORITHM:  This function does the following.
        1. Allocates a new stack of the requested size
           using the Python memory allocator.
        2. Uses the PARI library function forcecopy to
           make a complete copy of the objects referenced
           by Python/PARI objects and any PARI objects
           they reference.
        3. Updates the references of all Python/PARI objects
           to point to their location on the new stack. 
        4. Frees the memory occupied by the old stack.
    """
    cdef pari_sp new_bot, new_top, new_avma
    cdef size_t amount, old_size, s

    global top, bot, avma

    am = 1048576*mb

    old_size = top-bot

    if am == 0:
        amount = old_size
    else:
        amount = am - old_size

    if <long> amount < 0:
        if <long>(old_size+amount) < <long>(top-avma):
            raise MemoryError, "Cannot shrink PARI stack that small, since some PARI objects would be destroyed."

    # Change the amount of memory used by PARI
    s = fix_size(old_size + amount)
    new_bot = <pari_sp> PyMem_Malloc(s)
    if not new_bot:
        raise MemoryError, "Unable to allocate %s bytes memory for PARI."%(<long>amount)
    # Record the top of the new stack
    new_top = new_bot + s
    new_avma = new_top - (top-avma)

    # Copy the contents of the old stack onto the new one.
    memcpy(<void*>new_avma, <void*>avma, (top-avma))

    # Update all the gens to reflect their new position in memory.
    # Again, this is only needed because PARI is upside down.
    shift_gens(new_top)

    # Free old stack
    #PyMem_Free(<void*> bot)

    # Update the avma pointer and current top.
    avma = new_avma
    top = new_top
    bot = new_bot

def stack_increase(mb=0):
    """
    Make another stack of size mb, where mb is a number of megabytes.
    If mb = 0, double the current stack size.    

    We use the following *trick*: As far as I can tell PARI does not
    require the stack to be contiguous.   It can work with multiple
    stacks, as long as the stack pointers (which are global variables)
    are set correctly.

    ALGORITHM:  This function does the following.
        1. Allocates a new stack of the requested size
           using the Python memory allocator.
        2. Points the stack pointers to it. 
    """
    cdef size_t s
    
    if mb == 0:
        s = (top-bot)*2
    else:
        s = 1048576*mb
    init_stack(s)
    

def my_sigint(signum, frame):
    raise KeyboardInterrupt

initialized = 0
_max_prime = 0
def init(long size=50000000, unsigned long maxprime=2000000):
    """
    Initialize the PARI system.
    
    INPUT:
        size -- long, the number of bytes given to PARI to work with
                (default: 50MB)
        maxprime -- unsigned long, upper limit on a precomputed prime
                    number table  (default: 2000000)
    
    If you do not want prime numbers, put maxprime=2, but be careful
    because many PARI functions require this table.  If you get the
    error message "not enough precomputed primes", increase this
    parameter.
    """
    global initialized, _max_prime
    if initialized:
        raise RuntimeError, "pari_init has already been called."
    else:
        #print "Initializing PARI (size=%s, maxprime=%s)"%(size,maxprime)
        #pari_init(size, maxprime)
        pari_init(1, maxprime)
        init_stack(size)
        import signal
        signal.signal(signal.SIGINT, my_sigint)
        initialized = 1
    _max_prime = maxprime

# This *has* to be here.  If this module is imported then some
# of the functions in this module assume the PARI interpreter
# has been started.  This doesn't take too much time. 
init()

############################################################
# conversion between GEN and string types
# Note that string --> GEN evaluates the string in PARI,
# where GEN_to_str returns a Python string. 
############################################################


cdef GEN str_to_GEN(char* s) except NULL:
    cdef int n
    n = setjmp(GP_DATA.env)
    if n: _error(n,s)
    return flisseq(s)

cdef object GEN_to_str(GEN g):
    cdef char* c
    cdef int n
    n = setjmp(GP_DATA.env)
    if n:
        _error(n,"converting a gen to a string")
    c = GENtostr(g)
    s = str(c)
    free(c)
    return s


############################################################
# Extension class that models the PARI GEN type.
############################################################

gens = []

############################################################
# Definition of the extension class.
############################################################

cdef class gen:
    cdef GEN g

    def __init__(self):
        #if avma - bot <= 10000:
        #    raise MemoryError, "PARI is out of memory."
        gens.append(self)

    def __repr__(self):
        
        return GEN_to_str(self.g)

    def refs(self):
        return gc.get_referrers(self)

    cdef GEN _gen(self):
        return self.g

    def __reduce__(self):
        z = str(self)
        import sage.pari.py_pari_py
        return sage.pari.py_pari_py.pari, (z,)


    ###########################################
    # ARITHMETIC
    ###########################################    
    def __add__(gen self, gen other):
        return new_gen(gadd(self.g, other.g))

    def __sub__(gen self, gen other):
        return new_gen(gsub(self.g, other.g))

    def __mul__(gen self, gen other):
        return new_gen(gmul(self.g, other.g))

    def __div__(gen self, gen other):
        return new_gen(gdiv(self.g, other.g))

    def __neg__(gen self):
        return new_gen(gneg(self.g))

    def __mod__(gen self, gen other):
        return new_gen(gmod(self.g, other.g))

    def __pow__(gen self, long n, m):
        return new_gen(gpowgs(self.g, n))

    def __rshift__(gen self, long n):
        return new_gen(ishiftr(self.g, n))

    def __lshift__(gen self, long n):
        return new_gen(gshift(self.g, n))

    ###########################################
    # ACCESS
    ###########################################
    def __getitem__(gen self, long n):
        """
        Return a *copy* of the nth entry.

        The indexing is 0-based, like in Python.  However, *copying*
        the nth entry instead of returning reference is different
        than what one usually does in Python.  However, we do it
        this way for consistency with the PARI/GP interpreter, which
        does make copies.
        """
        if n < 0 or n >= glength(self.g):
            raise IndexError, "index (%s) must be between 0 and %s"%(n,glength(self.g)-1)
        return new_gen(<GEN> (self.g[n+1]))

    def __getslice__(self, i, j):
        raise NotImplementedError

    def __setitem__(gen self, long n, gen x):
        """
        Set the nth entry to a *copy* of x.
        
        The indexing is 0-based, like in Python.  However, setting
        the nth entry to a *copy* instead of a reference is different
        than what one usually does in Python.  However, we do it
        this way for consistency with the PARI/GP interpreter, which
        does make copies.
        """
        cdef gen c
        if n < 0 or n >= glength(self.g):
            raise IndexError, "index (%s) must be between 0 and %s"%(n,glength(self.g)-1)
        c = x.copy()
        (self.g)[n+1] = <long>(c.g)

    def __len__(gen self):
        return glength(self.g)


    ###########################################
    # ?
    ###########################################
    
    def __cmp__(gen self, gen other):
        return gcmp(self.g, other.g)

    def copy(gen self):
        return new_gen(gcopy(self.g))


    ###########################################
    # Conversion --> Python
    # Try to convert to a meaningful python object
    # in various ways
    ###########################################
    def eval(self):
        return eval(str(self))

    def list(self):
        s = str(self)
        if s[:4] == "Mat(":
            s = "[" + s[4:-1] + "]"
        s = s.replace("~","")
        if s.find(";") != -1:
            s = s.replace(";","], [")
            s = "[" + s + "]"
            return eval(s)
        else:
            return eval(s)
        
    def __int__(self):
        #return itos(self.g)
        return gtolong(self.g)

    def __long__(self):
        return long(GEN_to_str(self.g))
    
    def __float__(self):
        #return rtodbl(self.g)
        return gtodouble(self.g)


    ###########################################
    # arith1.c
    ###########################################
    def isprime(gen self, flag=0):
        """
        isprime(x, flag=0): Returns true (1) if x is a PROVEN prime
        number, and false (0) otherwise. 

        INPUT:
            flag -- int 
                    0 (default): use a combination of algorithms.
                    1: certify primality using the Pocklington-Lehmer Test.
                    2: certify primality using the APRCL test.
        OUTPUT:
            gen -- 0 or 1
        """
        return new_gen(gisprime(self.g, flag))


    ###########################################
    # 1: Standard monadic or dyadic OPERATORS
    ###########################################
    def divrem(gen x, gen y, long v=0):
        """
        divrem(x, y, long v): Euclidean division of x by y giving as a
        2-dimensional column vector the quotient and the remainder,
        with respect to v (to main variable if v is omitted).
        """
        return new_gen(divrem(x.g, y.g, v))

    def lex(gen x, gen y):
        """
        lex(x,y): Compare x and y lexicographically (1 if x>y, 0 if x==y, -1 if x<y)
        """
        return lexcmp(x.g, y.g)

    def max(gen x, gen y):
        """
        max(x,y): Return the maximum of x and y.
        """
        return new_gen(gmax(x.g, y.g))

    def min(gen x, gen y):
        """
        min(x,y): Return the minumum of x and y.
        """
        return new_gen(gmin(x.g, y.g))

    def shift(gen x, long n):
        """
        shift(x,n): shift x left n bits if n>=0, right -n bits if n<0.
        """
        return new_gen(gshift(x.g, n))

    def shiftmul(gen x, long n):
        """
        shiftmul(x,n): Return the product of x by 2^n (n>=0 or n<0).
        """
        return new_gen(gmul2n(x.g, n))

    def sign(gen x):
        """
        sign(x): Return the sign of x, where x is of type integer, real
        or fraction.
        """
        return gsigne(x.g)

    def vecmax(gen x):
        """
        vecmax(x): Return the maximum of the elements of the vector/matrix x,
        """
        return new_gen(vecmax(x.g))

        
    def vecmin(gen x):
        """
        vecmin(x): Return the maximum of the elements of the vector/matrix x,
        """
        return new_gen(vecmin(x.g))
    


    ###########################################
    # 2: CONVERSIONS and similar elementary functions
    ###########################################

    def Col(gen x):
        """
        Col(x): Transforms the object x into a column vector. 
        """
        #return new_gen(gtocol(x.g))
        # TODO -- when I uncomment the above I get a missing symbol error.  It is very
        # weird!!  Maybe there is a weird name collision ??
        raise NotImplementedError
        
    def List(gen x):
        """
        List(x): transforms the PARI vector or list x into a list.

        EXAMPLES:
            >>> v = pari([1,2,3])
            >>> v
            [1, 2, 3]
            >>> v.type()
            t_VEC
            >>> w = v.List()
            >>> w
            List([1, 2, 3])
            >>> w.type()
            t_LIST
            
            >>> func.List(v)
            List([1, 2, 3])
        """
        return new_gen(gtolist(x.g))

    def Mat(gen x):
        """
        Mat(x): transforms any PARI x into a matrix.

        EXAMPLES:
            >>> x = pari(5)
            >>> x.type()
            t_INT
            >>> y = x.Mat()
            >>> y
            Mat(5)
            >>> y.type()
            t_MAT
            >>> x = pari('[1,2;3,4]')
            >>> x.type()
            t_MAT
            >>> x = pari('[1,2,3,4]')
            >>> x.type()
            t_VEC
            >>> y = x.Mat()
            >>> y
            Mat([1, 2, 3, 4])
            >>> y.type()
            t_MAT
            
            >>> func.Mat(pari(5))
            Mat(5)
        """
        return new_gen(gtomat(x.g))
    
    def Mod(gen x, gen y):
        """
        Mod(x, y): creates the object x modulo y.

        INPUT:
            x -- gen
            y -- gen
            
        EXAMPLES:
            >>> z = pari(3)
            >>> x = z.Mod(pari(7))
            >>> x
            Mod(3, 7)
            >>> x**2
            Mod(2, 7)
            >>> x**100
            Mod(4, 7)
            >>> x.type()
            t_INTMOD

            >>> f = pari("x^2 + x + 1")
            >>> g = pari("x")
            >>> a = g.Mod(f)
            >>> a
            Mod(x, x^2 + x + 1)
            >>> a*a
            Mod(-x - 1, x^2 + x + 1)
            >>> a.type()
            t_POLMOD

            >>> func.Mod(pari(2), pari(3))
            Mod(2, 3)
        """
        return new_gen(gmodulcp(x.g,y.g))
    
    def Pol(gen x):
        """
        Pol(x): convert x (usually a vector or a power series) into a
        polynomial, starting with the leading coefficient

        INPUT:
            x -- gen
        OUTPUT:
            gen -- a polynomial
        EXAMPLES:
            >>> v = pari("[1,2,3,4]")
            >>> f = v.Pol()
            >>> f
            x^3 + 2*x^2 + 3*x + 4
            >>> f*f
            x^6 + 4*x^5 + 10*x^4 + 20*x^3 + 25*x^2 + 24*x + 16
            
            >>> v = pari("[1,2;3,4]")
            >>> v.Pol()
            [1, 3]~*x + [2, 4]~

            >>> func.Pol(pari('[1,2,3]'))
            x^2 + 2*x + 3
        """
        cdef long t

        t = typ(x.g)
        if t != t_POL and t != t_SER and t != t_POL and t != t_RFRAC and \
           t != t_QFR and t != t_QFI and t != t_VEC and t != t_COL and t != t_MAT:
            raise TypeError, "Unable to convert PARI object %s to a polynomial."%x
        return new_gen(gtopoly(x.g,0))
    
    def Polrev(gen x):
        """
        Polrev(x): convert x (usually a vector or a power series) into a
        polynomial, starting with the constant

        INPUT:
            x -- gen
        OUTPUT:
            gen -- a polynomial
        EXAMPLES:
            >>> v = pari("[1,2,3,4]")
            >>> f = v.Polrev()
            >>> f
            4*x^3 + 3*x^2 + 2*x + 1
            >>> f*f
            16*x^6 + 24*x^5 + 25*x^4 + 20*x^3 + 10*x^2 + 4*x + 1
            
        Note that Polrev does *not* reverse the coefficients of a polynomial!
            >>> f.Polrev()
            4*x^3 + 3*x^2 + 2*x + 1
            >>> v = pari("[1,2;3,4]")
            >>> v.Polrev()
            [2, 4]~*x + [1, 3]~
            
            >>> func.Polrev(v)
            [2, 4]~*x + [1, 3]~
        """
        
        cdef long t
        t = typ(x.g)
        if t != t_POL and t != t_SER and t != t_POL and t != t_RFRAC and \
           t != t_QFR and t != t_QFI and t != t_VEC and t != t_COL and t != t_MAT:
            raise TypeError, "Unable to convert PARI object %s to a polynomial."%x
        return new_gen(gtopolyrev(x.g,0))
    
    
    def Qfb(gen a, gen b, gen c, gen D=ZERO, long prec=PREC):
        """
        Qfb(a,b,c,{D=0.}): binary quadratic form a*x^2+b*x*y+c*y^2.    D is optional
                           (0.0 by default) and initializes Shanks's distance if b^2-4*a*c>0

        INPUT:
            a -- gen
            b -- gen
            c -- gen
            D -- gen
            prec -- long (default to py_pari.PREC)
        OUTPUT:
            gen -- binary quadratic form
        EXAMPLES:
            >>> a = pari(3)
            >>> a.Qfb(pari(7), pari(2))
            Qfb(3, 7, 2, 0.E-288)
            
        The method notation is awkward, so one can also use functional notation.
            >>> Qfb = func.Qfb
            >>> Qfb(pari(3), pari(7), pari(2))
            Qfb(3, 7, 2, 0.E-288)
        """
        return new_gen(Qfb0(a.g, b.g, c.g, D.g, prec))
        
    
    def Ser(gen x):
        """
        """
        raise NotImplementedError
    
    def Set(gen x):
        """
        """
        raise NotImplementedError
    
    def Str(gen x):
        """
        """
        raise NotImplementedError
    
    def Strchr(gen x):
        """
        """
        raise NotImplementedError
    
    def Strexpand(gen x):
        """
        """
        raise NotImplementedError
    
    def Strtex(gen x):
        """
        """
        raise NotImplementedError
    
    def Vec(gen x):
        """
        """
        return new_gen(gtovec(x.g))
    
    def Vecsmall(gen x):
        """
        """
        cdef long t
        t = typ(x.g)
        if not (t == t_VECSMALL or t == t_INT or t == t_STR):
            raise TypeError, "x must have PARI type one of t_VECSMALL, t_INT or t_STR"
        return new_gen(gtovecsmall(x.g))
    
    def binary(gen x):
        """
        """
        raise NotImplementedError
    
    def bitand(gen x):
        """
        """
        raise NotImplementedError
    
    def bitneg(gen x):
        """
        """
        raise NotImplementedError
    
    def bitnegimply(gen x):
        """
        """
        raise NotImplementedError
    
    def bitor(gen x):
        """
        """
        raise NotImplementedError
    
    def bittest(gen x):
        """
        """
        raise NotImplementedError
    
    def bitxor(gen x):
        """
        """
        raise NotImplementedError
    
    def ceil(gen x):
        """
        """
        raise NotImplementedError
    
    def centerlift(gen x):
        """
        """
        raise NotImplementedError
    
    def changevar(gen x):
        """
        """
        raise NotImplementedError
    
    def component(gen x):
        """
        """
        raise NotImplementedError
    
    def conj(gen x):
        """
        """
        raise NotImplementedError
    
    def conjvec(gen x):
        """
        """
        raise NotImplementedError
    
    def denominator(gen x):
        """
        """
        raise NotImplementedError
    
    def floor(gen x):
        """
        """
        return new_gen(gfloor(x.g))
    
    def frac(gen x):
        """
        """
        raise NotImplementedError
    
    def imag(gen x):
        """
        """
        raise NotImplementedError
    
    def length(gen x):
        """
        """
        raise NotImplementedError
    
    def lift(gen x):
        """
        """
        raise NotImplementedError
    
    def norm(gen x):
        """
        """
        raise NotImplementedError
    
    def norml2(gen x):
        """
        """
        raise NotImplementedError
    
    def numerator(gen x):
        """
        """
        raise NotImplementedError
    
    def numtoperm(gen x):
        """
        """
        raise NotImplementedError
    
    def padicprec(gen x):
        """
        """
        raise NotImplementedError
    
    def permtonum(gen x):
        """
        """
        raise NotImplementedError
    
    def precision(gen x):
        """
        """
        raise NotImplementedError
    
    def random(gen x):
        """
        """
        raise NotImplementedError
    
    def real(gen x):
        """
        """
        raise NotImplementedError
    
    def round(gen x):
        """
        """
        raise NotImplementedError
    
    def simplify(gen x):
        """
        """
        raise NotImplementedError
    
    def sizebyte(gen x):
        """
        """
        raise NotImplementedError
    
    def sizedigit(gen x):
        """
        """
        raise NotImplementedError
    
    def truncate(gen x):
        """
        """
        raise NotImplementedError
    
    def valuation(gen x):
        """
        """
        raise NotImplementedError
    
    def variable(gen x):
        """
        """
        raise NotImplementedError
    


    ###########################################
    # 3: TRANSCENDENTAL functions
    ###########################################

    ###########################################
    # 4: NUMBER THEORETICAL functions
    ###########################################

    ###########################################
    # 5: Functions related to ELLIPTIC CURVES
    ###########################################

    def elladd(gen self, gen z1, gen z2):
        """
        e.elladd(z1, z2)

        Sum of the points z1 and z2 on the elliptic curve e.

        INPUT:
            e -- elliptic curve E
            z1 -- point on E
            z2 -- point on E
            
        OUTPUT:
            point on E

        EXAMPLES:
            >>> e = pari("ellinit([0, 1, 1, -2, 0])")
            >>> z1 = pari("[1,0,1]")
            >>> z2 = pari("[-1,1,1]")
            >>> e.elladd(z1, z2)
            [-3/4, -15/8]
        """
        return new_gen(addell(self.g, z1.g, z2.g))

    def ellak(gen self, gen n):
        """
        e.ellak(n)

        Returns the coefficient a_n of the L-function of the elliptic
        curve e, i.e. the coefficient of a newform of weight 2 newform.
        The curve e must be a medium or long vector of the type given
        by ellinit. For this function to work for every n and not
        just those prime to the conductor, e must be a minimal Weierstrass
        equation. If this is not the case, use the function
        ellminimalmodel first before using ellak.

        INPUT:
            n -- integer ..
        """
        return new_gen(akell(self.g, n.g))

    def ellan(gen self, long n):
        return new_gen(anell(self.g, n))

    #def ellap0(gen self, gen p, flag=0):
    #return new_gen(ellap0(self.g, p.g, flag))

    def ellap(gen self, gen p):
        return new_gen(apell(self.g, p.g))

    #def ellap2(gen self, gen p):
    #    return new_gen(apell2(self.g, p.g))

    def ellbil(gen self, gen z1, gen z2, long prec):
        return new_gen(bilhell(self.g, z1.g, z2.g, prec))

    def ellchangecurve(gen e, gen ch):
        """
        """
        return new_gen(coordch(e.g, ch.g))

    def ellchangepoint(gen x, gen y):
        """
        ellchangepoint(x,y)

        change data on point or vector of points x on an elliptic curve according
        to y=[u,r,s,t]
        """
        return new_gen(pointch(x.g, y.g))

    def elleisnum(gen self, long k, flag=0, prec=PREC):
        return new_gen(elleisnum(self.g, k, flag, prec))

    def elleta(gen self, prec=PREC):
        return new_gen(elleta(self.g, prec.g))

    def ellglobalred(gen e1):
        return new_gen(globalreduction(e1.g))

    def ellheight(gen e, gen a, flag=0, prec=PREC):
        return new_gen(ellheight0(e.g, a.g, flag, prec))

    def ellheightmatrix(gen e, gen x, long prec=PREC):
        """
        ellheightmatrix(e,x)

        Returns the height matrix for vector of points x on elliptic curve e using
        theta functions.
        """
        return new_gen(mathell(e.g, x.g, prec))

    #def ellheight(gen e, gen a, long prec=PREC):
    #    return new_gen(ghell(e.g, a.g, prec))
    #def ellheight2(gen e, gen a, long prec=PREC):
    #    return new_gen(ghell2(e.g, a.g, prec))

    def ellinit(gen self, flag=0, prec=PREC):
        """
        ellinit(x,{flag=0}): x being the vector [a1,a2,a3,a4,a6], gives
        the vector: [a1,a2,a3,a4,a6,
                    b2,b4,b6,b8,
                    c4,c6,
                    delta,j,
                    [e1,e2,e3],
                    w1,w2,eta1,eta2,area].
        If the curve is defined over a p-adic field, the last six components
        are replaced by root,u^2,u,q,w,0. If optional flag is 1, omit them
        altogether.
        """
        return new_gen(ellinit0(self.g, flag, prec))

    def ellisoncurve(gen e, gen x):
        return oncurve(e.g, x.g)

    def ellj(gen e, long prec=PREC):
        return new_gen(jell(e.g, prec))

    def elllocalred(gen e, gen p):
        return new_gen(localreduction(e.g, p.g))

    def elllseries(gen e, gen s, gen A=ONE, long prec=4):
        if prec <= 2:
            raise RuntimeError, "prec must be >= 3."
        return new_gen(lseriesell(e.g, s.g, A.g, prec))

    def ellminimalmodel(gen e):
        cdef gen x, y
        x = gen()
        y = gen()
        x.g = ellminimalmodel(e.g, &y.g)
        return x, y

    def ellorder(gen e, gen x):
        return new_gen(orderell(e.g, x.g))

    def ellordinate(gen e, gen x, long prec=PREC):
        return new_gen(ordell(e.g, x.g, prec))

    def ellpointtoz(gen e, gen z, long prec=PREC):
        return new_gen(zell(e.g, z.g, prec))

    def ellpow(gen e, gen z, gen n):
        return new_gen(powell(e.g, z.g, n.g))
    
    def ellrootno(gen e, gen p=ONE):
        return ellrootno(e.g, p.g)

    def ellsigma(gen om, gen z, flag=0, prec=PREC):
        return new_gen(ellsigma(om.g, z.g, flag, prec))

    def ellsub(gen e, gen z1, gen z2):
        return new_gen(subell(e.g, z1.g, z2.g))

    def elltaniyama(gen e):
        return new_gen(taniyama(e.g))

    def elltors(gen e, flag=0):
        return new_gen(elltors0(e.g, flag))

    def ellwp(gen e, gen z, long flag=0, long prec=PREC):
        return new_gen(ellwp0(e.g, z.g, flag, prec, prec))

    def ellzeta(gen om, gen z, long prec=PREC):
        return new_gen(ellzeta(om.g, z.g, prec))

    def ellztopoint(gen e, gen z, long prec=PREC):
        return new_gen(pointell(e.g, z.g, prec))

    

    ###########################################
    # 6: Functions related to general NUMBER FIELDS
    ###########################################

    ###########################################
    # 7: POLYNOMIALS and power series
    ###########################################

    ###########################################
    # 8: Vectors, matrices, LINEAR ALGEBRA and sets
    ###########################################

    ###########################################
    # 9: SUMS, products, integrals and similar functions
    ###########################################

    ###########################################
    # 11. PROGRAMMING under GP
    ###########################################

    ###########################################
    # polarit2.c 
    ###########################################
    def factor(gen self, limit=-1):
        return new_gen(factor0(self.g, limit))


    ###########################################
    # misc (classify when I know where they go)
    ###########################################
    def exp(gen self, prec=PREC):
        return new_gen(gexp(self.g, prec))

    def log(gen self, prec=PREC):
        return new_gen(glog(self.g, prec))

    def type(gen self):
        return pari(type_name(typ(self.g)))
    

###########################################
# small values
###########################################
cdef gen ZERO, ONE, TWO
ZERO = new_gen(stoi(0))
ONE = new_gen(stoi(1))
TWO = new_gen(stoi(2))
_tmp = [ZERO,ONE,TWO]   # so they won't be garbage collected.

###########################################
# Create a gen from a GEN object.
# This *steals* a reference to the GEN.
###########################################
cdef gen new_gen(GEN x):
    cdef gen z
    z = gen()
    z.g = x
    return z

def pari(s):
    """
    Create the PARI object got by evaluating s using PARI.
    """
    cdef gen p
    
    t = str(s)
    p = gen()
    again = True
    while again:
        try:
            p.g = str_to_GEN(t)
            again = False
        except MemoryError:
            again = True
    return p

def stack_pointer():
    return avma

def stack_top():
    return top

def stack_bottom():
    return bot

def stack_size():
    """
    Stack size, in MEGABYTES.
    """
    return int((top-bot)/1048576)

###########################################
# misc (classify when I know where they go)
###########################################

cdef shift_gens(pari_sp new_top):
    cdef gen y
    cdef pari_sp offset
    
    global gens, top
    
    offset = (new_top - top)/sizeof(pari_sp)
    #print "offset = ", <unsigned long> offset
    
    for x in gens:
        y = x
        if y.g >= <GEN>bot and y.g <= <GEN> top:  # only move data on the stack, not the heap
            y.g = y.g + offset
            #print " -->", long(<unsigned long>y.g)
        

def stack_usage():
    """
    How much memory in megabytes is currently being used on the PARI stack.
    """
    return (top-avma)/1048576

def stack_available():
    """
    How much memory in megabytes is currently available on the PARI stack.
    """
    return (avma-bot)/1048576

def os(gen g, unsigned long k):
    return new_gen(g.g + k)


######################################################
# Useful direct interface to PARI C-level interpreter
######################################################

def execute(s):
    cdef int n
    n = setjmp(GP_DATA.env)
    if n:
        _error(n, s)
        execute(s)   # if call returns, then was memory error so try again

    # Parse s in PARI.
    s = s + "\n" 
    flisseq(s)
    
def eval(s):
    cdef int n
    n = setjmp(GP_DATA.env)
    if n:
        _error(n, s)
        return eval(s)   # if call returns, then was memory error so try again

    cdef GEN x
    cdef char *c

    # Parse s in pari and evaluate result.
    s = s + "\n"
    x = flisseq(s)
    # Get result as a C-string
    c = GENtostr(x)
    # De-callocate memory used by x:
    cgiv(x)
    # Convert c to a Python string.
    s = str(c)
    # Free memory used by c
    free(c)
    return s

def _error(n, s=""):
    if n == 64: # out of memory
        stack_resize()
        raise MemoryError, "Pari ran out of memory; automatically doubling the stack."
    if n == 65:
        raise OverflowError, "Object too big to fit in a codeword: %s"%s
    raise RuntimeError, "Pari error number %s: %s"%(n,s)



############################################################
# Garbage collection
############################################################
def refcounts():
    cdef int n
    
    global gens
    R = []
    for g in gens:
        n = 0
        for x in g.refs():
            if type(x) != type(g.refs):
                n = n + 1 
        R.append(n)
    return R

def garbage_collect(verbose=False):
    """
    Garbage collect the current PARI stack.

    This frees up space on the PARI stack by removing all Python/PARI
    objects that have gone out of scope (i.e., the PARI objects that
    are represented by Python objects that are no longer referenced).
    (Note that if you called stack_increase, then objects created on
    the previous stack will never be garbage collected.)

    This function must be called manually because PARI's garbage
    collector is very simplistic, and it is expensive to go through
    and recopy everything on the stack saving only certain objects.

    INPUT:
        verbose -- (default: False) if True, print out messages
                   about what is freed

    OUTPUT:
        tuple -- (number of bytes, number of objects) that were freed.
    
    """
    cdef GEN** gptr
    cdef int i, n
    cdef gen G
    cdef pari_sp orig_avma, freed_space

    global gens

    C = refcounts()
    K = []  # ones to keep
    for i from 0<=i<len(C):
        if C[i] > 1:
            K.append(i)
    n = len(K)
    if n == len(C):
        return (0,0)

    gptr = <GEN**>PyMem_Malloc(n * sizeof(GEN*))

    if verbose: print "starting avma =", avma
    orig_avma = avma
    remaining_gens = []        
    for i from 0 <= i < n:
        G = gens[K[i]]
        remaining_gens.append(G)
        gptr[i] = &G.g
        
    gerepilemanysp(top, avma, gptr, n)
    
    PyMem_Free(gptr)
    freed_space = avma - orig_avma
    num_freed = len(gens) - len(remaining_gens)
    if verbose:
        print "deleted %s objects"%num_freed
        print "new avma      =", avma
        print "freed space   = %s bytes"%freed_space
    gens = remaining_gens
    return (int(freed_space), num_freed)


def min(x,y):
    return x.min(y)

def max(x,y):
    return x.max(y)

###########################################
# Functional Notation
###########################################
cdef class _func:
        ###########################################
        # Miscellaneou
        ###########################################
    def log(self, gen x):
        return x.log()

    def vector(self, int n):
        """
        0 vector of length n.
        """
        if n < 0:
            raise ArithmeticError, "Vectors must have nonnegative length."
        return pari("vector(%s)"%n)

        ###########################################
        # 2: CONVERSIONS and similar elementary functions
        ###########################################

    def Col(self, gen x):
        return x.Col()

    def List(self, gen x):
        return x.List()

    def Mat(self, gen x):
        return x.Mat()

    def Mod(self, gen x, gen y):
        return x.Mod(y)

    def Pol(self, gen x):
        return x.Pol()

    def Polrev(self, gen x):
        return x.Polrev()

    def Qfb(self, gen a, gen b, gen c, gen D=ZERO, long prec=PREC):
        return a.Qfb(b,c,D,prec)

    def floor(self, gen x):
        return x.floor()


        ###########################################
        # 4: NUMBER THEORETICAL functions
        ###########################################
    def prime(self, long n):
        """
        prime(n): returns the n-th prime, where n is a C-int
        """
        if n <= 0:
            raise ArithmeticError, "nth prime meaningless for negative n (=%s)"%n
        cdef GEN g
        z = setjmp(GP_DATA.env)
        if z: _error(z, "Not enough precomputed primes")
        g = prime(n)
        return new_gen(g)

    def init_primes(self, ulong maxnum):
        initprimes(maxnum)

# func -- functional notation
func = _func()
