"""
Basic arithmetic functions
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import math

import sage.misc.misc as misc
import sage.misc.search 
import sage.pari.py_pari
pari = sage.pari.py_pari.pari

##################################################################
# Elementary Arithmetic
##################################################################
def sqrt(x, prec=0):
    """
    Returns the square root of x compute to precision prec.

    The precision is the real precision of the PARI system.
    
    EXAMPLES:
    
    """
    return pari(x).sqrt(prec).python()

def factorial(n):
    """
    Compute the factorial of n.

    >>> factorial(0)
    1
    >>> factorial(4)
    24
    >>> factorial(10)
    3628800

    @param n: A nonnegative integer.
    @type n: int
    @return: The factorial of n.
    @returntype: int
    """
    assert n>=0
    return misc.mul(range(1,n+1))

def is_prime(n, flag=0):
    """
    Returns True if x is a PROVEN prime number, and False otherwise. 
    INPUT:
        flag -- int 
                0 (default): use a combination of algorithms.
                1: certify primality using the Pocklington-Lehmer Test.
                2: certify primality using the APRCL test.
    OUTPUT:
        bool -- True or False

    EXAMPLES::
        >>> is_prime(389)
        True
        >>> is_prime(2000)
        False
        >>> is_prime(2)
        True
        >>> is_prime(-1)   
        False
        >>> factor(-6)
        [(2, 1), (3, 1)]
        >>> is_prime(1)
        False
        >>> is_prime(-2)
        False

    IMPLEMENTATION: Calls the PARI isprime function.
    """
    return pari(n).isprime()

    ## if n in [2,3,5,7,11,13,17,19,23,29,31,37,41,43,47]:
    ##         return True
    ##     if n <= 50:
    ##         return False
    ##     return (power_mod(2,n-1,n) == 1) and \
    ##            (power_mod(3,n-1,n) == 1) and \
    ##            (power_mod(5,n-1,n) == 1) and \
    ##            (power_mod(7,n-1,n) == 1)

def valuation(m, p):
    """
    The exact power of p>0 that divides the integer m.
    We do not require that p be prime, and if m is 0,
    then this function returns rings.infinity.
    
    EXAMPLES::

        >>> valuation(512,2)
        9
        >>> valuation(1,2)
        0

    Valuation of 0 is defined, but valuation with respect to 0 is not::

        >>> valuation(0,7)
        Infinity
        >>> valuation(3,0)
        Traceback (most recent call last):
        ...
        ValueError: valuation at 0 not defined

    Here are some other example::
    
        >>> valuation(100,10)
        2
        >>> valuation(200,10)
        2
        >>> valuation(243,3)
        5
        >>> valuation(243*10007,3)
        5
        >>> valuation(243*10007,10007)
        1
    """
    if p <= 0:
        raise ValueError, "valuation at 0 not defined"
    if m == 0:
        import sage.rings.rings
        return sage.rings.rings.infinity
    r=0
    power=p
    while m%power==0:
        r += 1
        power *= p
    return r




def prime_range(start, stop=None):
    """
    List of all primes between start and stop-1, inclusive.  If the
    second argument is omitted, returns the primes up to the first
    argument.

    Use this function when both start and stop are not too large,
    since in all cases this function makes a table of primes up to
    stop.  If both are large, use the primes iterator function
    instead.

    EXAMPLES:
        >>> prime_range(0,10)
        [2, 3, 5, 7]
        >>> prime_range(2000,2020)
        [2003, 2011, 2017]
        >>> prime_range(2,2)
        []
        >>> prime_range(2,3)
        [2]
        >>> prime_range(10)
        [2, 3, 5, 7]
    """
    if stop == None:
        start, stop = 2, start
    w = eratosthenes(stop-1)
    if start <= 2:
        return w
    _, i = sage.misc.search.search(w, start)
    return w[i:]

#
# This is from
#    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/366178
# It's impressively fast given that it's in Pure Python.
#
def eratosthenes(n):
    """
    Return a list of the primes <= n.
    """
    n = int(n)
    if n == 2:
        return [2]
    elif n<2:
        return []
    s = range(3,n+3,2)
    mroot = n ** 0.5
    half = (n+1)/2
    i = 0
    m = 3
    while m <= mroot:
        if s[i]:
            j = (m*m-3)/2
            s[j] = 0
            while j < half:
                s[j] = 0
                j += m
        i = i+1
        m = 2*i+3
    return [2] + [x for x in s if x and x <= n]

# My old versions; not as fast as the above.
## def eratosthenes(n):
##     """
##     Returns a list of the primes up to n, computed 
##     using the Sieve of Eratosthenes.
##     Input:
##         n -- a positive integer
##     Output:
##         list -- a list of the primes up to n
##     Examples:
##     >>> eratosthenes(7)
##     [2, 3, 5, 7]
##     >>> eratosthenes(45)
##     [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43]
##     """
##     if n <= 1: return []
##     X = [i for i in range(3,n+1) if i%2 != 0]     
##     P = [2]                                       
##     sqrt_n = sqrt(n)                         
##     while len(X) > 0 and X[0] <= sqrt_n:          
##         p = X[0]                                  
##         P.append(p)                               
##         X = [a for a in X if a%p != 0]            
##     return P + X                                  

prange = prime_range

def primes(start, stop=None):
    """
    Returns an iterator over all primes between start and stop-1,
    inclusive.  This is much slower than prime_range, but potentially
    uses less memory. 
    
    This command is like the xrange command, except it only iterates
    over primes.  In some cases it is better to use primes than
    prime_range, because primes does not build a list of all primes in
    the range in memory all at once.  However it is much slower since
    it simply calls the next_prime function repeatedly.

    EXAMPLES:
        >>> from arith import *
        >>> for p in primes(10):
        ...     print p
        ...
        2
        3
        5
        7
        >>> for p in primes(5,10):
        ...     print p
        ...
        5
        7
        >>> list(primes(10000000000, 10000000100))
        [10000000019L, 10000000033L, 10000000061L, 10000000069L, 10000000097L]
    """
    
    start = int(start)
    if stop == None:
        stop = start
        start = 2
    else:
        stop = int(stop)
    n = start - 1
    while True:
        n = next_prime(n)
        if n <= stop:
            yield n
        else:
            return
        

def next_prime(n, proof=True):
    """
    The next prime greater than the integer n.  If n is prime, then
    this function does not return n, but the next prime after n.  If
    the optional argument proof is false (the default), this function
    only returns a pseudo-prime, as defined by the PARI nextprime
    function.

    INPUT:
        n -- integer
        proof -- bool (default: True)

    EXAMPLES:
        >>> next_prime(-100)
        2
    
    Notice that the next_prime(5) is not 5 but 7.
        >>> next_prime(5)
        7
        >>> next_prime(2004)
        2011
    """
    if n < 2:   # negatives are not prime. 
        return 2
    if n == 2:
        return 3
    if not proof:  # pari nextprime is probabilistic (according to their docs)
        return eval(str(pari(n+1).nextprime()))

    if n % 2 == 0:
        n += 1
    else:
        n += 2
    while not is_prime(n):  # pari isprime is provably correct
        n += 2
    return n

def prev_prime(n):
    """
    The largest prime < n.  The result is provably
    correct.   If n <= 2, this function returns -p,
    where p is prime and -p < n and no larger negative
    of a prime has this property.

    EXAMPLES:
        >>> prev_prime(10)
        7
        >>> prev_prime(7)
        5
        >>> prev_prime(8)
        7
        >>> prev_prime(7)
        5
        >>> prev_prime(5)
        3
        >>> prev_prime(3)
        2
        >>> prev_prime(2)
        1
        >>> prev_prime(1)
        1
        >>> prev_prime(-20)
        -23
    """
    n = int(n)-1
    if n in [2,3,-2]:
        return n
    if n in [0,1,-1]:
        return -2
    if n%2 == 0:
        n -= 1
    while not is_prime(abs(n)):
        n -= 2
    return n
    

def divisors(n):
    """
    Returns a list of all positive integer divisors 
    of the nonzero integer n.
    
    EXAMPLES:
        >>> divisors(-3)
        [1, 3]
        >>> divisors(6)
        [1, 2, 3, 6]
        >>> divisors(28)
        [1, 2, 4, 7, 14, 28]
        >>> divisors(2**5)
        [1, 2, 4, 8, 16, 32]
        >>> divisors(100)
        [1, 2, 4, 5, 10, 20, 25, 50, 100]
        >>> divisors(1)
        [1]
        >>> divisors(0)
        Traceback (most recent call last):
        ...
        ValueError: n must be nonzero
        >>> divisors(2**3*3**2*17)
        [1, 2, 3, 4, 6, 8, 9, 12, 17, 18, 24, 34, 36, 51, 68, 72, 102, 136, 153, 204, 306, 408, 612, 1224]
    """
    if n == 0:
        raise ValueError, "n must be nonzero"
    if n < 0:  # make positive, or, since (-1,1) is a factor, will get wrong answer.
        n*=-1
    F = factor(n)
    r = [0 for i in range(len(F))]  
    e = [m for _, m in F]     # max exponents
    ans = []
    x = 1
    while r != e:
        ans.append(x)
        r[0] += 1
        if r[0] <= e[0]:
            x *= F[0][0]
        else:  # carry
            i = 0
            while i < len(F) and r[i] > e[i]:
                x /= F[i][0]**F[i][1]
                r[i] = 0
                i += 1
                if i < len(F):
                    r[i] += 1
                    if r[i] <= e[i]:
                        x *= F[i][0]
        #endif
    #endwhile
    ans.append(n)
    ans.sort()
    return ans

def sigma(k,n):
    """
    Return the sum of the k-th powers of the divisors of n.
    """
    return sum([d**k for d in divisors(n)])

def GCD(a, b=0):
    """
    The greatest commond divisor of a and b.

    >>> GCD(97,100)
    1
    >>> GCD(97 * 10**15, 19**20 * 97**2)
    97
    """
    if isinstance(a,list):
        return __GCD_list(a)
    return (pari(a).gcd(pari(b))).python()

gcd = GCD

def LCM(a, b=None):
    """The least common multiple of a and b.
    
    >>> LCM(97,100)
    9700
    >>> LCM(0,2)
    0
    >>> LCM(-3,-5)
    15
    """
    if isinstance(a,list):
        return __LCM_list(a)
    return (pari(a).lcm(pari(b))).python()

lcm = LCM

def __LCM_list(v):
    if len(v) == 0:
        return 1
    x = v[0]
    for i in range(1,len(v)):
        x = LCM(x, v[i])
    return x

def GCD_python(a, b=0):
    """This function should behave exactly the same as GCD,
    but is implemented in pure python."""
    if isinstance(a,list):
        return __GCD_list(a)
    if a == 0:
        return abs(b)
    if b == 0:
        return abs(a)
    if a < 0:
        a = -a
    if b < 0:
        b = -b
    while b != 0:
        c = a % b
        a = b; b = c
    return a

def __GCD_list(v):
    if len(v) == 0:
        return 1
    if len(v) == 1:
        return v[0]
    g = v[0]
    for i in range(1,len(v)):
        g = GCD(g, v[i])
    return g

def XGCD(a, b):
    """
    Returns triple of integers (g,s,t) such that g = s*a+t*b =
    GCD(a,b).
    
        >>> XGCD(56, 44)
        (4, 4, -5)
        >>> 4*56 + (-5)*44
        4
    """
    a = pari(a)
    b = pari(b)
    g,s,t = a.bezout(b)
    return g.python(), s.python(), t.python()

def XGCD_python(a, b):
    """Returns triple (g,p,q) such that g = p*a+b*q = GCD(a,b).
    This function should behave exactly the same as XGCD,
    but is implemented in pure python."""
    if a == 0 and b == 0:
        return (0,0,1)
    if a == 0:
        return (abs(b), 0, b/abs(b))
    if b == 0:
        return (abs(a), a/abs(a), 0)
    psign = 1
    qsign = 1
    if a < 0:
        a = -a
        psign = -1
    if b < 0:
        b = -b
        qsign = -1
    p = 1; q = 0; r = 0; s = 1
    while b != 0:
        c = a % b
        quot = a/b
        a = b; b = c
        new_r = p - quot*r
        new_s = q - quot*s
        p = r; q = s
        r = new_r; s = new_s
    return (a, p*psign, q*qsign)

def inverse_mod(a, m):
    """
    The inverse of the integer a modulo the integer m.
    >>> inverse_mod(7,1)
    0
    >>> inverse_mod(5,14)
    3
    >>> inverse_mod(3,-5)
    2
    """
    if m<0:
        m *= -1
    if m==1:
        return 0
    return (~(pari(a).Mod(m))).lift().python()

# def sqrt_mod(a, m):
#     """A square root of a modulo m."""

# def xxx_inverse_mod(a, m):
#     """The inverse of a modulo m."""
#     g,s,t = XGCD(a,m)
#     if g != 1:
#         raise "inverse_mod(a=%s,m=%s), error since GCD=%s"%(a,m,g)
#     return s

def power_mod(a,m,n):
    """The m-th power of a modulo the integer n.
    >>> power_mod(0,0,5)
    1
    >>> power_mod(2,390,391)
    285
    >>> power_mod(2,-1,7)
    4
    """
    if n==0:
        raise ZeroDivisionError, "Modulus must be nonzero."
    if n==1:
        return 0
    if m < 0:
        ainv = inverse_mod(a,n)
        return power_mod(ainv, -m, n)
    if m==0:
        return 1
    power = 1
    i = 0
    apow2 = a
    while ((m>>i) > 0):
        if((m>>i) & 1):
            power = (power * apow2) % n
        apow2 = (apow2 * apow2) % n
        i += 1
    return power

def generic_power(a, m):
    """
    The m-th power of a, where m is a non-negative
    integer and a is a Python object on which 
    multiplication is defined.  The exponentiation
    is done using the standard binary powering algorithm.
    
    EXAMPLES:
        >>> generic_power(2,5)
        32
        >>> generic_power(2.5,4)
        39.0625
        >>> generic_power(0,0)
        1
        >>> generic_power(2,-3)
        Traceback (most recent call last):
        ...
        ArithmeticError: 2 cannot be raised to the negative power -3
    """
    if a == 1:
        return a
    if m < 0:
        raise ArithmeticError, "%s cannot be raised to the negative power %s"%(a,m)
        m = -m
    if m == 0:
        return 1
    power = 1
    i = 0
    apow2 = a
    while (m>>i) > 0:
        if (m>>i) & 1:
            power *= apow2
        apow2 *= apow2
        i += 1
    return power



"""
 RATIONAL RECONSTRUCTION:                                                    
     Input:  Integer x and a modulus N.                                      
     Output: Numerator and denominator n, d of the unique rational           
             number r=n/d, if it exists, with                                
             |n| and |d| <= sqrt(N/2).                                       
             Return (0,0) if no such number exists.                          

 The algorithm for rational reconstruction is described
 (with a complete nontrivial proof)
 on pages 656-657 of Knuth, Vol 2, 3rd ed. as the solution to
 exercise 51 on page 379.  See in particular the conclusion
 paragraph right in the middle of page 657, which describes
 the algorithm thus:

    This discussion proves that the problem can be solved efficiently by
    applying Algorithm 4.5.2X with u=m and v=a, but with the following
    replacement for step X2: If v3<=sqrt(m/2), the algorithm terminates.
    The pair (x,y)=(|v2|,v3*sign(v2)) is then the unique solution,
 provided that x and y are coprime and x<=sqrt(m/2); otherwise there is
    no solution.   (Alg 4.5.2X is the extended Euclidean algorithm.)

 Knuth says this algorithm is
 due to Wang, Kornerup, and Gregory from around 1983.
"""
def rational_reconstruction(a, m):
    """
    This function tries to compute a pair (x,y), where x/y is a
    rational number is lowest terms such that reduction of x/y modulo
    m is equal to a and the absolute values of x and y are both <=
    sqrt(m/2).  If such a pair exists, that pair is unique and this
    function returns it.  If no such pair exists, this function return
    the pair (0,0).
    
    The efficient algorithm for computing rational reconstruction is very
    similar to the extended Euclidean algorithm.  For more details, see Knuth,
    Vol 2, 3rd ed, pages 656-657.

    EXAMPLES::
    
     >>> m = 100000
     >>> (119*inverse_mod(53,m))%m
     11323
     >>> rational_reconstruction(11323,m)
     (119, 53)
    """
    a %= m
    if a == 0 or m==0:
        return (0,1)
    if m < 0:
        m = -m
    if a < 0:
        a = m-a
    if a == 1:
        return (1,1)
    u = m
    v = a
    bnd = sqrt(m/2)
    U = (1,0,u)
    V = (0,1,v)
    while abs(V[2]) > bnd:
        q = U[2]/V[2]  # floor is implicit
        T = (U[0]-q*V[0], U[1]-q*V[1], U[2]-q*V[2])
        U = V
        V = T
    x = abs(V[1])
    y = V[2]
    if V[1] < 0:
        y *= -1
    if x <= bnd and GCD(x,y) == 1:
        return (y,x)
    return (0,0)

def mqrr_rational_reconstruction(u, m, T):
    """Maximal Quotient Rational Reconstruction.

    Input:
        u, m, and T are integers and
        m > u>=0, T>0.
    Output:
        Either integer n,d such that d>0, gcd(n,d)=1, n/d=u (mod m),
        and T*abs(n)*d < m, or None.

    Reference: Monagan, Maximal Quotient Rational Reconstruction: An
               Almost Optimal Algorithm for Rational Reconstruction (page 11)
    This algorithm is probabilistic.
    """
    if u == 0:
        if m > T:
            return (0,1)
        else:
            return None
    n, d = 0, 0
    t0, r0 = 0, m
    t1, r1 = 1, u
    while r1 != 0 and r0 > T:
        q = r0/r1   # C division implicit floor
        if q > T:
            n, d, T = r1, t1, q
        r0, r1 = r1, r0 - q*r1
        t0, t1 = t1, t0 - q*t1
    if d != 0 and GCD(n,d) == 1:
        return (n,d)
    return None
    

######################
    

def trial_division(n, bound=None):
    """
    Return the smallest prime divisor <= bound of the 
    positive integer n, or n if there is no such prime.  
    If the optional argument bound is omitted, then bound=n.
    Input:
        n -- a positive integer
        bound - (optional) a positive integer
    Output:
        int -- a prime p<=bound that divides n, or n if
               there is no such prime.
    Examples:
    >>> trial_division(15)
    3
    >>> trial_division(91)
    7
    >>> trial_division(11)
    11
    >>> trial_division(387833, 300)   
    387833
    >>> # 300 is not big enough to split off a 
    >>> # factor, but 400 is.
    >>> trial_division(387833, 400)  
    389
    """
    if n == 1: return 1
    for p in [2, 3, 5]:
        if n%p == 0: return p
    if bound == None: bound = n
    dif = [6, 4, 2, 4, 2, 4, 6, 2]
    m = 7; i = 1
    while m <= bound and m*m <= n:
        if n%m == 0:
            return m
        m += dif[i%8]
        i += 1
    return n

def __factor_using_trial_division(n):
    """
    Returns the factorization of the integer n as 
    a sorted list of tuples (p,e).
    Input:
        n -- an integer
    Output:
        list -- factorization of n
    Examples:
    >>> factor(500)
    [(2, 2), (5, 3)]
    >>> factor(-20)
    [(2, 2), (5, 1)]
    >>> factor(1)
    []
    >>> factor(2004)
    [(2, 2), (3, 1), (167, 1)]
    """
    if n in [-1, 0, 1]: return []
    if n < 0: n = -n
    F = []
    while n != 1:
        p = trial_division(n)
        e = 1
        n /= p
        while n%p == 0:
            e += 1; n /= p
        F.append((p,e))
    F.sort()
    return F

def __factor_using_pari(n, int_=False):
    if int_:
        #Z = lambda x: int(str(x))
        Z = int
    else:
        import sage.rings.integer_ring
        Z = sage.rings.integer_ring.IntegerRing()
    F = pari(n).factor()
    B = F[0]
    e = F[1]
    return [(Z(B[i]),Z(e[i])) for i in xrange(len(B))]

#todo: add a limit option to factor, so it will only split off 
# primes at most a given limit.

def factor(n, proof=True, int_=False):
    """
    Returns the factorization of the integer n as a sorted list of
    tuples (p,e).
    
    INPUT:
        n -- an integer
    OUTPUT:
        list -- factorization of n
    EXAMPLES:
        >>> factor(500)
        [(2, 2), (5, 3)]
        >>> factor(-20)
        [(2, 2), (5, 1)]
        >>> factor(0)
        []
        >>> factor(1)
        []
        >>> factor(-1)
        []
        >>> factor(2004)
        [(2, 2), (3, 1), (167, 1)]
    """
    if not isinstance(n,(int,long)):
        try:
            return n.factor()
        except AttributeError:
            raise TypeError, "unable to factor %s"%n
    n = abs(n)
    if n == 0 or n == 1:
        return []
    #if n < 10000000000: return __factor_using_trial_division(n)
    return __factor_using_pari(n, int_=int_)

def prime_divisors(n):    
    """
    The prime divisors of the integer n, sorted in increasing order.  If n
    is negative, we do *not* include -1 among the prime divisors, since -1 is
    not a prime number.
    
    >>> prime_divisors(1)
    []
    >>> prime_divisors(100)
    [2, 5]
    >>> prime_divisors(-100)
    [2, 5]
    >>> prime_divisors(2004)
    [2, 3, 167]
    """
    return [p for p,_ in factor(n) if p != -1]

def odd_part(n):
    """
    The odd part of the integer n.  This is n / 2^v,
    where v = valuation(n,2).
    """
    n = int(n)
    v = valuation(n,2)
    return n / (2**v)


def prime_to_m_part(n,m):
    """
    Returns the prime-to-m part of n, i.e., the largest divisor
    of n that is coprime to m.

    INPUT:
        n -- int (nonzero)
        m -- int
    OUTPUT:
        int
    """
    n = int(n); m = int(m)
    if n == 0:
        raise ValueError, "n (=%s) must be nonzero."%n
    if m == 0:
        return 1
    while True:
        g = gcd(n,m)
        if g == 1:
            return n
        n = n/g
    

def squarefree_part(n):
    """
    Returns the result of dividing n by the largest square of an
    integer that divides n.
    
    >>> squarefree_part(17*37*37)
    17
    >>> squarefree_part(-17*32)
    -34
    >>> squarefree_part(1)
    1
    >>> squarefree_part(-2)
    -2
    >>> squarefree_part(-4)
    -1
    >>>
    """
    if n==0:
        return n
    a = misc.mul([p for p,r in factor(n) if r%2 != 0])
    if n<0:
        a *= -1
    return a

def is_square(n, root=False):
    """
    Returns whether or not n is square, and if n is a square
    also returns the square root.  If n is not square, also
    returns None.
    INPUT:
        n -- an integer
        root -- whether or not to also return a square root (default: False)
    OUTPUT:
        bool -- whether or not a square
        object -- 
    """
    t, x = pari(n).issquare()
    if root:
        if t:
            if hasattr(n, 'parent'):
                x = n.parent()(str(x))
            else:
                x = x.python()
        return t, x
    return t


def is_squarefree(n):
    """
    Returns True if and only if n is not divisible by the square of an integer > 1.
    """
    if n==0:
        return False
    for p, r in factor(n):
        if r>1:
            return False
    return True


#################################################################
# Euler phi function
#################################################################
def eulerphi(n):
    """Euler phi function of the integer n.   We defined this to
    be the number of positive integers <= n that are relatively
    prime to n.  Thus if n<=0 then eulerphi(n) is defined and 
    equals 0.
    
    >>> eulerphi(1)
    1
    >>> eulerphi(2)
    1
    >>> eulerphi(3)
    2
    >>> eulerphi(12)
    4
    >>> eulerphi(37)
    36
    
    Notice that eulerphi is defined to be 0 on negative numbers and 0.
    >>> eulerphi(-1)  
    0
    >>> eulerphi(0)
    0
    """
    if n<=0:
        return 0
    if n<=2:
        return 1
    return misc.mul([(p-1)*p**(r-1) for p, r in factor(n)])

def CRT(a,b=0,m=1,n=1):
    """
    Use the Chinese Remainder Theorem to find some integer x such
    that x=a (mod m) and x=b (mod n).   Note that x is only well-defined
    modulo m*n. 

    >>> CRT(2, 1, 3, 5)
    -4
    >>> CRT(13,20,100,301)
    -2087
    """
    if isinstance(a,list):
        return CRT_list(a,b)
    g, alpha, beta = XGCD(m,n)
    if g != 1:
        raise ValueError, \
        "arguments a(=%s) and b(=%s) must be coprime"%(m,n) \
        + " but they have gcd=%s"%g
    return a+(b-a)*alpha*m

def CRT_list(v, moduli):
    if len(v) == 0:
        return 0
    x = v[0]
    m = moduli[0]
    for i in range(1,len(v)):
        x = CRT(x,v[i],m,moduli[i])
        m *= moduli[i]
    return x%m

def binom(n,m):
    """
    The binomial coefficient 'n choose m', where m and n are
    integers.  If m<0 then this function returns 0.

    INPUT::
        n -- int
        m -- int
        
    OUTPUT::
        int
        
    EXAMPLES::
    
    >>> binom(5,2)
    10
    >>> binom(2,0)
    1
    >>> binom(3,-1)
    0
    >>> binom(20,10)
    184756
    """
    if m > n or m < 0:
        return 0
    if n-m<m:
        m = n-m
    num = 1
    for i in range(n-m+1,n+1):
        num *= i
    denom = 1
    for i in range(2,m+1):
        denom *= i
    return int(num/denom)

def kronecker(x,y):
    """
    The Kronecker symbol (x|y).
    
    >>> kronecker(3,5)
    -1
    >>> kronecker(3,15)
    0
    >>> kronecker(2,15)
    1
    >>> kronecker(-2,15)
    -1
    """
    return pari(x).kronecker(y).python()

def Max(x):
    """
    The maximum of a list of objects, on which a binary max operation
    is defined.
    """
    assert isinstance(x,list), "Argument must be a list."
    if len(x)==0:
        return 0
    m=x[0]
    for i in range(1,len(x)):
        m=max(m,x[i])
    return m

def Min(x):
    """
    The minimum of a list of objects, on which a binary min operation
    is defined.
    """
    assert isinstance(x,list), "Argument must be a list."
    if len(x)==0:
        return 0
    m=x[0]
    for i in range(1,len(x)):
        m=min(m,x[i])
    return m

def moebius(n):
    """
    Returns the value of the Moebius function of abs(n), where n is an integer.

    DEFINITION:
        mu(n) is 0 if n is not square free, and otherwise equals (-1)^r, where n
        has r distinct prime factors.

    INPUT:
        n -- an integer
    OUTPUT:
        0, 1, or -1
    EXAMPLES:
        >>> moebius(-5)
        -1
        >>> moebius(9)
        0
        >>> moebius(12)
        0
        >>> moebius(-35)
        1
        >>> moebius(-1)
        1
        >>> moebius(7)
        -1
    """
    if n < 0:
        n = -n
    F = factor(n)
    for _, e in F:
        if e >= 2:
            return 0
    return (-1)**len(F)

def kronecker(a, b):
    return pari(a).kronecker(b)


def farey(v, lim):
    """
    Return the Farey sequence associated to the floating point
    number v.

    INPUT:
       v -- float (automatically converted to a float)
       lim --  maximum denominator.
    OUTPUT:
       Results are (numerator, denominator); (1, 0) is"infinity".

    AUTHOR: Scott David Daniels, Python Cookbook, 2nd Ed., Recipe 18.13
    """
    v = float(v)
    if v < 0:
        n, d = farey(-v, lim)
        return -n, d
    z = lim - lim    # Get a "0 of the right type" for denominator
    lower, upper = (z, z+1), (z+1, z)
    while True:
        mediant = (lower[0] + upper[0]), (lower[1] + upper[1])
        if v * mediant[1] > mediant[0]:
            if lim < mediant[1]:
                return upper
            lower = mediant
        elif v * mediant[1] == mediant[0]:
            if lim >= mediant[1]:
                return mediant
            if lower[1] < upper[1]:
                return lower
            return upper
        else:
            if lim < mediant[1]:
                return lower
            upper = mediant

def partitions(n):
    """
    Generate the partitions of an integer.

    INPUT:
        n -- int

    EXAMPLES:
        >>> partitions(3)
        <generator object at 0xab3b3eac>
        >>> list(partitions(3))
        [(1, 1, 1), (1, 2), (3,)]

    AUTHOR: David Eppstein, Jan Van lent, George Yoshida; Python Cookbook 2, Recipe 19.16.
    """
    n == int(n)
    # base case of the recursion: zero is the sum of the empty tuple
    if n == 0:
        yield ( )
        return
    # modify the partitions of n-1 to form the partitions of n
    for p in partitions(n-1):
        yield (1,) + p
        if p and (len(p) < 2 or p[1] > p[0]):
            yield (p[0] + 1,) + p[1:]


if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])


