"""
A finite field
"""

#*****************************************************************************
#
#   SAGE:  System for Arithmetic Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import random, weakref

import arith
import ring
import sage.rings.rings
import sage.structure.gens as gens

import sage.pari.py_pari as py_pari

import finite_field_element



_objsFiniteField = {}
class _uniqFiniteField(object):
    def __new__(cls, modulus, var="a"):
        key = (modulus,var)
        if _objsFiniteField.has_key(key):
            x = _objsFiniteField[key]()
            if x != None: return x
        R = object.__new__(cls)
        _objsFiniteField[key] = weakref.ref(R)
        return R

class FiniteField(_uniqFiniteField, ring.Ring, gens.Gens):
    """
    Finite field of order q, where q is a prime power.
    
    Create with the command
          FiniteField(modulus)

    INPUT:
        q -- int, prime power, order of the finite field
    OUTPUT:
        FiniteField -- finite field of order q.

    EXAMPLES:
        >>> k = FiniteField(9)
        >>> k
        Finite field of size 3^2
        >>> k.is_field()
        True
        >>> k.characteristic()
        3
        >>> a = k.gen()
        >>> a 
        a
        >>> a.parent()
        Finite field of size 3^2
        >>> a.charpoly()
        x^2 + x + 2
        >>> [a**i for i in range(8)]
        [1, a, 2*a + 1, 2*a + 2, 2, 2*a, a + 2, a + 1]

    Fields can be coerced into sets or list and iterated over:
        >>> list(k)
        [1, a, 2*a + 1, 2*a + 2, 2, 2*a, a + 2, a + 1, 0]
        >>> set(k)
        set([a, 2*a + 1, a + 2, a + 1, 0, 1, 2*a + 2, 2, 2*a])
        >>> for x in k: print x," ",
        1   a   2*a + 1   2*a + 2   2   2*a   a + 2   a + 1   0  
        
    Next we compute with the finite field of order 16, where
    the variable is named b.
        >>> k16 = FiniteField(16, "b")
        >>> z = k16.gen()
        >>> z
        b
        >>> z.charpoly()
        x^4 + x + 1
        >>> k16.is_field()
        True
        >>> k16.characteristic() 
        2
        >>> z.order()
        15

    Of course one can also make prime finite fields.
        >>> k = FiniteField(7)
        >>> k.gen()
        3
        >>> k.gen().order()
        6
    """
    def __init__(self, q, var="a"):
        """
        Create finite field of order q with variable var.

        INPUT:
            q -- integer, size of the finite field
            var -- (optional, defaults to 'a') variable used for
                   printing element of the finite field.  Also,
                   two finite fields are considered equivalent
                   if they have the same variable.
        OUTPUT:
            FiniteField -- a finite field of order q with variable var.
        EXAMPLES:
            >>> FiniteField(17)
            Finite field of size 17
            >>> FiniteField(2**10)
            Finite field of size 2^10
            >>> FiniteField(3**5, "b")
            Finite field of size 3^5
            >>> FiniteField(3**5, "b").gen()
            b

        You can also create a finite field using GF, which is a synonym 
        for FiniteField.
            >>> GF(19**2)
            Finite field of size 19^2
        """
        q = int(q)
        var = str(var)
        if q < 2:
            raise ArithmeticError, "q (=%s) must be a prime power"%q
        F = arith.factor(q, int_=True)
        if len(F) > 1:
            raise ArithmeticError, "q (=%s) must be a prime power"%q
        ring.Ring.__init__(self)
        self.__var = var
        self.__char = F[0][0]
        self.__pari_one = py_pari.pari(1).Mod(self.__char)
        self.__degree = F[0][1]
        self.__order = q
        self.__is_field = True
        self.__pari_modulus = py_pari.finitefield_init(self.__char, self.__degree, var)
        self.__gen = finite_field_element.FiniteFieldElement(self, py_pari.pari(var))

    def __contains__(self, x):
        if not isinstance(x, finite_field_element.FiniteFieldElement):
            return False
        return x.parent() == self

    def _pari_one(self):
        """
        The PARI object Mod(1,p).  This is implementation specific
        and should be ignored by users.
        """
        return self.__pari_one

    def _pari_modulus(self):
        """
        The polynomial mod p that defines the finite field, as a PARI
        object.  This is implementation specific, and some finite fields
        might not be implemented using PARI, so you should avoid using
        this function.

        INPUT:  nothing
        OUTPUT:
            gen -- a pari polynomial gen
        EXAMPLES:
            >>> GF(19**2)._pari_modulus()
            Mod(1, 19)*a^2 + Mod(1, 19)*a + Mod(14, 19)
            >>> GF(13**3)._pari_modulus()
            Mod(1, 13)*a^3 + Mod(9, 13)*a^2 + Mod(1, 13)*a + Mod(7, 13)
            >>> FiniteField(2**4, "b")._pari_modulus()
            Mod(1, 2)*b^4 + Mod(1, 2)*b + Mod(1, 2)
        """
        return self.__pari_modulus

    def gen(self, n=0):
        """
        Return chosen generator of the finite field.  This generator
        is a root of the defining polynomial of the finite field, and
        is guaranteed to be a generator for the multiplicative group.  

        INPUT:
            nothing
        
        OUTPUT:
            FiniteFieldElement -- field generator of finite field
            
        EXAMPLES:
            >>> FiniteField(2**4, "b").gen()
            b
            >>> k = FiniteField(3**4, "alpha")
            >>> a = k.gen()
            >>> a
            alpha
            >>> a**4
            2*alpha + 1
        """
        return self.__gen

    def ngens(self):
        """
        The number of generators of the finite field.  Always 1.

        EXAMPLES:
            >>> k = FiniteField(3**4)
            >>> k.ngens()
            1
        """
        return 1
        
    def is_field(self):
        """
        Returns whether or not the finite field is a field, i.e.,
        always returns True.

        EXAMPLES:
            >>> k = FiniteField(3**4)
            >>> k.is_field()
            True
        """
        return True
        
    def characteristic(self):
        """
        Returns the characteristic of the finite field, which is a
        prime int.

        EXAMPLES:
            >>> k = FiniteField(3**4)
            >>> k.characteristic()
            3
        """
        return self.__char
    
    def degree(self):
        """

        Returns the degree of the finite field, which is a positive
        integer.

        EXAMPLES:
            >>> FiniteField(3).degree()
            1
            >>> FiniteField(3**20).degree()
            20
        """
        return self.__degree
    
    def __repr__(self):
        s = "Finite field of size %s"%self.__char
        if self.__degree > 1:
            s = s + "^%s"%self.__degree
        return s
    
    def name(self):
        """
        Short description string that determines the finite field.
        EXAMPLES:
            >>> FiniteField(3**4).name()
            'GF3_4'
        """
        return "GF%s_%s"%(self.__char,self.__degree)
    

    def __call__(self, x):
        """
        Coerce x into the finite field.
        
        INPUT:
            x -- object

        OUTPUT:
            FiniteFieldElement -- if possible, makes a finite field element from x.
            
        EXAMPLES:
            >>> k = GF(3**4,'a')
            >>> b = k(5)
            >>> b.parent()
            Finite field of size 3^4
            >>> a = k.gen()
            >>> k(a + 2)
            a + 2
        
        WARNING: Finite fields are currently implemented somewhat
        naively using polynomials modulo p and the PARI ffinit
        function.  In particular, we do not use Conway polynomials and
        define natural consistent inclusion maps between different
        finite fields.
        """
        if isinstance(x, finite_field_element.FiniteFieldElement):
            if x.parent() is self:
                return x
            else:
                # This is where we *would* do coercion from one finite field to another...
                raise TypeError, "No coercion of %s into %s defined."%(x, self)
            
        if isinstance(x, (int, long, sage.rings.rings.Integer, sage.rings.rings.Rational, py_pari.gen)):
            return finite_field_element.FiniteFieldElement(self, x)
        raise TypeError, "No coercion of %s into %s defined."%(x, self)
    
    def __cmp__(self, other):
        """
        Compares this finite field with other.  Two finite fields are
        equal if and only if they have the same cardinality *and* the
        defining polynomials are the same.

        EXAMPLES:
            >>> FiniteField(3**2) == FiniteField(3**3)
            False
            >>> FiniteField(3**2) == FiniteField(3**2)
            True
            >>> FiniteField(3**2,'beta') == FiniteField(3**2,'alpha')
            False
            >>> FiniteField(3**2,'beta') == FiniteField(3**2,'beta')
            True
        """
        if self is other: return 0
        if not isinstance(other, FiniteField):
            return -1
        if self.__char < other.__char:
            return -1
        elif self.__char > other.__char:
            return 1
        if self.__order < other.__order:
            return -1
        elif self.__order== other.__order and self.__pari_modulus == other.__pari_modulus:
            return 0
        return 1

    def __getitem__(self, n):
        """

        Returns the n-th element of the field.  The elements of the
        field are

             [a, a^2, a^3, ... , 0]

        where a is the generator for the multiplicative group returned
        by unit_gen().

        EXAMPLES:
            >>> k = GF(8)
            >>> k[0]
            1
            >>> k[1]
            a
            >>> k[7]
            0
            >>> for i in xrange(8): print k[i]," ",
            1   a   a^2   a^2 + 1   a^2 + a + 1   a + 1   a^2 + a   0  
        """
        if n < 0 or n >= self.__order:
            raise IndexError, "n (=%s) must be between 0 and the order %s of the field."%(\
                n, self.__order)
        if n == self.__order-1:
            return self(0)
        return self.__gen**n

    def __iter__(self):
        """

        Return iterator over elements of the finite field, ending with 0.

        EXAMPLES:
            >>> k = GF(9)
            >>> i = 0
            >>> for x in k: print x, k[i]; i+= 1
            1 1
            a a
            a + 1 a + 1
            2*a + 1 2*a + 1
            2 2
            2*a 2*a
            2*a + 2 2*a + 2
            a + 2 a + 2
            0 0
        """
        a = self.__gen
        b = self(1)
        yield b
        for i in xrange(self.__order-2):
            b = b*a
            yield b
        yield self(0)

    def __len__(self):
        """
        The number of elements of the finite field.

        EXAMPLES:
            >>> k = GF(2**10)
            >>> k
            Finite field of size 2^10
            >>> len(k)
            1024
        """
        return self.__order

    def order(self):
        """
        The number of elements of the finite field.

        EXAMPLES:
            >>> k = GF(2**10)
            >>> k
            Finite field of size 2^10
            >>> k.order()
            1024
        """
        return self.__order

    def unit_group_exponent(self):
        """
        The exponent of the unit group of the finite field.  For a
        finite field, this is always the order minus 1.

        EXAMPLES:
            >>> k = GF(2**10)
            >>> k.order()
            1024
            >>> k.unit_group_exponent()
            1023
        """
        return self.__order - 1
        
    def random(self):
        """
        A random element of the finite field.

        EXAMPLES:
            >> k = GF(2**10)
            >> k.random()
            a^9 + a
        """
        return self[random.randrange(0,self.__order)]

    def modulus(self):
        """
        Return the modulus of this finite field, i.e., the polynomial f(x) so elements
        of the finite field as elements modulo f.

        EXAMPLES:
            >>> k = FiniteField(17)
            >>> k.modulus()
            x + 14
            >>> k = FiniteField(9)
            >>> k.modulus()
            x^2 + 2*x + 2
        """
        try:
            return self.__modulus
        except:
            R = sage.rings.rings.PolynomialRing(FiniteField(self.characteristic()))
            self.__modulus = R(self._pari_modulus())
        return self.__modulus

    polynomial = modulus

    def prime_field(self):
        """
        Returns the prime subfield of this field.
        
        EXAMPLES:
            >>> k = GF(9)
            >>> k.prime_field()
            Finite field of size 3
        """
        return FiniteField(self.__char)

    prime_subfield = prime_field

    def polynomial_ring(self):
        """
        Returns the polynomial ring over the prime subfield in the
        same variable as this finite field.
        
        EXAMPLES:
            >>> k = FiniteField(3**4, "alpha")
            >>> k.polynomial_ring()
            Univariate Polynomial Ring in alpha over Finite field of size 3
        """
        try:
            return self.__polynomial_ring
        except AttributeError:
            self.__polynomial_ring = sage.rings.rings.PolynomialRing(
                FiniteField(self.__char), self.__var)
            return self.__polynomial_ring
    
##################################################################
    
GF = FiniteField


    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
