"""
Element of finite field.
"""

import operator

import element
import finite_field
import arith
import rational
import sage.rings.rings
from sage.pari.py_pari import pari, gen
from sage.rings.coerce import bin_op

class FiniteFieldElement(element.Element):
    """
    An element of a finite field.

    Create elements by first defining the finite field F, then use
    the notation F(n), for n an integer. or let a = F.gen() and
    write the element in terms of a.
    """
    def __init__(self, parent, value):
        """
        Create element of a finite field.

        EXAMPLES:
            >>> k = GF(9)
            >>> a = k(11); a
            2
            >>> a.parent()
            Finite field of size 3^2
        """
        #  This causes doctest problems for some reason...
        # element.Element.__init__(self, parent, sage.rings.finite_field.FiniteField)
        element.Element.__init__(self, parent)   # just don't do the check; OK.
        self.__parent = parent
        try:
            if isinstance(value, gen):
                if value.type()[-3:] == "MOD" :
                    self.__value = value
                else:
                    self.__value = value.Mod(parent._pari_modulus())*parent._pari_one()
                return
            elif isinstance(value, FiniteFieldElement):
                if parent != value.parent():
                    raise TypeError, "No coercion of element %s into %s defined."%(value, parent)
                else:
                    self.__value = value.__value
                    return
            self.__value = pari(value).Mod(parent._pari_modulus())*parent._pari_one()
        except:
            raise TypeError, 'Unable to coerce %s into %s.'%(value, parent)

    def polynomial(self):
        """
        Elements of a finite field are represented as a polynomial
        modulo a modulus.  This functions returns the representating
        polynomial as an element of the polynomial ring over the
        prime finite field, with the same variable as the finite field.
        
        EXAMPLES:
        The default variable is a:
            >>> k = GF(3**2)
            >>> k.gen().polynomial()
            a
            
        The variable can be any string.
            >>> k = FiniteField(3**4, "alpha")
            >>> a = k.gen()
            >>> a.polynomial()
            alpha
            >>> (a**2 + 1).polynomial()
            alpha^2 + 1
            >>> (a**2 + 1).polynomial().parent()
            Univariate Polynomial Ring in alpha over Finite field of size 3
        """
        return self.parent().polynomial_ring()(self.__value.lift())

    def is_square(self):
        """
        Returns True if and only if this element is a perfect square.

        EXAMPLES:
            >>> k = GF(3**2)
            >>> a = k.gen()
            >>> a.is_square()
            False
            >>> (a**2).is_square()
            True
            >>> k = GF(2**2)
            >>> a = k.gen()
            >>> (a**2).is_square()
            True
            >>> k = GF(17**5); a = k.gen()
            >>> (a**2).is_square()
            True
            >>> a.is_square()
            False
        """
        K = self.parent()
        if K.characteristic() == 2:
            return True
        n = K.order() - 1
        a = self**(n / 2)
        return a == 1
        
    def rational_reconstruction(self):
        """
        If the parent field is a prime field, uses rational reconstruction to
        try to find a lift of this element to the rational numbers.

        EXAMPLES:
            >>> k = GF(97)
            >>> a = k(Q('2/3'))
            >>> a
            33
            >>> a.rational_reconstruction()
            2/3
        """
        if self.parent().degree() != 1:
            raise ArithmeticError, "The finite field must be prime, but has degree %s"%self.parent().degree()
        t = arith.rational_reconstruction(int(self), self.parent().characteristic())
        if t == None or t[1] == 0:
            raise ZeroDivisionError, "Unable to compute rational reconstruction."
        return rational.Rational((t[0],t[1]))

    def order(self):
        """
        Returns the order of self.
        """
        # todo: this is currently computed using a very *very* stupid
        # algorithm!  change to use factorization of order of group.
        try:
            return self.__order
        except:
            x = self.copy()
            one = self.__parent(1)
            self.__order = 1
            while x != one:
                self.__order += 1
                x *= self
                if self.__order == 100:
                    print "WARNING: Using IDIOTIC order algorithm in FiniteFieldElement.order!"
            return self.__order
    
    def copy(self):
        """
        Return a copy of this element.

        EXAMPLES:
            >>> k = GF(3**3)
            >>> a = k(5)
            >>> a
                2
            >>> a.copy()
                2
            >>> b = a.copy()
            >>> a == b
            True
            >>> a is b
            False
            >>> a is a
            True
        """
        return FiniteFieldElement(self.__parent, self.__value)

    def pari(self):
        """
        Return PARI object corresponding to this finite field element.

        EXAMPLES:
            >>> k = GF(3**3)
            >>> a = k.gen()
            >>> b = a^2 + 2*a + 1
            >>> b.pari()
            Mod(Mod(1, 3)*a^2 + Mod(2, 3)*a + Mod(1, 3), Mod(1, 3)*a^3 + Mod(2, 3)*a + Mod(1, 3))

        Looking at the PARI representation of a finite field element, it's no wonder people
        find PARI difficult to work with directly.  Compare our representation:
        
            >>> b
            a^2 + 2*a + 1
            >>> b.parent()
            Finite field of size 3^3
        """
        return self.__value

    def charpoly(self):
        """
        Returns the characteristic polynomial of this element.

        EXAMPLES:
            >> k = GF(3**3)
            >> a = k.gen()
            >> a.charpoly()
            x^3 + 2*x^2 + x + 1
            >> k.modulus()
            x^3 + 2*x^2 + x + 1
            >> b = a^2 + 1
            >> b.charpoly()
            x^3 + x^2 + x + 2
        """
        R = sage.rings.rings.PolynomialRing(self.parent().prime_field())
        return R(self.__value.charpoly())

    def trace(self):
        """
        Returns the trace of this element.

        EXAMPLES:
            >> k = GF(3**3); a = k.gen()
            >> b = a**2 + 2
            >> b.charpoly()
            x^3 + x^2 + 2*x + 1
            >> b.trace()
            2
            >> b.norm()
            1
        """
        return self.parent().prime_field()(self.__value.trace().lift())

    def norm(self):
        """
        Returns the norm of this element, which is the constant term
        of the characteristic polynomial, i.e., the determinant of left
        multiplication. 

        EXAMPLES:
            >> k = GF(3**3); a = k.gen()
            >> b = a**2 + 2
            >> b.charpoly()
            x^3 + x^2 + 2*x + 1
            >> b.trace()
            2
            >> b.norm()
            1

        """
        return self.charpoly()[0]
        
    def __repr__(self):
        return "%s"%(self.__value.lift().lift())
    
    def __compat(self, other):
        if self.parent() != other.parent():
            raise TypeError, "Parents of finite field elements %s and %s must be equal."%(self, other)
        
    def __add__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.add)
        self.__compat(right)
        return FiniteFieldElement(self.__parent, self.__value + right.__value)
    
    def __radd__(self, left):
        return self.parent()(left) + self
    
    def __sub__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.sub)
        self.__compat(right)        
        return FiniteFieldElement(self.__parent, self.__value - right.__value)
    
    def __rsub__(self, left):
        return self.parent()(left) - self        

    def __mul__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.mul)
        self.__compat(right)                
        return FiniteFieldElement(self.__parent, self.__value*right.__value)

    def __rmul__(self, left):
        return self.parent()(left)*self

    def __div__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.div)
        self.__compat(right)
        if right.__value == 0:
            raise ZeroDivisionError
        return FiniteFieldElement(self.__parent, self.__value / right.__value)

    def __int__(self):
        try:
            return int(self.__value.lift().lift())
        except:
            raise TypeError, "Cannot coerce %s to int"%self

    def __long__(self):
        try:
            return long(self.__value.lift().lift())
        except:
            raise TypeError, "Cannot coerce %s to int"%self

    def __float__(self):
        try:
            return float(self.__value.lift().lift())
        except:
            raise TypeError, "Cannot coerce %s to float"%self

    def __rdiv__(self, left):
        return self.parent()(left)/self
    
    def __pow__(self, right):
        right = int(right)
        return FiniteFieldElement(self.__parent, self.__value**right)

    def __neg__(self):
        return FiniteFieldElement(self.__parent, -self.__value)
    
    def __pos__(self):
        return self
    
    def __abs__(self):
        raise ArithmeticError, "Absolute value of element (=%s) of %s not defined."%(self, self.parent())
    
    def __invert__(self):
        """
        EXAMPLES:
            >>> k = GF(9); a = k.gen()
            >>> ~a
            a + 1
            >>> (a+1)*a
            1
        """

        if self.__value == 0:
            raise ZeroDivisionError, "Cannot invert 0"
        return FiniteFieldElement(self.__parent, ~self.__value)
    
    def lift(self):
        """
        If this element lies in a prime finite field, return a lift of this
        element to an integer.

        EXAMPLES:
            >>> k = GF(next_prime(10**10))
            >>> a = k(17/19)
            >>> b = a.lift(); b
            7894736858
            >>> b.parent()
            Integer Ring
        """
        return sage.rings.rings.IntegerRing()(self.__value.lift().lift())
    
    
    def __cmp__(self, other):
        """

        Compare an element of a finite field with other.  If other is
        not an element of a finite field, an attempt is made to coerce
        it so it is one.

        EXAMPLES:
            >>> k = GF(3**3); a = k.gen()
            >>> a == 1
            False
            >>> a**0 == 1
            True
            >>> a == a
            True
            >>> a < a^2
            True
            >>> a > a^2
            False
        """
        if self is other: return 0
        if not isinstance(other, FiniteFieldElement):
            try:
                other = self.parent()(other)
            except TypeError:
                return -1
        #if self.parent() != other.parent():
        #    return -1
        if self.__value == other.__value:
            return 0
        return -1
    
    
if __name__ ==  '__main__':
    import doctest, sys
    from finite_field import *
    from rings import *
    doctest.testmod(sys.modules[__name__])
