"""
The integers.
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

include "gmp.pxi"

cdef class Integer(_element.Element)

import sage.rings.integer_ring
import sage.rings.coerce
import _rational as rational
import _py_pari as py_pari

cdef public int set_mpz(Integer self, mpz_t value):
    mpz_set(self.value, value)

cdef set_from_Integer(Integer self, Integer other):
    mpz_set(self.value, other.value)

cdef set_from_int(Integer self, int other):
    mpz_set_si(self.value, other)

cdef public mpz_t* get_value(Integer self):
    return &self.value

cdef class Integer(_element.Element):
    def __new__(self, x=None):
        mpz_init(self.value)

    def __init__(self, x=None):
        if x != None:
            if isinstance(x, Integer):
                set_from_Integer(self, x) 
            elif isinstance(x, int):
                mpz_set_si(self.value, x)
            elif isinstance(x, long):
                s = "%x"%x
                mpz_set_str(self.value, s, 16)                
            elif isinstance(x, str):
                if x.isdigit() or x[1:].isdigit():
                    mpz_set_str(self.value, x, 0)
            elif isinstance(x, rational.Rational):
                if x.denominator() != 1:
                    raise TypeError, "Unable to coerce rational (=%s) to an Integer."%x
                set_from_Integer(self, x.numer())
            elif isinstance(x, sage.pari.py_pari.gen):
                # TODO: figure out how to convert to pari integer in base 16
                s = str(x)
                mpz_set_str(self.value, s, 0)
            else:
                raise TypeError, "Unable to coerce %s to an Integer."%x

    def __reduce__(self):
        # This single line below took me HOURS to figure out.
        # It is the *trick* needed to pickle pyrex extension types.
        # The trick is that you must Python a pure Python function
        # as the first argument, and that function must return
        # the result of unpickling with the argument in the second
        # tuple as input. All kinds of problems happen
        # if we don't do this.
        return sage.rings.integer_ring.Z, (long(self),)

    cdef int cmp(self, Integer x):
        cdef int i
        i = mpz_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def __cmp__(Integer self, Integer x):
        return self.cmp(x)

    def __richcmp__(Integer self, x, int op):
        cdef int n
        if not isinstance(x, Integer):
            try:
                x = Integer(x)
            except TypeError:
                n = -x.__cmp__(self)   # negative, because switch order
            else:
                n = self.cmp(x)
        else:
            n = self.cmp(x)
        if op == 0:
            return bool(n < 0)
        elif op == 1:
            return bool(n <= 0)
        elif op == 2:
            return bool(n == 0)
        elif op == 3:
            return bool(n != 0)
        elif op == 4:
            return bool(n > 0)
        elif op == 5:
            return bool(n >= 0)

    def copy(self):
        """
        Return a copy of the integer.
        """
        cdef Integer z
        z = Integer()
        set_mpz(z,self.value)
        return z

    def  __dealloc__(self):
        mpz_clear(self.value)
        
    def __repr__(self):
        return self.str()
        
    def str(self, int base=10):
        cdef char *s
        s = mpz_get_str(NULL, base, self.value)
        t = str(s)
        free(s)
        return t
    
    def set_si(self, signed long int n):
        """
        Coerces n to a C signed integer if possible, and sets self
        equal to n.
        """
        mpz_set_si(self.value, n)

    def set_str(self, s, base=10):
        """
        Set self equal to the number defined by the s in the given
        base.
        """
        valid = mpz_set_str(self.value, s, base)
        if valid != 0:
            raise ValueError, "invalid literal:" + s

    cdef void set_from_mpz(Integer self, mpz_t value):
        mpz_set(self.value, value)

    cdef mpz_t* get_value(Integer self):
        return &self.value

    def __add(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_add(x.value, self.value, other.value)
        return x
        
    def __add__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__add(y)
        return sage.rings.coerce.bin_op(x, y, operator.add)

    def __sub(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_sub(x.value, self.value, other.value)
        return x
        
    def __sub__(x, y):
        """
        EXAMPLES:
            >>> a = Integer(5); b = Integer(3)
            >>> b - a
            -2
        """
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__sub(y)
        return sage.rings.coerce.bin_op(x, y, operator.sub)

    def __mul(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_mul(x.value, self.value, other.value)
        return x
            
    def __mul__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__mul(y)
        return sage.rings.coerce.bin_op(x, y, operator.mul)
        
    def __div(Integer self, Integer other):
        cdef Integer x
        #if mpz_divisible_p(self.value, other.value):
        #    x = Integer()
        #    mpz_divexact(x.value, self.value, other.value)
        #    return x
        #else:
        return rational.Rational(self)/rational.Rational(other)
        #raise ArithmeticError, "Exact division impossible."
        
    def __div__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__div(y)
        return sage.rings.coerce.bin_op(x, y, operator.div)

    def __floordiv(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_fdiv_q(x.value, self.value, other.value)
        return x

    def __floordiv__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__floordiv(y)
        return sage.rings.coerce.bin_op(x, y, operator.floordiv)

    def __pow__(self, n, dummy):
        cdef Integer _self, _n
        cdef unsigned int _nval
        _n = Integer(n)
        if _n < 0:
            return Integer(1)/(self**(-_n))
        _self = integer(self)
        cdef Integer x
        x = Integer()
        _nval = _n
        mpz_pow_ui(x.value, _self.value, _nval)
        return x

    def __pos__(self):
        return self

    def __neg__(self):
        cdef Integer x
        x = Integer()
        mpz_neg(x.value, self.value)
        return x

    def __abs__(self):
        """
        """
        cdef Integer x
        x = Integer()
        mpz_abs(x.value, self.value)
        return x

    def __mod__(self, modulus):
        cdef Integer _modulus, _self
        _modulus = integer(modulus)
        if not _modulus:
            raise ZeroDivisionError, "Integer modulo by zero"            
        _self = integer(self)
        
        cdef Integer x
        x = Integer()
        mpz_mod(x.value, _self.value, _modulus.value)
        return x

    def powermod(self, exp, mod):
        """
        powermod(self, Integer exp, Integer mod):
        Compute self**exp modulo mod.
        """
        cdef Integer x, _exp, _mod
        _exp = Integer(exp); _mod = Integer(mod)
        x = Integer()
        mpz_powm(x.value, self.value, _exp.value, _mod.value)
        return x

    def powermodm_ui(self, unsigned long int exp, mod):
        cdef Integer x, _mod
        _mod = Integer(mod)
        x = Integer()
        mpz_powm_ui(x.value, self.value, exp, _mod.value)
        return x

    def __int__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        n = int(s,16)
        free(s)
        return n

    def __nonzero__(self):
        return not self.is_zero()

    def __long__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        n = long(s,16)
        free(s)
        return n

    def __float__(self):
        return mpz_get_d(self.value)

    def __hash__(self):
        return hash(self.str(16))

    def factor(self):
        """
        Return the prime factorization of self as a list of
        pairs (p,e).
        """
        return sage.rings.integer_ring.factor(self)

    def lcm(self, n):
        """
        Returns the least common multiple of self and n.
        """
        cdef Integer _n
        cdef mpz_t x

        _n = Integer(n)
        mpz_init(x)
        mpz_lcm(x, self.value, _n.value)
        cdef Integer z
        z = Integer()
        mpz_set(z.value,x)
        mpz_clear(x)
        return z
        
    def denominator(self):
        return ONE

    def is_one(self):
        return mpz_cmp_si(self.value, 1) == 0

    def is_zero(self):
        return mpz_cmp_si(self.value, 0) == 0

    def is_square_free(self):
        """
        Returns True if this integer is not divisible by the square of
        any prime and False otherwise.
        EXAMPLES:

        """
        return self.pari().issquarefree()

    def pari(self):
        if self._pari is None:
            # better to do in hex, but I can't figure out
            # how to input a number in hex in PARI!!
            self._pari = sage.pari.py_pari.pari(str(self))
        return self._pari

    def parent(self):
        """
        Return IntegerRing().
        """
        return sage.rings.integer_ring.Z

    def sqrt(self):
        """
        Returns the integer floor of the square root of self.
        EXAMPLE:
            >>> a = Integer(5)
            >>> a.sqrt()
            2
        """
        cdef Integer x
        x = Integer()
        mpz_sqrt(x.value, self.value)
        return x

    def xgcd(self, n):
        """
        Return a triple g, s, t of Integers such that
               g = s * self + t * n.
        """
        cdef mpz_t g, s, t
        cdef object g0, s0, t0
        cdef Integer _n
        _n = Integer(n)
        
        mpz_init(g)
        mpz_init(s)
        mpz_init(t)
        mpz_gcdext(g, s, t, self.value, _n.value)
        g0 = Integer()
        s0 = Integer()
        t0 = Integer()
        set_mpz(g0,g)
        set_mpz(s0,s)
        set_mpz(t0,t)
        mpz_clear(g)
        mpz_clear(s)
        mpz_clear(t)
        return g0, s0, t0

    def __lshift__(Integer self, n):
        """
        """
        n = int(n)
        return self*(Integer(2)**n)   # todo: optimize

    def __rshift__(Integer self, n):
        """
        """
        n = int(n)
        return self.__floordiv(Integer(2)**n)    # todo: optimize

    def _and(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_and(x.value, self.value, other.value)
        return x

    def __and__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x._and(y)
        return sage.rings.coerce.bin_op(x, y, operator.and_)


    def _or(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_ior(x.value, self.value, other.value)
        return x

    def __or__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x._or(y)
        return sage.rings.coerce.bin_op(x, y, operator.or_)


    def __invert__(self):
        """
        """
        return Integer(1)/self    # todo: optimize
        

    def inverse_mod(self, n):
        """
        Returns the inverse of self modulo n, if this inverse exists.
        Otherwise, raises a ZeroDivisionError exception.
        
        INPUT:
           self -- Integer
           n -- Integer
        OUTPUT:
           x -- Integer such that x*self = 1 (mod m), or
                raises ZeroDivisionError.
        IMPLEMENTATION:
           Call the mpz_invert GMP library function.
        EXAMPLES:
        >>> a = Integer(189)
        >>> a.inverse_mod(10000)
        4709
        >>> a.inverse_mod(-10000)
        4709
        >>> a.inverse_mod(1890)
        Traceback (most recent call last):
        ...
        ZeroDivisionError: Inverse of 189 mod 1890 does not exist.
        """
        cdef mpz_t x
        cdef object ans
        cdef int r
        cdef Integer m
        m = Integer(n)
        
        mpz_init(x)
        r = mpz_invert(x, self.value, m.value)
        if r == 0:
            raise ZeroDivisionError, "Inverse of %s mod %s does not exist."%(self,m)
        ans = Integer()
        set_mpz(ans,x)
        mpz_clear(x)
        return ans
        
    def gcd(self, n):
        """
        Return the greatest common divisor of self and n.
        """
        cdef Integer _n
        cdef mpz_t g
        cdef object g0

        _n = Integer(n)
        mpz_init(g)
        mpz_gcd(g, self.value, _n.value)
        g0 = Integer()
        set_mpz(g0,g)
        mpz_clear(g)
        return g0

    def crt(self, y, m, n):
        """
        Return unique integer between 0 and m*n that is self modulo m
        and y modulo n.  We assume that m and n are coprime.
        """
        cdef object g, s, t
        cdef Integer _y, _m, _n
        _y = Integer(y); _m = Integer(m); _n = Integer(n)
        g, s, t = _m.xgcd(_n)
        if not g.is_one():
            raise ArithmeticError, "CRT requires that gcd of moduli is 1."
        # Now s*m + t*n = 1, so the answer is x + (y-x)*s*m, where x=self.
        return (self + (_y-self)*s*_m) % (_m*_n)
        
ONE = Integer(1)

def integer(x):
    if isinstance(x, Integer):
        return x
    return Integer(x)

def factorial(unsigned long int n):
    """
    Return the factorial 1 *2 * 3 * ... * n.
    """
    cdef mpz_t x
    cdef Integer z
    
    mpz_init(x)
    mpz_fac_ui(x, n)
    z = Integer()
    set_mpz(z, x)
    mpz_clear(x)
    return z

def _test():
    import doctest, sys
    doctest.testmod(sys.modules[__name__], verbose=True)

if __name__ ==  '__main__':
    _test()
