"""
Element of integers modulo n.
"""

import operator

import element
import integer_mod_ring
import arith
import rational
from sage.rings.coerce import bin_op

class IntegerMod(element.Element):
    def __init__(self, parent, value):
        element.Element.__init__(self, parent, integer_mod_ring.IntegerModRing)
        self.__modulus = parent.modulus()
        self.__parent = parent
        self.__value = long(value % parent.modulus())
        self.__order = None

    def modulus(self):
        return self.__modulus

    def is_square(self):
        if not arith.is_prime(self.__modulus):
            raise NotImplementedError, "Is square only implemented for prime modulus."
        p = self.__modulus
        if p == 2:
            return True
        x = self**((p-1)/2) 
        if x == self.parent()(-1):
            return False
        elif x == self.parent()(1):
            return True
        assert False, "Bug in is_square."
        
    def rational_reconstruction(self):
        """
        EXAMPLES:
            >>> from integer_mod_ring import *
            >>> R = IntegerModRing(97)
            >>> a = R(2) / R(3)
            >>> a
            33
            >>> a.rational_reconstruction()
            2/3
        """
        t = arith.rational_reconstruction(self.__value, self.__modulus)
        if t == None or t[1] == 0:
            raise ZeroDivisionError, "Unable to compute rational reconstruction."
        return rational.Rational((t[0],t[1]))

    def crt(self, other):
        """
        Use the Chinese Remainder Theorem to find an element of the
        integers modulo the product of the moduli that reduces to self
        and to other.  The modulus of other must be coprime to the
        modulus of self.
        """
        assert isinstance(other, IntegerMod)
        if arith.GCD(self.modulus(), other.modulus()) != 1:
            raise errors.UsageError("Moduli must be coprime.")
        x = arith.CRT(self.lift(), other.lift(), self.modulus(), other.modulus())
        return Mod(x, self.modulus()*other.modulus())
    
    def order(self):
        """
        Returns the order of self.
        """
        # todo: this is currently computed using a very *very* stupid
        # algorithm!  change to use factorization of order of group.
        if self.__order != None:
            return self.__order
        x = self.copy()
        one = self.__parent(1)
        self.__order = 1
        while x != one:
            self.__order += 1
            x *= self
            if self.__order == 100:
                print "WARNING: Using IDIOTIC order algorithm in IntegerMod.order!"
        return self.__order
    
    def copy(self):
        return IntegerMod(self.__parent, self.__value)
    
    def __repr__(self):
        return "%s"%self.__value
    
    def __compat_moduli(self, other):
        if self.__modulus != other.__modulus:
            raise TypeError, "Moduli %s and %s should be equal"%(self.__modulus, other.__modulus)
        
    def __add__(self, right):
        if not isinstance(right, IntegerMod):
            return bin_op(self, right, operator.add)
        self.__compat_moduli(right)
        return IntegerMod(self.__parent, self.__value + right.__value)
    
    def __radd__(self, left):
        return self.parent()(left) + self
    
    def __sub__(self, right):
        if not isinstance(right, IntegerMod):
            return bin_op(self, right, operator.sub)
        self.__compat_moduli(right)        
        return IntegerMod(self.__parent, self.__value - right.__value)
    
    def __rsub__(self, left):
        return self.parent()(left) - self        

    def __mul__(self, right):
        if not isinstance(right, IntegerMod):
            return bin_op(self, right, operator.mul)
        self.__compat_moduli(right)                
        return IntegerMod(self.__parent, self.__value*right.__value)

    def __rmul__(self, left):
        return self.parent()(left)*self

    def __div__(self, right):
        if not isinstance(right, IntegerMod):
            return bin_op(self, right, operator.div)
        self.__compat_moduli(right)
        if right.__value == 0:
            raise ZeroDivisionError
        inv = arith.inverse_mod(right.__value, self.__modulus)
        return IntegerMod(self.__parent, inv * self.__value)

    def __int__(self):
        return self.__value

    def __long__(self):
        return long(self.__value)

    def __rdiv__(self, left):
        return self.parent()(left)/self
    
    def __mod__(self, right):
        """Returns the int got by reducing a lift of this modulo right."""
        if self.__modulus % right != 0:
            raise ZeroDivisionError, "Error - reduction mod %s not defined."%right
        return self.__value % right
    def __pow__(self, right):
        right = int(right)
        return IntegerMod(self.__parent, \
                          arith.power_mod(self.__value, right, self.__modulus))
    def __neg__(self):
        return IntegerMod(self.__parent, -self.__value)
    
    def __pos__(self):
        return self
    
    def __abs__(self):
        return self
    
    def __invert__(self):
        if self.__value == 0:
            raise ZeroDivisionError, "Cannot invert 0"
        return IntegerMod(self.__parent,\
                          arith.inverse_mod(self.__value, self.__modulus))
    def lift(self):
        return self.__value
    
    def __float__(self):
        return float(self.__value)
    
    def __cmp__(self, other):
        if not isinstance(other, IntegerMod):
            other = self.parent()(other)
        if self.__modulus != other.__modulus:
            return (self.__modulus).__cmp__(other.__modulus)
        if self.__value < other.__value:
            return -1
        elif self.__value == other.__value:
            return 0
        else:
            return 1

def Mod(n, m):
    return IntegerMod(integer_mod_ring.IntegerModRing(m),n)
mod=Mod

    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
