"""
Ring of integers modulo n.
"""

#*****************************************************************************
#
#   SAGE:  System for Arithmetic Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import weakref

import ring
import arith
import integer_mod

_objsIntegerModRing = {}
class _uniqIntegerModRing(object):
    def __new__(cls, modulus):
        if _objsIntegerModRing.has_key(modulus):
            x = _objsIntegerModRing[modulus]()
            if x != None: return x
        R = object.__new__(cls)
        _objsIntegerModRing[modulus] = weakref.ref(R)
        return R

class IntegerModRing(_uniqIntegerModRing, ring.Ring):
    """
    The ring of integers modulo N, e.g., when N is prime
    this is a prime finite field.
    
    Create with the command
          IntegerModRing(modulus)

    INPUT:
        modulus -- an integer > 1
    OUTPUT:
        IntegerModRing -- the ring of integers modulo N.

    EXAMPLES:
        >>> from sage.rings.rings import *
        >>> FF = IntegerModRing(17)
        >>> FF
        Ring of integers modulo 17
        >>> type(FF)
        <class 'sage.rings.integer_mod_ring.IntegerModRing'>
        >>> FF.is_field()
        True
        >>> FF.characteristic()
        17
        >>> FF.modulus()
        17
        >>> gens = FF.unit_gens()
        >>> a = gens[0]
        >>> a
        3
        >>> type(a)
        <class 'sage.rings.integer_mod.IntegerMod'>
        >>> a.is_square()
        False
        >>> def pow(i): return a**i
        >>> [pow(i) for i in range(16)]
        [1, 3, 9, 10, 13, 5, 15, 11, 16, 14, 8, 7, 4, 12, 2, 6]
        
        >>> Z16 = IntegerModRing(16)
        >>> Z16.is_field()
        False
        >>> Z16.modulus()
        16
        >>> Z16.characteristic() 
        16
        >>> gens = Z16.unit_gens()
        >>> gens
        [15, 5]
        >>> a = gens[0]
        >>> b = gens[1]
        >>> def powa(i): return a**i
        >>> def powb(i): return b**i
        >>> gp_exp = FF.unit_group_exponent()
        >>> gp_exp
        16
        >>> [powa(i) for i in range(15)]
        [1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1]
        >>> [powb(i) for i in range(15)]
        [1, 5, 9, 13, 1, 5, 9, 13, 1, 5, 9, 13, 1, 5, 9]
        >>> a.order()
        2
        >>> b.order()
        4
    """
    def __init__(self, modulus):
        modulus = int(modulus)
        if modulus == 0:
            raise ZeroDivisionError, "modulus must be nonzero"
        ring.Ring.__init__(self)
        self.__modulus = modulus
        self.__is_field = None
        self.__unit_group_exponent = None
        self.__factored_modulus = None
        
    def is_field(self):
        if self.__is_field != None:
            return self.__is_field
        self.__is_field = arith.is_prime(self.__modulus)
        return self.__is_field
        
    def factored_modulus(self):
        if self.__factored_modulus != None:
            return self.__factored_modulus
        self.__factored_modulus = arith.factor(self.__modulus, int_=True)
        return self.__factored_modulus
    
    def characteristic(self):
        return self.__modulus
    
    def __repr__(self):
        return "Ring of integers modulo %s"%self.__modulus
    
    def modulus(self):
        return self.__modulus
    
    def name(self):
        return "Zmod%s"%self.__modulus
    
    def __call__(self, x):
        return integer_mod.IntegerMod(self, x)
    
    def __cmp__(self, other):
        if not isinstance(other, IntegerModRing):
            return -1
        if self.__modulus < other.__modulus:
            return -1
        elif self.__modulus > other.__modulus:
            return 1
        return 0
    
    def __getitem__(self, n):
        if n < 0 or n >= self.modulus():
            raise IndexError
        return integer_mod.IntegerMod(self, n)

    # The following __unit_gens functions are here since I just factored
    # them out from the unit_gens function.  They are only called by
    # the unit_gens function.
    def __unit_gens_primecase(self, p):
        if p==2:
            return integer_mod.Mod(1,p)
        P = arith.prime_divisors(p-1)
        ord = p-1
        one = integer_mod.Mod(1,p)
        x = 2
        while x < p:
            generator = True
            z = integer_mod.Mod(x,p)
            for q in P:
                if z**(ord/q) == one:
                    generator = False
                    break
            if generator:
                z._IntegerMod__order = p-1
                return z
            x += 1
        #end for
        assert False, "didn't find primitive root for p=%s"%p

    def __unit_gens_primepowercase(self, p, r):
        """Find smallest generator for (Z/p^rZ)^*."""
        if r==1:
            return [self.__unit_gens_primecase(p)]
        elif p==2:
            if r==1:
                return []
            elif r==2:
                return [integer_mod.Mod(-1,2**r)]
            elif r>=3:
                pr=2**r
                a = integer_mod.Mod(5, pr)
                a._IntegerMod__order = 2**(r-2)
                return [integer_mod.Mod(-1,pr), a]
            assert False, "p=2, r=%s should be >=1"%r
        else:  # odd prime
            pr = p**r
            R = IntegerModRing(pr)
            x = R(self.__unit_gens_primecase(p).lift())
            n = p**(r-2)*(p-1)
            one = integer_mod.Mod(1,pr)
            for b in range(0,p):
                z = x+R(b*p)
                if z**n != one:
                    a = integer_mod.Mod(z,pr)
                    a._IntegerMod__order = p*n
                    return [a]
            assert False, "p=%s, r=%s, couldn't find generator"%(p,r)
        
    def unit_gens(self):
        """
        Returns generators for the (Z/NZ)^*.
        
        We compute the list of generators using a deterministic
        algorithm, so the generators list will always be the same.
        Each generator corresponds to a prime divisor of N (or
        possibly two prime divisors for p=2).

        INPUT: (none)
        OUTPUT:
            list -- a list of elements of self
        """
        try:
            return self.__unit_gens
        except:
            self.__unit_gens = []
        n = self.__modulus
        if n == 1:
            self.__unit_gens = [self(1)]
            return self.__unit_gens
        for p,r in self.factored_modulus():
            m = n/(p**r)
            for g in self.__unit_gens_primepowercase(p, r):
                x = g.crt(integer_mod.Mod(1,m))
                x._IntegerMod__order = g.order()
                self.__unit_gens.append(x)
        return self.__unit_gens
    
    def unit_group_exponent(self):
        if self.__unit_group_exponent != None:
            return self.__unit_group_exponent
        a = []
        for p, r in self.factored_modulus():
            if p != 2:
                a.append((p-1)*(p**(r-1)))   # phi(p**r)
            else:  # p=2
                if r==2:
                    a.append(2)
                elif r>2:
                    a.append(2**(r-2))
            #endif
        #endfor
        self.__unit_group_exponent = int(arith.LCM(a))
        return self.__unit_group_exponent
        
    def random(self):
        a = random.randrange(0,self.__modulus-1)
        return integer_mod.IntegerMod(self, a)
    
Zmod = IntegerModRing

def GF(p):
    if not arith.is_prime(p):
        raise NotImplementedError, "Only prime fields currently implemented."
    return IntegerModRing(p)

def crt(v):
    """CRT(v): v is a lift of elements of rings.IntegerMod(n), for
               various coprime moduli n.
    """
    if len(v) == 0:
        return integer_mod.Mod(1,1)
    x = v[0]
    for i in range(1,len(v)):
        x = x.crt(v[i])
    return x

    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
