"""
The ring of rational integers 
"""

#*****************************************************************************
#
#   SAGE:  System for Arithmetic Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import random

import sage.rings.arith 
import sage.misc.misc as misc
import sage.rings.ring as ring
import sage.rings.integer

_obj_int = {}
class _uniq_int(object):
    def __new__(cls):
        if _obj_int.has_key(0):
            return _obj_int[0]
        O = object.__new__(cls)
        _obj_int[0] = O
        return O

class IntegerRing(_uniq_int, ring.Ring):
    """
    The ring of integers.
    
    EXAMPLES:
    We create the rational numbers and call a few functions:
        >>> Z = IntegerRing(); Z
        Integer Ring
        >>> Z.characteristic()
        0
        >>> Z.is_field()
        False
        
    We illustrate arithmetic in Z.
        >>> a = Z(1234); b = Z(5678); print a, b
        1234 5678
        >>> type(a)
        <type '_integer.Integer'>
        >>> a + b
        6912
        >>> b + a
        6912
        >>> a * b
        7006652
        >>> b * a
        7006652
        >>> a - b
        -4444
        >>> b - a
        4444
        
    When we divide, the result is automatically coerced to the rationals,
    even if the result is an integer.
        >>> a / b
        617/2839
        >>> type(a/b)
        <type '_rational.Rational'>
        >>> a/a
        1
        >>> type(a/a)
        <type '_rational.Rational'>
        
    Next we illustrate arithmetic with automatic coercion.
    The types that coerce are: str, int, long, Integer.
        >>> a + 17
        1251
        >>> a * 374
        461516
        >>> 374 * a
        461516
        >>> a/19
        1234/19
        >>> 0 + Z(-64)
        -64
        
    Integers can be coerced:
        >>> a = Z(-64)
        >>> int(a)
        -64
        >>> long(a)
        -64L
    """
    
    
    def __contains__(self, x):
        return isinstance(x, sage.rings.integer.Integer)
    
    def __repr__(self):
        return "Integer Ring"
    
    def __call__(self, x):
        return sage.rings.integer.Integer(x)
        #z.set_str("%s"%hex(x).replace("0x","").replace("L",""),16)
        #return z

    def is_atomic_repr(self):
        """
        Return True, since elements of the integers do not have
        to be printed with paranethesis around them, when they
        are coefficients, e.g., in a polynomial.
        """
        return True
    
    def is_field(self):
        """
        Return False.
        """
        return False
    
    def name(self):
        """
        Return the string "Z".
        """
        return "Z"
    
    def characteristic(self):
        """
        Return 0 as a Python int.
        """
        return 0
    
    def random(self, bound=5):
        """
        Return a random integer between -bound and bound,
        including both endpoints.
        """
        return self(random.randint(-bound,bound))
    
    def __cmp__(self, other):
        if isinstance(other, IntegerRing):
            return 0
        return -1
    
    def zeta(self):
        """
        Return the integer -1, which is the root of unity of largest
        order in Z.
        """
        return sage.rings.integer.Integer(-1)


Z = IntegerRing()

def factor(n):
    return [(Z(p),e) for p, e in sage.rings.arith.factor(long(n))]
    
def crt_basis(X, xgcd=None):
    """
    Compute and return a Chinese Remainder Theorem basis for the list
    X of coprime integers.
    
    INPUT:
        X -- a list of Integers that are coprime in pairs
    OUTPUT:
        E -- a list of Integers such that E[i] = 1 (mod X[i])
             and E[i] = 0 (mod X[j]) for all j!=i.

    The E[i] have the property that if A is a list of objects, e.g.,
    integers, vectors, matrices, etc., where A[i] is moduli X[i], then
    a CRT lift of A is simply
                       sum E[i] * A[i].

    ALGORITHM:
    To compute E[i], compute integers s and t such that
    
                s * X[i] + t * (prod over i!=j of X[j]) = 1.   (*)

    Then E[i] = t * (prod over i!=j of X[j]).  Notice that equation
    (*) implies that E[i] is congruent to 1 modulo X[i] and to 0
    modulo the other X[j] for j!=i.

    COMPLEXITY: We compute len(X) extended GCD's.
    """
    if not isinstance(X, list):
        raise TypeError, "X must be a list"
    if len(X) == 0:
        return []

    P = misc.mul(X)

    Y = []
    # 2. Compute extended GCD's
    ONE=X[0].parent()(1)
    for i in range(len(X)):
        p = X[i]
        prod = P/p
        g,s,t = p.xgcd(prod)
        if g != ONE:
            raise ArithmeticError, "The elements of the list X must be coprime in pairs."
        Y.append(t*prod)
    return Y
             
    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
