"""
Basic interval arithmetic class
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import math
import operator

from coerce import bin_op
import sage.rings.ring as ring
import sage.rings.rings 
import sage.rings.element as element

def _adapt(x):
    if isinstance(x, Interval):
        return x
    if isinstance(x, (int, long, float, sage.rings.rings.Integer,
                      sage.rings.rings.Rational, sage.rings.rings.RealNumber)):
        return Interval(x)
    raise TypeError, "cannot construct interval from %s"%x

_obj = {}
class _uniq(object):
    def __new__(cls):
        if _obj.has_key(0):
            return _obj[0]
        O = object.__new__(cls)
        _obj[0] = O
        return O

class IntervalRing(ring.Ring, _uniq):
    def __contains__(self, x):
        return isinstance(x, Interval)
    
    def __repr__(self):
        return 'Float interval arithmetic pseudoring.'

    def __call__(self, x, y=None):
        if y == None:
            return _adapt(x)
        else:
            return Interval(x,y)

    def is_atomic_repr(self):
        return True

    def is_field(self):
        return True

    def name(self):
        return "interval"

    def characteristic(self):
        return 0

    def random(self, bound=0):
        return Interval(0)

    def __cmp__(self, other):
        if isinstance(other, IntervalRing):
            return 0
        return -1

    def zeta(self):
        return Interval(-1)

_inst = IntervalRing()  # make sure there is at least one instance.


class Interval(element.Element):

    def __init__(self, min, max=None):
        if max == None: max = min
        self.__min, self.__max = float(min), float(max)
        assert self.__min <= self.__max
        element.Element.__init__(self, _inst)
        
    def __repr__(self):
        return "[%s,%s]"%(self.__min, self.__max)

    def min(self):
        return self.__min
    lower = min

    def max(self):
        return self.__max
    upper = max

    def __add__(self, other):
        if not isinstance(other, Interval):
            return bin_op(self, other, operator.add)
        return Interval(self.__min+other.__min, self.__max + other.__max)

    def __sub__(self, other):
        if not isinstance(other, Interval):
            return bin_op(self, other, operator.sub)
        return Interval(self.__min+- other.__max, self.__max - other.__min)

    def __neg__(self):
        return Interval(-self.__max, -self.__min)

    def __mul__(self, other):
        if not isinstance(other, Interval):
            return bin_op(self, other, operator.mul)
        xl,xu,yl,yu = self.__min, self.__max, other.__min, other.__max
        return Interval(min(xl*yl,xl*yu,xu*yl,xu*yu), \
                        max(xl*yl,xl*yu,xu*yl,xu*yu))

    def __invert__(self):
        if 0 in self:
            raise ZeroDivisionError, "cannot invert interval %s"%self
        return Interval(1/self.__max, 1/self.__min)

    def __div__(self, other):
        if not isinstance(other, Interval):
            return bin_op(self, other, operator.div)
        return self * ~other

    def __contains__(self, x):
        x = float(x)
        return self.__min <= x and x <= self.__max
    
    def __cmp__(self, other):
        try:
            other = _adapt(other)
        except TypeError:
            return -1
        if self.__max < other.__min:
            return -1
        elif self.__min > other.__max:
            return 1
        elif self.__min == other.__min and self.__max == other.__max:
            return 0
        else:
            return -1
        #   raise ArithmeticError, "Unable to compare intervals %s and %s"%(self, other)

    def sqrt(self):
        return Interval(math.sqrt(self.__min), math.sqrt(self.__max))

    def length(self):
        return self.__max - self.__min

    def __int__(self):
        """
        If there is a unique integer in the interval, return it.
        Otherwise raise a ValueError exception.
        """
        x = math.ceil(self.__min)
        y = math.floor(self.__max)
        if x != y:
            raise ValueError, "Cannot coerce to int because there is not a unique integer in the interval %s."%self
        return int(x)

    def is_int(self):
        try:
            x = int(self)
        except ValueError:
            return False, None
        return True, x
            
# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from sage.all import *
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])
