"""
A Laurent series
"""

import operator

from element import Element
import power_series
import power_series_ring
import polynomial
import laurent_series_ring

class LaurentSeries(Element):
    def __init__(self, a, b=None, n=0):
        """
        Create a Laurent series.

        LaurentSeries(f)
           f: Polynomial, PowerSeries, or LaurentSeries

        LaurentSeries(parent, f, n)
           parent: LaurentSeriesRing
           f: Polynomial, PowerSeries, or LaurentSeries
           n: int   (power of t)
        """
        if isinstance(a, polynomial.Polynomial):
            u = power_series.PowerSeries(a)
            R = a.parent()
            parent = laurent_series_ring.LaurentSeriesRing(R.base_ring(),R.variable())
        elif isinstance(a, LaurentSeries):
            u = a.unit_part()
            parent = a.parent()
            n = a.valuation()
        else:
            parent = a
            u = b
            if not isinstance(parent, laurent_series_ring.LaurentSeriesRing):
                raise TypeError, "First argument must be a LaurentSeriesRing."
            if not isinstance(u, power_series.PowerSeries):
                try:
                    u = power_series_ring.PowerSeriesRing(parent.base_ring(),parent.variable())(u)
                except TypeError:
                    raise TypeError, "%s not coercible"%u
        try:
            n = int(n)
        except:
            raise TypeError, "Argument three (%s) must be an integer."%n
        
        #    parent: LaurentSeriesRing
        #    u: PowerSeries
        #    n: int
        Element.__init__(self, parent, laurent_series_ring.LaurentSeriesRing)
        self.__repr = None
        self.__u = PowerSeries(u)
        self.__n = n
        self.__normalize()

    def is_zero(self):
        return self.__u.is_zero()
        
    def __call__(self, x):
        """Compute value of this Laurent series at x."""
        return (x**self.valuation())*self.unit_part()(x)

    def __normalize(self):
        """
        A Laurent series is a pair (u(t), n), where either u=0 (to
        some precision) or u is a unit.  This pair corresponds to
        t^n*u(t).
        """
        if self.is_zero():
            return
        v = self.__u.valuation()
        if v == 0:
            return
        self.__n += v
        self.__u = self.__u.unit_part()
        
    def __repr__(self):
        return self.__u._PowerSeries__print_helper(self.__n)

    def __getitem__(self, i):
        return self.__u[i-self.__n]

    def __setitem__(self, i, value):
        j = i - self.__n
        if j >= 0:
            self.__u[j] = value
        else: # off to the left
            if value != 0:
                self.__n = self.__n + j
                R = self.parent().base_ring()
                coeffs = [value] + [R(0) for _ in range(1,-j)] + self.__u.coeffs()
                self.__u = self.__u.parent()(coeffs)
        self.__normalize()

    def __add__(self, right):
        if not isinstance(right, LaurentSeries):
            return bin_op(self, right, operator.add)
        # Algorithm: Make a copy of self then modify coefficients
        # using those of right.
        if right.is_zero():
            return self.add_bigoh(right.prec())
        if self.is_zero():
            return right.add_bigoh(self.prec())
        s = LaurentSeries(self)
        prec = min(self.prec(),right.prec())
        if prec != infinity:
            m = prec
        else:
            m = max(self.degree(), right.degree())+1
        for i in range(right.valuation(), m):
            s[i] += right[i]
        return s.add_bigoh(prec)

    def __radd__(self, left):
        return bin_op(left, self, operator.add)

    def add_bigoh(self, prec):
        if prec == infinity or prec >= self.prec():
            return LaurentSeries(self)
        u = self.__u.add_bigoh(prec - self.__n)
        return LaurentSeries(self.parent(), u, self.__n)

    def degree(self):
        return self.__u.degree() + self.valuation()

    def __sub__(self, right):
        return self + (-1)*right

    def __neg__(self):
        return (-1)*self

    def __mul__(self, right):
        if not isinstance(right, LaurentSeries):
            return bin_op(self, right, operator.mul)
        return LaurentSeries(self.parent(),
                             self.__u * right.__u,
                             self.__n + right.__n)
    def __rmul__(self, left):
        return bin_op(left, self, operator.mul)

    def __pow__(self, right):
        right=int(right)
        return LaurentSeries(self.parent(), self.__u**right, self.__n*right)

        
    def __div__(self, right):
        if not isinstance(right, LaurentSeries):
            return bin_op(self, right, operator.div)
        return LaurentSeries(self.parent(),
                             self.__u / right.__u,
                             self.__n - right.__n)

    def __rdiv__(self, left):
        return bin_op(left, self, operator.div)
        

    def __cmp__(self, other):
        if not isinstance(other, LaurentSeries):
            return cmp(self, other)
        if self.__n < other.__n:
            return -1
        elif self.__n > other.__n:
            return 1
        return cmp(self.__u, other.__u)

    def unit_part(self):
        return self.__u

    def valuation(self):
         return self.__n

    def variable(self):
        return self.parent().variable()

    def prec(self):
        """
        This function returns the n so that the Laurent series is
        of the form (stuff) + O(t^n).  It doesn't matter how many
        negative powers appear in the expansion.  In particular,
        prec could be negative.
        
        """
        return self.__u.prec() + self.valuation()
