"""
Number field element
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import coerce
import element
import infinity
import polynomial
import polynomial_ring
import rational_field
import arith
import sage.misc.misc as misc
import number_field
from sage.pari.py_pari import gen
QQ = rational_field.RationalField()

class NumberFieldElement(element.Element):
    """
    An element of a number field.
    
    EXAMPLES:
    The following examples illustrate creation of elements of
    number fields, and some basic arithmetic.
    
    First we define a polynomial over Q.
        >>> from rational_field import RationalField
        >>> from polynomial_ring import PolynomialRing
        >>> from number_field import NumberField
        >>> x = PolynomialRing(RationalField()).gen()
        >>> f = x**2 + 1
        
    Next we use f to define the number field.
        >>> K = NumberField(f, "a"); K
        Number Field with defining polynomial x^2 + 1
        >>> a = K.gen()
        >>> a**2
        -1
        >>> (a+1)**2
        2*a
        >>> a**2
        -1
        >>> z = K(5); 1/z
        1/5
    
    We create a cube root of 2.
        >>> K = NumberField(x**3 - 2, "b")
        >>> b = K.gen()
        >>> b**3
        2
        >>> (b**2 + b + 1)**3
        12*b^2 + 15*b + 19
    """
    def __init__(self, parent, f):
        if isinstance(f, gen):
            f = f.lift()
            f = parent.polynomial_ring()(f)
        if not isinstance(f, polynomial.Polynomial):
            f = parent.polynomial_ring()(f)
        if f.degree() >= parent.degree():
            f %= parent.polynomial()
        element.Element.__init__(self, parent)
        self.__element = f
        self.__order = None

    def __repr__(self):
        x = self.__element
        return str(x).replace(x.parent().variable(),self.parent().variable())

    def pari(self, var=None):
        """
        Return PARI representation of self.
        """
        try:
            return self.__pari
        except:
            if var == None:
                var = self.parent().variable()
            f = self.__element.pari().subst("x",var)
            g = self.parent().polynomial().pari().subst("x",var)
            return f.Mod(g)
            

    def __cmp__(self, other):
        if not isinstance(other, NumberFieldElement):
            return coerce.cmp(self, other)
        if self.parent() != other.parent():
            return -1
        return misc.generic_cmp(self.__element, other.__element)

    def __add__(self, other):
        if not isinstance(other, NumberFieldElement):
            return coerce.bin_op(self, other, operator.add)
        return NumberFieldElement(self.parent(), self.__element+other.__element)

    def __radd__(self, left):
        return self.parent()(left) + self
    
    def __mul__(self, other):
        """
        Returns the product of self and other as elements of a number field.

        NOTES: In LiDIA, they build a multiplication table for the
        number field, so it's not necessary to reduce modulo the
        defining polynomial every time:
             src/number_fields/algebraic_num/order.cc: compute_table
        
        """
        if not isinstance(other, NumberFieldElement):
            return coerce.bin_op(self, other, operator.mul)
        return NumberFieldElement(self.parent(), self.__element * other.__element)
    
    def __rmul__(self, left):
        return self.parent()(left)*self

    def __pow__(self, right):
        right = int(right)
        if right < 0:
            x = self.__invert__()
            right *= -1
            return arith.generic_power(x, right)
        return arith.generic_power(self, right)
    
    def __sub__(self, other):
        if not isinstance(other, NumberFieldElement):
            return coerce.bin_op(self, other, operator.sub)
        return NumberFieldElement(self.parent(), self.__element - other.__element)
    
    def __rsub__(self, left):
        return self.parent()(left) - self
    
    def __div__(self, other):
        if not isinstance(other, NumberFieldElement):
            return coerce.bin_op(self, other, operator.div)
        return self * other.__invert__()
    
    def __rdiv__(self, left):
        return self.parent()(left) / self        

    def __neg__(self):
        return NumberFieldElement(self.parent(), -self.__element)

    def __int__(self):
        return int(self.__element)

    def __long__(self):
        return long(self.__element)
    
    def __invert__(self):
        if self.__element.is_zero():
            raise ZeroDivisionError
        g, _, a = self.parent().polynomial().xgcd(self.__element)
        assert g.degree() == 0
        c = g[0]
        return NumberFieldElement(self.parent(), (1/c)*a)

    def polynomial(self):
        return self.__element

    def order(self):
        if self.__order != None:
            return self.__order
        if self.__element == 1:
            self.__order = 1
            return self.__order
        if self.__element == -1:
            self.__order = 2
            return self.__order
        if isinstance(self.parent(), number_field.CyclotomicField):
            t = self.parent().order_table()
            f = self.__element
            if t.has_key(f):
                self.__order = t[f]
                return self.__order
            
        ####################################################################
        # Algorithm to compute the order of an element x of a number field K.
        #
        # 1. Find an integer B such that if n>=B then phi(n) > deg(K).
        #    For this use that for n>6 we have phi(n) >= log_2(n)
        #    (to see this thing about the worst prime factorization
        #    in the multiplicative formula for phi.)
        # 2. Compute x, x^2, ..., x^B in order to determine the order.
        #
        # todo-- Alternative: Only do the above if we don't require an optional
        # argument which gives a multiple of the order, which is usually
        # something available in any actual appication.
        ####################################################################
        d = self.parent().degree()
        B = max(7, 2**d+1)
        x = self
        for i in range(1,B):
            if x == 1:
                self.__order = i
                return self.__order
            x *= self
        self.__order = infinity.infinity
        return infinity.infinity

    def trace(self):
        return QQ(self.pari().trace())
        # return self.matrix().trace()

    def norm(self):
        return QQ(self.pari().norm())
        #return self.matrix().determinant()

    def charpoly(self):
        R = polynomial_ring.PolynomialRing(QQ)
        return R(self.pari().charpoly())
        # return self.matrix().charpoly()

    def minpoly(self):
        # The minimal polynomial is square-free and
        # divisible by same irreducible factors as
        # the characteristic polynomial.
        # TODO: factoring to find the square-free part is idiotic.
        # Instead use a GCD algorithm!
        f = polynomial_ring.PolynomialRing(QQ)(1)
        for g, _ in self.charpoly().factor():
            f *= g
        return f

    def matrix(self):
        """
        The matrix of left multiplication by the element
        on the power basis 1, x, x^2, ..., x^(d-1) for the
        number field.
        """
        # Mutiply each power of field generator on
        # the left by this element; make matrix
        # whose rows are the coefficients of the result,
        # and transpose.
        K = self.parent()
        v = []
        x = K.gen()
        a = K(1)
        d = K.degree()
        for n in range(d):
            v += (self*a).list()
            a *= x
        Q = rational_field.RationalField()
        import sage.linalg.linalg
        M = sage.linalg.linalg.MatrixSpace(rational_field.RationalField(), d)
        A = M(v)
        return M(v).transpose()

    def list(self):
        return self.__element[:self.parent().degree()]


if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
