"""
p-adic Ring
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import operator

from element import Element
from infinity import infinity

import arith
import coerce
import integer
import padic_field
import sage.rings.arith
import rational

class pAdic(Element):
    """
    The field of p-adic numbers of either finite or infinite precision.
    We represent p-adic numbers as product (p**r)*unit, where r is an integer
    or Infinity and unit is a p-adic unit known to some precision.
    If r=Infinity, then (p**r)*unit is the 0 element. 

    Binary operations on two elements of Qp reduce the precision of
    the unit part of the argument with larger precision to that of the
    one with lesser precision.  This applies to all operations,
    including equality testing, so, e.g., the element O(p) is equal to
    every p-adic integer, since comparison will truncate the other
    p-adic integer to precision O(p).
    """

    def __init__(self, parent, x, big_oh=infinity, ordp=None):
        """The parent is a p-adic field.  The second argument, x, is
        anything that can be coerced to a p-adic number.  The third
        argument big_oh is such that, e.g. 3^(-1)+1+2*3+O(3^2)
        has big_oh equal to 2."""
        Element.__init__(self, parent, padic_field.pAdicField)
        if isinstance(x, pAdic):
            self.__parent = x.__parent
            self.__p = x.__p
            self.__unit = x.__unit
            self.__ordp = x.__ordp
            if big_oh == infinity:
                self.__prec = x.__prec
            else:
                if x.__ordp != infinity:
                    self.__prec = min(x.__prec, big_oh-x.__ordp)
                else:
                    self.__prec = min(x.__prec, big_oh)                    
            return

        self.__p = int(parent.residue_characteristic())
        self.__parent = parent
        
        if isinstance(x, integer.Integer):
            x = int(x)
        elif isinstance(x, rational.Rational):
            if x == 0:
                x=0
            else:
                ordp = x.valuation(self.__p)
                ppow = rational.Rational(self.__p)**ordp
                x = x/ppow
                if x.denominator() == 1:
                    x = int(x)
                else:
                    pr = min(parent.prec(), big_oh - ordp)
                    x = int(x.numerator())*\
                        arith.inverse_mod(int(x.denominator()), self.__p**pr)
                    big_oh = pr + ordp
        if x==0:
            unit = 1
            ordp = infinity
        elif ordp != None:
            unit = int(x)   
        else:  # ord_p is None, so we compute it from x, which need not be a unit.
            p = self.__p
            if isinstance(x, (int, long)):
                ordp = sage.rings.arith.valuation(x,p)
                unit = x/(p**ordp)
            elif isinstance(x, rational.Rational):
                ordp = x.valuation(p)
                unit = int(x/(p**ordp))
            else:
                raise NotImplementedError
        self.__unit = unit
        self.__ordp = ordp
        if big_oh == infinity and ordp == infinity:
            self.__prec = infinity
        else:
            if ordp != infinity:
                self.__prec = big_oh - ordp  # since internal prec is of the unit part
            else:
                self.__prec = big_oh
        self.__order = None

    def denominator(self):
        return self.__p**(-min(self.__ordp,0))

    def ordp(self):
        return self.__ordp

    def valuation(self):
        return self.__ordp

    def order(self):
        """
        The order of this as an element of the multiplicative group.
        If the element is known to infinite precision, then it is
        truncated to the parent precision before the order is computed.
        
        >>> K = Qp(13)
        >>> a = K(-1)
        >>> a.order()
        2

        We immediately know that 13 has infinite order, since it is 0
        modulo 13.
        >>> b = K(13)
        >>> b.order()
        Infinity

        The following element has finite order modulo 5**2.
        >>> c = 3 + 3*5 + 2*5**2 + O(5**3)
        >>> c.order()
        4
        """
        if self.__ordp != 0:
            return infinity
        if self == 1:
            return 1
        if self == -1:
            return 2
        if self.__prec == infinity:
            prec = min(self.__prec, self.parent().prec())
        else:
            prec = self.__prec
        return Mod(self.__unit, self.__p**prec).order()

    def copy(self):
        return pAdic(self.__parent, self.__unit, self.__prec, self.__ordp)

    def big_oh(self):
        if self.__ordp == infinity:
            return self.__prec
        return self.__prec + self.__ordp

    def __repr__(self):
        if not self.parent().series_print():
            return "%s^%s * (%s + O(%s^%s))"%(self.__p, self.__ordp, \
                                              self.__unit, self.__p, self.__prec)
        # series printing
        if self.__ordp == infinity:
            return "0"
            #return "0 + O(%o^Infinity)"%(self.__p)
        if self.__ordp == 0 and self.__prec == infinity and self.__unit == 1:
            return "1"
        s     = ""
        u     = self.__unit
        exp   = self.__ordp
        p     = self.__p
        prec  = min(self.__prec, self.parent().print_prec()-self.__ordp)
        if prec == infinity:
            prec = self.parent().prec()
        u   %= self.__p ** prec
        while u != 0:
            coeff = u % p
            if coeff != 0:
                if exp == 0:
                    s += "%s + "%coeff
                else:
                    var = "%s"%p
                    if exp != 1:
                        var += "^%s"%exp
                    if coeff != 1:
                        s += "%s*%s + "%(coeff,var)
                    else:
                        s += "%s + "%var
            exp += 1
            u = (u-coeff)/p
        s += "O(%s"%(p)
        if self.big_oh() == 1:
            s += ")"
        else:
            s += "^%s)"%self.big_oh()
        return s

    def __add__(self, right):
        if not isinstance(right, pAdic):
            return coerce.bin_op(self, right, operator.add)
        if self.__p != right.__p:
            raise errors.MathError("Addition of %s and %s not defined."%(self,right))
        if self.__ordp <= right.__ordp:
            x = self; y = right
        else:
            x = right; y = self
        big_oh = min(x.big_oh(), y.big_oh())
        if y == 0:
            return pAdic(self.__parent, x, big_oh)
        p = x.__p
        a = x.__unit + p**(y.__ordp - x.__ordp)*y.__unit
        n = arith.valuation(a,p)
        if n == infinity:    # 0
            return pAdic(self.__parent, 0, big_oh)
        a /= p**n
        prec = big_oh - n - x.__ordp
        if prec != infinity:
            a %= p**prec
        if a==0:
            return pAdic(self.__parent, 0, big_oh)
        return pAdic(self.__parent, a, big_oh ,x.__ordp + n)

    def __radd__(self, left):
        return self.parent()(left) + self

    def __sub__(self, right):
        if not isinstance(right, pAdic):
            return coerce.bin_op(self, right, operator.sub)
        if self.__p != right.__p:
            raise errors.MathError("Addition of %s and %s not defined."%(self,right))
        return self + (-right)

    def __rsub__(self, left):
        return self.parent()(left) - self        

    def __mul__(self, right):
        if not isinstance(right, pAdic):
            return coerce.bin_op(self, right, operator.mul)
        if self.__p != right.__p:
            raise errors.MathError("Addition of %s and %s not defined."%(self,right))
        prec = min(self.__prec, right.__prec)
        ordp = self.__ordp + right.__ordp
        unit = self.__unit * right.__unit
        if prec != infinity:
            unit %= self.__p**prec
        return pAdic(self.__parent, unit, prec+ordp, ordp)

    def __rmul__(self, left):
        return self.parent()(left)*self

    def __div__(self, right):
        if not isinstance(right, pAdic):
            return coerce.bin_op(self, right, operator.div)
        if self.__p != right.__p:
            raise errors.MathError("Addition of %s and %s not defined."%(self,right))
        return self*right.__invert__()

    def __rdiv__(self, left):
        return self.parent()(left)/self        

    def __mod__(self, right):
        if self.__ordp < 0:
            raise ZeroDivisionError, \
                  "Reduction of %s not defined (there is a denominator)."%(self)
        if self.__ordp >= 0 and isinstance(right,(int,long)):
            n = arith.valuation(right,self.__p)
            if self.__p**n == right:  # i.e., right is a p-power
                if n > self.__prec + self.__ordp:
                    raise errors.MathError("%s not known to large enough precision "%self + \
                                           " to reduce modulo %s^%s."%(self.__p, n))
                p = Mod(self.__p,right)
                u = Mod(self.__unit, right)
                return p**self.__ordp * u
        raise ZeroDivisionError, "Reduction of %s mod %s not defined."%(self,right)

    def __pow__(self, right):
        right = int(right)
        if right < 0:
            inv = 1/self
            return inv**(-right)
        if right == 0:
            return pAdic(self.__parent, 1)
        ordp = right * self.__ordp
        if self.__prec == infinity:
            prec = self.parent().prec()
        else:
            prec = self.__prec
        unit = arith.power_mod(self.__unit, right, self.__p**prec)
        return pAdic(self.__parent, unit, prec + ordp, ordp)

    def __neg__(self):
        u = - self.__unit
        return pAdic(self.__parent, u, self.__prec+self.__ordp, self.__ordp)

    def __pos__(self):
        return self

    def __invert__(self):
        if self.__prec == infinity:
            prec = self.parent().prec()
        else:
            prec = self.__prec
        if prec <= 0:
            raise ZeroDivisionError, "Can not invert %s"%self
        unit = arith.inverse_mod(self.__unit, self.__p**prec)
        big_oh = prec - self.__ordp
        return pAdic(self.__parent, unit, big_oh,  -self.__ordp)

    def lift(self):
        if self.is_zero():
            return frac(0,1)
        p = self.__p
        if self.__prec == infinity:
            return frac(p,1)**self.__ordp * frac(self.__unit,1)
        return self.__unit*(p**self.__prec)

    def __cmp__(self, other):
        if not isinstance(other, pAdic):
            other = self.parent()(other)
        if self.__p != other.__p:
            return (self.__p).__cmp__(other.__p)
        m = min(self.big_oh(), other.big_oh())
        x_ordp = self.__ordp
        if x_ordp >= m :
            x_ordp = infinity
        y_ordp = other.__ordp
        if y_ordp >= m :
            y_ordp = infinity
        if x_ordp < y_ordp:
            return -1
        elif x_ordp > y_ordp:
            return 1
        else:  # equal ordp
            if x_ordp == infinity:
                return 0 # since both are zero
        p = self.__p
        a = self.__unit
        b = other.__unit
        prec = min(self.__prec, other.__prec)
        if prec != infinity:
            ppow = p**prec
            a %= ppow
            b %= ppow
        if a < b:
            return -1
        elif a == b:
            return 0
        else:
            return 1

    def unit_part(self):
        """The unit part of (p**r)*u, which is simply u.
        >>> x = 9*(2+3+O(3**7))
        >>> x.unit_part()
        2 + 3 + O(3^7)
        """
        return pAdic(self.__parent, self.__unit, self.__prec, 0)

    def is_zero(self):
        return self.__ordp >= self.big_oh()

    def is_unit(self):
        return self.__ordp == 0

    def rational_reconstruction(self):
        """Try to lift the p-adic number to the rationals
        using rational reconstruction, as follows:
        Suppose the p-adic number is p^r*(u+O(p^n)), where u
        is a unit.  Using rational reconstruction, try to
        find the unique rational number a/b such that a/b is
        congruent to u modulo p^n, and abs(a), abs(b) are
        both at most sqrt{p/2}.  If such a/b exists, return
        p^r*(a/b).
        """
        if self.is_zero():
            return frac(0,1)
        p = self.__p
        if self.__prec == infinity:
            return frac(p,1)**self.__ordp * frac(self.__unit,1)
        alpha = self.__unit
        m = p**self.__prec
        a, b = arith.rational_reconstruction(alpha, m)
        if b==0:
            return (0,0)
        return (frac(p,1)**self.__ordp)*frac(a, b)

    def log(self):
        """Compute the p-adic logarithm of a unit in Z_p.
        
        The usual power series for log with values in the additive
        group of Q_p only converges for 1-units (units congruent to 1
        modulo p).  However, there is a unique extension of log to a
        homomorphism defined on all the units.  If u = a*v is a
        unit with v = 1 (mod p), then we define log(u) = log(v).
        This is the correct extension because the units U of Z_p
        splits as a product U = V x <w>, where V is the subgroup
        of 1-units and w is a (p-1)st root of unity.  The <w> factor
        is torsion, so must go to 0 under any homomorphism to the
        torsion free group (Q_p, +).

        Notes -- What some other systems do:
           PARI:  Seems to define log the same way as we do.
           MAGMA: Gives an error when unit is not a 1-unit.

        Algorithm:
           Input: Some p-adic unit u.
           1. Check that the input p-adic number is really a unit
              (i.e., valuation 0)
           2. Let 1-x = u**(p-1), which is a 1-unit.
           3. Use the series expansion
           
                log(1-x) = F(x) = -x - 1/2*x^2 - 1/3*x^3 - 1/4*x^4 - 1/5*x^5 - ...
                
              to compute the logarithm log(u**(p-1)).  Use enough
              terms so that terms added on are zero (to the default
              precision, if the input has infinite precision).
           4. Then log(u) = log(u**(p-1))/(p-1) = F(1-u**(p-1))/(p-1).
        
        Examples:

        """

        # Step 1 -- a unit?
        if self.is_unit() and self.__unit % self.__p == 1:
            # It's already a 1-unit, so just use the series (base case of "induction")
            x = 1 - self
            xpow = x
            ans = self.parent()(0)
            if self.__prec == infinity:
                prec = self.parent().prec()
            else:
                prec = self.__prec
            for n in range(1,prec):  
                ans -= xpow/n
                xpow *= x
            return pAdic(self.__parent, ans, prec)
        else:
            z = self.unit_part()
            p = self.__p
            return (z**(p-1)).log()/(p-1)
