"""
Univariate Polynomial
"""

import operator

import sage.rings.rational_field 
import arith
import element
import integer_ring
import polynomial_pyx
import rational_field
import complex_field
from infinity import infinity
import sage.misc.misc as misc
from sage.pari.py_pari import pari

from coerce import bin_op

class Polynomial(element.Element):
    """
    The following examples illustrate creation of elements of
    polynomial rings, and some basic arithmetic.
    
    First we make a polynomial over the integers and do some arithmetic:
        >>> x = PolynomialRing(IntegerRing()).gen()
        >>> f = x**5 + 2*x**2 + (-1); f
        x^5 + 2*x^2 - 1
        >>> f**2
        x^10 + 4*x^7 - 2*x^5 + 4*x^4 - 4*x^2 + 1
    
    Next we do arithmetic in a sparse polynomial ring over the integers:
        >>> R = PolynomialRing(IntegerRing(), "x"); x = R.gen(); R
        Univariate Polynomial Ring in x over Integer Ring
        >>> S = PolynomialRing(R, "Z"); Z = S.gen(); S
        Univariate Polynomial Ring in Z over Univariate Polynomial Ring in x over Integer Ring
        >>> f = Z**3 + (x**2-2*x+1)*Z - 3; f
        Z^3 + (x^2 - 2*x + 1)*Z + -3
        >>> f*f
        Z^6 + (2*x^2 - 4*x + 2)*Z^4 + (-6)*Z^3 + (x^4 - 4*x^3 + 6*x^2 - 4*x + 1)*Z^2 + (-6*x^2 + 12*x - 6)*Z + 9
        >>> f**3 == f*f*f
        True
        
    To have the element print as 'y', give 'y' as the
    second argument to the PolynomialRing constructor.
        >>> y = PolynomialRing(IntegerRing(), 'y').gen()
        >>> y**3 - 2*y
        y^3 - 2*y
    """
    
    def __init__(self, parent, is_gen = False): 
        element.Element.__init__(self, parent)
        self._is_gen = is_gen

    def __add__(self, right):
        if not isinstance(right, Polynomial) or self.parent() != right.parent():
            return bin_op(self, right, operator.add)

        if self.degree() >= right.degree():
            x = list(self.list())
            y = right.list()
        else:
            x = list(right.list())
            y = self.list()
            
        for i in xrange(len(y)):
            x[i] += y[i]
        
        return self.polynomial(x)
        
    def __call__(self, x):
        """
        Compute value of this polynomial at x.
        """
        try:
            P = x.parent()
            ans = P(self[0])
        except AttributeError:
            ans = self[0]
        powr = x
        d = self.degree()
        for i in range(1, d+1):
            if self[i] != 0:
                ans += powr*self[i]
            if i <= d:
                powr *= x
        return ans

    def __cmp__(self, other):
        if not isinstance(other, Polynomial):
            try:
                other = self.parent()(other)
            except TypeError:
                return -1
        if self.parent() != other.parent():
            return -1
        elif self.degree() < other.degree():
            return -1
        elif self.degree() > other.degree():
            return 1
        if self.list() < other.list():
            return -1
        elif self.list() > other.list():
            return 1
        else:
            return 0
            
    def __div__(self, other):
        """
        Division with remainder.  Returns a tuple (quotient, remainder).
        """
        #       idea: do with ints modulo several primes, and use CRT.
        #             what is the bound on coefficients of quotient and remainder?
        #             note -- can also just multiply out to see that results is correct.
        if not isinstance(other,Polynomial) or self.parent() != other.parent():
            return bin_op(self, other, operator.div)
        if other.is_zero(): 
            raise ZeroDivisionError
            
        # This is algorithm 3.1.1 in Cohen GTM 138
        A = self
        B = other
        R = A
        Q = self.polynomial(0)
        X = self.parent().gen()
        while R.degree() >= B.degree():
            S =  (R.leading()/B.leading()) * X**(R.degree()-B.degree())
            Q += S
            R -= S*B
            
        return (Q, R)

    def __floordiv__(self,right):
        """
        Quotient of division of self by other.  This is denoted //.
        """
        Q, _ = self/right
        return Q
        
    def __getitem__(self, n):
        raise NotImplementedError
        
    def __hash__(self):
        return hash((self.degree(), self[0]))
        
    def __float__(self):
         if self.degree() > 0:
             raise ValueError, "cannot coerce nonconstant polynomial to float"
         return float(self[0])

    def __int__(self):
        if self.degree() > 0:
            raise ValueError, "cannot coerce nonconstant polynomial to int"
        return int(self[0])

    def __long__(self):
        if self.degree() > 0:
            raise ValueError, "cannot coerce nonconstant polynomial to long"
        return long(self[0])

    def __mod__(self, other):
        """
        Remainder of division of self by other.
        EXAMPLES:
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> x % (x+1)
            -1
            >>> (x**3 + x - 1) % (x**2 - 1)
            2*x - 1
        """
        _, R = self/other
        return R
        
    def __mul__(self, right):
        """
        EXAMPLES:
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> (x - 4)*(x**2 - 8*x + 16)
            x^3 - 12*x^2 + 48*x - 64
        """
        if right in self.base_ring():
            v = list(self.list())
            for i in range(len(v)):
                v[i] *= right
            return self.polynomial(v)
        if not isinstance(right,Polynomial) or self.parent() != right.parent():
            return bin_op(self, right, operator.mul)
        if right == 0 or self == 0:
            return self.polynomial(0)
        return self._mul_karatsuba(right)
        
    def __neg__(self):
        return (self.base_ring()(-1))*self

    def __pos__(self):
        return self

    def __pow__(self, right):
        right = int(right)        
        if right < 0:
            raise ValueError, "negative power of polynomial not defined"
        if self._is_gen:   # special case x**n should be faster!
            z = self.parent()(0)
            z[right] = 1
            return z
        return arith.generic_power(self, right)
    
    def __radd__(self, left):
        return bin_op(left, self, operator.add)

    def __rdiv__(self, left):
        return bin_op(left, self, operator.div)

    def __repr__(self):
        s = " "
        m = self.degree() + 1
        r = reversed(xrange(m))
        X = self.parent().variable()
        atomic_repr = self.parent().base_ring().is_atomic_repr()
        coeffs = self.list()
        for n in reversed(xrange(m)):
            x = coeffs[n]
            if x != 0:
                if atomic_repr:
                    if x > 0:
                        sign = "+"
                    else:
                        sign = "-"
                        x = -x
                    if n != m-1:
                        s += " %s "%sign
                    elif sign=="-":
                        s += "%s"%sign
                else:
                    if n != m-1:
                        s += " + "
                    x = "%s"%x
                    if n > 0 and (x.find("+") != -1 or x.find("-") != -1):
                        x = "(%s)"%x
                #end if
                if n > 1:
                    var = "*%s^%s"%(X,n)
                elif n==1:
                    var = "*%s"%X
                else:
                    var = ""
                s += "%s%s"%(x,var)
        s = s.replace(" 1*"," ")
        s = s.replace(" -1*", " -")
        if s==" ":
            return "0"
        return s[1:]
       
    def __rmul__(self, left):
        return bin_op(left, self, operator.mul)
          
    def __rsub__(self, left):
        return bin_op(left, self, operator.sub)
        
    def __setitem__(self, n, x):
        raise NotImplentedError

    def __sub__(self, right):
        """
        The difference self - right.
        
        EXAMPLES:
            >>> import rings
            >>> Q = rings.RationalField()
            >>> x = rings.PolynomialRing(Q).gen()
            >>> x - (x+1)
            -1
        """
        if not isinstance(right,Polynomial):
            return bin_op(self, right, operator.sub)
        return self + (-right)
        
    def _mul_generic(self, right):
        d1 = self.degree()
        d2 = right.degree()
        d = d1 + d2
        w = [sum([self[i]*right[k-i] for i in range(0,min(d1,k)+1) if \
                  i <= d1 and k-i <= d2 and self[i]!=0 and right[k-i]!=0]) \
                for k in range(d+1)]
        return Polynomial(self.parent(), w)

    def _mul_karatsuba(self, right):
        """
        Returns the product of two polynomials using the Karatsuba 
        divide and conquer multiplication algorithm.

        INPUT:
           self: Polynomial
           right: Polynomial (over same base ring as self)

        OUTPUT: Polynomial
           The product self*right.

        ALGORITHM:
           The basic idea is to use that
           
               (a*X+b)*(c*X+d) = ac*X^2+((a+b)*(c+d)-ac-bd)*X+bd
               
           where ac=a*c and bd=b*d, which requires three
           multiplications instead of the naive four.  (In my examples,
           strangely just doing the above with four multiplications
           does tend to speed things up noticeably.)
           Given f and g of arbitrary degree bigger than one, let e
           be min(deg(f),deg(g))/2.  Write
           
                  f = a*X^e + b    and    g = c*X^e + d
                  
           and use the identity
           
                 (a*X^e+b)*(c*X^e+d) = ac*X^(2e) +((a+b)*(c+d)-ac-bd)*X^e + bd
                 
           to recursively compute f*g.

        TIMINGS:
        On a Pentium M 1.8Ghz laptop:
           f=R.random(1000,bound=100)
           g=R.random(1000,bound=100)
           time h=f._mul_karatsuba(g)
           Time: 0.42 seconds
           The naive multiplication algorithm takes 14.58 seconds.
           In contrast, MAGMA does this sort of product almost
           instantly, and can easily deal with degree 5000.  Basically
           MAGMA is 100 times faster at polynomial multiplication.
           Moving the below into C with GMP might make it competitive.

        NOTES:
         * Karatsuba multiplication of polynomials is also implemented in PARI in
                src/basemath/polarit3.c
         * The MAGMA documentation appears to give no information about how
           polynomial multiplication is implemented.
        """
        
        def sum(v,w):
            if len(v)>=len(w):
                x = list(v)
                y = w
            else:
                x = list(w)
                y = v
            for i in range(len(y)):
                x[i] += y[i]
            return x
        def dif(v,w):
            if len(v)>=len(w):
                x = list(v)
                y = w
            else:
                x = list(w)
                y = v
            for i in range(len(y)):
                x[i] -= y[i]
            return x
        def do_karatsuba(left, right):
            if len(left) == 0 or len(right) == 0:
                return []
            if len(left) == 1:
                return [left[0]*a for a in right]
            if len(right) == 1:
                return [right[0]*a for a in left]
            if len(left) == 2 and len(right) == 2:
                b = left[0]
                a = left[1]
                d = right[0]
                c = right[1]
                ac = a*c
                bd = b*d
                return [bd,(a+b)*(c+d)-ac-bd,ac]
            e = min(len(left), len(right))/2
            assert e>=1, "bug in karatsuba"
            a, b = left[e:], left[:e]
            c, d = right[e:], right[:e]
            ac = do_karatsuba(a,c)
            bd = do_karatsuba(b,d)
            zeros = [0 for _ in range(e)]
            t2 = zeros + zeros + ac
            t1 = zeros + dif(do_karatsuba(sum(a,b),sum(c,d)),sum(ac,bd))
            t0 = bd
            return sum(t0,sum(t1,t2))
        return self.parent()(do_karatsuba(self.list(), right.list()))
        
    def base_ring(self):
        """
        Return the base ring of the parent of self.
        
        EXAMPLES:
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> x.base_ring()
            Integer Ring
            >>> (2*x+3).base_ring()
            Integer Ring
        """
        return self.parent().base_ring()
        
    def copy(self):
        """
        Return a copy of self.
        
        EXAMPLES:
        We create the polynomial f=x+3, then set g=f, and
        change the coefficient of x in g, which also changes
        the coefficient of x in f.  If we instead copy f, then
        changing the coefficient of x of g does not change f.
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> f = x+3
            >>> g = f
            >>> g[1]=3
            >>> f
            3*x + 3
            >>> g = f.copy()
            >>> g[1]=5
            >>> f
            3*x + 3
            >>> g
            5*x + 3
        """
        return self.polynomial(self)

    def degree(self):
        """
        Return the degree of this polynomial.  The zero polynomial
        has degree -1.
        EXAMPLES:
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> f = x**93 + 2*x + 1
            >>> f.degree()
            93
            >>> x = PolynomialRing(RationalField(), sparse=True).gen()
            >>> f = x**100000 
            >>> f.degree()
            100000
        """
        raise NotImplementedError

    def denominator(self):
        """
        Return the least common multiple of the denominators of
        the entries of self, when this makes sense, i.e., when the
        coefficients have a denominator function.  
        
        WARNING: This is not the denominator of the rational function 
        defined by self, which would always be 1 since self is a polynomial.
        
        EXAMPLES:
        First we compute the denominator of a polynomial with integer
        coefficients, which is of course 1.
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> f = x**3 + 17*x + 1
            >>> f.denominator()
            1
            
        Next we compute the denominator of a polynomial with rational coefficients.
            >>> x = PolynomialRing(RationalField()).gen()
            >>> f = '1/17'*x**19 - '2/3'*x + '1/3'; f
            1/17*x^19 - 2/3*x + 1/3
            >>> f.denominator()
            51
            
        Finally, we try to compute the denominator of a polynomial with
        coefficients in the real numbers, which is a ring whose elements
        do not have a denominator method.
            >>> x = PolynomialRing(RealField()).gen()
            >>> f = x + '1/3'; f
            1.0*x + 0.333333333333333333333
            >>> f.denominator()
            Traceback (most recent call last):
            ...
            AttributeError: 'RealNumber_mpf' object has no attribute 'denominator'
        """
        if self.degree() == -1:
            return 1
        R = self.base_ring()
        x = self.list()
        d = x[0].denominator()
        for y in x:
            d = d.lcm(y.denominator())
        return d
        
    def derivative(self):
        return self.polynomial([n*self[n] for n in xrange(1,self.degree()+1)])
        
    def dict(self):
        X = {}
        Y = self.list()
        for i in xrange(len(Y)):
            X[i] = Y[i]
        return X
        
#     def factor_ntl(self):
#         """
#         NOTE:
#             I should change this to use PARI, because in many tests with SMALL
#             degree PARI is
#             substantially better than NTL.  (And magma is better yet.)  And
#             the timing difference has nothing to do with moving Python data
#             to NTL and back.
#             For large degree ( > 1500) in the one test I tried, NTL was 
#             much better than MAGMA, and far better than PARI.  So probably
#             NTL's implementation is asymptotically better.  I could use
#             PARI for smaller degree over other rings besides Z, and use
#             NTL in general.
#         """
#         R = self.parent().base_ring()
#         if isinstance(R, (integer_ring.IntegerRing,
#                           rational_field.RationalField)):
#             f = self.list()
#             if isinstance(R, rational_field.RationalField):
#                 d = self.denominator()
#                 if d > 1:
#                     f = (d*self).list()
#             s="[ "
#             for x in f:
#                 s+=" %s "%x
#             s += " ]"
#             s = s.replace("L","")
#             fac = []
#             # Note: The time to use a string to get a polynomial into 
#             # NTL from Python data space is trivial compared to the time
#             # required to actually factor the polynomial.
#             ntl = cxx.polyfactor.NTLPolynomialOverZ(s)
#             t = misc.verbose("start factoring using ntl")
#             ntl.number_of_factors()
#             misc.verbose("finished factoring",t)
#             for n in range(ntl.number_of_factors()):
#                 F = cxx.polyfactor.str(ntl.nth_factor(n))
#                 F = eval(F.replace(" ",","))
#                 F = self.polynomial(F)
#                 e = ntl.nth_exponent(n)
#                 fac.append((F,e))
#             fac.sort()
#             return fac
#         else:
#             raise NotImplementedError

    def factor(self):
        """
        Return polynomials f1, ..., fn and exponents e1, ..., en
        such that the gcd fo the coefficients of the fi is 1, and
        prod fi**ei is equal to a scalar multiple of self.
        """

        # PERFORMANCE NOTE:
        #     In many tests with SMALL degree PARI is substantially
        #     better than NTL.  (And magma is better yet.)  And the
        #     timing difference has nothing to do with moving Python
        #     data to NTL and back.
        #     For large degree ( > 1500) in the one test I tried, NTL was 
        #     much better than MAGMA, and far better than PARI.  So probably
        #     NTL's implementation is asymptotically better.  I could use
        #     PARI for smaller degree over other rings besides Z, and use
        #     NTL in general.
        
        R = self.parent().base_ring()
        if self.degree() <= 0:
            return []
        G = None
        import sage.rings.number_field
        if isinstance(R, (integer_ring.IntegerRing,
                          rational_field.RationalField)):
            G = list(self.pari().factor())
        elif isinstance(R, sage.rings.number_field.NumberField):
            v = [x.pari("a") for x in reversed(self.list())]
            f = pari(v).Pol()
            G = list(f.factor())
        if G is None:
            raise NotImplementedError
        pols = G[0]
        exps = G[1]
        F = []
        R = self.parent()
        for i in xrange(len(pols)):
            F.append((R(pols[i]),int(exps[i])))
        return F
 
    def gcd(self, other):
        """
        Greatest common divisor of self and polynomial other.
        """
        # Algorithm 3.2.1 in Cohen, GTM 138
        A = self
        B = other
        while B.is_nonzero():
            Q, R = A/B
            A = B
            B = R
        return A

    def is_monic(self):
        """
        Returns True if this polynomial is monic.  The zero
        polynomial is by definition not monic.
        """
        return not self.is_zero() and self[self.degree()] == 1
       
    def is_gen(self):
        return self._is_gen

    def is_nonzero(self):
        return self.degree() != -1
    
    def is_zero(self):
        return self.degree() == -1

    def leading(self):
        return self[self.degree()]
        
    def list(self):
        raise NotImplementedError
        
    def polynomial(self, *args, **kwds):
        return self.parent().polynomial(*args, **kwds)

    def complex_roots(self, flag=0):
        """
        Returns the complex roots of this polynomial.
        INPUT:
            flag -- optional, and can be
                    0: (default), uses Schonhage's method modified by Gourdon,
                    1: uses a modified Newton method.
        OUTPUT:
            list of complex roots of this polynomial, counted with multiplicities.
            
        NOTE: Calls the pari function polroots.

        EXAMPLE:
        We compute the roots of the characteristic polynomial of some Salem numbers:
            >>> Q = rational_field.RationalField()
            >>> R = PolynomialRing(Q); x = R.gen()
            >>> f = 1 - x**2 - x**3 - x**4 + x**6
            >>> f.complex_roots()[0]
            0.713639173536900883414
        """
        R = self.pari().polroots(flag)
        C = complex_field.CC
        return [C(a) for a in R]

    def pari(self):
        try:
            return self.__pari
        except AttributeError:
            v = list(reversed(self.list()))
            try:
                v = [x.pari() for x in v]
            except AttributeError:
                pass
            self.__pari = pari(v).Pol(self.parent().variable())
            return self.__pari
        
    def resultant(self, other, flag=0):
        raise NotImplementedError
            
        ## This should be switched to use NTL, which can apparently compute 
        ## resultants!
        ##        void XGCD(ZZ& r, ZZX& s, ZZX& t, const ZZX& a, const ZZX& b, 
        ##          long deterministic=0);
        ##// r = resultant of a and b; if r != 0, then computes s and t such
        ##// that: a*s + b*t = r; otherwise s and t not affected.  if
        ##// !deterministic, then resultant computation may use a randomized
        ##// strategy that errs with probability no more than 2^{-80}.
        #m = magma.Magma()
        #cmd = "R<%s> := PolynomialRing(RationalField()); "%self.parent().variable() + \
        #      "Resultant(%s, %s);"%(self,other)
        #s = m.cmd(cmd)
        #i = s.find("\r")
        #return eval(s[:i])

    def reverse(self):
        v = list(self.list())
        v.reverse()
        return Polynomial(self.parent(), v)

    def valuation(self):
        """
        If f = a_r x^r + a_{r+1}x^{r+1} + higher terms ..., with
        a_r nonzero then the valuation of f is r.  The valuation of
        the zero polynomial is rings.infinity.
        """
        if self.is_zero():
            return infinity
        for i in xrange(self.degree()+1):
            if self[i] != 0:
                return i
        raise RuntimeError, "bug in computing valuation of polynomial"
            
    def variable(self):
        return self.parent().variable()

    def xgcd(self, other):
        """
        Extended gcd of self and polynomial other.
        """
        # Algorithm 3.2.2 of Cohen, GTM 138
        R = self.parent()
        A = self
        B = other
        U = R(1)
        G = A
        V1 = R(0)
        V3 = B
        while V3.is_nonzero():
            Q, R = G/V3
            T = U - V1*Q
            U = V1
            G = V3
            V1 = T
            V3 = R
        V = (G-A*U)//B
        return G, U, V

    def is_irreducible(self):
        try:
            f = self.pari()
            return f.polisirreducible()
        except:
            F = self.factor()
            if len(F) > 1 or F[0][1] > 1:
                return False
            return True     
        
    
class Polynomial_generic_dense(Polynomial):
    def __init__(self, parent, x=None, check=True, is_gen=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        if x == None:
            self.__coeffs = []
            return
        R = parent.base_ring()
#        if isinstance(x, Polynomial) and x.parent() == self.parent():
#            x = list(x.list())
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                x = list(x.list())
            elif x.parent() == R:
                x = [x]
            else:
                raise TypeError, "Cannot coerce %s into %s."%(x, parent)
        elif isinstance(x, dict):
            zero = R(0)
            n = max(x.keys())
            v = [zero for _ in xrange(n+1)]
            for i, z in x.iteritems():
                v[i] = z
            x = v
        elif isinstance(x, sage.pari.py_pari.gen):
            x = [R(w) for w in reversed(x.Vec())]
            check = True
        elif not isinstance(x, list):
            x = [x]   # constant polynomials
        if check:
            self.__coeffs = [R(z) for z in x]
        else:
            self.__coeffs = x
        if check:
            self.__normalize()
        
    def __reduce__(self):
        return Polynomial_generic_dense, (self.parent(), self.list(), False)

    def __normalize(self):
        x = self.__coeffs
        n = len(x)-1
        while n>=0 and x[n] == 0:
            del x[n]
            n -= 1
        
    def __getitem__(self,n):
        if n < 0 or n >= len(self.__coeffs):
            return self.base_ring()(0)
        return self.__coeffs[n]

    def __getslice__(self, i, j):
        if i < 0:
            i = 0
        v = self.__coeffs[i:j]
        zero = self.base_ring()(0)
        for k in xrange(len(self.__coeffs),j):
            v.append(zero)
        return v

    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        value = self.base_ring()(value)
        if n >= 0 and n < len(self.__coeffs):
            self.__coeffs[n] = value
            if n == len(self.__coeffs) and value == 0:
                self.__normalize()  
        elif n < 0:
            raise IndexError, "polynomial coefficient index must be nonnegative"
        elif value != 0:
            zero = self.base_ring()(0)
            for _ in xrange(len(self.__coeffs), n):
                self.__coeffs.append(zero)
            self.__coeffs.append(value)
            
    def list(self):
        return self.__coeffs

    def degree(self):
        return len(self.__coeffs) - 1


class Polynomial_generic_sparse(Polynomial):
    def __init__(self, parent, x=None, check=True, is_gen=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        if x == None:
            self.__coeffs = {}
            return
        R = parent.base_ring()
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                x = dict(x.dict())
            elif x.parent() == R:
                x = {0:x}
            else:
                raise TypeError, "Cannot coerce %s into %s."%(x, parent)
        elif isinstance(x, list):
            y = {}
            for i in xrange(len(x)):
                if x[i] != 0:
                    y[i] = x[i]
            x = y
        elif not isinstance(x, dict):
            x = {0:x}   # constant polynomials
        elif isinstance(x, sage.pari.py_pari.gen):
            x = [R(w) for w in reversed(x.Vec())]
            check = True
        if check:
            self.__coeffs = {}
            for i, z in x.iteritems():
                self.__coeffs[i] = R(z)
        else:
            self.__coeffs = x
        if check:
            self.__normalize()
        
    def __reduce__(self):
        return Polynomial_generic_sparse, (self.parent(), self.dict(), False)

    def __normalize(self):
        x = self.__coeffs
        zero = self.base_ring()(0)
        D = [n for n, z in x.iteritems() if z == 0]
        for n in D:
            del x[n]
        
    def __getitem__(self,n):
        if not self.__coeffs.has_key(n):
            return self.base_ring()(0)
        return self.__coeffs[n]

    def __getslice__(self, i, j):
        if i < 0:
            i = 0
        zero = self.base_ring()(0)
        v = [zero for _ in xrange(i,j)]
        x = self.__coeffs
        for k in set(x.keys()).intersection(set(xrange(i,j))):
            v[k] = x[k]
        return v

    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        value = self.base_ring()(value)
        x = self.__coeffs
        if n < 0:
            raise IndexError, "polynomial coefficient index must be nonnegative"
        if value == 0:
            if x.has_key(n):
                del x[n]
        else:
            x[n] = value
            
    def list(self):
        zero = self.base_ring()(0)
        v = [zero for _ in xrange(self.degree()+1)]
        for n, x in self.__coeffs.iteritems():
            v[n] = x
        return v

    def degree(self):
        v = self.__coeffs.keys()
        if len(v) == 0:
            return -1
        return max(v)


Q = rational_field.RationalField()        
class Polynomial_rational_dense(Polynomial):
    def __init__(self, parent, x=None, check=True, is_gen=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        self.__poly = polynomial_pyx.Polynomial_rational()
        if x == None: return
        if isinstance(x, polynomial_pyx.Polynomial_rational):
            self.__poly = x
            return
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                x = list(x.list())
            elif x.parent() == R:
                x = [x]
            else:
                raise TypeError, "Cannot coerce %s into %s."%(x, parent)
        elif isinstance(x, dict):
            zero = Q(0)
            n = max(x.keys())
            v = [zero for _ in xrange(n+1)]
            for i, z in x.iteritems():
                v[i] = z
            x = v
        elif isinstance(x, sage.pari.py_pari.gen):
            x = [Q(str(w)) for w in reversed(x.Vec())]
            check = True
        elif not isinstance(x, list):
            x = [x]   # constant polynomials
        if check:
            x = [Q(z) for z in x]
        self.__poly.set_from_list(x)
           
    def __reduce__(self):
        return Polynomial_rational_dense, \
               (self.parent(), self.list(), False, self._is_gen)
        
    def __getitem__(self,n):
        return Q(self.__poly[n])

    def __getslice__(self, i, j):
        return [Q(self.__poly[k]) for k in range(i,j)]
     
    def __add__(self, right):
        if not isinstance(right, Polynomial_rational_dense):
            return bin_op(self, right, operator.add)
        return Polynomial_rational_dense(self.parent(), 
                                   self.__poly + right.__poly)
                                   
    def __mul__(self, right):
        """
        EXAMPLES:
            >>> import rings
            >>> Q = rings.RationalField()
            >>> x = rings.PolynomialRing(Q).gen()
            >>> (x - '2/3')*(x**2 - 8*x + 16)
            x^3 - 26/3*x^2 + 64/3*x - 32/3
        """
        if not isinstance(right, Polynomial):
            right = Q(right)
            v = list(self.list())
            for i in range(len(v)):
                v[i] *= right
            return self.polynomial(v)
        if not isinstance(right, Polynomial_rational_dense):
            return bin_op(self, right, operator.mul)
        return Polynomial_rational_dense(self.parent(), 
                                   self.__poly * right.__poly)        

    def __sub__(self, right):
        if not isinstance(right, Polynomial_rational_dense):
            return bin_op(self, right, operator.sub)
        return Polynomial_rational_dense(self.parent(), 
                                   self.__poly - right.__poly)
                               
    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        self.__poly[n] = value
        if hasattr(self, "__list"):
            del self.__list
     
    def copy(self):
        f = Polynomial_rational_dense(self.parent())
        f.__poly = self.__poly.copy()
        return f
        
    def degree(self):
        return self.__poly.degree()

    def discriminant(self):
        s = sage.pari.py_pari.eval("poldisc(%s)"%self)
        return rational_field.RationalField()(s)
      
    def list(self):
        try:
            return self.__list
        except AttributeError:
            self.__list = [Q(x) for x in self.__poly.list()]
            return self.__list

    def resultant(self, other):
        """
        Returns the resultant of self and other, which must lie in the same
        polynomial ring.

        INPUT:
            other -- a polynomial
            flag -- integer
               0: default, assumes that the polynomials have exact
                  entries (uses the subresultant algorithm)
               1: for arbitrary polynomials, using Sylvester's matrix
               2: using a Ducos's modified subresultant algorithm.
        OUTPUT:
            an element of the base ring of the polynomial ring
        NOTES:
            Implemented using pari's polresultant function.
        EXAMPLES:
            >>> x = PolynomialRing(IntegerRing()).gen()
            >>> f=x**3 + x + 1;  g=x**3 - x - 1
            >>> f.resultant(g)
            -8
        """
        if not isinstance(other, Polynomial) or self.parent() != other.parent():
            other = self.polynomial(other)
        K = self.base_ring()
        return K(str(self.pari().polresultant(other.pari())))
    resultant.__doc__ = Polynomial.resultant.__doc__


# doctests
if __name__ ==  '__main__':
    import doctest, sys
    
    # The doctests assume that the following names have been imported:
    from integer_ring import IntegerRing
    from rational_field import RationalField
    from real_field import RealField
    from polynomial_ring import PolynomialRing
    
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])
    
    

