"""
Classes that define maps and homomorphisms
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.misc.errors as errors
import sage.linalg.linalg as linalg


class Function:
    def __init__(self, f, domain, codomain, f_inverse=None):
        """
        The first argument f is a function of one argument that
        takes objects in domain and returns objects in codomain.  The
        Map object, when called with argument x, tries to coerce it
        into the domain, if possible, and raises an error if it is not.
        """
        self.__domain = domain
        self.__codomain = codomain
        self.__f = f
        self.__f_inverse = f_inverse
        
    def __repr__(self):
        return "Function: %s --> %s"%(self.domain(),self.codomain())
        
    def domain(self):
        return self.__domain
        
    def codomain(self):
        return self.__codomain
        
    def __invert__(self):  # notation in python is (~f) for the inverse of f.
        assert self.__f_inverse != None, "Inverse of %s not known."%self
        return Function(elf.__f_inverse, self.__codomain, \
                        self.__domain, self.__f)
                        
    def __call__(self, x):
        try:
            y = self.__domain(x)
        except errors.CoercionError:
            raise errors.CoercionError("%s must be coercible into %s"%(x,self.domain()))
        return self.__f(y)


# A matrix function is a function that is defined by multiplication by a
# matrix.  Elements of domain must either have a method "vector()" that
# returns a vector that the defining matrix can hit from the left, or
# be coercible into vector space of appropriate dimension.

class MatrixFunction(Function):

    def __init__(self, matrix, domain, codomain, name=""):
        self.__matrix = matrix
        self.__domain_vs = linalg.VectorSpace(\
            matrix.parent().base_ring(),matrix.ncols())
        self.__name = name
        Function.__init__(self, None, domain, codomain, None)
        
    def name(self, new=None):
        if new == None:
            return self.__name
        self.__name = new
        
    def __repr__(self):
        #return "Function: %s\n%s --> %s\nDefined by\n%s"%(\
        #    self.__name, self.domain(), self.codomain(), linalg.Matrix.__repr__(self))
        return "Linear function defined by matrix:\n%s"%self.__matrix
                
    def __call__(self, x):
        if x.parent() != self.domain():
            x = self.domain()(x)
        try:
            x = x.vector()
        except AttributeError:
            try:
                x = self.__domain_vs(x)
            except errors.CoercionError:
                raise errors.CoercionError("%s must be coercible into %s"%\
                                           (x,self.domain()))
        return self.codomain()(x*self)
    
    def __invert__(self):
        if self.nrows() != self.ncols():
            raise errors.MathError("Inverse of morphism not defined."%self)
        try:
            B = linalg.Matrix.__invert__(self) 
        except ZeroDivisionError:
            raise ZeroDivisionError, "Inverse does not exist."
        return MatrixFunction(B, self.codomain(), self.domain())
            
    def __mul__(self, right):
        if isinstance(right, MatrixFunction):
            return MatrixFunction(linalg.Matrix.__mul__(self,right), \
                                  right.domain(), self.codomain())
        return linalg.Matrix.__mul__(self, right)
        
    def charpoly(self):
        if not self.is_endomorphism():
            raise ArithmeticError, "charpoly only defined for endomorphisms " +\
                    "(i.e., domain = range)"
        return self.__matrix.charpoly()
        
    def decompose(self):    
        D = self.domain()
        return [D(V) for V, _ in self.matrix().decompose()]
        
    def fcp(self):
        return self.charpoly().factor()

    def kernel(self):
        V = self.matrix().kernel()
        return self.domain()(V)
        
    def is_endomorphism(self):
        return self.domain() == self.codomain()
        
    def matrix(self):
        return self.__matrix

    def restrict_domain(self, sub):
        """Restrict this matrix function to a subspace sub of the domain.
        The subspace sub should have a basis() method and elements of the basis
        should be coercible into domain.
        """
        D = self.domain()
        B = sub.basis()
        ims = [self*(D(b).vector()) for b in B]
        A = linalg.MatrixSpace(self.base_ring(),len(B),self.nrows())(ims).transpose()
        return MatrixFunction(A, sub, self.codomain())
    
# import rings
# def testfun():
#     def f(x):
#         return x+1
#     def finv(x):
#         return x-1
#     return Function(f, rings.Q, rings.Q, finv)

# import linalg
# def testfun2():
#     M = linalg.MatrixSpace(rings.Q,2,2)
#     A = M([1,2,3,4])
#     V = linalg.VectorSpace(rings.Q,2)
#     return MatrixFunction(A, V, V)
