"""nodoctest

John Jones's tables of number fields with bounded ramification and
degree.

"""

#*****************************************************************************
#       SAGE: System for Arithmetic Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import os

from sage.rings.rings import NumberField, x  # x is generator of QQ[x].
from sage.misc.misc import powerset
import sage.tables.db 

class Database(sage.tables.db.Database):
    def __init__(self, read_only=True):
        sage.tables.db.Database.__init__(self,
                  name="jones", read_only=read_only)

    def __repr__(self):
        return "John Jones's table of number fields with bounded ramification and degree <= 6."
    
    def _load(self, path, filename):
        print filename
        i = 0
        while filename[i].isalpha():
            i += 1
        j = len(filename)-1
        while filename[j].isalpha() or filename[j] in [".", "_"]:
            j -= 1
        S = [eval(z) for z in filename[i:j+1].split("-")]
        S.sort()
        data = open(path + "/" + filename).read()
        data = data.replace("^","**")
        v = eval(data)
        s = tuple(S)
        if self.root.has_key(s):
            self.root[s] += v
            self.root[s].sort()
        else:
            self.root[s] = v
            

    def _create_from_jones_data(self,
                                path = \
           "/home/was/people/jones/tables/hobbes.la.asu.edu/Number_Fields"):
        """
        Create the database from scratch from the PARI files on John
        Jone's web page.
        INPUT:
            -- path must be the path to Jone's Number_Fields directory, e.g.,
                 ~/people/jones/tables/hobbes.la.asu.edu/Number_Fields
               These files should have been downloaded using wget.
        EXAMPLE:
            This is how to create the database from scratch, assuming
            that the number fields are in the default directory above:
            >> import jones
            >> j=jones.Jones(read_only=False)
            >> j._create_from_jones_data()
            >> j.commit()
        """
        n = 0
        self.root[tuple([])] = [x-1]
        if not os.path.exists(path):
            raise IOError, "Path %s does not exist."%path
        for X in os.listdir(path):
            if X[-4:] == "solo":
                Z = path + "/" + X
                print X
                for Y in os.listdir(Z):
                    if Y[-3:] == ".gp":
                        self._load(Z, Y)

    def unramified_outside(self, S, d=None):
        """
        Return iterator over fields in the database of degree d
        unramified outside S.  If d is omitted, return fields of any
        degree.
        INPUT:
            S -- list or set of primes
            d -- None (default) or an integer 
        EXAMPLES:
            >>> from sage.all import tables
            >>> J = tables.jones()
            >>> J.unramified_outside([101,119])
            [Number Field with defining polynomial x - 1,
             Number Field with defining polynomial x^2 - 101,
             Number Field with defining polynomial x^4 - x^3 + 13*x^2 - 19*x + 361,
             Number Field with defining polynomial x^5 + 2*x^4 + 7*x^3 + 4*x^2 + 11*x - 6,
             Number Field with defining polynomial x^5 + x^4 - 6*x^3 - x^2 + 18*x + 4,
             Number Field with defining polynomial x^5 - x^4 - 40*x^3 - 93*x^2 - 21*x + 17]
        """
        try:
            S = list(S)
        except TypeError:
            S = [S]
        Z = []
        for X in powerset(S):
            Z += [k for k in self.ramified_at(X, d)]
        Z.sort()
        return Z

    def __getitem__(self, S):
        try:
            S = list(S)
        except TypeError:
            S = [S]
        S.sort()
        s = tuple(S)
        if not self.root.has_key(s):
            return []
        return [NumberField(f, check=False) for f in self.root[s]]

    def ramified_at(self, S, d=None):
        """
        Return all fields in the database of degree d ramified
        exactly at the primes in S.
        INPUT:
            S -- list or set of primes
            d -- None (default) or an integer 
        EXAMPLES:
            >>> from sage.all import tables
            >>> J = tables.jones()
            >>> J.ramified_at([101,119])
            >>> J.ramified_at([119])
            >>> J.ramified_at(101)
            [Number Field with defining polynomial x^2 - 101,
             Number Field with defining polynomial x^4 - x^3 + 13*x^2 - 19*x + 361,
             Number Field with defining polynomial x^5 + 2*x^4 + 7*x^3 + 4*x^2 + 11*x - 6,
             Number Field with defining polynomial x^5 + x^4 - 6*x^3 - x^2 + 18*x + 4,
             Number Field with defining polynomial x^5 - x^4 - 40*x^3 - 93*x^2 - 21*x + 17]
        """
        Z = self[S]
        if d == None:
            return Z
        return [k for k in self[S] if k.degree() == d]
        

# not done since for some reason it messes up the ZODB stuff.
if __name__ ==  '__main__':
    import doctest, sys
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])
